/*******************************************************************************
*                                                                              *
*   (C) 1997-2002 by Ernst W. Mayer.                                           *
*                                                                              *
*  This program is free software; you can redistribute it and/or modify it     *
*  under the terms of the GNU General Public License as published by the       *
*  Free Software Foundation; either version 2 of the License, or (at your      *
*  option) any later version.                                                  *
*                                                                              *
*  This program is distributed in the hope that it will be useful, but WITHOUT *
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
*  more details.                                                               *
*                                                                              *
*  You should have received a copy of the GNU General Public License along     *
*  with this program; see the file GPL.txt.  If not, you may view one at       *
*  http://www.fsf.org/licenses/licenses.html, or obtain one by writing to the  *
*  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     *
*  02111-1307, USA.                                                            *
*                                                                              *
*******************************************************************************/

#include "Mlucas.h"

/* Use for testing higher-accuracy version of the twiddles computation */
#define HIACC 1

/***************/

void radix16_dif_pass(double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int index[], int nloops, int incr)
{

/*
!...Acronym: DIF = Decimation In Frequency
!
!...Subroutine to perform a single radix-16 complex DIF FFT pass on the data in the length-N real vector A.
!
!   The data are stored in a 1-D zero-offset array, with 2^PAD_BITS 8-byte padding elements inserted
!   between every block of 2^DAT_BITS contiguous data. The array padding is to prevent data being accessed
!   in strides that are large powers of two and thus to minimize cache thrashing
!   (in cache-based microprocessor architectures) or bank conflicts (in supercomputers.)
*/
	int i,j,j1,jt,jlo,jhi,k,m,iroot_prim,iroot,k1,k2,nrtm1,p1,p2,p3,p4,p8,p12;
	static double c = 0.9238795325112867561281831, s = 0.3826834323650897717284599;	/* exp[i*(twopi/16)]*/
	double rt,it
		,c1,c2,c3,c4,c5,c6,c7,c8,c9,c10,c11,c12,c13,c14,c15
		,s1,s2,s3,s4,s5,s6,s7,s8,s9,s10,s11,s12,s13,s14,s15
	,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,t30,t31,t32;
#if PFETCH
	double *add0;
	int prefetch_offset;
#endif
/*
!   Here's how our padded-array indexing works: given an unpadded-array index J,
!   we convert that to a padded-array index J1 using the formula
!
!	J1 = floor[J/(NBLOCK)] x [NBLOCK+PAD] + mod[J, NBLOCK],
!
!   where NBLOCK is the number of 8-byte array data in each contiguous-data block = 2^DAT_BITS,
!   and PAD is the number of 8-byte padding data between data block = 2^PAD_BITS.
!   Since N is an integer, we use the fact that mod[J, NBLOCK] = J - [NBLOCK] x floor[J/(NBLOCK)] to derive
!
!	J1 = J + PAD x floor[J/(NBLOCK)],
!
!   which expresses the fact that J1-J is PAD times the number of padding blocks we skip over.
!
!   Since NBLOCK is a power of 2, i.e. DAT_BITS = log_2(NBLOCK) is an integer, then floor[J/(NBLOCK)] = J >> DAT_BITS,
!   and J1 can be calculated using a shift, an integer multiply, and an add:
!
!	J1 = J + PAD x (J >> DAT_BITS).
!
!   If both NBLOCK and PAD are powers of 2 (as here), this can be done using two shifts and an add:
!
!	J1 = J + (J >> DAT_BITS) << PAD_BITS.
!
!   If PAD is 4 or 8, this can be done using a shift and a scaled add on Alpha.
!
!   Note that we only need to do this kind of calculation for the first elements of each block of 16 complex data.
!
!...The radix-16 pass is here. The data are processed in NLOOPS blocks of INCR real elements each
!   stride between elements processed together.
*/
	nrtm1 = nrt - 1;

	p1 = incr >> 4;
	p2 = p1 +p1;
	p3 = p2 +p1;
	p4 = p3 +p1;
	p8 = p4 +p4;
	p12= p8 +p4;

	p1 = p1 + ( (p1 >> DAT_BITS) << PAD_BITS );
	p2 = p2 + ( (p2 >> DAT_BITS) << PAD_BITS );
	p3 = p3 + ( (p3 >> DAT_BITS) << PAD_BITS );
	p4 = p4 + ( (p4 >> DAT_BITS) << PAD_BITS );
	p8 = p8 + ( (p8 >> DAT_BITS) << PAD_BITS );
	p12= p12+ ( (p12>> DAT_BITS) << PAD_BITS );

	k=0;				/* init sincos array index */
	jlo=0; jhi=(incr >> 4)-1;	/* init inner loop bounds */
	iroot_prim=(incr >> 5);		/* (incr/2)/radix_now */
/*
printf("\n radix16_dif_pass: nloops = %d  incr = %d  twiddles = %d th roots of unity.\n",nloops,incr,n/(2*iroot_prim));
*/
	for(m=0; m < nloops; m++)	/* NLOOPS may range from 1 (if first pass radix = 16) to P*N/32 (last pass radix = 16).     */
	{				/* NLOOPS satisfies the identity NLOOPS * INCR = P*N, which says that we process the entire */
					/* array each time this subroutine is executed (since P*N = vector length, sans padding.)   */
/*
printf("%d...",k);
*/
/*	here are the needed sincos data - these are processed below in bit-reversed order. */
	  iroot = index[k]*iroot_prim;
	  i = iroot;
/* In the DIF pass we may be able to afford a more-accurate twiddles computation,
   since we only compute each set of twiddles once and re-use it for many data blocks. */
#if HIACC
	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 2*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c1=t1*t3-t2*t4;	s1=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 3*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c2=t1*t3-t2*t4;	s2=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 4*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c3=t1*t3-t2*t4;	s3=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 5*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c4=t1*t3-t2*t4;	s4=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 6*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c5=t1*t3-t2*t4;	s5=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 7*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c6=t1*t3-t2*t4;	s6=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 8*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c7=t1*t3-t2*t4;	s7=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 9*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c8=t1*t3-t2*t4;	s8=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/*10*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c9=t1*t3-t2*t4;	s9=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/*11*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c10=t1*t3-t2*t4;	s10=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/*12*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c11=t1*t3-t2*t4;	s11=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/*13*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c12=t1*t3-t2*t4;	s12=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/*14*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c13=t1*t3-t2*t4;	s13=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/*15*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c14=t1*t3-t2*t4;	s14=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c15=t1*t3-t2*t4;	s15=t1*t4+t2*t3;
#else
	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 2*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c1=t1*t3-t2*t4;	s1=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += i;				/* 4*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c2=t1*t3-t2*t4;	s2=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += i;				/* 8*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c4=t1*t3-t2*t4;	s4=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += (iroot << 2)+iroot;		/* 13*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c8=t1*t3-t2*t4;	s8=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c13=t1*t3-t2*t4;	s13=t1*t4+t2*t3;

	    /* c3,5 */
	    t1=c1*c4; t2=c1*s4; t3=s1*c4; t4=s1*s4;
	    c3=t1+t4; s3=t2-t3; c5=t1-t4; s5=t2+t3;

	    /* c6,7,9,10 */
	    t1=c1*c8; t2=c1*s8; t3=s1*c8; t4=s1*s8;
	    c7=t1+t4; s7=t2-t3; c9=t1-t4; s9=t2+t3;

	    t1=c2*c8; t2=c2*s8; t3=s2*c8; t4=s2*s8;
	    c6=t1+t4; s6=t2-t3; c10=t1-t4; s10=t2+t3;

	    /* c11,12,14,15 */
	    t1=c1*c13; t2=c1*s13; t3=s1*c13; t4=s1*s13;
	    c12=t1+t4; s12=t2-t3; c14=t1-t4; s14=t2+t3;

	    t1=c2*c13; t2=c2*s13; t3=s2*c13; t4=s2*s13;
	    c11=t1+t4; s11=t2-t3; c15=t1-t4; s15=t2+t3;
#endif

	  for(j=jlo; j < jhi; j += 2)
	  {

	    j1=j + ( (j>> DAT_BITS) << PAD_BITS );	/* padded-array fetch index is here */

/* gather the needed data (8 64-bit complex, i.e. 16 64-bit reals) and do the first set of four length-4 transforms.
   We process the sincos data in bit-reversed order.	*/
#if PFETCH
prefetch_offset = ((j >> 1) & 3)*p4 + 4;	/* Cycle among p0, p4, p8 and p12. */
#endif
/*...Block 1: */

	jt = j1;

#if PFETCH
prefetch_data(a, jt+prefetch_offset);
#endif
	    t1 =a[jt    ];			t2 =a[jt+1    ];
	    rt =a[jt+p8 ]*c8 -a[jt+1+p8 ]*s8 ;	it =a[jt+1+p8 ]*c8 +a[jt+p8 ]*s8;
	    t3 =t1 -rt;		t1 =t1 +rt;
	    t4 =t2 -it;		t2 =t2 +it;

	    t5 =a[jt+p4 ]*c4 -a[jt+1+p4 ]*s4 ;	t6 =a[jt+1+p4 ]*c4 +a[jt+p4 ]*s4;
	    rt =a[jt+p12]*c12-a[jt+1+p12]*s12;	it =a[jt+1+p12]*c12+a[jt+p12]*s12;
	    t7 =t5 -rt;		t5 =t5 +rt;
	    t8 =t6 -it;		t6 =t6 +it;

	    rt =t5;	t5 =t1 -rt;	t1 =t1 +rt;
	    it =t6;	t6 =t2 -it;	t2 =t2 +it;

	    rt =t7;	t7 =t3 +t8;	t3 =t3 -t8;
			t8 =t4 -rt;	t4 =t4 +rt;

/*...Block 2: */

	jt = j1 + p2;

#if PFETCH
prefetch_data(a, jt+prefetch_offset);
#endif
	    t9 =a[jt    ]*c2 -a[jt+1    ]*s2 ;	t10=a[jt+1    ]*c2 +a[jt    ]*s2;
	    rt =a[jt+p8 ]*c10-a[jt+1+p8 ]*s10;	it =a[jt+1+p8 ]*c10+a[jt+p8 ]*s10;
	    t11=t9 -rt;		t9 =t9 +rt;
	    t12=t10-it;		t10=t10+it;

	    t13=a[jt+p4 ]*c6 -a[jt+1+p4 ]*s6 ;	t14=a[jt+1+p4 ]*c6 +a[jt+p4 ]*s6;
	    rt =a[jt+p12]*c14-a[jt+1+p12]*s14;	it =a[jt+1+p12]*c14+a[jt+p12]*s14;
	    t15=t13-rt;		t13=t13+rt;
	    t16=t14-it;		t14=t14+it;

	    rt =t13;	t13=t9 -rt;	t9 =t9 +rt;
	    it =t14;	t14=t10-it;	t10=t10+it;

	    rt =t15;	t15=t11+t16;	t11=t11-t16;
			t16=t12-rt;	t12=t12+rt;

/*...Block 3: */

	jt = j1 + p1;

#if PFETCH
prefetch_data(a, jt+prefetch_offset);
#endif
	    t17=a[jt    ]*c1 -a[jt+1    ]*s1 ;	t18=a[jt+1    ]*c1 +a[jt    ]*s1;
	    rt =a[jt+p8 ]*c9 -a[jt+1+p8 ]*s9 ;	it =a[jt+1+p8 ]*c9 +a[jt+p8 ]*s9;
	    t19=t17-rt;		t17=t17+rt;
	    t20=t18-it;		t18=t18+it;

	    t21=a[jt+p4 ]*c5 -a[jt+1+p4 ]*s5 ;	t22=a[jt+1+p4 ]*c5 +a[jt+p4 ]*s5;
	    rt =a[jt+p12]*c13-a[jt+1+p12]*s13;	it =a[jt+1+p12]*c13+a[jt+p12]*s13;
	    t23=t21-rt;		t21=t21+rt;
	    t24=t22-it;		t22=t22+it;

	    rt =t21;	t21=t17-rt;	t17=t17+rt;
	    it =t22;	t22=t18-it;	t18=t18+it;

	    rt =t23;	t23=t19+t24;	t19=t19-t24;
			t24=t20-rt;	t20=t20+rt;

/*...Block 4: */

	jt = j1 + p3;

#if PFETCH
prefetch_data(a, jt+prefetch_offset);
#endif
	    t25=a[jt    ]*c3 -a[jt+1    ]*s3 ;	t26=a[jt+1    ]*c3 +a[jt    ]*s3;
	    rt =a[jt+p8 ]*c11-a[jt+1+p8 ]*s11;	it =a[jt+1+p8 ]*c11+a[jt+p8 ]*s11;
	    t27=t25-rt;		t25=t25+rt;
	    t28=t26-it;		t26=t26+it;

	    t29=a[jt+p4 ]*c7 -a[jt+1+p4 ]*s7 ;	t30=a[jt+1+p4 ]*c7 +a[jt+p4 ]*s7;
	    rt =a[jt+p12]*c15-a[jt+1+p12]*s15;	it =a[jt+1+p12]*c15+a[jt+p12]*s15;
	    t31=t29-rt;		t29=t29+rt;
	    t32=t30-it;		t30=t30+it;

	    rt =t29;	t29=t25-rt;	t25=t25+rt;
	    it =t30;	t30=t26-it;	t26=t26+it;

	    rt =t31;	t31=t27+t32;	t27=t27-t32;
			t32=t28-rt;	t28=t28+rt;
/*
!...and now do four more radix-4 transforms, including the internal twiddle factors:
!	1, exp(i* 1*twopi/16) =        ( c, s), exp(i* 2*twopi/16) = isqrt2*( 1, 1), exp(i* 3*twopi/16) =        ( s, c) (for inputs to transform block 2)
!	1, exp(i* 2*twopi/16) = isqrt2*( 1, 1), exp(i* 4*twopi/16) =        ( 0, 1), exp(i* 6*twopi/16) = isqrt2*(-1, 1) (for inputs to transform block 3)
!	1, exp(i* 3*twopi/16) =        ( s, c), exp(i* 6*twopi/16) = isqrt2*(-1, 1), exp(i* 9*twopi/16) =        (-c,-s) (for inputs to transform block 4).
!  (This is 4 real*complex and 4 complex*complex multiplies (= 24 FMUL), compared to 6 and 4, respectively (= 28 FMUL) for my old scheme.)
!   I.e. do similar as above, except inputs a[j1  +p0:15:1) are replaced by t0:30:2,
!                                           a[j1+1+p0:15:1) are replaced by t1:31:2, and v.v. for outputs,
!   and only the last 3 inputs to each of the radix-4 transforms 2 through 4 are multiplied by non-unity twiddles.
*/
/*...Block 1: t1,9,17,25 */

	jt = j1;

	    rt =t9;	t9 =t1 -rt;	t1 =t1 +rt;
	    it =t10;	t10=t2 -it;	t2 =t2 +it;

	    rt =t25;	t25=t17-rt;	t17=t17+rt;
	    it =t26;	t26=t18-it;	t18=t18+it;

	    a[jt    ]=t1+t17;	a[jt+1    ]=t2+t18;
	    a[jt+p1 ]=t1-t17;	a[jt+1+p1 ]=t2-t18;

	    a[jt+p2 ]=t9 -t26;	a[jt+1+p2 ]=t10+t25;	/* mpy by E^4=i is inlined here... */
	    a[jt+p3 ]=t9 +t26;	a[jt+1+p3 ]=t10-t25;

/*...Block 3: t5,13,21,29 */

	jt = j1 + p4;

	    rt =t13;	t13=t5 +t14;	t5 =t5 -t14;		/* twiddle mpy by E^4 = I */
			t14=t6 -rt;	t6 =t6 +rt;

	    rt =(t21-t22)*ISRT2;t22=(t21+t22)*ISRT2;	t21=rt;	/* twiddle mpy by E^2; could also do mpy by ISRT2 directly on t21,22,29,30 */
	    rt =(t30+t29)*ISRT2;it =(t30-t29)*ISRT2;		/* twiddle mpy by -E^6 is here... */
	    t29=t21+rt;		t21=t21-rt;			/* ...and get E^6=(i-1)/sqrt by flipping signs here. */
	    t30=t22+it;		t22=t22-it;

	    a[jt    ]=t5+t21;	a[jt+1    ]=t6+t22;
	    a[jt+p1 ]=t5-t21;	a[jt+1+p1 ]=t6-t22;

	    a[jt+p2 ]=t13-t30;	a[jt+1+p2 ]=t14+t29;	/* mpy by E^4=i is inlined here... */
	    a[jt+p3 ]=t13+t30;	a[jt+1+p3 ]=t14-t29;

/*...Block 2: t3,11,19,27 */

	jt = j1 + p8;

	    rt =(t11-t12)*ISRT2;it =(t11+t12)*ISRT2;		/* twiddle mpy by E^2; could also do mpy by ISRT2 directly on t11,12 */
	    t11=t3 -rt;		t3 =t3 +rt;
	    t12=t4 -it;		t4 =t4 +it;

	    rt =t19*c - t20*s;	t20=t20*c + t19*s;	t19=rt;	/* twiddle mpy by E^1 */
	    rt =t27*s - t28*c;	it =t28*s + t27*c;		/* twiddle mpy by E^3 */
	    t27=t19-rt;		t19=t19+rt;
	    t28=t20-it;		t20=t20+it;

	    a[jt    ]=t3+t19;	a[jt+1    ]=t4+t20;
	    a[jt+p1 ]=t3-t19;	a[jt+1+p1 ]=t4-t20;

	    a[jt+p2 ]=t11-t28;	a[jt+1+p2 ]=t12+t27;	/* mpy by E^4=i is inlined here... */
	    a[jt+p3 ]=t11+t28;	a[jt+1+p3 ]=t12-t27;

/*...Block 4: t7,15,23,31 */

	jt = j1 + p12;

	    rt =(t16+t15)*ISRT2;it =(t16-t15)*ISRT2;		/* twiddle mpy by -E^6 is here... */
	    t15=t7 +rt;		t7 =t7 -rt;			/* ...and get E^6=(i-1)/sqrt by flipping signs here. */
	    t16=t8 +it;		t8 =t8 -it;

	    rt =t23*s - t24*c;	t24=t24*s + t23*c;	t23=rt;	/* twiddle mpy by E^3 */
	    rt =t31*c - t32*s;	it =t32*c + t31*s;		/* twiddle mpy by E^1 = -E^9... */
	    t31=t23+rt;		t23=t23-rt;			/* ...and get E^9 by flipping signs here. */
	    t32=t24+it;		t24=t24-it;			/* Note: t23+rt = t23*(s+1) */

	    a[jt    ]=t7+t23;	a[jt+1    ]=t8+t24;
	    a[jt+p1 ]=t7-t23;	a[jt+1+p1 ]=t8-t24;

	    a[jt+p2 ]=t15-t32;	a[jt+1+p2 ]=t16+t31;	/* mpy by E^4=i is inlined here... */
	    a[jt+p3 ]=t15+t32;	a[jt+1+p3 ]=t16-t31;

	  }

	  k=k+1;
	  jlo=jlo+incr; jhi=jhi+incr;

	}
/*
printf("\n");
*/
}

/***************/

void radix16_dit_pass(double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int nloops, int incr)
{
/*
!
!...Acronym: DIT = Decimation In Time
!
!...Subroutine to perform a single radix-16 complex inverse DIT FFT pass on the data in the length-N real vector A.
!
!   See the documentation in radix16_dif_pass for further details.
*/
	int i,j,j1,jt,jlo,jhi,k,m,iroot_prim,iroot,k1,k2,nrtm1,p1,p2,p3,p4,p8,p12;
	static double c = 0.9238795325112867561281831, s = 0.3826834323650897717284599;/* exp[i*(twopi/16)]*/
	double rt,it
	,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,t30,t31,t32;
#if PFETCH
	double *add0;
	int prefetch_offset;
#endif
/*
!...The radix-16 pass is here.
!   stride between elements processed together.
*/
	nrtm1 = nrt - 1;

	p1 = (incr >> 4);
	p2 = p1 +p1;
	p3 = p2 +p1;
	p4 = p3 +p1;
	p8 = p4 +p4;
	p12= p8 +p4;

	p1 = p1 + ( (p1 >> DAT_BITS) << PAD_BITS );
	p2 = p2 + ( (p2 >> DAT_BITS) << PAD_BITS );
	p3 = p3 + ( (p3 >> DAT_BITS) << PAD_BITS );
	p4 = p4 + ( (p4 >> DAT_BITS) << PAD_BITS );
	p8 = p8 + ( (p8 >> DAT_BITS) << PAD_BITS );
	p12= p12+ ( (p12>> DAT_BITS) << PAD_BITS );

	k=0;				/* init sincos array index */
	jlo=0; jhi=(incr >> 4)-1;	/* init inner loop bounds */
/*	iroot_prim = nloops;	replace this with something that doesn't depend on nloops, since the latter varies, depending on
				how many pieces we break the full-length array into for the purposes of the cache-friendly scheme. */
	iroot_prim = n/incr;
/*
printf("\n radix16_dit_pass: nloops = %d  incr = %d  using %d th roots of unity.\n",nloops,incr,n/(2*iroot_prim));
*/
	for(m=0; m < nloops; m++)
	{
	  for(j=jlo; j < jhi; j += 2)
	  {
	    j1 = j + ( (j>> DAT_BITS) << PAD_BITS );	/* padded-array fetch index is here */
/*
if(k==0)printf("%d - %d...", k, (incr >> 5)-1 );
*/
/*	here are the needed sincos data - these will be processed below in bit-reversed order. */
	    iroot = k*iroot_prim;
	    i = iroot;

	      k1=(i & nrtm1);
	      k2=(i >> nrt_bits);
	      i += iroot;			/* 2*iroot */
	      t1=rt0[k1].re;	t2=rt0[k1].im;
	      rt=rt1[k2].re;	it=rt1[k2].im;
	      t17=t1*rt-t2*it;	t18=t1*it+t2*rt;

	      k1=(i & nrtm1);
	      k2=(i >> nrt_bits);
	      i += (iroot << 1);		/* 4*iroot */
	      t1=rt0[k1].re;	t2=rt0[k1].im;
	      rt=rt1[k2].re;	it=rt1[k2].im;
	      t9=t1*rt-t2*it;	t10=t1*it+t2*rt;

	      k1=(i & nrtm1);
	      k2=(i >> nrt_bits);
	      i += (iroot << 2);		/* 8*iroot */
	      t1=rt0[k1].re;	t2=rt0[k1].im;
	      rt=rt1[k2].re;	it=rt1[k2].im;
	      t5=t1*rt-t2*it;	t6=t1*it+t2*rt;

	      k1=(i & nrtm1);
	      k2=(i >> nrt_bits);
	      i += (iroot << 2) + iroot;	/* 13*iroot */
	      t1=rt0[k1].re;	t2=rt0[k1].im;
	      rt=rt1[k2].re;	it=rt1[k2].im;
	      t3=t1*rt-t2*it;	t4=t1*it+t2*rt;

	      k1=(i & nrtm1);
	      k2=(i >> nrt_bits);
	      t1=rt0[k1].re;	t2=rt0[k1].im;
	      rt=rt1[k2].re;	it=rt1[k2].im;
	      t23=t1*rt-t2*it;	t24=t1*it+t2*rt;

	    /* c3,5 */
	      t1=t17*t5 ; t2=t17*t6 ; rt=t18*t5 ; it=t18*t6;
	      t25=t1 +it; t26=t2 -rt; t21=t1 -it; t22=t2 +rt;

	    /* c6,7,9,10 */
	      t1=t17*t3 ; t2=t17*t4 ; rt=t18*t3 ; it=t18*t4;
	      t29=t1 +it; t30=t2 -rt; t19=t1 -it; t20=t2 +rt;

	      t1=t9 *t3 ; t2=t9 *t4 ; rt=t10*t3 ; it=t10*t4;
	      t13=t1 +it; t14=t2 -rt; t11=t1 -it; t12=t2 +rt;

	    /* c11,12,14,15 */
	      t1=t17*t23; t2=t17*t24; rt=t18*t23; it=t18*t24;
	      t7 =t1 +it; t8 =t2 -rt; t15=t1 -it; t16=t2 +rt;

	      t1=t9 *t23; t2=t9 *t24; rt=t10*t23; it=t10*t24;
	      t27=t1 +it; t28=t2 -rt; t31=t1 -it; t32=t2 +rt;

/*       gather the needed data (16 64-bit complex, i.e. 32 64-bit reals) and do the first set of four length-4 transforms. */
#if PFETCH
prefetch_offset = ((j >> 1) & 3)*p1 + 4;	/* Cycle among p0, p1, p2 and p3. */
#endif
/*...Block 1: */

	jt = j1;

#if PFETCH
prefetch_data(a, jt+prefetch_offset);
#endif
	    t1 =a[jt    ];			t2 =a[jt+1    ];
	    rt =a[jt+p1 ]*t3 +a[jt+1+p1 ]*t4;	it =a[jt+1+p1 ]*t3 -a[jt+p1 ]*t4;
	    t3 =t1 -rt;		t1 =t1 +rt;
	    t4 =t2 -it;		t2 =t2 +it;

	    rt =a[jt+p2 ]*t5 +a[jt+1+p2 ]*t6;	t6 =a[jt+1+p2 ]*t5 -a[jt+p2 ]*t6;	t5 =rt;
	    rt =a[jt+p3 ]*t7 +a[jt+1+p3 ]*t8;	it =a[jt+1+p3 ]*t7 -a[jt+p3 ]*t8;
	    t7 =t5 -rt;		t5 =t5 +rt;
	    t8 =t6 -it;		t6 =t6 +it;

	    rt =t5;	t5 =t1 -rt;	t1 =t1 +rt;
	    it =t6;	t6 =t2 -it;	t2 =t2 +it;

	    rt =t7;	t7 =t3 -t8;	t3 =t3 +t8;
			t8 =t4 +rt;	t4 =t4 -rt;

/*...Block 2: */

	jt = j1 + p4;

#if PFETCH
prefetch_data(a, jt+prefetch_offset);
#endif
	    rt =a[jt    ]*t9 +a[jt+1    ]*t10;	t10=a[jt+1    ]*t9 -a[jt    ]*t10;	t9 =rt;
	    rt =a[jt+p1 ]*t11+a[jt+1+p1 ]*t12;	it =a[jt+1+p1 ]*t11-a[jt+p1 ]*t12;
	    t11=t9 -rt;		t9 =t9 +rt;
	    t12=t10-it;		t10=t10+it;

	    rt =a[jt+p2 ]*t13+a[jt+1+p2 ]*t14;	t14=a[jt+1+p2 ]*t13-a[jt+p2 ]*t14;	t13=rt;
	    rt =a[jt+p3 ]*t15+a[jt+1+p3 ]*t16;	it =a[jt+1+p3 ]*t15-a[jt+p3 ]*t16;
	    t15=t13-rt;		t13=t13+rt;
	    t16=t14-it;		t14=t14+it;

	    rt =t13;	t13=t9 -rt;	t9 =t9 +rt;
	    it =t14;	t14=t10-it;	t10=t10+it;

	    rt =t15;	t15=t11-t16;	t11=t11+t16;
			t16=t12+rt;	t12=t12-rt;

/*...Block 3: */

	jt = j1 + p8;

#if PFETCH
prefetch_data(a, jt+prefetch_offset);
#endif
	    rt =a[jt    ]*t17+a[jt+1    ]*t18;	t18=a[jt+1    ]*t17-a[jt    ]*t18;	t17=rt;
	    rt =a[jt+p1 ]*t19+a[jt+1+p1 ]*t20;	it =a[jt+1+p1 ]*t19-a[jt+p1 ]*t20;
	    t19=t17-rt;		t17=t17+rt;
	    t20=t18-it;		t18=t18+it;

	    rt =a[jt+p2 ]*t21+a[jt+1+p2 ]*t22;	t22=a[jt+1+p2 ]*t21-a[jt+p2 ]*t22;	t21=rt;
	    rt =a[jt+p3 ]*t23+a[jt+1+p3 ]*t24;	it =a[jt+1+p3 ]*t23-a[jt+p3 ]*t24;
	    t23=t21-rt;		t21=t21+rt;
	    t24=t22-it;		t22=t22+it;

	    rt =t21;	t21=t17-rt;	t17=t17+rt;
	    it =t22;	t22=t18-it;	t18=t18+it;

	    rt =t23;	t23=t19-t24;	t19=t19+t24;
			t24=t20+rt;	t20=t20-rt;

/*...Block 4: */

	jt = j1 + p12;

#if PFETCH
prefetch_data(a, jt+prefetch_offset);
#endif
	    rt =a[jt    ]*t25+a[jt+1    ]*t26;	t26=a[jt+1    ]*t25-a[jt    ]*t26;	t25=rt;
	    rt =a[jt+p1 ]*t27+a[jt+1+p1 ]*t28;	it =a[jt+1+p1 ]*t27-a[jt+p1 ]*t28;
	    t27=t25-rt;		t25=t25+rt;
	    t28=t26-it;		t26=t26+it;

	    rt =a[jt+p2 ]*t29+a[jt+1+p2 ]*t30;	t30=a[jt+1+p2 ]*t29-a[jt+p2 ]*t30;	t29=rt;
	    rt =a[jt+p3 ]*t31+a[jt+1+p3 ]*t32;	it =a[jt+1+p3 ]*t31-a[jt+p3 ]*t32;
	    t31=t29-rt;		t29=t29+rt;
	    t32=t30-it;		t30=t30+it;

	    rt =t29;	t29=t25-rt;	t25=t25+rt;
	    it =t30;	t30=t26-it;	t26=t26+it;

	    rt =t31;	t31=t27-t32;	t27=t27+t32;
			t32=t28+rt;	t28=t28-rt;
/*
!...and now do four more radix-4 transforms, including the internal twiddle factors:
!	1, exp(-i* 1*twopi/16) =       ( c,-s), exp(-i* 2*twopi/16) = ISRT2*( 1,-1), exp(-i* 3*twopi/16) =       ( s,-c) (for inputs to transform block 2)
!	1, exp(-i* 2*twopi/16) = ISRT2*( 1,-1), exp(-i* 4*twopi/16) =       ( 0,-1), exp(-i* 6*twopi/16) = ISRT2*(-1,-1) (for inputs to transform block 3)
!	1, exp(-i* 3*twopi/16) =       ( s,-c), exp(-i* 6*twopi/16) = ISRT2*(-1,-1), exp(-i* 9*twopi/16) =       (-c, s) (for inputs to transform block 4).
!  (This is 4 real*complex and 4 complex*complex multiplies (= 24 FMUL), compared to 6 and 4, respectively (= 28 FMUL) for my old scheme.)
!   I.e. do similar as above, except inputs a[jt  +p0:15:1) are replaced by t0:30:2,
!                                           a[jt+1+p0:15:1) are replaced by t1:31:2, and v.v. for outputs,
!   and only the last 3 inputs to each of the radix-4 transforms 2 through 4 are multiplied by non-unity twiddles.
*/
/*...Block 1: t1,9,17,25 */

	jt = j1;

	    rt =t9;	t9 =t1 -rt;	t1 =t1 +rt;
	    it =t10;	t10=t2 -it;	t2 =t2 +it;

	    rt =t25;	t25=t17-rt;	t17=t17+rt;
	    it =t26;	t26=t18-it;	t18=t18+it;

	    a[jt    ]=t1+t17;	a[jt+1    ]=t2+t18;
	    a[jt+p8 ]=t1-t17;	a[jt+1+p8 ]=t2-t18;

	    a[jt+p4 ]=t9 +t26;	a[jt+1+p4 ]=t10-t25;	/* mpy by E^-4 = -I is inlined here... */
	    a[jt+p12]=t9 -t26;	a[jt+1+p12]=t10+t25;

/*...Block 3: t5,13,21,29 */

	jt = j1 + p2;

	    rt =t13;	t13=t5 -t14;	t5 =t5 +t14;		/* twiddle mpy by E^4 =-I */
			t14=t6 +rt;	t6 =t6 -rt;

	    rt =(t22+t21)*ISRT2;t22=(t22-t21)*ISRT2;	t21=rt;	/* twiddle mpy by E^-2 */
	    rt =(t29-t30)*ISRT2;it =(t29+t30)*ISRT2;		/* twiddle mpy by E^2 = -E^-6 is here... */
	    t29=t21+rt;		t21=t21-rt;			/* ...and get E^-6 by flipping signs here. */
	    t30=t22+it;		t22=t22-it;

	    a[jt    ]=t5+t21;	a[jt+1    ]=t6+t22;
	    a[jt+p8 ]=t5-t21;	a[jt+1+p8 ]=t6-t22;

	    a[jt+p4 ]=t13+t30;	a[jt+1+p4 ]=t14-t29;	/* mpy by E^-4 = -I is inlined here... */
	    a[jt+p12]=t13-t30;	a[jt+1+p12]=t14+t29;

/*...Block 2: t3,11,19,27 */

	jt = j1 + p1;

	    rt =(t12+t11)*ISRT2;it =(t12-t11)*ISRT2;		/* twiddle mpy by E^-2 */
	    t11=t3 -rt;		t3 =t3 +rt;
	    t12=t4 -it;		t4 =t4 +it;

	    rt =t19*c + t20*s;	t20=t20*c - t19*s;	t19=rt;	/* twiddle mpy by E^-1 */
	    rt =t27*s + t28*c;	it =t28*s - t27*c;		/* twiddle mpy by E^-3 */
	    t27=t19-rt;		t19=t19+rt;
	    t28=t20-it;		t20=t20+it;

	    a[jt    ]=t3+t19;	a[jt+1    ]=t4+t20;
	    a[jt+p8 ]=t3-t19;	a[jt+1+p8 ]=t4-t20;

	    a[jt+p4 ]=t11+t28;	a[jt+1+p4 ]=t12-t27;	/* mpy by E^-4 = -I is inlined here... */
	    a[jt+p12]=t11-t28;	a[jt+1+p12]=t12+t27;

/*...Block 4: t7,15,23,31 */

	jt = j1 + p3;

	    rt =(t15-t16)*ISRT2;it =(t15+t16)*ISRT2;		/* twiddle mpy by E^2 = -E^-6 is here... */
	    t15=t7 +rt;		t7 =t7 -rt;			/* ...and get E^6=(i-1)/sqrt by flipping signs here. */
	    t16=t8 +it;		t8 =t8 -it;

	    rt =t23*s + t24*c;	t24=t24*s - t23*c;	t23=rt;	/* twiddle mpy by E^-3 */
	    rt =t31*c + t32*s;	it =t32*c - t31*s;		/* twiddle mpy by E^-1 = -E^-9... */
	    t31=t23+rt;		t23=t23-rt;			/* ...and get E^9 by flipping signs here. */
	    t32=t24+it;		t24=t24-it;

	    a[jt    ]=t7+t23;	a[jt+1    ]=t8+t24;
	    a[jt+p8 ]=t7-t23;	a[jt+1+p8 ]=t8-t24;

	    a[jt+p4 ]=t15+t32;	a[jt+1+p4 ]=t16-t31;	/* mpy by E^-4 = -I is inlined here... */
	    a[jt+p12]=t15-t32;	a[jt+1+p12]=t16+t31;

	    k=k+1; if(k == (incr >> 5)) k=0;	/* increment sincos array index; recycle whenever counter reaches kmax. */
	  }
	  jlo=jlo+incr; jhi=jhi+incr;
	}
/*
printf("\n");
printf("exiting radix8_dif_pass with jlo = %d\n",jlo);
*/
}

/***************/

void radix16_dit_passB(double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int index[], int nloops, int incr)
{
/*
!...Post-twiddles implementation of radix16_dit_pass.
*/
	int i,j,j1,jt,jlo,jhi,k,m,iroot_prim,iroot,k1,k2,nrtm1,p1,p2,p3,p4,p8,p12;
	static double c = 0.9238795325112867561281831, s = 0.3826834323650897717284599;	/* exp[i*(twopi/16)]*/
	double rt,it
		,c1,c2,c3,c4,c5,c6,c7,c8,c9,c10,c11,c12,c13,c14,c15
		,s1,s2,s3,s4,s5,s6,s7,s8,s9,s10,s11,s12,s13,s14,s15
	,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,t30,t31,t32;
#if PFETCH
	double *add0;
	int prefetch_offset;
#endif

	nrtm1 = nrt - 1;

	p1 = incr >> 4;
	p2 = p1 +p1;
	p3 = p2 +p1;
	p4 = p3 +p1;
	p8 = p4 +p4;
	p12= p8 +p4;

	p1 = p1 + ( (p1 >> DAT_BITS) << PAD_BITS );
	p2 = p2 + ( (p2 >> DAT_BITS) << PAD_BITS );
	p3 = p3 + ( (p3 >> DAT_BITS) << PAD_BITS );
	p4 = p4 + ( (p4 >> DAT_BITS) << PAD_BITS );
	p8 = p8 + ( (p8 >> DAT_BITS) << PAD_BITS );
	p12= p12+ ( (p12>> DAT_BITS) << PAD_BITS );

	k=0;				/* init sincos array index */
	jlo=0; jhi=(incr >> 4)-1;	/* init inner loop bounds */
	iroot_prim=(incr >> 5);		/* (incr/2)/radix_now */
/*
printf("\n radix16_dif_pass: nloops = %d  incr = %d  twiddles = %d th roots of unity.\n",nloops,incr,n/(2*iroot_prim));
*/
	for(m=0; m < nloops; m++)	/* NLOOPS may range from 1 (if first pass radix = 16) to P*N/32 (last pass radix = 16).     */
	{				/* NLOOPS satisfies the identity NLOOPS * INCR = P*N, which says that we process the entire */
					/* array each time this subroutine is executed (since P*N = vector length, sans padding.)   */
/*
printf("%d...",k);
*/
/*	here are the needed sincos data - these are processed below in bit-reversed order. */
	  iroot = index[k]*iroot_prim;
	  i = iroot;
/* In the DIF pass we may be able to afford a more-accurate twiddles computation,
   since we only compute each set of twiddles once and re-use it for many data blocks. */
#if HIACC
	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 2*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c1=t1*t3-t2*t4;	s1=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 3*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c2=t1*t3-t2*t4;	s2=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 4*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c3=t1*t3-t2*t4;	s3=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 5*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c4=t1*t3-t2*t4;	s4=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 6*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c5=t1*t3-t2*t4;	s5=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 7*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c6=t1*t3-t2*t4;	s6=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 8*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c7=t1*t3-t2*t4;	s7=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 9*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c8=t1*t3-t2*t4;	s8=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/*10*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c9=t1*t3-t2*t4;	s9=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/*11*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c10=t1*t3-t2*t4;	s10=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/*12*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c11=t1*t3-t2*t4;	s11=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/*13*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c12=t1*t3-t2*t4;	s12=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/*14*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c13=t1*t3-t2*t4;	s13=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/*15*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c14=t1*t3-t2*t4;	s14=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c15=t1*t3-t2*t4;	s15=t1*t4+t2*t3;
#else
	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 2*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c1=t1*t3-t2*t4;	s1=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += i;				/* 4*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c2=t1*t3-t2*t4;	s2=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += i;				/* 8*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c4=t1*t3-t2*t4;	s4=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += (iroot << 2)+iroot;		/* 13*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c8=t1*t3-t2*t4;	s8=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c13=t1*t3-t2*t4;	s13=t1*t4+t2*t3;

	    /* c3,5 */
	    t1=c1*c4; t2=c1*s4; t3=s1*c4; t4=s1*s4;
	    c3=t1+t4; s3=t2-t3; c5=t1-t4; s5=t2+t3;

	    /* c6,7,9,10 */
	    t1=c1*c8; t2=c1*s8; t3=s1*c8; t4=s1*s8;
	    c7=t1+t4; s7=t2-t3; c9=t1-t4; s9=t2+t3;

	    t1=c2*c8; t2=c2*s8; t3=s2*c8; t4=s2*s8;
	    c6=t1+t4; s6=t2-t3; c10=t1-t4; s10=t2+t3;

	    /* c11,12,14,15 */
	    t1=c1*c13; t2=c1*s13; t3=s1*c13; t4=s1*s13;
	    c12=t1+t4; s12=t2-t3; c14=t1-t4; s14=t2+t3;

	    t1=c2*c13; t2=c2*s13; t3=s2*c13; t4=s2*s13;
	    c11=t1+t4; s11=t2-t3; c15=t1-t4; s15=t2+t3;
#endif

	  for(j=jlo; j < jhi; j += 2)
	  {

	    j1=j + ( (j>> DAT_BITS) << PAD_BITS );	/* padded-array fetch index is here */

/*       gather the needed data (16 64-bit complex, i.e. 32 64-bit reals) and do the first set of four length-4 transforms. */
#if PFETCH
prefetch_offset = ((j >> 1) & 3)*p1 + 4;	/* Cycle among p0, p1, p2 and p3. */
#endif
/*...Block 1: */

	jt = j1;

#if PFETCH
prefetch_data(a, jt+prefetch_offset);
#endif
	    t1 =a[jt    ];	t2 =a[jt+1    ];
	    rt =a[jt+p1 ];	it =a[jt+1+p1 ];
	    t3 =t1 -rt;		t1 =t1 +rt;
	    t4 =t2 -it;		t2 =t2 +it;

	    rt =a[jt+p2 ];	t6 =a[jt+1+p2 ];	t5 =rt;
	    rt =a[jt+p3 ];	it =a[jt+1+p3 ];
	    t7 =t5 -rt;		t5 =t5 +rt;
	    t8 =t6 -it;		t6 =t6 +it;

	    rt =t5;	t5 =t1 -rt;	t1 =t1 +rt;
	    it =t6;	t6 =t2 -it;	t2 =t2 +it;

	    rt =t7;	t7 =t3 -t8;	t3 =t3 +t8;
			t8 =t4 +rt;	t4 =t4 -rt;

/*...Block 2: */

	jt = j1 + p4;

#if PFETCH
prefetch_data(a, jt+prefetch_offset);
#endif
	    rt =a[jt    ];	t10=a[jt+1    ];	t9 =rt;
	    rt =a[jt+p1 ];	it =a[jt+1+p1 ];
	    t11=t9 -rt;		t9 =t9 +rt;
	    t12=t10-it;		t10=t10+it;

	    rt =a[jt+p2 ];	t14=a[jt+1+p2 ];	t13=rt;
	    rt =a[jt+p3 ];	it =a[jt+1+p3 ];
	    t15=t13-rt;		t13=t13+rt;
	    t16=t14-it;		t14=t14+it;

	    rt =t13;	t13=t9 -rt;	t9 =t9 +rt;
	    it =t14;	t14=t10-it;	t10=t10+it;

	    rt =t15;	t15=t11-t16;	t11=t11+t16;
			t16=t12+rt;	t12=t12-rt;

/*...Block 3: */

	jt = j1 + p8;

#if PFETCH
prefetch_data(a, jt+prefetch_offset);
#endif
	    rt =a[jt    ];	t18=a[jt+1    ];	t17=rt;
	    rt =a[jt+p1 ];	it =a[jt+1+p1 ];
	    t19=t17-rt;	t17=t17+rt;
	    t20=t18-it;	t18=t18+it;

	    rt =a[jt+p2 ];	t22=a[jt+1+p2 ];	t21=rt;
	    rt =a[jt+p3 ];	it =a[jt+1+p3 ];
	    t23=t21-rt;		t21=t21+rt;
	    t24=t22-it;		t22=t22+it;

	    rt =t21;	t21=t17-rt;	t17=t17+rt;
	    it =t22;	t22=t18-it;	t18=t18+it;

	    rt =t23;	t23=t19-t24;	t19=t19+t24;
			t24=t20+rt;	t20=t20-rt;

/*...Block 4: */

	jt = j1 + p12;

#if PFETCH
prefetch_data(a, jt+prefetch_offset);
#endif
	    rt =a[jt    ];	t26=a[jt+1    ];	t25=rt;
	    rt =a[jt+p1 ];	it =a[jt+1+p1 ];
	    t27=t25-rt;	t25=t25+rt;
	    t28=t26-it;	t26=t26+it;

	    rt =a[jt+p2 ];	t30=a[jt+1+p2 ];	t29=rt;
	    rt =a[jt+p3 ];	it =a[jt+1+p3 ];
	    t31=t29-rt;		t29=t29+rt;
	    t32=t30-it;		t30=t30+it;

	    rt =t29;	t29=t25-rt;	t25=t25+rt;
	    it =t30;	t30=t26-it;	t26=t26+it;

	    rt =t31;	t31=t27-t32;	t27=t27+t32;
			t32=t28+rt;	t28=t28-rt;
/*
!...and now do four more radix-4 transforms, including the internal twiddle factors:
!	1, exp(-i* 1*twopi/16) =       ( c,-s), exp(-i* 2*twopi/16) = ISRT2*( 1,-1), exp(-i* 3*twopi/16) =       ( s,-c) (for inputs to transform block 2)
!	1, exp(-i* 2*twopi/16) = ISRT2*( 1,-1), exp(-i* 4*twopi/16) =       ( 0,-1), exp(-i* 6*twopi/16) = ISRT2*(-1,-1) (for inputs to transform block 3)
!	1, exp(-i* 3*twopi/16) =       ( s,-c), exp(-i* 6*twopi/16) = ISRT2*(-1,-1), exp(-i* 9*twopi/16) =       (-c, s) (for inputs to transform block 4).
!  (This is 4 real*complex and 4 complex*complex multiplies (= 24 FMUL), compared to 6 and 4, respectively (= 28 FMUL) for my old scheme.)
!   I.e. do similar as above, except inputs a[jt  +p0:15:1) are replaced by t0:30:2,
!                                           a[jt+1+p0:15:1) are replaced by t1:31:2, and v.v. for outputs,
!   and only the last 3 inputs to each of the radix-4 transforms 2 through 4 are multiplied by non-unity twiddles.
*/
/*...Block 1: t1,9,17,25 */

	jt = j1;

	    rt =t9;	t9 =t1 -rt;	t1 =t1 +rt;
	    it =t10;	t10=t2 -it;	t2 =t2 +it;

	    rt =t25;	t25=t17-rt;	t17=t17+rt;
	    it =t26;	t26=t18-it;	t18=t18+it;

	    a[jt    ]=t1+t17;		a[jt+1    ]=t2+t18;
	    t1       =t1-t17;		t2         =t2-t18;
	    a[jt+p8 ]=t1 *c8 +t2 *s8;	a[jt+1+p8 ]=t2 *c8 -t1 *s8;

	    rt       =t9 +t26;		it         =t10-t25;	/* mpy by E^-4 = -I is inlined here... */
	    t9       =t9 -t26;		t10        =t10+t25;
	    a[jt+p4 ]=rt *c4 +it *s4;	a[jt+1+p4 ]=it *c4 -rt *s4;
	    a[jt+p12]=t9 *c12+t10*s12;	a[jt+1+p12]=t10*c12-t9 *s12;

/*...Block 3: t5,13,21,29 */

	jt = j1 + p2;

	    rt =t13;	t13=t5 -t14;	t5 =t5 +t14;		/* twiddle mpy by E^4 =-I */
			t14=t6 +rt;	t6 =t6 -rt;

	    rt =(t22+t21)*ISRT2;t22=(t22-t21)*ISRT2;	t21=rt;	/* twiddle mpy by E^-2 */
	    rt =(t29-t30)*ISRT2;it =(t29+t30)*ISRT2;		/* twiddle mpy by E^2 = -E^-6 is here... */
	    t29=t21+rt;		t21=t21-rt;			/* ...and get E^-6 by flipping signs here. */
	    t30=t22+it;		t22=t22-it;

	    rt       =t5 +t21;		it         =t6 +t22;
	    t5       =t5 -t21;		t6         =t6 -t22;
	    a[jt    ]=rt *c2 +it *s2;	a[jt+1    ]=it *c2 -rt *s2;
	    a[jt+p8 ]=t5 *c10+t6 *s10;	a[jt+1+p8 ]=t6 *c10-t5 *s10;

	    rt       =t13+t30;		it         =t14-t29;	/* mpy by E^-4 = -I is inlined here... */
	    t13      =t13-t30;		t14        =t14+t29;
	    a[jt+p4 ]=rt *c6 +it *s6;	a[jt+1+p4 ]=it *c6 -rt *s6;
	    a[jt+p12]=t13*c14+t14*s14;	a[jt+1+p12]=t14*c14-t13*s14;

/*...Block 2: t3,11,19,27 */

	jt = j1 + p1;

	    rt =(t12+t11)*ISRT2;it =(t12-t11)*ISRT2;		/* twiddle mpy by E^-2 */
	    t11=t3 -rt;		t3 =t3 +rt;
	    t12=t4 -it;		t4 =t4 +it;

	    rt =t19*c + t20*s;	t20=t20*c - t19*s;	t19=rt;	/* twiddle mpy by E^-1 */
	    rt =t27*s + t28*c;	it =t28*s - t27*c;		/* twiddle mpy by E^-3 */
	    t27=t19-rt;		t19=t19+rt;
	    t28=t20-it;		t20=t20+it;

	    rt       =t3 +t19;		it         =t4 +t20;
	    t3       =t3 -t19;		t4         =t4 -t20;
	    a[jt    ]=rt *c1 +it *s1;	a[jt+1    ]=it *c1 -rt *s1;
	    a[jt+p8 ]=t3 *c9 +t4 *s9;	a[jt+1+p8 ]=t4 *c9 -t3 *s9;

	    rt       =t11+t28;		it         =t12-t27;	/* mpy by E^-4 = -I is inlined here... */
	    t11      =t11-t28;		t12        =t12+t27;
	    a[jt+p4 ]=rt *c5 +it *s5;	a[jt+1+p4 ]=it *c5 -rt *s5;
	    a[jt+p12]=t11*c13+t12*s13;	a[jt+1+p12]=t12*c13-t11*s13;

/*...Block 4: t7,15,23,31 */

	jt = j1 + p3;

	    rt =(t15-t16)*ISRT2;it =(t15+t16)*ISRT2;		/* twiddle mpy by E^2 = -E^-6 is here... */
	    t15=t7 +rt;		t7 =t7 -rt;			/* ...and get E^6=(i-1)/sqrt by flipping signs here. */
	    t16=t8 +it;		t8 =t8 -it;

	    rt =t23*s + t24*c;	t24=t24*s - t23*c;	t23=rt;	/* twiddle mpy by E^-3 */
	    rt =t31*c + t32*s;	it =t32*c - t31*s;		/* twiddle mpy by E^-1 = -E^-9... */
	    t31=t23+rt;		t23=t23-rt;			/* ...and get E^9 by flipping signs here. */
	    t32=t24+it;		t24=t24-it;

	    rt       =t7 +t23;		it         =t8 +t24;
	    t7       =t7 -t23;		t8         =t8 -t24;
	    a[jt    ]=rt *c3 +it *s3;	a[jt+1    ]=it *c3 -rt *s3;
	    a[jt+p8 ]=t7 *c11+t8 *s11;	a[jt+1+p8 ]=t8 *c11-t7 *s11;

	    rt       =t15+t32;		it         =t16-t31;	/* mpy by E^-4 = -I is inlined here... */
	    t15      =t15-t32;		t16        =t16+t31;
	    a[jt+p4 ]=rt *c7 +it *s7;	a[jt+1+p4 ]=it *c7 -rt *s7;
	    a[jt+p12]=t15*c15+t16*s15;	a[jt+1+p12]=t16*c15-t15*s15;

	  }

	  k=k+1;
	  jlo=jlo+incr; jhi=jhi+incr;

	}
/*
printf("\n");
*/
}

