/*******************************************************************************
*                                                                              *
*   (C) 1997-2004 by Ernst W. Mayer.                                           *
*                                                                              *
*  This program is free software; you can redistribute it and/or modify it     *
*  under the terms of the GNU General Public License as published by the       *
*  Free Software Foundation; either version 2 of the License, or (at your      *
*  option) any later version.                                                  *
*                                                                              *
*  This program is distributed in the hope that it will be useful, but WITHOUT *
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
*  more details.                                                               *
*                                                                              *
*  You should have received a copy of the GNU General Public License along     *
*  with this program; see the file GPL.txt.  If not, you may view one at       *
*  http://www.fsf.org/licenses/licenses.html, or obtain one by writing to the  *
*  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     *
*  02111-1307, USA.                                                            *
*                                                                              *
*******************************************************************************/

#include "Mlucas.h"

/***************/

void radix20_dif_pass1(double a[], int n)
{
/*
!...Acronym: DIF = Decimation In Frequency
!
!...Subroutine to perform an initial radix-20 complex DIF FFT pass on the data in the length-N real vector A.
!
!   Uses an optimized radix-5 transform a la Nussbaumer (2nd ed., p.146).
*/
	int j,j1;
	static int n20, first_entry=TRUE
		    ,p01,p02,p03,p04,p05,p06,p07,p08,p09,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19;
	static double cc1 = -1.25000000000000000000,	/* [cos(u)+cos(2u)]/2-1 = -5/4; u = 2*pi/5 */
		      cc2 =  0.55901699437494742409,	/* [cos(u)-cos(2u)]/2 */
		      s2  =  0.95105651629515357211,	/*  sin(u) */
		      ss1 =  1.53884176858762670130,	/* [sin(u)+sin(2u)] */
		      ss2 =  0.36327126400268044292;	/* [sin(u)-sin(2u)] */
	double rt,it
		,t00,t01,t02,t03,t04,t05,t06,t07,t08,t09
		,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19
		,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29
		,t30,t31,t32,t33,t34,t35,t36,t37,t38,t39;

	if(!first_entry && (n/20) != n20)	/* New runlength?	*/
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  first_entry=FALSE;
	  n20=n/20;

/*   constant index offsets for array load/stores are here.	*/

	  p01 = n20;
	  p02 = p01 + p01;
	  p03 = p02 + p01;
	  p04 = p03 + p01;
	  p05 = p04 + p01;
	  p06 = p05 + p01;
	  p07 = p06 + p01;
	  p08 = p07 + p01;
	  p09 = p08 + p01;
	  p10 = p09 + p01;
	  p11 = p10 + p01;
	  p12 = p11 + p01;
	  p13 = p12 + p01;
	  p14 = p13 + p01;
	  p15 = p14 + p01;
	  p16 = p15 + p01;
	  p17 = p16 + p01;
	  p18 = p17 + p01;
	  p19 = p18 + p01;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p05 = p05 + ( (p05 >> DAT_BITS) << PAD_BITS );
	  p06 = p06 + ( (p06 >> DAT_BITS) << PAD_BITS );
	  p07 = p07 + ( (p07 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p09 = p09 + ( (p09 >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p11 = p11 + ( (p11 >> DAT_BITS) << PAD_BITS );
	  p12 = p12 + ( (p12 >> DAT_BITS) << PAD_BITS );
	  p13 = p13 + ( (p13 >> DAT_BITS) << PAD_BITS );
	  p14 = p14 + ( (p14 >> DAT_BITS) << PAD_BITS );
	  p15 = p15 + ( (p15 >> DAT_BITS) << PAD_BITS );
	  p16 = p16 + ( (p16 >> DAT_BITS) << PAD_BITS );
	  p17 = p17 + ( (p17 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	  p19 = p19 + ( (p19 >> DAT_BITS) << PAD_BITS );
	}

/*...The radix-20 pass is here.	*/

      for(j=0; j < n20; j += 2)
      {
	  j1 = j  + ( (j  >> DAT_BITS) << PAD_BITS );
/*
Twiddleless version requires us to swap inputs as follows:
indices  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19
      -> 0,15,10, 5,16,11, 6  1,12, 7, 2,17, 8, 3,18,13, 4,19,14, 9
I.e. start out with first quintet of indices {0,4,8,12,16}, permute those according to
{0,4,8,12,16}*19%20 = {0,16,12,8,4}, then each is head of a length-4 list of indices with decrement 5.
*/

/* here is the normal radix-5 input ordering:
	  t00=a[j1    ];		t01=a[j1+1    ];
	  t02=a[j1+p04]+a[j1+p16];	t03=a[j1+1+p04]+a[j1+1+p16];
	  t04=a[j1+p08]+a[j1+p12];	t05=a[j1+1+p08]+a[j1+1+p12];
	  t06=a[j1+p08]-a[j1+p12];	t07=a[j1+1+p08]-a[j1+1+p12];
	  t08=a[j1+p04]-a[j1+p16];	t09=a[j1+1+p04]-a[j1+1+p16];
*/

#if 1
/*...gather the needed data (20 64-bit complex, i.e. 40 64-bit reals) and do 4 radix-5 transforms...*/
                     /*                                                inputs                                                   */ /*                 outputs                   */
	RADIX_05_DFT(a[j1    ],a[j1+1    ],a[j1+p16],a[j1+1+p16],a[j1+p12],a[j1+1+p12],a[j1+p08],a[j1+1+p08],a[j1+p04],a[j1+1+p04],t00,t01,t02,t03,t04,t05,t06,t07,t08,t09,rt,it);
	RADIX_05_DFT(a[j1+p15],a[j1+1+p15],a[j1+p11],a[j1+1+p11],a[j1+p07],a[j1+1+p07],a[j1+p03],a[j1+1+p03],a[j1+p19],a[j1+1+p19],t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,rt,it);
	RADIX_05_DFT(a[j1+p10],a[j1+1+p10],a[j1+p06],a[j1+1+p06],a[j1+p02],a[j1+1+p02],a[j1+p18],a[j1+1+p18],a[j1+p14],a[j1+1+p14],t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,rt,it);
	RADIX_05_DFT(a[j1+p05],a[j1+1+p05],a[j1+p01],a[j1+1+p01],a[j1+p17],a[j1+1+p17],a[j1+p13],a[j1+1+p13],a[j1+p09],a[j1+1+p09],t30,t31,t32,t33,t34,t35,t36,t37,t38,t39,rt,it);

/*...and now do 5 radix-4 transforms...*/
                     /*          inputs           */ /*                                      outputs                                      */
	RADIX_04_DIF(t00,t01,t10,t11,t20,t21,t30,t31,a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p03],a[j1+1+p03],a[j1+p02],a[j1+1+p02],rt,it);
	RADIX_04_DIF(t02,t03,t12,t13,t22,t23,t32,t33,a[j1+p16],a[j1+1+p16],a[j1+p17],a[j1+1+p17],a[j1+p19],a[j1+1+p19],a[j1+p18],a[j1+1+p18],rt,it);
	RADIX_04_DIF(t04,t05,t14,t15,t24,t25,t34,t35,a[j1+p14],a[j1+1+p14],a[j1+p15],a[j1+1+p15],a[j1+p12],a[j1+1+p12],a[j1+p13],a[j1+1+p13],rt,it);
	RADIX_04_DIF(t06,t07,t16,t17,t26,t27,t36,t37,a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],a[j1+p10],a[j1+1+p10],a[j1+p11],a[j1+1+p11],rt,it);
	RADIX_04_DIF(t08,t09,t18,t19,t28,t29,t38,t39,a[j1+p07],a[j1+1+p07],a[j1+p06],a[j1+1+p06],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],rt,it);

#else

/*...First radix-5 transform:	*/
	/* Swap 4<->16, 8<->12 below to get permuted version of j1 + p00,04,08,12,16: */
	  t00=a[j1    ];		t01=a[j1+1    ];
	  t02=a[j1+p16]+a[j1+p04];	t03=a[j1+1+p16]+a[j1+1+p04];
	  t04=a[j1+p12]+a[j1+p08];	t05=a[j1+1+p12]+a[j1+1+p08];
	  t06=a[j1+p12]-a[j1+p08];	t07=a[j1+1+p12]-a[j1+1+p08];
	  t08=a[j1+p16]-a[j1+p04];	t09=a[j1+1+p16]-a[j1+1+p04];
	  rt = t02+t04;			it = t03+t05;
	  t00= t00+rt;			t01= t01+it;		/* y0	*/
	  rt = t00+cc1*rt;		it = t01+cc1*it;
	  t04= cc2*(t02-t04);		t05= cc2*(t03-t05);
	  t02= rt+t04;			t03= it+t05;
	  t04= rt-t04;			t05= it-t05;
	  rt = s2 *(t08-t06);		it = s2 *(t09-t07);
	  t06= ss1* t06;		t07= ss1* t07;
	  t08= ss2* t08;		t09= ss2* t09;
	  t06= rt+t06;			t07= it+t07;
	  t08= rt-t08;			t09= it-t09;
	  rt = t06;			it = t07;
	  t06= t02+it;			t07= t03-rt;		/* y4 - note swap with y3 below!	*/
	  t02= t02-it;			t03= t03+rt;		/* y1	*/
	  rt = t08;			it = t09;
	  t08= t04+it;			t09= t05-rt;		/* y3	*/
	  t04= t04-it;			t05= t05+rt;		/* y2	*/

/*...Second radix-5 transform:	*/
	/* Swap 01->15, 05->11, 09->07, 13->03, 17->19 below to get permuted version of j1 + p01,05,09,13,17: */
	  t10=a[j1+p15];		t11=a[j1+1+p15];
	  t12=a[j1+p11]+a[j1+p19];	t13=a[j1+1+p11]+a[j1+1+p19];
	  t14=a[j1+p07]+a[j1+p03];	t15=a[j1+1+p07]+a[j1+1+p03];
	  t16=a[j1+p07]-a[j1+p03];	t17=a[j1+1+p07]-a[j1+1+p03];
	  t18=a[j1+p11]-a[j1+p19];	t19=a[j1+1+p11]-a[j1+1+p19];
	  rt = t12+t14;			it = t13+t15;
	  t10= t10+rt;			t11= t11+it;		/* y0	*/
	  rt = t10+cc1*rt;		it = t11+cc1*it;
	  t14= cc2*(t12-t14);		t15= cc2*(t13-t15);
	  t12= rt+t14;			t13= it+t15;
	  t14= rt-t14;			t15= it-t15;
	  rt = s2 *(t18-t16);		it = s2 *(t19-t17);
	  t16= ss1* t16;		t17= ss1* t17;
	  t18= ss2* t18;		t19= ss2* t19;
	  t16= rt+t16;			t17= it+t17;
	  t18= rt-t18;			t19= it-t19;
	  rt = t16;			it = t17;
	  t16= t12+it;			t17= t13-rt;		/* y4 - note swap with y3 below!	*/
	  t12= t12-it;			t13= t13+rt;		/* y1	*/
	  rt = t18;			it = t19;
	  t18= t14+it;			t19= t15-rt;		/* y3	*/
	  t14= t14-it;			t15= t15+rt;		/* y2	*/

/*...Third radix-5 transform:	*/
	/* Swap 02<->10, 14<->18 below to get permuted version of j1 + p02,06,10,14,18: */
	  t20=a[j1+p10];		t21=a[j1+1+p10];
	  t22=a[j1+p06]+a[j1+p14];	t23=a[j1+1+p06]+a[j1+1+p14];
	  t24=a[j1+p02]+a[j1+p18];	t25=a[j1+1+p02]+a[j1+1+p18];
	  t26=a[j1+p02]-a[j1+p18];	t27=a[j1+1+p02]-a[j1+1+p18];
	  t28=a[j1+p06]-a[j1+p14];	t29=a[j1+1+p06]-a[j1+1+p14];
	  rt = t22+t24;			it = t23+t25;
	  t20= t20+rt;			t21= t21+it;		/* y0	*/
	  rt = t20+cc1*rt;		it = t21+cc1*it;
	  t24= cc2*(t22-t24);		t25= cc2*(t23-t25);
	  t22= rt+t24;			t23= it+t25;
	  t24= rt-t24;			t25= it-t25;
	  rt = s2 *(t28-t26);		it = s2 *(t29-t27);
	  t26= ss1* t26;		t27= ss1* t27;
	  t28= ss2* t28;		t29= ss2* t29;
	  t26= rt+t26;			t27= it+t27;
	  t28= rt-t28;			t29= it-t29;
	  rt = t26;			it = t27;
	  t26= t22+it;			t27= t23-rt;		/* y4 - note swap with y3 below!	*/
	  t22= t22-it;			t23= t23+rt;		/* y1	*/
	  rt = t28;			it = t29;
	  t28= t24+it;			t29= t25-rt;		/* y3	*/
	  t24= t24-it;			t25= t25+rt;		/* y2	*/

/*...Fourth radix-5 transform:	*/
	/* Swap 03->05, 07->01, 11->17, 15->13, 19->09 below to get permuted version of j1 + p03,07,11,15,19: */
	  t30=a[j1+p05];		t31=a[j1+1+p05];
	  t32=a[j1+p01]+a[j1+p09];	t33=a[j1+1+p01]+a[j1+1+p09];
	  t34=a[j1+p17]+a[j1+p13];	t35=a[j1+1+p17]+a[j1+1+p13];
	  t36=a[j1+p17]-a[j1+p13];	t37=a[j1+1+p17]-a[j1+1+p13];
	  t38=a[j1+p01]-a[j1+p09];	t39=a[j1+1+p01]-a[j1+1+p09];
	  rt = t32+t34;			it = t33+t35;
	  t30= t30+rt;			t31= t31+it;		/* y0	*/
	  rt = t30+cc1*rt;		it = t31+cc1*it;
	  t34= cc2*(t32-t34);		t35= cc2*(t33-t35);
	  t32= rt+t34;			t33= it+t35;
	  t34= rt-t34;			t35= it-t35;
	  rt = s2 *(t38-t36);		it = s2 *(t39-t37);
	  t36= ss1* t36;		t37= ss1* t37;
	  t38= ss2* t38;		t39= ss2* t39;
	  t36= rt+t36;			t37= it+t37;
	  t38= rt-t38;			t39= it-t39;
	  rt = t36;			it = t37;
	  t36= t32+it;			t37= t33-rt;		/* y4 - note swap with y3 below!	*/
	  t32= t32-it;			t33= t33+rt;		/* y1	*/
	  rt = t38;			it = t39;
	  t38= t34+it;			t39= t35-rt;		/* y3	*/
	  t34= t34-it;			t35= t35+rt;		/* y2	*/

/*...and now do five radix-4 transforms:	*/
	    rt =t20;	t20=t00-rt;	t00=t00+rt;
	    it =t21;	t21=t01-it;	t01=t01+it;

	    rt =t30;	t30=t10-rt;	t10=t10+rt;
	    it =t31;	t31=t11-it;	t11=t11+it;

	    a[j1    ]=t00+t10;	a[j1+1    ]=t01+t11;
	    a[j1+p01]=t00-t10;	a[j1+1+p01]=t01-t11;
	    a[j1+p03]=t20-t31;	a[j1+1+p03]=t21+t30;	/* mpy by I is inlined here...	*/
	    a[j1+p02]=t20+t31;	a[j1+1+p02]=t21-t30;

/*...Block 2: */
	    rt =t22;	t22=t02-rt;	t02=t02+rt;
	    it =t23;	t23=t03-it;	t03=t03+it;

	    rt =t32;	t32=t12-rt;	t12=t12+rt;
	    it =t33;	t33=t13-it;	t13=t13+it;

	    a[j1+p16]=t02+t12;	a[j1+1+p16]=t03+t13;
	    a[j1+p17]=t02-t12;	a[j1+1+p17]=t03-t13;
	    a[j1+p19]=t22-t33;	a[j1+1+p19]=t23+t32;	/* mpy by I is inlined here...	*/
	    a[j1+p18]=t22+t33;	a[j1+1+p18]=t23-t32;

/*...Block 3: */
	    rt =t24;	t24=t04-rt;	t04=t04+rt;
	    it =t25;	t25=t05-it;	t05=t05+it;

	    rt =t34;	t34=t14-rt;	t14=t14+rt;
	    it =t35;	t35=t15-it;	t15=t15+it;

	    a[j1+p14]=t04+t14;	a[j1+1+p14]=t05+t15;
	    a[j1+p15]=t04-t14;	a[j1+1+p15]=t05-t15;
	    a[j1+p12]=t24-t35;	a[j1+1+p12]=t25+t34;	/* mpy by I is inlined here...	*/
	    a[j1+p13]=t24+t35;	a[j1+1+p13]=t25-t34;

/*...Block 4: */
	    rt =t26;	t26=t06-rt;	t06=t06+rt;
	    it =t27;	t27=t07-it;	t07=t07+it;

	    rt =t36;	t36=t16-rt;	t16=t16+rt;
	    it =t37;	t37=t17-it;	t17=t17+it;

	    a[j1+p07]=t06+t16;	a[j1+1+p07]=t07+t17;
	    a[j1+p06]=t06-t16;	a[j1+1+p06]=t07-t17;
	    a[j1+p05]=t26-t37;	a[j1+1+p05]=t27+t36;	/* mpy by I is inlined here...	*/
	    a[j1+p04]=t26+t37;	a[j1+1+p04]=t27-t36;

/*...Block 5: */
	    rt =t28;	t28=t08-rt;	t08=t08+rt;
	    it =t29;	t29=t09-it;	t09=t09+it;

	    rt =t38;	t38=t18-rt;	t18=t18+rt;
	    it =t39;	t39=t19-it;	t19=t19+it;

	    a[j1+p09]=t08+t18;	a[j1+1+p09]=t09+t19;
	    a[j1+p08]=t08-t18;	a[j1+1+p08]=t09-t19;
	    a[j1+p10]=t28-t39;	a[j1+1+p10]=t29+t38;	/* mpy by I is inlined here...	*/
	    a[j1+p11]=t28+t39;	a[j1+1+p11]=t29-t38;
						/* Totals: (4*34 + 5*16) = 216 FADD, 4*10 = 40 FMUL	*/
#endif
	}
}

/* unpermuted-index output order is as follows:
ActualNeed                Actual                             Need
---- ------ -------------------------------    -------------------------------
   0 (   0)  107.0000000000    87.0000000000   107.0000000000    87.0000000000
   1 (   1)  -29.0000000000   -11.0000000000   -29.0000000000   -11.0000000000
   2 (   3)  -14.0000000000   -12.0000000000   -12.0000000000   -16.0000000000
   3 (   2)  -12.0000000000   -16.0000000000   -14.0000000000   -12.0000000000
*
   4 (  16)    3.8975770799    -1.7742323052   -11.8647898507     6.5977392870
   5 (  17)   -6.1639601065    -0.2162986542    -2.6779551960    15.1374471603
   6 (  19)   -1.5023846915     1.7166548059    10.4000280840   -11.4574634246
   7 (  18)    9.5729937832   -25.6321811408     2.3384908976    -2.8438016835
*
   8 (  14)   14.5070279402   -12.6894179137    -2.8854468251   -23.7084688127
   9 (  15)    2.6493788476    -3.6177691085   -12.7430959177    10.3074519025
  10 (  12)    8.1391134275    -9.5889261410    10.0412264600    19.7348241747
  11 (  13)  -20.9443792061     5.5722500297     5.2361752736     4.4621918240
*
  12 (  07)    2.3384908976    -2.8438016835     8.1391134275    -9.5889261410
  13 (  06)   10.4000280840   -11.4574634246   -20.9443792061     5.5722500297
  14 (  05)   -2.6779551960    15.1374471603    14.5070279402   -12.6894179137
  15 (  04)  -11.8647898507     6.5977392870     2.6493788476    -3.6177691085
*
  16 (  09)  -12.7430959177    10.3074519025     3.8975770799    -1.7742323052
  17 (  08)   -2.8854468251   -23.7084688127    -6.1639601065    -0.2162986542
  18 (  10)   10.0412264600    19.7348241747     9.5729937832   -25.6321811408
  19 (  11)    5.2361752736     4.4621918240    -1.5023846915     1.7166548059
*/

/**************/

int radix20_ditN_cy_dif1(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[], double baseinv[], int iter, double *fracmax, int p)
{
/*
!...Acronym: DWT = Discrete Weighted Transform, DIT = Decimation In Time, DIF = Decimation In Frequency
!
!...Performs a final radix-20 complex DIT pass, an inverse DWT weighting, a carry propagation,
!   a forward DWT weighting, and an initial radix-20 complex DIF pass on the data in the length-N real vector A.
!
!   Data enter and are returned in the A-array.
!
!   See the documentation in mers_mod_square and radix16_dif_pass for further details on the array
!   storage scheme, and radix8_ditN_cy_dif1 for details on the reduced-length weights array scheme.
*/
	int n20, bjmodn00,bjmodn01,bjmodn02,bjmodn03,bjmodn04,bjmodn05,bjmodn06,bjmodn07,bjmodn08,bjmodn09
		,bjmodn10,bjmodn11,bjmodn12,bjmodn13,bjmodn14,bjmodn15,bjmodn16,bjmodn17,bjmodn18,bjmodn19
		,i,j,j1,jstart,jhi,iroot,root_incr,k,khi,l,outer;
	static int psave=0,bw,sw,bjmodnini,nrtm1,p01,p02,p03,p04,p05,p06,p07,p08,p09,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19;
	const  double one_half[3] = {1.0, 0.5, 0.25};	/* Needed for small-weights-tables scheme */
	static double cc1 = -1.25000000000000000000,	/* [cos(u)+cos(2u)]/2-1 = -5/4; u = 2*pi/5 */
		      cc2 =  0.55901699437494742409,	/* [cos(u)-cos(2u)]/2 */
		      s2  =  0.95105651629515357211,	/*  sin(u) */
		      ss1 =  1.53884176858762670130,	/* [sin(u)+sin(2u)] */
		      ss2 =  0.36327126400268044292;	/* [sin(u)-sin(2u)] */
	static double radix_inv, n2inv;
	double rt,it
		,t00,t01,t02,t03,t04,t05,t06,t07,t08,t09
		,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19
		,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29
		,t30,t31,t32,t33,t34,t35,t36,t37,t38,t39
	,a1p00r,a1p01r,a1p02r,a1p03r,a1p04r,a1p05r,a1p06r,a1p07r,a1p08r,a1p09r,a1p10r,a1p11r,a1p12r,a1p13r,a1p14r,a1p15r,a1p16r,a1p17r,a1p18r,a1p19r
	,a1p00i,a1p01i,a1p02i,a1p03i,a1p04i,a1p05i,a1p06i,a1p07i,a1p08i,a1p09i,a1p10i,a1p11i,a1p12i,a1p13i,a1p14i,a1p15i,a1p16i,a1p17i,a1p18i,a1p19i
		,cy00,cy01,cy02,cy03,cy04,cy05,cy06,cy07,cy08,cy09,cy10,cy11,cy12,cy13,cy14,cy15,cy16,cy17,cy18,cy19,temp,frac,scale;
	double maxerr = 0.0;
#if PFETCH
	double *add0, *add;
#endif
	int err;
	FILE *f1,*f2;
	static int first_entry=TRUE;

/*...stuff for the reduced-length DWT weights array is here:	*/
	int n_div_nwt;
	int col,co2,co3,m,m2,n_minus_sil,n_minus_silp1,sinwt,sinwtm1;
	double wt,wtinv,wtl,wtlp1,wtn,wtnm1,wtA,wtB,wtC;

/*...change n20 and n_div_wt to non-static to work around a gcc compiler bug. */
	n20   = n/20;
	n_div_nwt = n20 >> nwt_bits;

	if((n_div_nwt << nwt_bits) != n20)
	{
	  if(INTERACT)printf("FATAL: iter = %10d; NWT_BITS does not divide N/20 in radix20_ditN_cy_dif1.\n",iter);
	  f1 = fopen(   OFILE,"a");
	  f2 = fopen(STATFILE,"a");
	  fprintf(f1,"FATAL: iter = %10d; NWT_BITS does not divide N/20 in radix20_ditN_cy_dif1.\n",iter);
	  fprintf(f2,"FATAL: iter = %10d; NWT_BITS does not divide N/20 in radix20_ditN_cy_dif1.\n",iter);
	  fclose(f1);
	  fclose(f2);
	  err=1;
	  return(err);
	}

	if(p != psave)
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  psave = p;
	  first_entry=FALSE;
	  radix_inv = (double)((long double)1.0L/(long double)20.0L);
	  n2inv     = (double)((long double)1.0L/(long double)(n/2));

	  bw    = p%n;		/* Number of bigwords in the Crandall/Fagin mixed-radix representation = (Mersenne exponent) mod (vector length).	*/
	  sw    = n - bw;	/* Number of smallwords.	*/

/*   constant index offsets for array load/stores are here.	*/

	  p01 = n20;
	  p02 = p01 + p01;
	  p03 = p02 + p01;
	  p04 = p03 + p01;
	  p05 = p04 + p01;
	  p06 = p05 + p01;
	  p07 = p06 + p01;
	  p08 = p07 + p01;
	  p09 = p08 + p01;
	  p10 = p09 + p01;
	  p11 = p10 + p01;
	  p12 = p11 + p01;
	  p13 = p12 + p01;
	  p14 = p13 + p01;
	  p15 = p14 + p01;
	  p16 = p15 + p01;
	  p17 = p16 + p01;
	  p18 = p17 + p01;
	  p19 = p18 + p01;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p05 = p05 + ( (p05 >> DAT_BITS) << PAD_BITS );
	  p06 = p06 + ( (p06 >> DAT_BITS) << PAD_BITS );
	  p07 = p07 + ( (p07 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p09 = p09 + ( (p09 >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p11 = p11 + ( (p11 >> DAT_BITS) << PAD_BITS );
	  p12 = p12 + ( (p12 >> DAT_BITS) << PAD_BITS );
	  p13 = p13 + ( (p13 >> DAT_BITS) << PAD_BITS );
	  p14 = p14 + ( (p14 >> DAT_BITS) << PAD_BITS );
	  p15 = p15 + ( (p15 >> DAT_BITS) << PAD_BITS );
	  p16 = p16 + ( (p16 >> DAT_BITS) << PAD_BITS );
	  p17 = p17 + ( (p17 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	  p19 = p19 + ( (p19 >> DAT_BITS) << PAD_BITS );

	  bjmodnini=0;
	  for(j=0; j < n20; j++)
	  {
	    bjmodnini -= sw; bjmodnini = bjmodnini + ( (-(int)((uint32)bjmodnini >> 31)) & n);
	  }
	}

/*...The radix-20 final DIT pass is here.	*/

	cy00=-2;		/* init carries	*/
	cy01= 0;
	cy02= 0;
	cy03= 0;
	cy04= 0;
	cy05= 0;
	cy06= 0;
	cy07= 0;
	cy08= 0;
	cy09= 0;
	cy10= 0;
	cy11= 0;
	cy12= 0;
	cy13= 0;
	cy14= 0;
	cy15= 0;
	cy16= 0;
	cy17= 0;
	cy18= 0;
	cy19= 0;

	*fracmax=0;	/* init max. fractional error	*/

	iroot = 0;	/* init sincos array index	*/
	root_incr = 1;	/* init sincos array index increment (set = 1 for normal carry pass, = 0 for wrapper pass)	*/
	scale = n2inv;	/* init inverse-weight scale factor  (set = 2/n for normal carry pass, = 1 for wrapper pass)	*/

	jstart = 0;
	jhi = jstart+nwt-1;
	khi = n_div_nwt;

for(outer=0; outer <= 1; outer++)
{
	i = 1;		/* Pointer to the BASE and BASEINV arrays. lowest-order digit is always a bigword (i = 1).	*/

	bjmodn00= 0;
	bjmodn01= bjmodnini;
	bjmodn02= bjmodn01+bjmodnini-n; bjmodn02= bjmodn02+ ( (-(int)((uint32)bjmodn02>> 31)) & n);
	bjmodn03= bjmodn02+bjmodnini-n; bjmodn03= bjmodn03+ ( (-(int)((uint32)bjmodn03>> 31)) & n);
	bjmodn04= bjmodn03+bjmodnini-n; bjmodn04= bjmodn04+ ( (-(int)((uint32)bjmodn04>> 31)) & n);
	bjmodn05= bjmodn04+bjmodnini-n; bjmodn05= bjmodn05+ ( (-(int)((uint32)bjmodn05>> 31)) & n);
	bjmodn06= bjmodn05+bjmodnini-n; bjmodn06= bjmodn06+ ( (-(int)((uint32)bjmodn06>> 31)) & n);
	bjmodn07= bjmodn06+bjmodnini-n; bjmodn07= bjmodn07+ ( (-(int)((uint32)bjmodn07>> 31)) & n);
	bjmodn08= bjmodn07+bjmodnini-n; bjmodn08= bjmodn08+ ( (-(int)((uint32)bjmodn08>> 31)) & n);
	bjmodn09= bjmodn08+bjmodnini-n; bjmodn09= bjmodn09+ ( (-(int)((uint32)bjmodn09>> 31)) & n);
	bjmodn10= bjmodn09+bjmodnini-n; bjmodn10= bjmodn10+ ( (-(int)((uint32)bjmodn10>> 31)) & n);
	bjmodn11= bjmodn10+bjmodnini-n; bjmodn11= bjmodn11+ ( (-(int)((uint32)bjmodn11>> 31)) & n);
	bjmodn12= bjmodn11+bjmodnini-n; bjmodn12= bjmodn12+ ( (-(int)((uint32)bjmodn12>> 31)) & n);
	bjmodn13= bjmodn12+bjmodnini-n; bjmodn13= bjmodn13+ ( (-(int)((uint32)bjmodn13>> 31)) & n);
	bjmodn14= bjmodn13+bjmodnini-n; bjmodn14= bjmodn14+ ( (-(int)((uint32)bjmodn14>> 31)) & n);
	bjmodn15= bjmodn14+bjmodnini-n; bjmodn15= bjmodn15+ ( (-(int)((uint32)bjmodn15>> 31)) & n);
	bjmodn16= bjmodn15+bjmodnini-n; bjmodn16= bjmodn16+ ( (-(int)((uint32)bjmodn16>> 31)) & n);
	bjmodn17= bjmodn16+bjmodnini-n; bjmodn17= bjmodn17+ ( (-(int)((uint32)bjmodn17>> 31)) & n);
	bjmodn18= bjmodn17+bjmodnini-n; bjmodn18= bjmodn18+ ( (-(int)((uint32)bjmodn18>> 31)) & n);
	bjmodn19= bjmodn18+bjmodnini-n; bjmodn19= bjmodn19+ ( (-(int)((uint32)bjmodn19>> 31)) & n);

	col=0;
	co2=(n >> nwt_bits)-1+20;
	co3=co2-20;		/* At the start of each new j-loop, co3=co2-radix(1)	*/

	for(k=1; k <= khi; k++)	/* Do n/(radix(1)*nwt) outer loop executions...	*/
	{
	  for(j=jstart; j<jhi; j += 2)	/* Each inner loop execution processes (radix(1)*nwt) array data.	*/
	  {
	    j1 = j + ( (j >> DAT_BITS) << PAD_BITS );	/* padded-array fetch index is here	*/
/*
!...gather the needed data (20 64-bit complex, i.e. 40 64-bit reals) and do a radix-20 DIT transform...
*/
#if 1
/*...gather the needed data (20 64-bit complex, i.e. 40 64-bit reals) and do 5 radix-4 transforms...*/
                     /*          inputs           */ /*                                      outputs                                      */
	RADIX_04_DIT(a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p03],a[j1+1+p03],a[j1+p02],a[j1+1+p02],t00,t01,t10,t11,t20,t21,t30,t31,rt,it);
	RADIX_04_DIT(a[j1+p07],a[j1+1+p07],a[j1+p06],a[j1+1+p06],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],t02,t03,t12,t13,t22,t23,t32,t33,rt,it);
	RADIX_04_DIT(a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],a[j1+p10],a[j1+1+p10],a[j1+p11],a[j1+1+p11],t04,t05,t14,t15,t24,t25,t34,t35,rt,it);
	RADIX_04_DIT(a[j1+p14],a[j1+1+p14],a[j1+p15],a[j1+1+p15],a[j1+p12],a[j1+1+p12],a[j1+p13],a[j1+1+p13],t06,t07,t16,t17,t26,t27,t36,t37,rt,it);
	RADIX_04_DIT(a[j1+p16],a[j1+1+p16],a[j1+p17],a[j1+1+p17],a[j1+p19],a[j1+1+p19],a[j1+p18],a[j1+1+p18],t08,t09,t18,t19,t28,t29,t38,t39,rt,it);

/*...and now do 4 radix-5 transforms...*/
                     /*                                                inputs                                                   */ /*                 outputs                   */
	RADIX_05_DFT(t00,t01,t02,t03,t04,t05,t06,t07,t08,t09,a1p00r,a1p00i,a1p16r,a1p16i,a1p12r,a1p12i,a1p08r,a1p08i,a1p04r,a1p04i,rt,it);
	RADIX_05_DFT(t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,a1p15r,a1p15i,a1p11r,a1p11i,a1p07r,a1p07i,a1p03r,a1p03i,a1p19r,a1p19i,rt,it);
	RADIX_05_DFT(t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,a1p10r,a1p10i,a1p06r,a1p06i,a1p02r,a1p02i,a1p18r,a1p18i,a1p14r,a1p14i,rt,it);
	RADIX_05_DFT(t30,t31,t32,t33,t34,t35,t36,t37,t38,t39,a1p05r,a1p05i,a1p01r,a1p01i,a1p17r,a1p17i,a1p13r,a1p13i,a1p09r,a1p09i,rt,it);
#else

/*...Block 1:	*/
	t00=a[j1    ];	t01=a[j1+1    ];
	rt =a[j1+p01];	it =a[j1+1+p01];
	t10=t00-rt;  	t00=t00+rt;
	t11=t01-it;	t01=t01+it;

	t20=a[j1+p03];	t21=a[j1+1+p03];
	rt =a[j1+p02];	it =a[j1+1+p02];
	t30=t20-rt;  	t20=t20+rt;
	t31=t21-it;  	t21=t21+it;

	rt =t20;	t20=t00-rt;	t00=t00+rt;
	it =t21;	t21=t01-it;	t01=t01+it;

	rt =t30;	t30=t10-t31;	t10=t10+t31;
			t31=t11+rt;	t11=t11-rt;

/*...Block 2:	*/
	t02=a[j1+p07];	t03=a[j1+1+p07];
	rt =a[j1+p06];	it =a[j1+1+p06];
	t12=t02-rt;  	t02=t02+rt;
	t13=t03-it;	t03=t03+it;

	t22=a[j1+p05];	t23=a[j1+1+p05];
	rt =a[j1+p04];	it =a[j1+1+p04];
	t32=t22-rt;  	t22=t22+rt;
	t33=t23-it;  	t23=t23+it;

	rt =t22;	t22=t02-rt;	t02=t02+rt;
	it =t23;	t23=t03-it;	t03=t03+it;

	rt =t32;	t32=t12-t33;	t12=t12+t33;
			t33=t13+rt;	t13=t13-rt;

/*...Block 3:	*/
	t04=a[j1+p09];	t05=a[j1+1+p09];
	rt =a[j1+p08];	it =a[j1+1+p08];
	t14=t04-rt;  	t04=t04+rt;
	t15=t05-it;	t05=t05+it;

	t24=a[j1+p10];	t25=a[j1+1+p10];
	rt =a[j1+p11];	it =a[j1+1+p11];
	t34=t24-rt;  	t24=t24+rt;
	t35=t25-it;  	t25=t25+it;

	rt =t24;	t24=t04-rt;	t04=t04+rt;
	it =t25;	t25=t05-it;	t05=t05+it;

	rt =t34;	t34=t14-t35;	t14=t14+t35;
			t35=t15+rt;	t15=t15-rt;

/*...Block 4:	*/
	t06=a[j1+p14];	t07=a[j1+1+p14];
	rt =a[j1+p15];	it =a[j1+1+p15];
	t16=t06-rt;  	t06=t06+rt;
	t17=t07-it;	t07=t07+it;

	t26=a[j1+p12];	t27=a[j1+1+p12];
	rt =a[j1+p13];	it =a[j1+1+p13];
	t36=t26-rt;  	t26=t26+rt;
	t37=t27-it;  	t27=t27+it;

	rt =t26;	t26=t06-rt;	t06=t06+rt;
	it =t27;	t27=t07-it;	t07=t07+it;

	rt =t36;	t36=t16-t37;	t16=t16+t37;
			t37=t17+rt;	t17=t17-rt;

/*...Block 5:	*/
	t08=a[j1+p16];	t09=a[j1+1+p16];
	rt =a[j1+p17];	it =a[j1+1+p17];
	t18=t08-rt;  	t08=t08+rt;
	t19=t09-it;	t09=t09+it;

	t28=a[j1+p19];	t29=a[j1+1+p19];
	rt =a[j1+p18];	it =a[j1+1+p18];
	t38=t28-rt;  	t28=t28+rt;
	t39=t29-it;  	t29=t29+it;

	rt =t28;	t28=t08-rt;	t08=t08+rt;
	it =t29;	t29=t09-it;	t09=t09+it;

	rt =t38;	t38=t18-t39;	t18=t18+t39;
			t39=t19+rt;	t19=t19-rt;

/*       ...and now do four radix-5 transforms.	*/
/*...First radix-5 transform:	*/
	    rt  =t08;			it  =t09;
	    t08 =t02 -rt;		t09 =t03 -it;
	    t02 =t02 +rt;		t03 =t03 +it;
	    rt  =t06;			it  =t07;
	    t06 =t04 -rt;		t07 =t05 -it;
	    t04 =t04 +rt;		t05 =t05 +it;
	    rt  = t02+t04;		it  = t03+t05;
	    t00 = t00+rt;		t01 = t01+it;
	    rt  = t00+cc1*rt;		it  = t01+cc1*it;
	    t04 = cc2*(t02-t04);	t05 = cc2*(t03-t05);
	    t02 = rt+t04;		t03 = it+t05;
	    t04 = rt-t04;		t05 = it-t05;
	    rt  = s2 *(t06-t08);	it  = s2 *(t07-t09);
	    t06 = ss1* t06;		t07 = ss1* t07;
	    t08 = ss2* t08;		t09 = ss2* t09;
	    t06 = rt-t06;		t07 = it-t07;
	    t08 = rt+t08;		t09 = it+t09;
	    a1p00r=t00;			a1p00i=t01;
	    a1p04r=t02-t07;		a1p04i=t03+t06;
	    a1p08r=t04-t09;		a1p08i=t05+t08;
	    a1p12r=t04+t09;		a1p12i=t05-t08;
	    a1p16r=t02+t07;		a1p16i=t03-t06;

/*...Second radix-5 transform:	*/
	    rt  =t18;			it  =t19;
	    t18 =t12 -rt;		t19 =t13 -it;
	    t12 =t12 +rt;		t13 =t13 +it;
	    rt  =t16;			it  =t17;
	    t16 =t14 -rt;		t17 =t15 -it;
	    t14 =t14 +rt;		t15 =t15 +it;
	    rt  = t12+t14;		it  = t13+t15;
	    t10 = t10+rt;		t11 = t11+it;
	    rt  = t10+cc1*rt;		it  = t11+cc1*it;
	    t14 = cc2*(t12-t14);	t15 = cc2*(t13-t15);
	    t12 = rt+t14;		t13 = it+t15;
	    t14 = rt-t14;		t15 = it-t15;
	    rt  = s2 *(t16-t18);	it  = s2 *(t17-t19);
	    t16 = ss1* t16;		t17 = ss1* t17;
	    t18 = ss2* t18;		t19 = ss2* t19;
	    t16 = rt-t16;		t17 = it-t17;
	    t18 = rt+t18;		t19 = it+t19;
	    a1p15r=t10;			a1p15i=t11;
	    a1p19r=t12-t17;		a1p19i=t13+t16;
	    a1p03r=t14-t19;		a1p03i=t15+t18;
	    a1p07r=t14+t19;		a1p07i=t15-t18;
	    a1p11r=t12+t17;		a1p11i=t13-t16;

/*...Third radix-5 transform:	*/
	    rt  =t28;			it  =t29;
	    t28 =t22 -rt;		t29 =t23 -it;
	    t22 =t22 +rt;		t23 =t23 +it;
	    rt  =t26;			it  =t27;
	    t26 =t24 -rt;		t27 =t25 -it;
	    t24 =t24 +rt;		t25 =t25 +it;
	    rt  = t22+t24;		it  = t23+t25;
	    t20 = t20+rt;		t21 = t21+it;
	    rt  = t20+cc1*rt;		it  = t21+cc1*it;
	    t24 = cc2*(t22-t24);	t25 = cc2*(t23-t25);
	    t22 = rt+t24;		t23 = it+t25;
	    t24 = rt-t24;		t25 = it-t25;
	    rt  = s2 *(t26-t28);	it  = s2 *(t27-t29);
	    t26 = ss1* t26;		t27 = ss1* t27;
	    t28 = ss2* t28;		t29 = ss2* t29;
	    t26 = rt-t26;		t27 = it-t27;
	    t28 = rt+t28;		t29 = it+t29;
	    a1p10r=t20;			a1p10i=t21;
	    a1p14r=t22-t27;		a1p14i=t23+t26;
	    a1p18r=t24-t29;		a1p18i=t25+t28;
	    a1p02r=t24+t29;		a1p02i=t25-t28;
	    a1p06r=t22+t27;		a1p06i=t23-t26;

/*...Fourth radix-5 transform:	*/
	    rt  =t38;			it  =t39;
	    t38 =t32 -rt;		t39 =t33 -it;
	    t32 =t32 +rt;		t33 =t33 +it;
	    rt  =t36;			it  =t37;
	    t36 =t34 -rt;		t37 =t35 -it;
	    t34 =t34 +rt;		t35 =t35 +it;
	    rt  = t32+t34;		it  = t33+t35;
	    t30 = t30+rt;		t31 = t31+it;
	    rt  = t30+cc1*rt;		it  = t31+cc1*it;
	    t34 = cc2*(t32-t34);	t35 = cc2*(t33-t35);
	    t32 = rt+t34;		t33 = it+t35;
	    t34 = rt-t34;		t35 = it-t35;
	    rt  = s2 *(t36-t38);	it  = s2 *(t37-t39);
	    t36 = ss1* t36;		t37 = ss1* t37;
	    t38 = ss2* t38;		t39 = ss2* t39;
	    t36 = rt-t36;		t37 = it-t37;
	    t38 = rt+t38;		t39 = it+t39;
	    a1p05r=t30;			a1p05i=t31;
	    a1p09r=t32-t37;		a1p09i=t33+t36;
	    a1p13r=t34-t39;		a1p13i=t35+t38;
	    a1p17r=t34+t39;		a1p17i=t35-t38;
	    a1p01r=t32+t37;		a1p01i=t33-t36;
#endif

/*...Now do the carries. Since the outputs would
    normally be getting dispatched to 20 separate blocks of the A-array, we need 20 separate carries.	*/

	    l= j & (nwt-1);
	    n_minus_sil   = n-si[l  ];
	    n_minus_silp1 = n-si[l+1];
	    sinwt   = si[nwt-l  ];
	    sinwtm1 = si[nwt-l-1];

	    wtl     =wt0[    l  ];
	    wtn     =wt0[nwt-l  ]*scale;	/* Include 1/(n/2) scale factor of inverse transform here...	*/
	    wtlp1   =wt0[    l+1];
	    wtnm1   =wt0[nwt-l-1]*scale;	/* ...and here.	*/

/*...set0 is slightly different from others:	*/
	   cmplx_carry_norm_errcheck0(a1p00r,a1p00i,cy00,bjmodn00   );
	    cmplx_carry_norm_errcheck(a1p01r,a1p01i,cy01,bjmodn01,1 );
	    cmplx_carry_norm_errcheck(a1p02r,a1p02i,cy02,bjmodn02,2 );
	    cmplx_carry_norm_errcheck(a1p03r,a1p03i,cy03,bjmodn03,3 );
	    cmplx_carry_norm_errcheck(a1p04r,a1p04i,cy04,bjmodn04,4 );
	    cmplx_carry_norm_errcheck(a1p05r,a1p05i,cy05,bjmodn05,5 );
	    cmplx_carry_norm_errcheck(a1p06r,a1p06i,cy06,bjmodn06,6 );
	    cmplx_carry_norm_errcheck(a1p07r,a1p07i,cy07,bjmodn07,7 );
	    cmplx_carry_norm_errcheck(a1p08r,a1p08i,cy08,bjmodn08,8 );
	    cmplx_carry_norm_errcheck(a1p09r,a1p09i,cy09,bjmodn09,9 );
	    cmplx_carry_norm_errcheck(a1p10r,a1p10i,cy10,bjmodn10,10);
	    cmplx_carry_norm_errcheck(a1p11r,a1p11i,cy11,bjmodn11,11);
	    cmplx_carry_norm_errcheck(a1p12r,a1p12i,cy12,bjmodn12,12);
	    cmplx_carry_norm_errcheck(a1p13r,a1p13i,cy13,bjmodn13,13);
	    cmplx_carry_norm_errcheck(a1p14r,a1p14i,cy14,bjmodn14,14);
	    cmplx_carry_norm_errcheck(a1p15r,a1p15i,cy15,bjmodn15,15);
	    cmplx_carry_norm_errcheck(a1p16r,a1p16i,cy16,bjmodn16,16);
	    cmplx_carry_norm_errcheck(a1p17r,a1p17i,cy17,bjmodn17,17);
	    cmplx_carry_norm_errcheck(a1p18r,a1p18i,cy18,bjmodn18,18);
	    cmplx_carry_norm_errcheck(a1p19r,a1p19i,cy19,bjmodn19,19);

	    i =((uint32)(sw - bjmodn00) >> 31);	/* get ready for the next set...	*/
	    co2=co3;	/* For all data but the first set in each j-block, co2=co3. Thus, after the first block of data is done
			   and only then: for all subsequent blocks it's superfluous), this assignment decrements co2 by radix(1).	*/

/*...The radix-20 DIF pass is here:	*/
#if PFETCH
add0 = &a[j1];
prefetch_p_doubles(add0);
#endif

#if 1
/*...gather the needed data (20 64-bit complex, i.e. 40 64-bit reals) and do 4 radix-5 transforms...*/
                     /*                                                inputs                                                   */ /*                 outputs                   */
#if PFETCH
	RADIX_05_DFT_PFETCH(a1p00r,a1p00i,a1p16r,a1p16i,a1p12r,a1p12i,a1p08r,a1p08i,a1p04r,a1p04i,t00,t01,t02,t03,t04,t05,t06,t07,t08,t09,rt,it,p01,p02);
	RADIX_05_DFT_PFETCH(a1p15r,a1p15i,a1p11r,a1p11i,a1p07r,a1p07i,a1p03r,a1p03i,a1p19r,a1p19i,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,rt,it,p03,p04);
	RADIX_05_DFT_PFETCH(a1p10r,a1p10i,a1p06r,a1p06i,a1p02r,a1p02i,a1p18r,a1p18i,a1p14r,a1p14i,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,rt,it,p05,p06);
	RADIX_05_DFT_PFETCH(a1p05r,a1p05i,a1p01r,a1p01i,a1p17r,a1p17i,a1p13r,a1p13i,a1p09r,a1p09i,t30,t31,t32,t33,t34,t35,t36,t37,t38,t39,rt,it,p07,p08);
#else
	RADIX_05_DFT       (a1p00r,a1p00i,a1p16r,a1p16i,a1p12r,a1p12i,a1p08r,a1p08i,a1p04r,a1p04i,t00,t01,t02,t03,t04,t05,t06,t07,t08,t09,rt,it);
	RADIX_05_DFT       (a1p15r,a1p15i,a1p11r,a1p11i,a1p07r,a1p07i,a1p03r,a1p03i,a1p19r,a1p19i,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,rt,it);
	RADIX_05_DFT       (a1p10r,a1p10i,a1p06r,a1p06i,a1p02r,a1p02i,a1p18r,a1p18i,a1p14r,a1p14i,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,rt,it);
	RADIX_05_DFT       (a1p05r,a1p05i,a1p01r,a1p01i,a1p17r,a1p17i,a1p13r,a1p13i,a1p09r,a1p09i,t30,t31,t32,t33,t34,t35,t36,t37,t38,t39,rt,it);
#endif
/*...and now do 5 radix-4 transforms...*/
                     /*          inputs           */ /*                                      outputs                                      */
#if PFETCH
	add = add0+p09;
	prefetch_p_doubles(add);

	RADIX_04_DIF_PFETCH(t00,t01,t10,t11,t20,t21,t30,t31,a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p03],a[j1+1+p03],a[j1+p02],a[j1+1+p02],rt,it,p10,p11);
	RADIX_04_DIF_PFETCH(t02,t03,t12,t13,t22,t23,t32,t33,a[j1+p16],a[j1+1+p16],a[j1+p17],a[j1+1+p17],a[j1+p19],a[j1+1+p19],a[j1+p18],a[j1+1+p18],rt,it,p12,p13);
	RADIX_04_DIF_PFETCH(t04,t05,t14,t15,t24,t25,t34,t35,a[j1+p14],a[j1+1+p14],a[j1+p15],a[j1+1+p15],a[j1+p12],a[j1+1+p12],a[j1+p13],a[j1+1+p13],rt,it,p14,p15);
	RADIX_04_DIF_PFETCH(t06,t07,t16,t17,t26,t27,t36,t37,a[j1+p07],a[j1+1+p07],a[j1+p06],a[j1+1+p06],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],rt,it,p16,p17);
	RADIX_04_DIF_PFETCH(t08,t09,t18,t19,t28,t29,t38,t39,a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],a[j1+p10],a[j1+1+p10],a[j1+p11],a[j1+1+p11],rt,it,p18,p19);
#else
	RADIX_04_DIF       (t00,t01,t10,t11,t20,t21,t30,t31,a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p03],a[j1+1+p03],a[j1+p02],a[j1+1+p02],rt,it);
	RADIX_04_DIF       (t02,t03,t12,t13,t22,t23,t32,t33,a[j1+p16],a[j1+1+p16],a[j1+p17],a[j1+1+p17],a[j1+p19],a[j1+1+p19],a[j1+p18],a[j1+1+p18],rt,it);
	RADIX_04_DIF       (t04,t05,t14,t15,t24,t25,t34,t35,a[j1+p14],a[j1+1+p14],a[j1+p15],a[j1+1+p15],a[j1+p12],a[j1+1+p12],a[j1+p13],a[j1+1+p13],rt,it);
	RADIX_04_DIF       (t06,t07,t16,t17,t26,t27,t36,t37,a[j1+p07],a[j1+1+p07],a[j1+p06],a[j1+1+p06],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],rt,it);
	RADIX_04_DIF       (t08,t09,t18,t19,t28,t29,t38,t39,a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],a[j1+p10],a[j1+1+p10],a[j1+p11],a[j1+1+p11],rt,it);
#endif

#else

/*...First radix-5 transform:	*/
	/* Swap 4<->16, 8<->12 below to get permuted version of j1 + p00,04,08,12,16: */
	  t00=a1p00r;			t01=a1p00i;
	  t02=a1p16r+a1p04r;		t03=a1p16i+a1p04i;
	  t04=a1p12r+a1p08r;		t05=a1p12i+a1p08i;
	  t06=a1p12r-a1p08r;		t07=a1p12i-a1p08i;
	  t08=a1p16r-a1p04r;		t09=a1p16i-a1p04i;
	  rt = t02+t04;			it = t03+t05;
	  t00= t00+rt;			t01= t01+it;		/* y0	*/
	  rt = t00+cc1*rt;		it = t01+cc1*it;
	  t04= cc2*(t02-t04);		t05= cc2*(t03-t05);
	  t02= rt+t04;			t03= it+t05;
	  t04= rt-t04;			t05= it-t05;
#if PFETCH
add = add0+p01;
prefetch_p_doubles(add);
#endif
	  rt = s2 *(t08-t06);		it = s2 *(t09-t07);
	  t06= ss1* t06;		t07= ss1* t07;
	  t08= ss2* t08;		t09= ss2* t09;
	  t06= rt+t06;			t07= it+t07;
	  t08= rt-t08;			t09= it-t09;
	  rt = t06;			it = t07;
	  t06= t02+it;			t07= t03-rt;		/* y4 - note swap with y3 below!	*/
	  t02= t02-it;			t03= t03+rt;		/* y1	*/
	  rt = t08;			it = t09;
	  t08= t04+it;			t09= t05-rt;		/* y3	*/
	  t04= t04-it;			t05= t05+rt;		/* y2	*/
#if PFETCH
add = add0+p02;
prefetch_p_doubles(add);
#endif

/*...Second radix-5 transform:	*/
	/* Swap 01->15, 05->11, 09->07, 13->03, 17->19 below to get permuted version of j1 + p01,05,09,13,17: */
	  t10=a1p15r;			t11=a1p15i;
	  t12=a1p11r+a1p19r;		t13=a1p11i+a1p19i;
	  t14=a1p07r+a1p03r;		t15=a1p07i+a1p03i;
	  t16=a1p07r-a1p03r;		t17=a1p07i-a1p03i;
	  t18=a1p11r-a1p19r;		t19=a1p11i-a1p19i;
	  rt = t12+t14;			it = t13+t15;
	  t10= t10+rt;			t11= t11+it;		/* y0	*/
	  rt = t10+cc1*rt;		it = t11+cc1*it;
	  t14= cc2*(t12-t14);		t15= cc2*(t13-t15);
	  t12= rt+t14;			t13= it+t15;
	  t14= rt-t14;			t15= it-t15;
#if PFETCH
add = add0+p03;
prefetch_p_doubles(add);
#endif
	  rt = s2 *(t18-t16);		it = s2 *(t19-t17);
	  t16= ss1* t16;		t17= ss1* t17;
	  t18= ss2* t18;		t19= ss2* t19;
	  t16= rt+t16;			t17= it+t17;
	  t18= rt-t18;			t19= it-t19;
	  rt = t16;			it = t17;
	  t16= t12+it;			t17= t13-rt;		/* y4 - note swap with y3 below!	*/
	  t12= t12-it;			t13= t13+rt;		/* y1	*/
	  rt = t18;			it = t19;
	  t18= t14+it;			t19= t15-rt;		/* y3	*/
	  t14= t14-it;			t15= t15+rt;		/* y2	*/
#if PFETCH
add = add0+p04;
prefetch_p_doubles(add);
#endif

/*...Third radix-5 transform:	*/
	/* Swap 02<->10, 14<->18 below to get permuted version of j1 + p02,06,10,14,18: */
	  t20=a1p10r;			t21=a1p10i;
	  t22=a1p06r+a1p14r;		t23=a1p06i+a1p14i;
	  t24=a1p02r+a1p18r;		t25=a1p02i+a1p18i;
	  t26=a1p02r-a1p18r;		t27=a1p02i-a1p18i;
	  t28=a1p06r-a1p14r;		t29=a1p06i-a1p14i;
	  rt = t22+t24;			it = t23+t25;
	  t20= t20+rt;			t21= t21+it;		/* y0	*/
	  rt = t20+cc1*rt;		it = t21+cc1*it;
	  t24= cc2*(t22-t24);		t25= cc2*(t23-t25);
	  t22= rt+t24;			t23= it+t25;
	  t24= rt-t24;			t25= it-t25;
#if PFETCH
add = add0+p05;
prefetch_p_doubles(add);
#endif
	  rt = s2 *(t28-t26);		it = s2 *(t29-t27);
	  t26= ss1* t26;		t27= ss1* t27;
	  t28= ss2* t28;		t29= ss2* t29;
	  t26= rt+t26;			t27= it+t27;
	  t28= rt-t28;			t29= it-t29;
	  rt = t26;			it = t27;
	  t26= t22+it;			t27= t23-rt;		/* y4 - note swap with y3 below!	*/
	  t22= t22-it;			t23= t23+rt;		/* y1	*/
	  rt = t28;			it = t29;
	  t28= t24+it;			t29= t25-rt;		/* y3	*/
	  t24= t24-it;			t25= t25+rt;		/* y2	*/
#if PFETCH
add = add0+p06;
prefetch_p_doubles(add);
#endif

/*...Fourth radix-5 transform:	*/
	/* Swap 03->05, 07->01, 11->17, 15->13, 19->09 below to get permuted version of j1 + p03,07,11,15,19: */
	  t30=a1p05r;			t31=a1p05i;
	  t32=a1p01r+a1p09r;		t33=a1p01i+a1p09i;
	  t34=a1p17r+a1p13r;		t35=a1p17i+a1p13i;
	  t36=a1p17r-a1p13r;		t37=a1p17i-a1p13i;
	  t38=a1p01r-a1p09r;		t39=a1p01i-a1p09i;
	  rt = t32+t34;			it = t33+t35;
	  t30= t30+rt;			t31= t31+it;		/* y0	*/
	  rt = t30+cc1*rt;		it = t31+cc1*it;
	  t34= cc2*(t32-t34);		t35= cc2*(t33-t35);
	  t32= rt+t34;			t33= it+t35;
	  t34= rt-t34;			t35= it-t35;
#if PFETCH
add = add0+p07;
prefetch_p_doubles(add);
#endif
	  rt = s2 *(t38-t36);		it = s2 *(t39-t37);
	  t36= ss1* t36;		t37= ss1* t37;
	  t38= ss2* t38;		t39= ss2* t39;
	  t36= rt+t36;			t37= it+t37;
	  t38= rt-t38;			t39= it-t39;
	  rt = t36;			it = t37;
	  t36= t32+it;			t37= t33-rt;		/* y4 - note swap with y3 below!	*/
	  t32= t32-it;			t33= t33+rt;		/* y1	*/
	  rt = t38;			it = t39;
	  t38= t34+it;			t39= t35-rt;		/* y3	*/
	  t34= t34-it;			t35= t35+rt;		/* y2	*/
#if PFETCH
add = add0+p08;
prefetch_p_doubles(add);
#endif

/*...and now do five radix-4 transforms:	*/
	    rt =t20;	t20=t00-rt;	t00=t00+rt;
	    it =t21;	t21=t01-it;	t01=t01+it;

	    rt =t30;	t30=t10-rt;	t10=t10+rt;
	    it =t31;	t31=t11-it;	t11=t11+it;
#if PFETCH
add = add0+p09;
prefetch_p_doubles(add);
#endif
	    a[j1    ]=t00+t10;	a[j1+1    ]=t01+t11;
	    a[j1+p01]=t00-t10;	a[j1+1+p01]=t01-t11;
	    a[j1+p03]=t20-t31;	a[j1+1+p03]=t21+t30;	/* mpy by I is inlined here...	*/
	    a[j1+p02]=t20+t31;	a[j1+1+p02]=t21-t30;
#if PFETCH
add = add0+p10;
prefetch_p_doubles(add);
#endif
/*...Block 2: */
	    rt =t22;	t22=t02-rt;	t02=t02+rt;
	    it =t23;	t23=t03-it;	t03=t03+it;

	    rt =t32;	t32=t12-rt;	t12=t12+rt;
	    it =t33;	t33=t13-it;	t13=t13+it;
#if PFETCH
add = add0+p11;
prefetch_p_doubles(add);
#endif
	    a[j1+p16]=t02+t12;	a[j1+1+p16]=t03+t13;
	    a[j1+p17]=t02-t12;	a[j1+1+p17]=t03-t13;
	    a[j1+p19]=t22-t33;	a[j1+1+p19]=t23+t32;	/* mpy by I is inlined here...	*/
	    a[j1+p18]=t22+t33;	a[j1+1+p18]=t23-t32;
#if PFETCH
add = add0+p12;
prefetch_p_doubles(add);
#endif
/*...Block 3: */
	    rt =t24;	t24=t04-rt;	t04=t04+rt;
	    it =t25;	t25=t05-it;	t05=t05+it;

	    rt =t34;	t34=t14-rt;	t14=t14+rt;
	    it =t35;	t35=t15-it;	t15=t15+it;
#if PFETCH
add = add0+p13;
prefetch_p_doubles(add);
#endif
	    a[j1+p14]=t04+t14;	a[j1+1+p14]=t05+t15;
	    a[j1+p15]=t04-t14;	a[j1+1+p15]=t05-t15;
	    a[j1+p12]=t24-t35;	a[j1+1+p12]=t25+t34;	/* mpy by I is inlined here...	*/
	    a[j1+p13]=t24+t35;	a[j1+1+p13]=t25-t34;
#if PFETCH
add = add0+p14;
prefetch_p_doubles(add);
#endif
/*...Block 4: */
	    rt =t26;	t26=t06-rt;	t06=t06+rt;
	    it =t27;	t27=t07-it;	t07=t07+it;

	    rt =t36;	t36=t16-rt;	t16=t16+rt;
	    it =t37;	t37=t17-it;	t17=t17+it;
#if PFETCH
add = add0+p15;
prefetch_p_doubles(add);
#endif
	    a[j1+p07]=t06+t16;	a[j1+1+p07]=t07+t17;
	    a[j1+p06]=t06-t16;	a[j1+1+p06]=t07-t17;
	    a[j1+p05]=t26-t37;	a[j1+1+p05]=t27+t36;	/* mpy by I is inlined here...	*/
	    a[j1+p04]=t26+t37;	a[j1+1+p04]=t27-t36;
#if PFETCH
add = add0+p16;
prefetch_p_doubles(add);
#endif
/*...Block 5: */
	    rt =t28;	t28=t08-rt;	t08=t08+rt;
	    it =t29;	t29=t09-it;	t09=t09+it;

	    rt =t38;	t38=t18-rt;	t18=t18+rt;
	    it =t39;	t39=t19-it;	t19=t19+it;
#if PFETCH
add = add0+p17;
prefetch_p_doubles(add);
#endif
	    a[j1+p09]=t08+t18;	a[j1+1+p09]=t09+t19;
	    a[j1+p08]=t08-t18;	a[j1+1+p08]=t09-t19;
	    a[j1+p10]=t28-t39;	a[j1+1+p10]=t29+t38;	/* mpy by I is inlined here...	*/
	    a[j1+p11]=t28+t39;	a[j1+1+p11]=t29-t38;
#if PFETCH
add = add0+p18;
prefetch_p_doubles(add);
#endif
#if PFETCH
add = add0+p19;
prefetch_p_doubles(add);
#endif

#endif	// endif(1)

	  iroot += root_incr;		/* increment sincos index.	*/

	  }

	  jstart += nwt;
	  jhi    += nwt;
	  col += 20;
	  co3 -= 20;

	}

	if(root_incr==0)break;

/*   Wraparound carry cleanup loop is here: ***
!
!   (1) Invert the radix-36 forward DIF FFT of the first block of 36 complex elements in A and unweight;
!   (2) Propagate cleanup carries among the real and imaginary parts of the 36 outputs of (1);
!   (3) Reweight and perform a radix-36 forward DIF FFT on the result of (2);
!   (4) If any of the exit carries from (2) are nonzero, advance to the next 36 elements and repeat (1-4).
*/
	t00 = cy19;
	cy19= cy18;
	cy18= cy17;
	cy17= cy16;
	cy16= cy15;
	cy15= cy14;
	cy14= cy13;
	cy13= cy12;
	cy12= cy11;
	cy11= cy10;
	cy10= cy09;
	cy09= cy08;
	cy08= cy07;
	cy07= cy06;
	cy06= cy05;
	cy05= cy04;
	cy04= cy03;
	cy03= cy02;
	cy02= cy01;
	cy01= cy00;
	cy00= t00 ;

	iroot = 0;
	root_incr = 0;
	scale = 1;

	jstart = 0;
	jhi = 3;
	khi = 1;

	for(j=0; j<=jhi; j++)
	{
	  a[j    ] *= radix_inv;
	  a[j+p01] *= radix_inv;
	  a[j+p02] *= radix_inv;
	  a[j+p03] *= radix_inv;
	  a[j+p04] *= radix_inv;
	  a[j+p05] *= radix_inv;
	  a[j+p06] *= radix_inv;
	  a[j+p07] *= radix_inv;
	  a[j+p08] *= radix_inv;
	  a[j+p09] *= radix_inv;
	  a[j+p10] *= radix_inv;
	  a[j+p11] *= radix_inv;
	  a[j+p12] *= radix_inv;
	  a[j+p13] *= radix_inv;
	  a[j+p14] *= radix_inv;
	  a[j+p15] *= radix_inv;
	  a[j+p16] *= radix_inv;
	  a[j+p17] *= radix_inv;
	  a[j+p18] *= radix_inv;
	  a[j+p19] *= radix_inv;
	}
}

	if(fabs(cy00)+fabs(cy01)+fabs(cy02)+fabs(cy03)+fabs(cy04)+fabs(cy05)+fabs(cy06)+fabs(cy07)+fabs(cy08)+fabs(cy09)+fabs(cy10)+fabs(cy11)+fabs(cy12)+fabs(cy13)+fabs(cy14)+fabs(cy15)+fabs(cy16)+fabs(cy17)+fabs(cy18)+fabs(cy19) != 0.0)
	{
	    if(INTERACT)printf("FATAL: iter = %10d; nonzero exit carry in radix20_ditN_cy_dif1.\n",iter);
	    f1 = fopen(   OFILE,"a");
	    f2 = fopen(STATFILE,"a");
	    fprintf(f1,"FATAL: iter = %10d; nonzero exit carry in radix20_ditN_cy_dif1.\n",iter);
	    fprintf(f2,"FATAL: iter = %10d; nonzero exit carry in radix20_ditN_cy_dif1.\n",iter);
	    fclose(f1);
	    fclose(f2);
	    err=1;
	    return(err);
	}
	*fracmax = maxerr;
	return(0);
}

/***************/

int radix20_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], double base[], double baseinv[], int iter                 , int p)
{
/*
!...Acronym: DWT = Discrete Weighted Transform, DIT = Decimation In Time, DIF = Decimation In Frequency
!
!...Performs a final radix-20 complex DIT pass, an inverse DWT weighting, a carry propagation,
!   a forward DWT weighting, and an initial radix-20 complex DIF pass on the data in the length-N real vector A.
!
!   Data enter and are returned in the A-array.
!
!   See the documentation in mers_mod_square and radix16_dif_pass for further details on the array
!   storage scheme, and radix8_ditN_cy_dif1 for details on the reduced-length weights array scheme.
*/
	int n20, bjmodn00,bjmodn01,bjmodn02,bjmodn03,bjmodn04,bjmodn05,bjmodn06,bjmodn07,bjmodn08,bjmodn09
		,bjmodn10,bjmodn11,bjmodn12,bjmodn13,bjmodn14,bjmodn15,bjmodn16,bjmodn17,bjmodn18,bjmodn19
		,i,j,j1,jstart,jhi,iroot,root_incr,k,khi,l,outer;
	static int psave=0,bw,sw,bjmodnini,nrtm1,p01,p02,p03,p04,p05,p06,p07,p08,p09,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19;
	const  double one_half[3] = {1.0, 0.5, 0.25};	/* Needed for small-weights-tables scheme */
	static double cc1 = -1.25000000000000000000,	/* [cos(u)+cos(2u)]/2-1 = -5/4; u = 2*pi/5 */
		      cc2 =  0.55901699437494742409,	/* [cos(u)-cos(2u)]/2 */
		      s2  =  0.95105651629515357211,	/*  sin(u) */
		      ss1 =  1.53884176858762670130,	/* [sin(u)+sin(2u)] */
		      ss2 =  0.36327126400268044292;	/* [sin(u)-sin(2u)] */
	static double radix_inv, n2inv;
	double rt,it
		,t00,t01,t02,t03,t04,t05,t06,t07,t08,t09
		,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19
		,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29
		,t30,t31,t32,t33,t34,t35,t36,t37,t38,t39
	,a1p00r,a1p01r,a1p02r,a1p03r,a1p04r,a1p05r,a1p06r,a1p07r,a1p08r,a1p09r,a1p10r,a1p11r,a1p12r,a1p13r,a1p14r,a1p15r,a1p16r,a1p17r,a1p18r,a1p19r
	,a1p00i,a1p01i,a1p02i,a1p03i,a1p04i,a1p05i,a1p06i,a1p07i,a1p08i,a1p09i,a1p10i,a1p11i,a1p12i,a1p13i,a1p14i,a1p15i,a1p16i,a1p17i,a1p18i,a1p19i
		,cy00,cy01,cy02,cy03,cy04,cy05,cy06,cy07,cy08,cy09,cy10,cy11,cy12,cy13,cy14,cy15,cy16,cy17,cy18,cy19,temp,scale;
#if PFETCH
	double *add0, *add;
#endif
	int err;
	FILE *f1,*f2;
	static int first_entry=TRUE;

/*...stuff for the reduced-length DWT weights array is here:	*/
	int n_div_nwt;
	int col,co2,co3,m,m2,n_minus_sil,n_minus_silp1,sinwt,sinwtm1;
	double wt,wtinv,wtl,wtlp1,wtn,wtnm1,wtA,wtB,wtC;

/*...change n20 and n_div_wt to non-static to work around a gcc compiler bug. */
	n20   = n/20;
	n_div_nwt = n20 >> nwt_bits;

	if((n_div_nwt << nwt_bits) != n20)
	{
	  if(INTERACT)printf("FATAL: iter = %10d; NWT_BITS does not divide N/20 in radix20_ditN_cy_dif1.\n",iter);
	  f1 = fopen(   OFILE,"a");
	  f2 = fopen(STATFILE,"a");
	  fprintf(f1,"FATAL: iter = %10d; NWT_BITS does not divide N/20 in radix20_ditN_cy_dif1.\n",iter);
	  fprintf(f2,"FATAL: iter = %10d; NWT_BITS does not divide N/20 in radix20_ditN_cy_dif1.\n",iter);
	  fclose(f1);
	  fclose(f2);
	  err=1;
	  return(err);
	}

	if(p != psave)
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  psave = p;
	  first_entry=FALSE;
	  radix_inv = (double)((long double)1.0L/(long double)20.0L);
	  n2inv     = (double)((long double)1.0L/(long double)(n/2));

	  bw    = p%n;		/* Number of bigwords in the Crandall/Fagin mixed-radix representation = (Mersenne exponent) mod (vector length).	*/
	  sw    = n - bw;	/* Number of smallwords.	*/

/*   constant index offsets for array load/stores are here.	*/

	  p01 = n20;
	  p02 = p01 + p01;
	  p03 = p02 + p01;
	  p04 = p03 + p01;
	  p05 = p04 + p01;
	  p06 = p05 + p01;
	  p07 = p06 + p01;
	  p08 = p07 + p01;
	  p09 = p08 + p01;
	  p10 = p09 + p01;
	  p11 = p10 + p01;
	  p12 = p11 + p01;
	  p13 = p12 + p01;
	  p14 = p13 + p01;
	  p15 = p14 + p01;
	  p16 = p15 + p01;
	  p17 = p16 + p01;
	  p18 = p17 + p01;
	  p19 = p18 + p01;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p05 = p05 + ( (p05 >> DAT_BITS) << PAD_BITS );
	  p06 = p06 + ( (p06 >> DAT_BITS) << PAD_BITS );
	  p07 = p07 + ( (p07 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p09 = p09 + ( (p09 >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p11 = p11 + ( (p11 >> DAT_BITS) << PAD_BITS );
	  p12 = p12 + ( (p12 >> DAT_BITS) << PAD_BITS );
	  p13 = p13 + ( (p13 >> DAT_BITS) << PAD_BITS );
	  p14 = p14 + ( (p14 >> DAT_BITS) << PAD_BITS );
	  p15 = p15 + ( (p15 >> DAT_BITS) << PAD_BITS );
	  p16 = p16 + ( (p16 >> DAT_BITS) << PAD_BITS );
	  p17 = p17 + ( (p17 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	  p19 = p19 + ( (p19 >> DAT_BITS) << PAD_BITS );

	  bjmodnini=0;
	  for(j=0; j < n20; j++)
	  {
	    bjmodnini -= sw; bjmodnini = bjmodnini + ( (-(int)((uint32)bjmodnini >> 31)) & n);
	  }
	}

/*...The radix-20 final DIT pass is here.	*/

	cy00=-2;		/* init carries	*/
	cy01= 0;
	cy02= 0;
	cy03= 0;
	cy04= 0;
	cy05= 0;
	cy06= 0;
	cy07= 0;
	cy08= 0;
	cy09= 0;
	cy10= 0;
	cy11= 0;
	cy12= 0;
	cy13= 0;
	cy14= 0;
	cy15= 0;
	cy16= 0;
	cy17= 0;
	cy18= 0;
	cy19= 0;

	iroot = 0;	/* init sincos array index	*/
	root_incr = 1;	/* init sincos array index increment (set = 1 for normal carry pass, = 0 for wrapper pass)	*/
	scale = n2inv;	/* init inverse-weight scale factor  (set = 2/n for normal carry pass, = 1 for wrapper pass)	*/

	jstart = 0;
	jhi = jstart+nwt-1;
	khi = n_div_nwt;

for(outer=0; outer <= 1; outer++)
{
	i = 1;		/* Pointer to the BASE and BASEINV arrays. lowest-order digit is always a bigword (i = 1).	*/

	bjmodn00= 0;
	bjmodn01= bjmodnini;
	bjmodn02= bjmodn01+bjmodnini-n; bjmodn02= bjmodn02+ ( (-(int)((uint32)bjmodn02>> 31)) & n);
	bjmodn03= bjmodn02+bjmodnini-n; bjmodn03= bjmodn03+ ( (-(int)((uint32)bjmodn03>> 31)) & n);
	bjmodn04= bjmodn03+bjmodnini-n; bjmodn04= bjmodn04+ ( (-(int)((uint32)bjmodn04>> 31)) & n);
	bjmodn05= bjmodn04+bjmodnini-n; bjmodn05= bjmodn05+ ( (-(int)((uint32)bjmodn05>> 31)) & n);
	bjmodn06= bjmodn05+bjmodnini-n; bjmodn06= bjmodn06+ ( (-(int)((uint32)bjmodn06>> 31)) & n);
	bjmodn07= bjmodn06+bjmodnini-n; bjmodn07= bjmodn07+ ( (-(int)((uint32)bjmodn07>> 31)) & n);
	bjmodn08= bjmodn07+bjmodnini-n; bjmodn08= bjmodn08+ ( (-(int)((uint32)bjmodn08>> 31)) & n);
	bjmodn09= bjmodn08+bjmodnini-n; bjmodn09= bjmodn09+ ( (-(int)((uint32)bjmodn09>> 31)) & n);
	bjmodn10= bjmodn09+bjmodnini-n; bjmodn10= bjmodn10+ ( (-(int)((uint32)bjmodn10>> 31)) & n);
	bjmodn11= bjmodn10+bjmodnini-n; bjmodn11= bjmodn11+ ( (-(int)((uint32)bjmodn11>> 31)) & n);
	bjmodn12= bjmodn11+bjmodnini-n; bjmodn12= bjmodn12+ ( (-(int)((uint32)bjmodn12>> 31)) & n);
	bjmodn13= bjmodn12+bjmodnini-n; bjmodn13= bjmodn13+ ( (-(int)((uint32)bjmodn13>> 31)) & n);
	bjmodn14= bjmodn13+bjmodnini-n; bjmodn14= bjmodn14+ ( (-(int)((uint32)bjmodn14>> 31)) & n);
	bjmodn15= bjmodn14+bjmodnini-n; bjmodn15= bjmodn15+ ( (-(int)((uint32)bjmodn15>> 31)) & n);
	bjmodn16= bjmodn15+bjmodnini-n; bjmodn16= bjmodn16+ ( (-(int)((uint32)bjmodn16>> 31)) & n);
	bjmodn17= bjmodn16+bjmodnini-n; bjmodn17= bjmodn17+ ( (-(int)((uint32)bjmodn17>> 31)) & n);
	bjmodn18= bjmodn17+bjmodnini-n; bjmodn18= bjmodn18+ ( (-(int)((uint32)bjmodn18>> 31)) & n);
	bjmodn19= bjmodn18+bjmodnini-n; bjmodn19= bjmodn19+ ( (-(int)((uint32)bjmodn19>> 31)) & n);

	col=0;
	co2=(n >> nwt_bits)-1+20;
	co3=co2-20;		/* At the start of each new j-loop, co3=co2-radix(1)	*/

	for(k=1; k <= khi; k++)	/* Do n/(radix(1)*nwt) outer loop executions...	*/
	{
	  for(j=jstart; j<jhi; j += 2)	/* Each inner loop execution processes (radix(1)*nwt) array data.	*/
	  {
	    j1 = j + ( (j >> DAT_BITS) << PAD_BITS );	/* padded-array fetch index is here	*/
/*
!...gather the needed data (20 64-bit complex, i.e. 40 64-bit reals) and do a radix-20 DIT transform...
*/
#if 1
/*...gather the needed data (20 64-bit complex, i.e. 40 64-bit reals) and do 5 radix-4 transforms...*/
                     /*          inputs           */ /*                                      outputs                                      */
	RADIX_04_DIT(a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p03],a[j1+1+p03],a[j1+p02],a[j1+1+p02],t00,t01,t10,t11,t20,t21,t30,t31,rt,it);
	RADIX_04_DIT(a[j1+p07],a[j1+1+p07],a[j1+p06],a[j1+1+p06],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],t02,t03,t12,t13,t22,t23,t32,t33,rt,it);
	RADIX_04_DIT(a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],a[j1+p10],a[j1+1+p10],a[j1+p11],a[j1+1+p11],t04,t05,t14,t15,t24,t25,t34,t35,rt,it);
	RADIX_04_DIT(a[j1+p14],a[j1+1+p14],a[j1+p15],a[j1+1+p15],a[j1+p12],a[j1+1+p12],a[j1+p13],a[j1+1+p13],t06,t07,t16,t17,t26,t27,t36,t37,rt,it);
	RADIX_04_DIT(a[j1+p16],a[j1+1+p16],a[j1+p17],a[j1+1+p17],a[j1+p19],a[j1+1+p19],a[j1+p18],a[j1+1+p18],t08,t09,t18,t19,t28,t29,t38,t39,rt,it);

/*...and now do 4 radix-5 transforms...*/
                     /*                                                inputs                                                   */ /*                 outputs                   */
	RADIX_05_DFT(t00,t01,t02,t03,t04,t05,t06,t07,t08,t09,a1p00r,a1p00i,a1p16r,a1p16i,a1p12r,a1p12i,a1p08r,a1p08i,a1p04r,a1p04i,rt,it);
	RADIX_05_DFT(t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,a1p15r,a1p15i,a1p11r,a1p11i,a1p07r,a1p07i,a1p03r,a1p03i,a1p19r,a1p19i,rt,it);
	RADIX_05_DFT(t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,a1p10r,a1p10i,a1p06r,a1p06i,a1p02r,a1p02i,a1p18r,a1p18i,a1p14r,a1p14i,rt,it);
	RADIX_05_DFT(t30,t31,t32,t33,t34,t35,t36,t37,t38,t39,a1p05r,a1p05i,a1p01r,a1p01i,a1p17r,a1p17i,a1p13r,a1p13i,a1p09r,a1p09i,rt,it);
#else

/*...Block 1:	*/
	t00=a[j1    ];	t01=a[j1+1    ];
	rt =a[j1+p01];	it =a[j1+1+p01];
	t10=t00-rt;  	t00=t00+rt;
	t11=t01-it;	t01=t01+it;

	t20=a[j1+p03];	t21=a[j1+1+p03];
	rt =a[j1+p02];	it =a[j1+1+p02];
	t30=t20-rt;  	t20=t20+rt;
	t31=t21-it;  	t21=t21+it;

	rt =t20;	t20=t00-rt;	t00=t00+rt;
	it =t21;	t21=t01-it;	t01=t01+it;

	rt =t30;	t30=t10-t31;	t10=t10+t31;
			t31=t11+rt;	t11=t11-rt;

/*...Block 2:	*/
	t02=a[j1+p07];	t03=a[j1+1+p07];
	rt =a[j1+p06];	it =a[j1+1+p06];
	t12=t02-rt;  	t02=t02+rt;
	t13=t03-it;	t03=t03+it;

	t22=a[j1+p05];	t23=a[j1+1+p05];
	rt =a[j1+p04];	it =a[j1+1+p04];
	t32=t22-rt;  	t22=t22+rt;
	t33=t23-it;  	t23=t23+it;

	rt =t22;	t22=t02-rt;	t02=t02+rt;
	it =t23;	t23=t03-it;	t03=t03+it;

	rt =t32;	t32=t12-t33;	t12=t12+t33;
			t33=t13+rt;	t13=t13-rt;

/*...Block 3:	*/
	t04=a[j1+p09];	t05=a[j1+1+p09];
	rt =a[j1+p08];	it =a[j1+1+p08];
	t14=t04-rt;  	t04=t04+rt;
	t15=t05-it;	t05=t05+it;

	t24=a[j1+p10];	t25=a[j1+1+p10];
	rt =a[j1+p11];	it =a[j1+1+p11];
	t34=t24-rt;  	t24=t24+rt;
	t35=t25-it;  	t25=t25+it;

	rt =t24;	t24=t04-rt;	t04=t04+rt;
	it =t25;	t25=t05-it;	t05=t05+it;

	rt =t34;	t34=t14-t35;	t14=t14+t35;
			t35=t15+rt;	t15=t15-rt;

/*...Block 4:	*/
	t06=a[j1+p14];	t07=a[j1+1+p14];
	rt =a[j1+p15];	it =a[j1+1+p15];
	t16=t06-rt;  	t06=t06+rt;
	t17=t07-it;	t07=t07+it;

	t26=a[j1+p12];	t27=a[j1+1+p12];
	rt =a[j1+p13];	it =a[j1+1+p13];
	t36=t26-rt;  	t26=t26+rt;
	t37=t27-it;  	t27=t27+it;

	rt =t26;	t26=t06-rt;	t06=t06+rt;
	it =t27;	t27=t07-it;	t07=t07+it;

	rt =t36;	t36=t16-t37;	t16=t16+t37;
			t37=t17+rt;	t17=t17-rt;

/*...Block 5:	*/
	t08=a[j1+p16];	t09=a[j1+1+p16];
	rt =a[j1+p17];	it =a[j1+1+p17];
	t18=t08-rt;  	t08=t08+rt;
	t19=t09-it;	t09=t09+it;

	t28=a[j1+p19];	t29=a[j1+1+p19];
	rt =a[j1+p18];	it =a[j1+1+p18];
	t38=t28-rt;  	t28=t28+rt;
	t39=t29-it;  	t29=t29+it;

	rt =t28;	t28=t08-rt;	t08=t08+rt;
	it =t29;	t29=t09-it;	t09=t09+it;

	rt =t38;	t38=t18-t39;	t18=t18+t39;
			t39=t19+rt;	t19=t19-rt;

/*       ...and now do four radix-5 transforms.	*/
/*...First radix-5 transform:	*/
	    rt  =t08;			it  =t09;
	    t08 =t02 -rt;		t09 =t03 -it;
	    t02 =t02 +rt;		t03 =t03 +it;
	    rt  =t06;			it  =t07;
	    t06 =t04 -rt;		t07 =t05 -it;
	    t04 =t04 +rt;		t05 =t05 +it;
	    rt  = t02+t04;		it  = t03+t05;
	    t00 = t00+rt;		t01 = t01+it;
	    rt  = t00+cc1*rt;		it  = t01+cc1*it;
	    t04 = cc2*(t02-t04);	t05 = cc2*(t03-t05);
	    t02 = rt+t04;		t03 = it+t05;
	    t04 = rt-t04;		t05 = it-t05;
	    rt  = s2 *(t06-t08);	it  = s2 *(t07-t09);
	    t06 = ss1* t06;		t07 = ss1* t07;
	    t08 = ss2* t08;		t09 = ss2* t09;
	    t06 = rt-t06;		t07 = it-t07;
	    t08 = rt+t08;		t09 = it+t09;
	    a1p00r=t00;		a1p00i=t01;
	    a1p04r=t02-t07;		a1p04i=t03+t06;
	    a1p08r=t04-t09;		a1p08i=t05+t08;
	    a1p12r=t04+t09;		a1p12i=t05-t08;
	    a1p16r=t02+t07;		a1p16i=t03-t06;

/*...Second radix-5 transform:	*/
	    rt  =t18;			it  =t19;
	    t18 =t12 -rt;		t19 =t13 -it;
	    t12 =t12 +rt;		t13 =t13 +it;
	    rt  =t16;			it  =t17;
	    t16 =t14 -rt;		t17 =t15 -it;
	    t14 =t14 +rt;		t15 =t15 +it;
	    rt  = t12+t14;		it  = t13+t15;
	    t10 = t10+rt;		t11 = t11+it;
	    rt  = t10+cc1*rt;		it  = t11+cc1*it;
	    t14 = cc2*(t12-t14);	t15 = cc2*(t13-t15);
	    t12 = rt+t14;		t13 = it+t15;
	    t14 = rt-t14;		t15 = it-t15;
	    rt  = s2 *(t16-t18);	it  = s2 *(t17-t19);
	    t16 = ss1* t16;		t17 = ss1* t17;
	    t18 = ss2* t18;		t19 = ss2* t19;
	    t16 = rt-t16;		t17 = it-t17;
	    t18 = rt+t18;		t19 = it+t19;
	    a1p15r=t10;		a1p15i=t11;
	    a1p19r=t12-t17;		a1p19i=t13+t16;
	    a1p03r=t14-t19;		a1p03i=t15+t18;
	    a1p07r=t14+t19;		a1p07i=t15-t18;
	    a1p11r=t12+t17;		a1p11i=t13-t16;

/*...Third radix-5 transform:	*/
	    rt  =t28;			it  =t29;
	    t28 =t22 -rt;		t29 =t23 -it;
	    t22 =t22 +rt;		t23 =t23 +it;
	    rt  =t26;			it  =t27;
	    t26 =t24 -rt;		t27 =t25 -it;
	    t24 =t24 +rt;		t25 =t25 +it;
	    rt  = t22+t24;		it  = t23+t25;
	    t20 = t20+rt;		t21 = t21+it;
	    rt  = t20+cc1*rt;		it  = t21+cc1*it;
	    t24 = cc2*(t22-t24);	t25 = cc2*(t23-t25);
	    t22 = rt+t24;		t23 = it+t25;
	    t24 = rt-t24;		t25 = it-t25;
	    rt  = s2 *(t26-t28);	it  = s2 *(t27-t29);
	    t26 = ss1* t26;		t27 = ss1* t27;
	    t28 = ss2* t28;		t29 = ss2* t29;
	    t26 = rt-t26;		t27 = it-t27;
	    t28 = rt+t28;		t29 = it+t29;
	    a1p10r=t20;		a1p10i=t21;
	    a1p14r=t22-t27;		a1p14i=t23+t26;
	    a1p18r=t24-t29;		a1p18i=t25+t28;
	    a1p02r=t24+t29;		a1p02i=t25-t28;
	    a1p06r=t22+t27;		a1p06i=t23-t26;

/*...Fourth radix-5 transform:	*/
	    rt  =t38;			it  =t39;
	    t38 =t32 -rt;		t39 =t33 -it;
	    t32 =t32 +rt;		t33 =t33 +it;
	    rt  =t36;			it  =t37;
	    t36 =t34 -rt;		t37 =t35 -it;
	    t34 =t34 +rt;		t35 =t35 +it;
	    rt  = t32+t34;		it  = t33+t35;
	    t30 = t30+rt;		t31 = t31+it;
	    rt  = t30+cc1*rt;		it  = t31+cc1*it;
	    t34 = cc2*(t32-t34);	t35 = cc2*(t33-t35);
	    t32 = rt+t34;		t33 = it+t35;
	    t34 = rt-t34;		t35 = it-t35;
	    rt  = s2 *(t36-t38);	it  = s2 *(t37-t39);
	    t36 = ss1* t36;		t37 = ss1* t37;
	    t38 = ss2* t38;		t39 = ss2* t39;
	    t36 = rt-t36;		t37 = it-t37;
	    t38 = rt+t38;		t39 = it+t39;
	    a1p05r=t30;		a1p05i=t31;
	    a1p09r=t32-t37;		a1p09i=t33+t36;
	    a1p13r=t34-t39;		a1p13i=t35+t38;
	    a1p17r=t34+t39;		a1p17i=t35-t38;
	    a1p01r=t32+t37;		a1p01i=t33-t36;
#endif

/*...Now do the carries. Since the outputs would
    normally be getting dispatched to 20 separate blocks of the A-array, we need 20 separate carries.	*/

	    l= j & (nwt-1);
	    n_minus_sil   = n-si[l  ];
	    n_minus_silp1 = n-si[l+1];
	    sinwt   = si[nwt-l  ];
	    sinwtm1 = si[nwt-l-1];

	    wtl     =wt0[    l  ];
	    wtn     =wt0[nwt-l  ]*scale;	/* Include 1/(n/2) scale factor of inverse transform here...	*/
	    wtlp1   =wt0[    l+1];
	    wtnm1   =wt0[nwt-l-1]*scale;	/* ...and here.	*/

/*...set0 is slightly different from others:	*/
	   cmplx_carry_norm_nocheck0(a1p00r,a1p00i,cy00,bjmodn00   );
	    cmplx_carry_norm_nocheck(a1p01r,a1p01i,cy01,bjmodn01,1 );
	    cmplx_carry_norm_nocheck(a1p02r,a1p02i,cy02,bjmodn02,2 );
	    cmplx_carry_norm_nocheck(a1p03r,a1p03i,cy03,bjmodn03,3 );
	    cmplx_carry_norm_nocheck(a1p04r,a1p04i,cy04,bjmodn04,4 );
	    cmplx_carry_norm_nocheck(a1p05r,a1p05i,cy05,bjmodn05,5 );
	    cmplx_carry_norm_nocheck(a1p06r,a1p06i,cy06,bjmodn06,6 );
	    cmplx_carry_norm_nocheck(a1p07r,a1p07i,cy07,bjmodn07,7 );
	    cmplx_carry_norm_nocheck(a1p08r,a1p08i,cy08,bjmodn08,8 );
	    cmplx_carry_norm_nocheck(a1p09r,a1p09i,cy09,bjmodn09,9 );
	    cmplx_carry_norm_nocheck(a1p10r,a1p10i,cy10,bjmodn10,10);
	    cmplx_carry_norm_nocheck(a1p11r,a1p11i,cy11,bjmodn11,11);
	    cmplx_carry_norm_nocheck(a1p12r,a1p12i,cy12,bjmodn12,12);
	    cmplx_carry_norm_nocheck(a1p13r,a1p13i,cy13,bjmodn13,13);
	    cmplx_carry_norm_nocheck(a1p14r,a1p14i,cy14,bjmodn14,14);
	    cmplx_carry_norm_nocheck(a1p15r,a1p15i,cy15,bjmodn15,15);
	    cmplx_carry_norm_nocheck(a1p16r,a1p16i,cy16,bjmodn16,16);
	    cmplx_carry_norm_nocheck(a1p17r,a1p17i,cy17,bjmodn17,17);
	    cmplx_carry_norm_nocheck(a1p18r,a1p18i,cy18,bjmodn18,18);
	    cmplx_carry_norm_nocheck(a1p19r,a1p19i,cy19,bjmodn19,19);

	    i =((uint32)(sw - bjmodn00) >> 31);	/* get ready for the next set...	*/
	    co2=co3;	/* For all data but the first set in each j-block, co2=co3. Thus, after the first block of data is done
			   and only then: for all subsequent blocks it's superfluous), this assignment decrements co2 by radix(1).	*/

/*...The radix-20 DIF pass is here:	*/
#if PFETCH
add0 = &a[j1];
prefetch_p_doubles(add0);
#endif

#if 1
/*...gather the needed data (20 64-bit complex, i.e. 40 64-bit reals) and do 4 radix-5 transforms...*/
                     /*                                                inputs                                                   */ /*                 outputs                   */
#if PFETCH
	RADIX_05_DFT_PFETCH(a1p00r,a1p00i,a1p16r,a1p16i,a1p12r,a1p12i,a1p08r,a1p08i,a1p04r,a1p04i,t00,t01,t02,t03,t04,t05,t06,t07,t08,t09,rt,it,p01,p02);
	RADIX_05_DFT_PFETCH(a1p15r,a1p15i,a1p11r,a1p11i,a1p07r,a1p07i,a1p03r,a1p03i,a1p19r,a1p19i,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,rt,it,p03,p04);
	RADIX_05_DFT_PFETCH(a1p10r,a1p10i,a1p06r,a1p06i,a1p02r,a1p02i,a1p18r,a1p18i,a1p14r,a1p14i,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,rt,it,p05,p06);
	RADIX_05_DFT_PFETCH(a1p05r,a1p05i,a1p01r,a1p01i,a1p17r,a1p17i,a1p13r,a1p13i,a1p09r,a1p09i,t30,t31,t32,t33,t34,t35,t36,t37,t38,t39,rt,it,p07,p08);
#else
	RADIX_05_DFT       (a1p00r,a1p00i,a1p16r,a1p16i,a1p12r,a1p12i,a1p08r,a1p08i,a1p04r,a1p04i,t00,t01,t02,t03,t04,t05,t06,t07,t08,t09,rt,it);
	RADIX_05_DFT       (a1p15r,a1p15i,a1p11r,a1p11i,a1p07r,a1p07i,a1p03r,a1p03i,a1p19r,a1p19i,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,rt,it);
	RADIX_05_DFT       (a1p10r,a1p10i,a1p06r,a1p06i,a1p02r,a1p02i,a1p18r,a1p18i,a1p14r,a1p14i,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,rt,it);
	RADIX_05_DFT       (a1p05r,a1p05i,a1p01r,a1p01i,a1p17r,a1p17i,a1p13r,a1p13i,a1p09r,a1p09i,t30,t31,t32,t33,t34,t35,t36,t37,t38,t39,rt,it);
#endif
/*...and now do 5 radix-4 transforms...*/
                     /*          inputs           */ /*                                      outputs                                      */
#if PFETCH
	add = add0+p09;
	prefetch_p_doubles(add);

	RADIX_04_DIF_PFETCH(t00,t01,t10,t11,t20,t21,t30,t31,a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p03],a[j1+1+p03],a[j1+p02],a[j1+1+p02],rt,it,p10,p11);
	RADIX_04_DIF_PFETCH(t02,t03,t12,t13,t22,t23,t32,t33,a[j1+p16],a[j1+1+p16],a[j1+p17],a[j1+1+p17],a[j1+p19],a[j1+1+p19],a[j1+p18],a[j1+1+p18],rt,it,p12,p13);
	RADIX_04_DIF_PFETCH(t04,t05,t14,t15,t24,t25,t34,t35,a[j1+p14],a[j1+1+p14],a[j1+p15],a[j1+1+p15],a[j1+p12],a[j1+1+p12],a[j1+p13],a[j1+1+p13],rt,it,p14,p15);
	RADIX_04_DIF_PFETCH(t06,t07,t16,t17,t26,t27,t36,t37,a[j1+p07],a[j1+1+p07],a[j1+p06],a[j1+1+p06],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],rt,it,p16,p17);
	RADIX_04_DIF_PFETCH(t08,t09,t18,t19,t28,t29,t38,t39,a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],a[j1+p10],a[j1+1+p10],a[j1+p11],a[j1+1+p11],rt,it,p18,p19);
#else
	RADIX_04_DIF       (t00,t01,t10,t11,t20,t21,t30,t31,a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p03],a[j1+1+p03],a[j1+p02],a[j1+1+p02],rt,it);
	RADIX_04_DIF       (t02,t03,t12,t13,t22,t23,t32,t33,a[j1+p16],a[j1+1+p16],a[j1+p17],a[j1+1+p17],a[j1+p19],a[j1+1+p19],a[j1+p18],a[j1+1+p18],rt,it);
	RADIX_04_DIF       (t04,t05,t14,t15,t24,t25,t34,t35,a[j1+p14],a[j1+1+p14],a[j1+p15],a[j1+1+p15],a[j1+p12],a[j1+1+p12],a[j1+p13],a[j1+1+p13],rt,it);
	RADIX_04_DIF       (t06,t07,t16,t17,t26,t27,t36,t37,a[j1+p07],a[j1+1+p07],a[j1+p06],a[j1+1+p06],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],rt,it);
	RADIX_04_DIF       (t08,t09,t18,t19,t28,t29,t38,t39,a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],a[j1+p10],a[j1+1+p10],a[j1+p11],a[j1+1+p11],rt,it);
#endif

#else

/*...First radix-5 transform:	*/
	/* Swap 4<->16, 8<->12 below to get permuted version of j1 + p00,04,08,12,16: */
	  t00=a1p00r;			t01=a1p00i;
	  t02=a1p16r+a1p04r;		t03=a1p16i+a1p04i;
	  t04=a1p12r+a1p08r;		t05=a1p12i+a1p08i;
	  t06=a1p12r-a1p08r;		t07=a1p12i-a1p08i;
	  t08=a1p16r-a1p04r;		t09=a1p16i-a1p04i;
	  rt = t02+t04;			it = t03+t05;
	  t00= t00+rt;			t01= t01+it;		/* y0	*/
	  rt = t00+cc1*rt;		it = t01+cc1*it;
	  t04= cc2*(t02-t04);		t05= cc2*(t03-t05);
	  t02= rt+t04;			t03= it+t05;
	  t04= rt-t04;			t05= it-t05;
#if PFETCH
add = add0+p01;
prefetch_p_doubles(add);
#endif
	  rt = s2 *(t08-t06);		it = s2 *(t09-t07);
	  t06= ss1* t06;		t07= ss1* t07;
	  t08= ss2* t08;		t09= ss2* t09;
	  t06= rt+t06;			t07= it+t07;
	  t08= rt-t08;			t09= it-t09;
	  rt = t06;			it = t07;
	  t06= t02+it;			t07= t03-rt;		/* y4 - note swap with y3 below!	*/
	  t02= t02-it;			t03= t03+rt;		/* y1	*/
	  rt = t08;			it = t09;
	  t08= t04+it;			t09= t05-rt;		/* y3	*/
	  t04= t04-it;			t05= t05+rt;		/* y2	*/
#if PFETCH
add = add0+p02;
prefetch_p_doubles(add);
#endif

/*...Second radix-5 transform:	*/
	/* Swap 01->15, 05->11, 09->07, 13->03, 17->19 below to get permuted version of j1 + p01,05,09,13,17: */
	  t10=a1p15r;			t11=a1p15i;
	  t12=a1p11r+a1p19r;		t13=a1p11i+a1p19i;
	  t14=a1p07r+a1p03r;		t15=a1p07i+a1p03i;
	  t16=a1p07r-a1p03r;		t17=a1p07i-a1p03i;
	  t18=a1p11r-a1p19r;		t19=a1p11i-a1p19i;
	  rt = t12+t14;			it = t13+t15;
	  t10= t10+rt;			t11= t11+it;		/* y0	*/
	  rt = t10+cc1*rt;		it = t11+cc1*it;
	  t14= cc2*(t12-t14);		t15= cc2*(t13-t15);
	  t12= rt+t14;			t13= it+t15;
	  t14= rt-t14;			t15= it-t15;
#if PFETCH
add = add0+p03;
prefetch_p_doubles(add);
#endif
	  rt = s2 *(t18-t16);		it = s2 *(t19-t17);
	  t16= ss1* t16;		t17= ss1* t17;
	  t18= ss2* t18;		t19= ss2* t19;
	  t16= rt+t16;			t17= it+t17;
	  t18= rt-t18;			t19= it-t19;
	  rt = t16;			it = t17;
	  t16= t12+it;			t17= t13-rt;		/* y4 - note swap with y3 below!	*/
	  t12= t12-it;			t13= t13+rt;		/* y1	*/
	  rt = t18;			it = t19;
	  t18= t14+it;			t19= t15-rt;		/* y3	*/
	  t14= t14-it;			t15= t15+rt;		/* y2	*/
#if PFETCH
add = add0+p04;
prefetch_p_doubles(add);
#endif

/*...Third radix-5 transform:	*/
	/* Swap 02<->10, 14<->18 below to get permuted version of j1 + p02,06,10,14,18: */
	  t20=a1p10r;			t21=a1p10i;
	  t22=a1p06r+a1p14r;		t23=a1p06i+a1p14i;
	  t24=a1p02r+a1p18r;		t25=a1p02i+a1p18i;
	  t26=a1p02r-a1p18r;		t27=a1p02i-a1p18i;
	  t28=a1p06r-a1p14r;		t29=a1p06i-a1p14i;
	  rt = t22+t24;			it = t23+t25;
	  t20= t20+rt;			t21= t21+it;		/* y0	*/
	  rt = t20+cc1*rt;		it = t21+cc1*it;
	  t24= cc2*(t22-t24);		t25= cc2*(t23-t25);
	  t22= rt+t24;			t23= it+t25;
	  t24= rt-t24;			t25= it-t25;
#if PFETCH
add = add0+p05;
prefetch_p_doubles(add);
#endif
	  rt = s2 *(t28-t26);		it = s2 *(t29-t27);
	  t26= ss1* t26;		t27= ss1* t27;
	  t28= ss2* t28;		t29= ss2* t29;
	  t26= rt+t26;			t27= it+t27;
	  t28= rt-t28;			t29= it-t29;
	  rt = t26;			it = t27;
	  t26= t22+it;			t27= t23-rt;		/* y4 - note swap with y3 below!	*/
	  t22= t22-it;			t23= t23+rt;		/* y1	*/
	  rt = t28;			it = t29;
	  t28= t24+it;			t29= t25-rt;		/* y3	*/
	  t24= t24-it;			t25= t25+rt;		/* y2	*/
#if PFETCH
add = add0+p06;
prefetch_p_doubles(add);
#endif

/*...Fourth radix-5 transform:	*/
	/* Swap 03->05, 07->01, 11->17, 15->13, 19->09 below to get permuted version of j1 + p03,07,11,15,19: */
	  t30=a1p05r;			t31=a1p05i;
	  t32=a1p01r+a1p09r;		t33=a1p01i+a1p09i;
	  t34=a1p17r+a1p13r;		t35=a1p17i+a1p13i;
	  t36=a1p17r-a1p13r;		t37=a1p17i-a1p13i;
	  t38=a1p01r-a1p09r;		t39=a1p01i-a1p09i;
	  rt = t32+t34;			it = t33+t35;
	  t30= t30+rt;			t31= t31+it;		/* y0	*/
	  rt = t30+cc1*rt;		it = t31+cc1*it;
	  t34= cc2*(t32-t34);		t35= cc2*(t33-t35);
	  t32= rt+t34;			t33= it+t35;
	  t34= rt-t34;			t35= it-t35;
#if PFETCH
add = add0+p07;
prefetch_p_doubles(add);
#endif
	  rt = s2 *(t38-t36);		it = s2 *(t39-t37);
	  t36= ss1* t36;		t37= ss1* t37;
	  t38= ss2* t38;		t39= ss2* t39;
	  t36= rt+t36;			t37= it+t37;
	  t38= rt-t38;			t39= it-t39;
	  rt = t36;			it = t37;
	  t36= t32+it;			t37= t33-rt;		/* y4 - note swap with y3 below!	*/
	  t32= t32-it;			t33= t33+rt;		/* y1	*/
	  rt = t38;			it = t39;
	  t38= t34+it;			t39= t35-rt;		/* y3	*/
	  t34= t34-it;			t35= t35+rt;		/* y2	*/
#if PFETCH
add = add0+p08;
prefetch_p_doubles(add);
#endif

/*...and now do five radix-4 transforms:	*/
	    rt =t20;	t20=t00-rt;	t00=t00+rt;
	    it =t21;	t21=t01-it;	t01=t01+it;

	    rt =t30;	t30=t10-rt;	t10=t10+rt;
	    it =t31;	t31=t11-it;	t11=t11+it;
#if PFETCH
add = add0+p09;
prefetch_p_doubles(add);
#endif
	    a[j1    ]=t00+t10;	a[j1+1    ]=t01+t11;
	    a[j1+p01]=t00-t10;	a[j1+1+p01]=t01-t11;
	    a[j1+p03]=t20-t31;	a[j1+1+p03]=t21+t30;	/* mpy by I is inlined here...	*/
	    a[j1+p02]=t20+t31;	a[j1+1+p02]=t21-t30;
#if PFETCH
add = add0+p10;
prefetch_p_doubles(add);
#endif
/*...Block 2: */
	    rt =t22;	t22=t02-rt;	t02=t02+rt;
	    it =t23;	t23=t03-it;	t03=t03+it;

	    rt =t32;	t32=t12-rt;	t12=t12+rt;
	    it =t33;	t33=t13-it;	t13=t13+it;
#if PFETCH
add = add0+p11;
prefetch_p_doubles(add);
#endif
	    a[j1+p16]=t02+t12;	a[j1+1+p16]=t03+t13;
	    a[j1+p17]=t02-t12;	a[j1+1+p17]=t03-t13;
	    a[j1+p19]=t22-t33;	a[j1+1+p19]=t23+t32;	/* mpy by I is inlined here...	*/
	    a[j1+p18]=t22+t33;	a[j1+1+p18]=t23-t32;
#if PFETCH
add = add0+p12;
prefetch_p_doubles(add);
#endif
/*...Block 3: */
	    rt =t24;	t24=t04-rt;	t04=t04+rt;
	    it =t25;	t25=t05-it;	t05=t05+it;

	    rt =t34;	t34=t14-rt;	t14=t14+rt;
	    it =t35;	t35=t15-it;	t15=t15+it;
#if PFETCH
add = add0+p13;
prefetch_p_doubles(add);
#endif
	    a[j1+p14]=t04+t14;	a[j1+1+p14]=t05+t15;
	    a[j1+p15]=t04-t14;	a[j1+1+p15]=t05-t15;
	    a[j1+p12]=t24-t35;	a[j1+1+p12]=t25+t34;	/* mpy by I is inlined here...	*/
	    a[j1+p13]=t24+t35;	a[j1+1+p13]=t25-t34;
#if PFETCH
add = add0+p14;
prefetch_p_doubles(add);
#endif
/*...Block 4: */
	    rt =t26;	t26=t06-rt;	t06=t06+rt;
	    it =t27;	t27=t07-it;	t07=t07+it;

	    rt =t36;	t36=t16-rt;	t16=t16+rt;
	    it =t37;	t37=t17-it;	t17=t17+it;
#if PFETCH
add = add0+p15;
prefetch_p_doubles(add);
#endif
	    a[j1+p07]=t06+t16;	a[j1+1+p07]=t07+t17;
	    a[j1+p06]=t06-t16;	a[j1+1+p06]=t07-t17;
	    a[j1+p05]=t26-t37;	a[j1+1+p05]=t27+t36;	/* mpy by I is inlined here...	*/
	    a[j1+p04]=t26+t37;	a[j1+1+p04]=t27-t36;
#if PFETCH
add = add0+p16;
prefetch_p_doubles(add);
#endif
/*...Block 5: */
	    rt =t28;	t28=t08-rt;	t08=t08+rt;
	    it =t29;	t29=t09-it;	t09=t09+it;

	    rt =t38;	t38=t18-rt;	t18=t18+rt;
	    it =t39;	t39=t19-it;	t19=t19+it;
#if PFETCH
add = add0+p17;
prefetch_p_doubles(add);
#endif
	    a[j1+p09]=t08+t18;	a[j1+1+p09]=t09+t19;
	    a[j1+p08]=t08-t18;	a[j1+1+p08]=t09-t19;
	    a[j1+p10]=t28-t39;	a[j1+1+p10]=t29+t38;	/* mpy by I is inlined here...	*/
	    a[j1+p11]=t28+t39;	a[j1+1+p11]=t29-t38;
#if PFETCH
add = add0+p18;
prefetch_p_doubles(add);
#endif
#if PFETCH
add = add0+p19;
prefetch_p_doubles(add);
#endif

#endif	/* endif(1) */

	  iroot += root_incr;		/* increment sincos index.	*/

	  }

	  jstart += nwt;
	  jhi    += nwt;
	  col += 20;
	  co3 -= 20;

	}

	if(root_incr==0)break;

/*   Wraparound carry cleanup loop is here: ***
!
!   (1) Invert the radix-36 forward DIF FFT of the first block of 36 complex elements in A and unweight;
!   (2) Propagate cleanup carries among the real and imaginary parts of the 36 outputs of (1);
!   (3) Reweight and perform a radix-36 forward DIF FFT on the result of (2);
!   (4) If any of the exit carries from (2) are nonzero, advance to the next 36 elements and repeat (1-4).
*/
	t00 = cy19;
	cy19= cy18;
	cy18= cy17;
	cy17= cy16;
	cy16= cy15;
	cy15= cy14;
	cy14= cy13;
	cy13= cy12;
	cy12= cy11;
	cy11= cy10;
	cy10= cy09;
	cy09= cy08;
	cy08= cy07;
	cy07= cy06;
	cy06= cy05;
	cy05= cy04;
	cy04= cy03;
	cy03= cy02;
	cy02= cy01;
	cy01= cy00;
	cy00= t00 ;

	iroot = 0;
	root_incr = 0;
	scale = 1;

	jstart = 0;
	jhi = 3;
	khi = 1;

	for(j=0; j<=jhi; j++)
	{
	  a[j    ] *= radix_inv;
	  a[j+p01] *= radix_inv;
	  a[j+p02] *= radix_inv;
	  a[j+p03] *= radix_inv;
	  a[j+p04] *= radix_inv;
	  a[j+p05] *= radix_inv;
	  a[j+p06] *= radix_inv;
	  a[j+p07] *= radix_inv;
	  a[j+p08] *= radix_inv;
	  a[j+p09] *= radix_inv;
	  a[j+p10] *= radix_inv;
	  a[j+p11] *= radix_inv;
	  a[j+p12] *= radix_inv;
	  a[j+p13] *= radix_inv;
	  a[j+p14] *= radix_inv;
	  a[j+p15] *= radix_inv;
	  a[j+p16] *= radix_inv;
	  a[j+p17] *= radix_inv;
	  a[j+p18] *= radix_inv;
	  a[j+p19] *= radix_inv;
	}
}

	if(fabs(cy00)+fabs(cy01)+fabs(cy02)+fabs(cy03)+fabs(cy04)+fabs(cy05)+fabs(cy06)+fabs(cy07)+fabs(cy08)+fabs(cy09)+fabs(cy10)+fabs(cy11)+fabs(cy12)+fabs(cy13)+fabs(cy14)+fabs(cy15)+fabs(cy16)+fabs(cy17)+fabs(cy18)+fabs(cy19) != 0.0)
	{
	    if(INTERACT)printf("FATAL: iter = %10d; nonzero exit carry in radix20_ditN_cy_dif1.\n",iter);
	    f1 = fopen(   OFILE,"a");
	    f2 = fopen(STATFILE,"a");
	    fprintf(f1,"FATAL: iter = %10d; nonzero exit carry in radix20_ditN_cy_dif1.\n",iter);
	    fprintf(f2,"FATAL: iter = %10d; nonzero exit carry in radix20_ditN_cy_dif1.\n",iter);
	    fclose(f1);
	    fclose(f2);
	    err=1;
	    return(err);
	}

	return(0);
}

/***************/

void radix20_dit_pass1(double a[], int n)
{
/*
!...Acronym: DIT = Decimation In Time
!
!...Subroutine to perform an initial radix-20 complex DIT FFT pass on the data in the length-N real vector A.
!
!   This routine is designed exclusively to undo the effects of radix20_dif_pass1,
!   i.e. to reobtain the raw all-integer residue vector at the end of an iteration cycle.
*/
	int j,j1;
	static int n20, first_entry=TRUE
		    ,p01,p02,p03,p04,p05,p06,p07,p08,p09,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19;
	static double cc1 = -1.25000000000000000000,	/* [cos(u)+cos(2u)]/2-1 = -5/4; u = 2*pi/5 */
		      cc2 =  0.55901699437494742409,	/* [cos(u)-cos(2u)]/2 */
		      s2  =  0.95105651629515357211,	/*  sin(u) */
		      ss1 =  1.53884176858762670130,	/* [sin(u)+sin(2u)] */
		      ss2 =  0.36327126400268044292;	/* [sin(u)-sin(2u)] */
	double rt,it
		,t00,t01,t02,t03,t04,t05,t06,t07,t08,t09
		,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19
		,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29
		,t30,t31,t32,t33,t34,t35,t36,t37,t38,t39;

	if(!first_entry && (n/20) != n20)	/* New runlength?	*/
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  first_entry=FALSE;
	  n20=n/20;

/*   constant index offsets for array load/stores are here.	*/

	  p01 = n20;
	  p02 = p01 + p01;
	  p03 = p02 + p01;
	  p04 = p03 + p01;
	  p05 = p04 + p01;
	  p06 = p05 + p01;
	  p07 = p06 + p01;
	  p08 = p07 + p01;
	  p09 = p08 + p01;
	  p10 = p09 + p01;
	  p11 = p10 + p01;
	  p12 = p11 + p01;
	  p13 = p12 + p01;
	  p14 = p13 + p01;
	  p15 = p14 + p01;
	  p16 = p15 + p01;
	  p17 = p16 + p01;
	  p18 = p17 + p01;
	  p19 = p18 + p01;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p05 = p05 + ( (p05 >> DAT_BITS) << PAD_BITS );
	  p06 = p06 + ( (p06 >> DAT_BITS) << PAD_BITS );
	  p07 = p07 + ( (p07 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p09 = p09 + ( (p09 >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p11 = p11 + ( (p11 >> DAT_BITS) << PAD_BITS );
	  p12 = p12 + ( (p12 >> DAT_BITS) << PAD_BITS );
	  p13 = p13 + ( (p13 >> DAT_BITS) << PAD_BITS );
	  p14 = p14 + ( (p14 >> DAT_BITS) << PAD_BITS );
	  p15 = p15 + ( (p15 >> DAT_BITS) << PAD_BITS );
	  p16 = p16 + ( (p16 >> DAT_BITS) << PAD_BITS );
	  p17 = p17 + ( (p17 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	  p19 = p19 + ( (p19 >> DAT_BITS) << PAD_BITS );
	}

/*...The radix-20 pass is here.	*/

      for(j=0; j < n20; j += 2)
      {
	  j1 = j  + ( (j  >> DAT_BITS) << PAD_BITS );
/*
Twiddleless version requires us to swap inputs as follows:
indices  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19
      -> 0,16,12, 8, 4,15,11, 7, 3,19,10, 6, 2,18,14, 5, 1,17,13, 9

I.e. start out with first quartet of indices {0,5,10,15}, permute those according to
  {0,5,10,15}*19%20 = {0,15,10,5}, then each is head of a length-5 list of indices with decrement 4

Remember, inputs to DIT are bit-reversed, so
a[0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19] contain
x[0,10, 5,15, 1,11, 6,16, 2,12, 7,17, 3,13, 8,18, 4,14, 9,19], which get swapped to
x[0,10,15, 5,16, 6,11, 1,12, 2, 7,17, 8,18, 3,13, 4,14,19, 9], which means the a-indices get swapped as
a[0, 1, 3, 2, 7, 6, 5, 4, 9, 8,10,11,14,15,12,13,16,17,19,18].
*/
#if 1
/*...gather the needed data (20 64-bit complex, i.e. 40 64-bit reals) and do 5 radix-4 transforms...*/
                     /*          inputs           */ /*                                      outputs                                      */
	RADIX_04_DIT(a[j1    ],a[j1+1    ],a[j1+p01],a[j1+1+p01],a[j1+p03],a[j1+1+p03],a[j1+p02],a[j1+1+p02],t00,t01,t10,t11,t20,t21,t30,t31,rt,it);
	RADIX_04_DIT(a[j1+p07],a[j1+1+p07],a[j1+p06],a[j1+1+p06],a[j1+p05],a[j1+1+p05],a[j1+p04],a[j1+1+p04],t02,t03,t12,t13,t22,t23,t32,t33,rt,it);
	RADIX_04_DIT(a[j1+p09],a[j1+1+p09],a[j1+p08],a[j1+1+p08],a[j1+p10],a[j1+1+p10],a[j1+p11],a[j1+1+p11],t04,t05,t14,t15,t24,t25,t34,t35,rt,it);
	RADIX_04_DIT(a[j1+p14],a[j1+1+p14],a[j1+p15],a[j1+1+p15],a[j1+p12],a[j1+1+p12],a[j1+p13],a[j1+1+p13],t06,t07,t16,t17,t26,t27,t36,t37,rt,it);
	RADIX_04_DIT(a[j1+p16],a[j1+1+p16],a[j1+p17],a[j1+1+p17],a[j1+p19],a[j1+1+p19],a[j1+p18],a[j1+1+p18],t08,t09,t18,t19,t28,t29,t38,t39,rt,it);

/*...and now do 4 radix-5 transforms...*/
                     /*                                                inputs                                                   */ /*                 outputs                   */
	RADIX_05_DFT(t00,t01,t02,t03,t04,t05,t06,t07,t08,t09,a[j1    ],a[j1+1    ],a[j1+p16],a[j1+1+p16],a[j1+p12],a[j1+1+p12],a[j1+p08],a[j1+1+p08],a[j1+p04],a[j1+1+p04],rt,it);
	RADIX_05_DFT(t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,a[j1+p15],a[j1+1+p15],a[j1+p11],a[j1+1+p11],a[j1+p07],a[j1+1+p07],a[j1+p03],a[j1+1+p03],a[j1+p19],a[j1+1+p19],rt,it);
	RADIX_05_DFT(t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,a[j1+p10],a[j1+1+p10],a[j1+p06],a[j1+1+p06],a[j1+p02],a[j1+1+p02],a[j1+p18],a[j1+1+p18],a[j1+p14],a[j1+1+p14],rt,it);
	RADIX_05_DFT(t30,t31,t32,t33,t34,t35,t36,t37,t38,t39,a[j1+p05],a[j1+1+p05],a[j1+p01],a[j1+1+p01],a[j1+p17],a[j1+1+p17],a[j1+p13],a[j1+1+p13],a[j1+p09],a[j1+1+p09],rt,it);
#else

/*...Block 1:	*/
	t00=a[j1    ];	t01=a[j1+1    ];
	rt =a[j1+p01];	it =a[j1+1+p01];
	t10=t00-rt;  	t00=t00+rt;
	t11=t01-it;	t01=t01+it;

	t20=a[j1+p03];	t21=a[j1+1+p03];
	rt =a[j1+p02];	it =a[j1+1+p02];
	t30=t20-rt;  	t20=t20+rt;
	t31=t21-it;  	t21=t21+it;

	rt =t20;	t20=t00-rt;	t00=t00+rt;
	it =t21;	t21=t01-it;	t01=t01+it;

	rt =t30;	t30=t10-t31;	t10=t10+t31;
			t31=t11+rt;	t11=t11-rt;

/*...Block 2:	*/
	t02=a[j1+p07];	t03=a[j1+1+p07];
	rt =a[j1+p06];	it =a[j1+1+p06];
	t12=t02-rt;  	t02=t02+rt;
	t13=t03-it;	t03=t03+it;

	t22=a[j1+p05];	t23=a[j1+1+p05];
	rt =a[j1+p04];	it =a[j1+1+p04];
	t32=t22-rt;  	t22=t22+rt;
	t33=t23-it;  	t23=t23+it;

	rt =t22;	t22=t02-rt;	t02=t02+rt;
	it =t23;	t23=t03-it;	t03=t03+it;

	rt =t32;	t32=t12-t33;	t12=t12+t33;
			t33=t13+rt;	t13=t13-rt;

/*...Block 3:	*/
	t04=a[j1+p09];	t05=a[j1+1+p09];
	rt =a[j1+p08];	it =a[j1+1+p08];
	t14=t04-rt;  	t04=t04+rt;
	t15=t05-it;	t05=t05+it;

	t24=a[j1+p10];	t25=a[j1+1+p10];
	rt =a[j1+p11];	it =a[j1+1+p11];
	t34=t24-rt;  	t24=t24+rt;
	t35=t25-it;  	t25=t25+it;

	rt =t24;	t24=t04-rt;	t04=t04+rt;
	it =t25;	t25=t05-it;	t05=t05+it;

	rt =t34;	t34=t14-t35;	t14=t14+t35;
			t35=t15+rt;	t15=t15-rt;

/*...Block 4:	*/
	t06=a[j1+p14];	t07=a[j1+1+p14];
	rt =a[j1+p15];	it =a[j1+1+p15];
	t16=t06-rt;  	t06=t06+rt;
	t17=t07-it;	t07=t07+it;

	t26=a[j1+p12];	t27=a[j1+1+p12];
	rt =a[j1+p13];	it =a[j1+1+p13];
	t36=t26-rt;  	t26=t26+rt;
	t37=t27-it;  	t27=t27+it;

	rt =t26;	t26=t06-rt;	t06=t06+rt;
	it =t27;	t27=t07-it;	t07=t07+it;

	rt =t36;	t36=t16-t37;	t16=t16+t37;
			t37=t17+rt;	t17=t17-rt;

/*...Block 5:	*/
	t08=a[j1+p16];	t09=a[j1+1+p16];
	rt =a[j1+p17];	it =a[j1+1+p17];
	t18=t08-rt;  	t08=t08+rt;
	t19=t09-it;	t09=t09+it;

	t28=a[j1+p19];	t29=a[j1+1+p19];
	rt =a[j1+p18];	it =a[j1+1+p18];
	t38=t28-rt;  	t28=t28+rt;
	t39=t29-it;  	t29=t29+it;

	rt =t28;	t28=t08-rt;	t08=t08+rt;
	it =t29;	t29=t09-it;	t09=t09+it;

	rt =t38;	t38=t18-t39;	t18=t18+t39;
			t39=t19+rt;	t19=t19-rt;

/*       ...and now do four radix-5 transforms.	*/
/*...First radix-5 transform:	*/
	    rt  =t08;			it  =t09;
	    t08 =t02 -rt;		t09 =t03 -it;
	    t02 =t02 +rt;		t03 =t03 +it;
	    rt  =t06;			it  =t07;
	    t06 =t04 -rt;		t07 =t05 -it;
	    t04 =t04 +rt;		t05 =t05 +it;
	    rt  = t02+t04;		it  = t03+t05;
	    t00 = t00+rt;		t01 = t01+it;
	    rt  = t00+cc1*rt;		it  = t01+cc1*it;
	    t04 = cc2*(t02-t04);	t05 = cc2*(t03-t05);
	    t02 = rt+t04;		t03 = it+t05;
	    t04 = rt-t04;		t05 = it-t05;
	    rt  = s2 *(t06-t08);	it  = s2 *(t07-t09);
	    t06 = ss1* t06;		t07 = ss1* t07;
	    t08 = ss2* t08;		t09 = ss2* t09;
	    t06 = rt-t06;		t07 = it-t07;
	    t08 = rt+t08;		t09 = it+t09;
	    a[j1    ]=t00;		a[j1+1    ]=t01;
	    a[j1+p04]=t02-t07;		a[j1+1+p04]=t03+t06;
	    a[j1+p08]=t04-t09;		a[j1+1+p08]=t05+t08;
	    a[j1+p12]=t04+t09;		a[j1+1+p12]=t05-t08;
	    a[j1+p16]=t02+t07;		a[j1+1+p16]=t03-t06;

/*...Second radix-5 transform:	*/
	    rt  =t18;			it  =t19;
	    t18 =t12 -rt;		t19 =t13 -it;
	    t12 =t12 +rt;		t13 =t13 +it;
	    rt  =t16;			it  =t17;
	    t16 =t14 -rt;		t17 =t15 -it;
	    t14 =t14 +rt;		t15 =t15 +it;
	    rt  = t12+t14;		it  = t13+t15;
	    t10 = t10+rt;		t11 = t11+it;
	    rt  = t10+cc1*rt;		it  = t11+cc1*it;
	    t14 = cc2*(t12-t14);	t15 = cc2*(t13-t15);
	    t12 = rt+t14;		t13 = it+t15;
	    t14 = rt-t14;		t15 = it-t15;
	    rt  = s2 *(t16-t18);	it  = s2 *(t17-t19);
	    t16 = ss1* t16;		t17 = ss1* t17;
	    t18 = ss2* t18;		t19 = ss2* t19;
	    t16 = rt-t16;		t17 = it-t17;
	    t18 = rt+t18;		t19 = it+t19;
	    a[j1+p15]=t10;		a[j1+1+p15]=t11;
	    a[j1+p19]=t12-t17;		a[j1+1+p19]=t13+t16;
	    a[j1+p03]=t14-t19;		a[j1+1+p03]=t15+t18;
	    a[j1+p07]=t14+t19;		a[j1+1+p07]=t15-t18;
	    a[j1+p11]=t12+t17;		a[j1+1+p11]=t13-t16;

/*...Third radix-5 transform:	*/
	    rt  =t28;			it  =t29;
	    t28 =t22 -rt;		t29 =t23 -it;
	    t22 =t22 +rt;		t23 =t23 +it;
	    rt  =t26;			it  =t27;
	    t26 =t24 -rt;		t27 =t25 -it;
	    t24 =t24 +rt;		t25 =t25 +it;
	    rt  = t22+t24;		it  = t23+t25;
	    t20 = t20+rt;		t21 = t21+it;
	    rt  = t20+cc1*rt;		it  = t21+cc1*it;
	    t24 = cc2*(t22-t24);	t25 = cc2*(t23-t25);
	    t22 = rt+t24;		t23 = it+t25;
	    t24 = rt-t24;		t25 = it-t25;
	    rt  = s2 *(t26-t28);	it  = s2 *(t27-t29);
	    t26 = ss1* t26;		t27 = ss1* t27;
	    t28 = ss2* t28;		t29 = ss2* t29;
	    t26 = rt-t26;		t27 = it-t27;
	    t28 = rt+t28;		t29 = it+t29;
	    a[j1+p10]=t20;		a[j1+1+p10]=t21;
	    a[j1+p14]=t22-t27;		a[j1+1+p14]=t23+t26;
	    a[j1+p18]=t24-t29;		a[j1+1+p18]=t25+t28;
	    a[j1+p02]=t24+t29;		a[j1+1+p02]=t25-t28;
	    a[j1+p06]=t22+t27;		a[j1+1+p06]=t23-t26;

/*...Fourth radix-5 transform:	*/
	    rt  =t38;			it  =t39;
	    t38 =t32 -rt;		t39 =t33 -it;
	    t32 =t32 +rt;		t33 =t33 +it;
	    rt  =t36;			it  =t37;
	    t36 =t34 -rt;		t37 =t35 -it;
	    t34 =t34 +rt;		t35 =t35 +it;
	    rt  = t32+t34;		it  = t33+t35;
	    t30 = t30+rt;		t31 = t31+it;
	    rt  = t30+cc1*rt;		it  = t31+cc1*it;
	    t34 = cc2*(t32-t34);	t35 = cc2*(t33-t35);
	    t32 = rt+t34;		t33 = it+t35;
	    t34 = rt-t34;		t35 = it-t35;
	    rt  = s2 *(t36-t38);	it  = s2 *(t37-t39);
	    t36 = ss1* t36;		t37 = ss1* t37;
	    t38 = ss2* t38;		t39 = ss2* t39;
	    t36 = rt-t36;		t37 = it-t37;
	    t38 = rt+t38;		t39 = it+t39;
	    a[j1+p05]=t30;		a[j1+1+p05]=t31;
	    a[j1+p09]=t32-t37;		a[j1+1+p09]=t33+t36;
	    a[j1+p13]=t34-t39;		a[j1+1+p13]=t35+t38;
	    a[j1+p17]=t34+t39;		a[j1+1+p17]=t35-t38;
	    a[j1+p01]=t32+t37;		a[j1+1+p01]=t33-t36;
#endif
	}
}

/* unpermuted-index output order is as follows:
ActualNeed                Actual                             Need
---- ------ -------------------------------    -------------------------------
   0 (   0)   107.0000000000   -87.0000000000   107.0000000000   -87.0000000000
   1 (  04)     3.8975770799     1.7742323052   -11.8647898507    -6.5977392870
   2 (  08)    14.5070279402    12.6894179137    -2.8854468251    23.7084688127
   3 (  12)   -12.7430959177   -10.3074519025     8.1391134275     9.5889261410
   4 (  16)     2.3384908976     2.8438016835     3.8975770799     1.7742323052
*
   5 (  15)   -14.0000000000    12.0000000000   -12.0000000000    16.0000000000
   6 (  19)    -1.5023846915    -1.7166548059    10.4000280840    11.4574634246
   7 (  03)     8.1391134275     9.5889261410    10.0412264600   -19.7348241747
   8 (  07)    10.0412264600   -19.7348241747    14.5070279402    12.6894179137
   9 (  11)    -2.6779551960   -15.1374471603     9.5729937832    25.6321811408
*
  10 (  10)   -29.0000000000    11.0000000000   -29.0000000000    11.0000000000
  11 (  14)    -6.1639601065     0.2162986542    -2.6779551960   -15.1374471603
  12 (  18)     2.6493788476     3.6177691085   -12.7430959177   -10.3074519025
  13 (  02)    -2.8854468251    23.7084688127   -20.9443792061    -5.5722500297
  14 (  06)    10.4000280840    11.4574634246    -6.1639601065     0.2162986542
*
  15 (  05)   -12.0000000000    16.0000000000   -14.0000000000    12.0000000000
  16 (  09)     9.5729937832    25.6321811408     2.3384908976     2.8438016835
  17 (  13)   -20.9443792061    -5.5722500297     5.2361752736    -4.4621918240
  18 (  17)     5.2361752736    -4.4621918240     2.6493788476     3.6177691085
  19 (  01)   -11.8647898507    -6.5977392870    -1.5023846915    -1.7166548059
*/
