/*******************************************************************************
*                                                                              *
*   (C) 1997-2002 by Ernst W. Mayer.                                           *
*                                                                              *
*  This program is free software; you can redistribute it and/or modify it     *
*  under the terms of the GNU General Public License as published by the       *
*  Free Software Foundation; either version 2 of the License, or (at your      *
*  option) any later version.                                                  *
*                                                                              *
*  This program is distributed in the hope that it will be useful, but WITHOUT *
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
*  more details.                                                               *
*                                                                              *
*  You should have received a copy of the GNU General Public License along     *
*  with this program; see the file GPL.txt.  If not, you may view one at       *
*  http://www.fsf.org/licenses/licenses.html, or obtain one by writing to the  *
*  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     *
*  02111-1307, USA.                                                            *
*                                                                              *
*******************************************************************************/

#include "Mlucas.h"

/* Use for testing higher-accuracy version of the twiddles computation */
#define HIACC 1

/***************/

void radix8_dif_pass(double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int index[], int nloops, int incr)
{
/*
!...Acronym: DIF = Decimation In Frequency
!
!...Subroutine to perform a single radix-8 complex DIF FFT pass on the data in the length-N real vector A.
!
!   See the documentation in radix16_dif_pass for further details.
*/
	int i,j,j1,jlo,jhi,k,m,iroot_prim,iroot,k1,k2,nrtm1,p1,p2,p3,p4,p5,p6,p7;
	double rt,it
		,c1,c2,c3,c4,c5,c6,c7
		,s1,s2,s3,s4,s5,s6,s7
		,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16;
#if PFETCH
	double *add0, *add;
#endif
#ifdef OKTHREADS
	/*int CHUNKSIZE = (incr>>3)/NTHREAD;*/
#endif
/*
!...The radix-8 pass is here. The data are processed in NLOOPS blocks of INCR real elements each,
!   where mloops = product of all preceding radices. The fundamental root of unity for this pass
!   is thus the primitive [(product of all preceding radices)*radix_now] =  [mloops*radix_now]th root,
!   and we get the proper table look-up index by multiplying the BR index by N/[mloops*radix_now] = (incr/2)/radix_now.
*/

	/*  stride between elements processed together. */

	nrtm1 = nrt - 1;

	p1 = incr>>3;
	p2 = p1 + p1;
	p3 = p2 + p1;
	p4 = p3 + p1;
	p5 = p4 + p1;
	p6 = p5 + p1;
	p7 = p6 + p1;

	p1 = p1 + ( (p1 >> DAT_BITS) << PAD_BITS );
	p2 = p2 + ( (p2 >> DAT_BITS) << PAD_BITS );
	p3 = p3 + ( (p3 >> DAT_BITS) << PAD_BITS );
	p4 = p4 + ( (p4 >> DAT_BITS) << PAD_BITS );
	p5 = p5 + ( (p5 >> DAT_BITS) << PAD_BITS );
	p6 = p6 + ( (p6 >> DAT_BITS) << PAD_BITS );
	p7 = p7 + ( (p7 >> DAT_BITS) << PAD_BITS );

	k=0;			/* init sincos array index */
	jlo=0; jhi=(incr >> 3);	/* init inner loop bounds */
	iroot_prim=(incr >> 4);	/* (incr/2)/radix_now */
/*
printf("radix8_dif_pass: nloops = %d  incr = %d  using %d th roots of unity.\n",nloops,incr,n/(2*iroot_prim));
*/
	for(m=0; m < nloops; m++)
	{
/*	here are the needed sincos data - these are processed below in bit-reversed order. */
	  iroot = index[m]*iroot_prim;
	  i = iroot;
/* In the DIF pass we may be able to afford a more-accurate twiddles computation,
   since we only compute each set of twiddles once and re-use it for many data blocks. */
#if HIACC
	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 2*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c1=t1*t3-t2*t4;	s1=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 3*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c2=t1*t3-t2*t4;	s2=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 4*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c3=t1*t3-t2*t4;	s3=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 5*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c4=t1*t3-t2*t4;	s4=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 6*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c5=t1*t3-t2*t4;	s5=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 7*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c6=t1*t3-t2*t4;	s6=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c7=t1*t3-t2*t4;	s7=t1*t4+t2*t3;
#else
	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c1=t1*t3-t2*t4;	s1=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += i + iroot;
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c2=t1*t3-t2*t4;	s2=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c5=t1*t3-t2*t4;	s5=t1*t4+t2*t3;

	    t1=c1*c5; t2=c1*s5; t3=s1*c5; t4=s1*s5;
	    c4=t1+t4; s4=t2-t3; c6=t1-t4; s6=t2+t3;

	    t1=c2*c5; t2=c2*s5; t3=s2*c5; t4=s2*s5;
	    c3=t1+t4; s3=t2-t3; c7=t1-t4; s7=t2+t3;
#endif
	/* Define the inner-loop parameters in terms of the outer-loop ones to make OpenMP's job easier: */
	  jlo = m*incr;
	  jhi = jlo+(incr >> 3);

/* Loop control depends on whether we're multithreading: */
#ifdef OKTHREADS
    #pragma omp parallel default(shared)
    {
    /*
    #pragma omp parallel shared(CHUNKSIZE,a,) private(j,j1,add0,add)
    {
    */
	NTHREADS = omp_get_num_procs();
if(jlo == 0)printf("NTHREADS = %d\n", NTHREADS);
	omp_set_dynamic( NTHREADS );
	omp_set_num_threads( NTHREADS );

	/*#pragma omp for schedule(dynamic,CHUNKSIZE) nowait*/
	#pragma omp for schedule(dynamic) nowait
	{
#endif
	  for(j=jlo; j < jhi; j += 2)
	  {
	    j1=j + ( (j >> DAT_BITS) << PAD_BITS );	/* padded-array fetch index is here */

/*       gather the needed data (8 64-bit complex, i.e. 16 64-bit reals) and combine to get the 4 length-2 transforms... */
#if PFETCH
add0 = &a[j1];
prefetch_p_doubles(add0);
#endif
	    t1 =a[j1   ];			t2 =a[j1+1   ];				/* r^0*x0 */
	    rt =a[j1+p4]*c4 -a[j1+1+p4]*s4;	it =a[j1+1+p4]*c4 +a[j1+p4]*s4;		/* r^4*x4 */
	    t3 =t1 -rt;				t4 =t2 -it;				/* y1 = x0-r^4*x4 */
	    t1 =t1 +rt;				t2 =t2 +it;				/* y0 = x0+r^4*x4 */
#if PFETCH
add = add0+p1;
prefetch_p_doubles(add);
#endif
	    t5 =a[j1+p2]*c2 -a[j1+1+p2]*s2;	t6 =a[j1+1+p2]*c2 +a[j1+p2]*s2;		/* r^2*x2 */
	    rt =a[j1+p6]*c6 -a[j1+1+p6]*s6;	it =a[j1+1+p6]*c6 +a[j1+p6]*s6;		/* r^6*x6 */
	    t7 =t5 -rt;				t8 =t6 -it;				/* r^2*y3 = r^2*(x2-r^4*x6) */
	    t5 =t5 +rt;				t6 =t6 +it;				/* r^2*y2 = r^2*(x2+r^4*x6) */
#if PFETCH
add = add0+p2;
prefetch_p_doubles(add);
#endif
	    t9 =a[j1+p1]*c1 -a[j1+1+p1]*s1;	t10=a[j1+1+p1]*c1 +a[j1+p1]*s1;		/* r^1*x1 */
	    rt =a[j1+p5]*c5 -a[j1+1+p5]*s5;	it =a[j1+1+p5]*c5 +a[j1+p5]*s5;		/* r^5*x5 */
	    t11=t9 -rt;				t12=t10-it;				/* r^1*y5 = r^1*(x1-r^4*x5) */
	    t9 =t9 +rt;				t10=t10+it;				/* r^1*y4 = r^1*(x1+r^4*x5) */
#if PFETCH
add = add0+p3;
prefetch_p_doubles(add);
#endif
	    t13=a[j1+p3]*c3 -a[j1+1+p3]*s3;	t14=a[j1+1+p3]*c3 +a[j1+p3]*s3;		/* r^3*x3 */
	    rt =a[j1+p7]*c7 -a[j1+1+p7]*s7;	it =a[j1+1+p7]*c7 +a[j1+p7]*s7;		/* r^7*x7 */
	    t15=t13-rt;				t16=t14-it;				/* r^3*y7 = r^3*(x3-r^4*x7) */
	    t13=t13+rt;				t14=t14+it;				/* r^3*y6 = r^3*(x3+r^4*x7) */
#if PFETCH
add = add0+p4;
prefetch_p_doubles(add);
#endif
/*       combine to get the 2 length-4 transforms... */
	    rt =t5;				it =t6;					/* r^2*y2 */
	    t5 =t1 -rt;				t6 =t2 -it;				/* z1 = y0-r^2*y2 */
	    t1 =t1 +rt;				t2 =t2 +it;				/* z0 = y0-r^2*y2 */

	    rt =t7;				it =t8;					/* r^2*y3 */
	    t7 =t3 +it;				t8 =t4 -rt;				/* z3 = y1-r^2*y3*I */
	    t3 =t3 -it;				t4 =t4 +rt;				/* z2 = y1+r^2*y3*I */
#if PFETCH
add = add0+p5;
prefetch_p_doubles(add);
#endif
	    rt =t13;				it =t14;				/* r^3*y6 */
	    t13=t9 -rt;				t14=t10-it;				/* r^1*z5 = r^1*(y4-r^2*y6) */
	    t9 =t9 +rt;				t10=t10+it;				/* r^1*z4 = r^1*(y4+r^2*y6) */

	    rt =t15;				it =t16;					/* r^3*y7 */
	    t15=t11+it;				t16=t12-rt;				/* r^1*z7 = r^1*(y5-r^4*y7*I) */
	    t11=t11-it;				t12=t12+rt;				/* r^1*z6 = r^1*(y5+r^4*y7*I) */
#if PFETCH
add = add0+p6;
prefetch_p_doubles(add);
#endif
/*       now combine the two half-transforms */
	    a[j1   ]=t1+t9;			a[j1+1   ]=t2+t10;
	    a[j1+p1]=t1-t9;			a[j1+1+p1]=t2-t10;

	    a[j1+p2]=t5-t14;			a[j1+1+p2]=t6+t13;
	    a[j1+p3]=t5+t14;			a[j1+1+p3]=t6-t13;
#if PFETCH
add = add0+p7;
prefetch_p_doubles(add);
#endif
	    rt =(t11-t12)*ISRT2;		it =(t11+t12)*ISRT2;
	    a[j1+p4]=t3+rt;			a[j1+1+p4]=t4+it;
	    a[j1+p5]=t3-rt;			a[j1+1+p5]=t4-it;

	    rt =(t15+t16)*ISRT2;		it =(t16-t15)*ISRT2;
	    a[j1+p6]=t7-rt;			a[j1+1+p6]=t8-it;
	    a[j1+p7]=t7+rt;			a[j1+1+p7]=t8+it;

	  }
#ifdef OKTHREADS
	}
    }
#endif
	  /*jlo=jlo+incr; jhi=jhi+incr; See my note about OpenMP above. */
	}
/*
printf("\n");
printf("exiting radix8_dif_pass with jlo = %d\n",jlo);
*/
}

/***************/

void radix8_dit_pass (double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int nloops, int incr)
{
/*
!...Acronym: DIT = Decimation In Time
!
!...Subroutine to perform a single radix-8 complex inverse DIT FFT pass on the data in the length-N real vector A.
!
!   See the documentation in radix16_dif_pass for further details.
*/

	int i,j,j1,jlo,jhi,k,m,iroot_prim,iroot,k1,k2,nrtm1,p1,p2,p3,p4,p5,p6,p7;
	double rt,it
		,c1,c2,c3,c4,c5,c6,c7
		,s1,s2,s3,s4,s5,s6,s7
		,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16;
#if PFETCH
	double *add0, *add;
#endif
/*
!...The radix-8 pass is here.
!   stride between elements processed together.
*/
	nrtm1 = nrt - 1;

	p1 = incr>>3;
	p2 = p1 + p1;
	p3 = p2 + p1;
	p4 = p3 + p1;
	p5 = p4 + p1;
	p6 = p5 + p1;
	p7 = p6 + p1;

	p1 = p1 + ( (p1 >> DAT_BITS) << PAD_BITS );
	p2 = p2 + ( (p2 >> DAT_BITS) << PAD_BITS );
	p3 = p3 + ( (p3 >> DAT_BITS) << PAD_BITS );
	p4 = p4 + ( (p4 >> DAT_BITS) << PAD_BITS );
	p5 = p5 + ( (p5 >> DAT_BITS) << PAD_BITS );
	p6 = p6 + ( (p6 >> DAT_BITS) << PAD_BITS );
	p7 = p7 + ( (p7 >> DAT_BITS) << PAD_BITS );

	k=0;				/* init sincos array index */
	jlo=0; jhi=(incr >> 3);	/* init inner loop bounds */
/*	iroot_prim = nloops;	replace this with something that doesn't depend on nloops, since the latter varies, depending on
				how many pieces we break the full-length array into for the purposes of the cache-friendly scheme. */
	iroot_prim = n/incr;
/*
printf("radix8_dit_pass: nloops = %d  incr = %d  using %d th roots of unity.\n",nloops,incr,n/(2*iroot_prim));
*/
	for(m=0; m < nloops; m++)
	{
	  for(j=jlo; j < jhi; j += 2)
	  {
	    j1 = j + ( (j >> DAT_BITS) << PAD_BITS );	/* padded-array fetch index is here */
/*
printf("%d...",k);
*/
/*	here are the needed sincos data - these will be processed below in bit-reversed order. */
	    iroot = k*iroot_prim;
	    i = iroot;

	      k1=(i & nrtm1);
	      k2=(i >> nrt_bits);
	      i += iroot;
	      t1=rt0[k1].re;	t2=rt0[k1].im;
	      t3=rt1[k2].re;	t4=rt1[k2].im;
	      c1=t1*t3-t2*t4;	s1=t1*t4+t2*t3;

	      k1=(i & nrtm1);
	      k2=(i >> nrt_bits);
	      i += i + iroot;
	      t1=rt0[k1].re;	t2=rt0[k1].im;
	      t3=rt1[k2].re;	t4=rt1[k2].im;
	      c2=t1*t3-t2*t4;	s2=t1*t4+t2*t3;

	      k1=(i & nrtm1);
	      k2=(i >> nrt_bits);
	      t1=rt0[k1].re;	t2=rt0[k1].im;
	      t3=rt1[k2].re;	t4=rt1[k2].im;
	      c5=t1*t3-t2*t4;	s5=t1*t4+t2*t3;

	      t1=c1*c5; t2=c1*s5; t3=s1*c5; t4=s1*s5;
	      c4=t1+t4; s4=t2-t3; c6=t1-t4; s6=t2+t3;

	      t1=c2*c5; t2=c2*s5; t3=s2*c5; t4=s2*s5;
	      c3=t1+t4; s3=t2-t3; c7=t1-t4; s7=t2+t3;

/*      gather the needed data (8 64-bit complex, i.e. 16 64-bit reals) and get the 4 length-2 transforms... */
#if PFETCH
add0 = &a[j1];
prefetch_p_doubles(add0);
#endif
	    t1 =a[j1   ];			t2 =a[j1+1   ];				/* r^0*x0 */
	    rt =a[j1+p1]*c4 +a[j1+1+p1]*s4;	it =a[j1+1+p1]*c4 -a[j1+p1]*s4;		/* r^4*x4 */
	    t3 =t1 -rt;				t4 =t2 -it;				/* r^0*y1 = r^0*x0 - r^4*x4 */
	    t1 =t1 +rt;				t2 =t2 +it;				/* r^0*y0 = r^0*x0 + r^4*x4 */
#if PFETCH
add = add0+p1;
prefetch_p_doubles(add);
#endif
	    t5 =a[j1+p2]*c2 +a[j1+1+p2]*s2;	t6 =a[j1+1+p2]*c2 -a[j1+p2]*s2;		/* r^2*x2 */
	    rt =a[j1+p3]*c6 +a[j1+1+p3]*s6;	it =a[j1+1+p3]*c6 -a[j1+p3]*s6;		/* r^6*x6 */
	    t7 =t5 -rt;				t8 =t6 -it;				/* r^2*y3 = r^2*x2 - r^6*x6 */
	    t5 =t5 +rt;				t6 =t6 +it;				/* r^2*y2 = r^2*x2 + r^6*x6 */
#if PFETCH
add = add0+p2;
prefetch_p_doubles(add);
#endif
	    t9 =a[j1+p4]*c1 +a[j1+1+p4]*s1;	t10=a[j1+1+p4]*c1 -a[j1+p4]*s1;		/* r^1*x1 */
	    rt =a[j1+p5]*c5 +a[j1+1+p5]*s5;	it =a[j1+1+p5]*c5 -a[j1+p5]*s5;		/* r^5*x5 */
	    t11=t9 -rt;				t12=t10-it;				/* r^1*y5 = r^1*x1 - r^5*x5 */
	    t9 =t9 +rt;				t10=t10+it;				/* r^1*y4 = r^1*x1 + r^5*x5 */
#if PFETCH
add = add0+p3;
prefetch_p_doubles(add);
#endif
	    t13=a[j1+p6]*c3 +a[j1+1+p6]*s3;	t14=a[j1+1+p6]*c3 -a[j1+p6]*s3;		/* r^3*x3 */
	    rt =a[j1+p7]*c7 +a[j1+1+p7]*s7;	it =a[j1+1+p7]*c7 -a[j1+p7]*s7;		/* r^7*x7 */
	    t15=t13-rt;				t16=t14-it;				/* r^3*y7 = r^3*x3 - r^7*x7 */
	    t13=t13+rt;				t14=t14+it;				/* r^3*y6 = r^3*x3 + r^7*x7 */
#if PFETCH
add = add0+p4;
prefetch_p_doubles(add);
#endif
/*      combine to get the 2 length-4 transform... */
	    rt =t5;	t5 =t1 -rt;	t1 =t1 +rt;
	    it =t6;	t6 =t2 -it;	t2 =t2 +it;

	    rt =t7;	t7 =t3 -t8;	t3 =t3 +t8;
			t8 =t4 +rt;	t4 =t4 -rt;
#if PFETCH
add = add0+p5;
prefetch_p_doubles(add);
#endif
	    rt =t13;	t13=t9 -rt;	t9 =t9 +rt;
	    it =t14;	t14=t10-it;	t10=t10+it;

	    rt =t15;	t15=t11-t16;	t11=t11+t16;
			t16=t12+rt;	t12=t12-rt;
#if PFETCH
add = add0+p6;
prefetch_p_doubles(add);
#endif
/*      now combine the two half-transforms */
	    a[j1   ]=t1+t9;			a[j1+1   ]=t2+t10;
	    a[j1+p4]=t1-t9;			a[j1+1+p4]=t2-t10;

	    rt=(t11+t12)*ISRT2;			it=(t11-t12)*ISRT2;
	    a[j1+p1]=t3+rt;			a[j1+1+p1]=t4-it;
	    a[j1+p5]=t3-rt;			a[j1+1+p5]=t4+it;
#if PFETCH
add = add0+p7;
prefetch_p_doubles(add);
#endif
	    a[j1+p2]=t5+t14;			a[j1+1+p2]=t6-t13;
	    a[j1+p6]=t5-t14;			a[j1+1+p6]=t6+t13;

	    rt=(t15-t16)*ISRT2;			it=(t15+t16)*ISRT2;
	    a[j1+p3]=t7-rt;			a[j1+1+p3]=t8-it;
	    a[j1+p7]=t7+rt;			a[j1+1+p7]=t8+it;

	    k=k+1; if(k == (incr >> 4)) k=0;	/* increment sincos array index; recycle whenever counter reaches kmax. */
	  }
	  jlo=jlo+incr; jhi=jhi+incr;
	}
/*
printf("\n");
printf("exiting radix8_dif_pass with jlo = %d\n",jlo);
*/
}

/***************/

void radix8_dit_passB(double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int index[], int nloops, int incr)
{
/*
!...Acronym: DIT = Decimation In Time
!
!...Subroutine to perform a single radix-8 complex DIT FFT pass on the data in the length-N real vector A.
!
!   See the documentation in radix16_dif_pass for further details.
*/
	int i,j,j1,jlo,jhi,k,m,iroot_prim,iroot,k1,k2,nrtm1,p1,p2,p3,p4,p5,p6,p7;
	double rt,it
		,c1,c2,c3,c4,c5,c6,c7
		,s1,s2,s3,s4,s5,s6,s7
		,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16;
#if PFETCH
	double *add0, *add;
#endif
/*
!...The radix-8 pass is here. The data are processed in NLOOPS blocks of INCR real elements each,
!   where mloops = product of all preceding radices. The fundamental root of unity for this pass
!   is thus the primitive [(product of all preceding radices)*radix_now] =  [mloops*radix_now]th root,
!   and we get the proper table look-up index by multiplying the BR index by N/[mloops*radix_now] = (incr/2)/radix_now.
*/

	/*  stride between elements processed together. */

	nrtm1 = nrt - 1;

	p1 = incr>>3;
	p2 = p1 + p1;
	p3 = p2 + p1;
	p4 = p3 + p1;
	p5 = p4 + p1;
	p6 = p5 + p1;
	p7 = p6 + p1;

	p1 = p1 + ( (p1 >> DAT_BITS) << PAD_BITS );
	p2 = p2 + ( (p2 >> DAT_BITS) << PAD_BITS );
	p3 = p3 + ( (p3 >> DAT_BITS) << PAD_BITS );
	p4 = p4 + ( (p4 >> DAT_BITS) << PAD_BITS );
	p5 = p5 + ( (p5 >> DAT_BITS) << PAD_BITS );
	p6 = p6 + ( (p6 >> DAT_BITS) << PAD_BITS );
	p7 = p7 + ( (p7 >> DAT_BITS) << PAD_BITS );

	k=0;				/* init sincos array index */
	jlo=0; jhi=(incr >> 3);		/* init inner loop bounds */
	iroot_prim=(incr >> 4);		/* (incr/2)/radix_now */
/*
printf("radix8_dit_passB: nloops = %d  incr = %d  using %d th roots of unity.\n",nloops,incr,n/(2*iroot_prim));
*/
	for(m=0; m < nloops; m++)
	{
/*
printf("%d...",k);
*/

/*!	here are the needed sincos data - these are processed below in bit-reversed order. */
	  iroot = index[k]*iroot_prim;
	  i = iroot;
/* In the DIF pass we may be able to afford a more-accurate twiddles computation,
   since we only compute each set of twiddles once and re-use it for many data blocks. */
#if HIACC
	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 2*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c1=t1*t3-t2*t4;	s1=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 3*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c2=t1*t3-t2*t4;	s2=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 4*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c3=t1*t3-t2*t4;	s3=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 5*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c4=t1*t3-t2*t4;	s4=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 6*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c5=t1*t3-t2*t4;	s5=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;			/* 7*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c6=t1*t3-t2*t4;	s6=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c7=t1*t3-t2*t4;	s7=t1*t4+t2*t3;
#else
	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += iroot;
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c1=t1*t3-t2*t4;	s1=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    i += i + iroot;
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c2=t1*t3-t2*t4;	s2=t1*t4+t2*t3;

	    k1=(i & nrtm1);
	    k2=(i >> nrt_bits);
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    t3=rt1[k2].re;	t4=rt1[k2].im;
	    c5=t1*t3-t2*t4;	s5=t1*t4+t2*t3;

	    t1=c1*c5; t2=c1*s5; t3=s1*c5; t4=s1*s5;
	    c4=t1+t4; s4=t2-t3; c6=t1-t4; s6=t2+t3;

	    t1=c2*c5; t2=c2*s5; t3=s2*c5; t4=s2*s5;
	    c3=t1+t4; s3=t2-t3; c7=t1-t4; s7=t2+t3;
#endif


/* Loop control depends on whether we're multithreading: */
#ifdef OKTHREADS
    #pragma omp parallel default(shared)
    {
	NTHREADS = omp_get_num_procs();
	omp_set_dynamic( NTHREADS );
	omp_set_num_threads( NTHREADS );

	#pragma omp for schedule(dynamic) nowait
	{
#endif
	  for(j=jlo; j < jhi; j += 2)
	  {
	    j1=j + ( (j >> DAT_BITS) << PAD_BITS );	/* padded-array fetch index is here */

/*      gather the needed data (8 64-bit complex, i.e. 16 64-bit reals) and get the 4 length-2 transforms... */
#if PFETCH
add0 = &a[j1];
prefetch_p_doubles(add0);
#endif
	    t1 =a[j1   ];	t2 =a[j1+1   ];		/* r^0*x0 */
	    rt =a[j1+p1];	it =a[j1+1+p1];		/* r^4*x4 */
	    t3 =t1 -rt;		t4 =t2 -it;		/* r^0*y1 = r^0*x0 - r^4*x4 */
	    t1 =t1 +rt;		t2 =t2 +it;		/* r^0*y0 = r^0*x0 + r^4*x4 */
#if PFETCH
add = add0+p1;
prefetch_p_doubles(add);
#endif
	    t5 =a[j1+p2];	t6 =a[j1+1+p2];		/* r^2*x2 */
	    rt =a[j1+p3];	it =a[j1+1+p3];		/* r^6*x6 */
	    t7 =t5 -rt;		t8 =t6 -it;		/* r^2*y3 = r^2*x2 - r^6*x6 */
	    t5 =t5 +rt;		t6 =t6 +it;		/* r^2*y2 = r^2*x2 + r^6*x6 */
#if PFETCH
add = add0+p2;
prefetch_p_doubles(add);
#endif
	    t9 =a[j1+p4];	t10=a[j1+1+p4];		/* r^1*x1 */
	    rt =a[j1+p5];	it =a[j1+1+p5];		/* r^5*x5 */
	    t11=t9 -rt;		t12=t10-it;		/* r^1*y5 = r^1*x1 - r^5*x5 */
	    t9 =t9 +rt;		t10=t10+it;		/* r^1*y4 = r^1*x1 + r^5*x5 */
#if PFETCH
add = add0+p3;
prefetch_p_doubles(add);
#endif
	    t13=a[j1+p6];	t14=a[j1+1+p6];		/* r^3*x3 */
	    rt =a[j1+p7];	it =a[j1+1+p7];		/* r^7*x7 */
	    t15=t13-rt;		t16=t14-it;		/* r^3*y7 = r^3*x3 - r^7*x7 */
	    t13=t13+rt;		t14=t14+it;		/* r^3*y6 = r^3*x3 + r^7*x7 */
#if PFETCH
add = add0+p4;
prefetch_p_doubles(add);
#endif
/*      combine to get the 2 length-4 transform... */
	    rt =t5;	t5 =t1 -rt;	t1 =t1 +rt;
	    it =t6;	t6 =t2 -it;	t2 =t2 +it;

	    rt =t7;	t7 =t3 -t8;	t3 =t3 +t8;
			t8 =t4 +rt;	t4 =t4 -rt;
#if PFETCH
add = add0+p5;
prefetch_p_doubles(add);
#endif
	    rt =t13;	t13=t9 -rt;	t9 =t9 +rt;
	    it =t14;	t14=t10-it;	t10=t10+it;

	    rt =t15;	t15=t11-t16;	t11=t11+t16;
			t16=t12+rt;	t12=t12-rt;
#if PFETCH
add = add0+p6;
prefetch_p_doubles(add);
#endif
/*      now combine the two half-transforms */
	    a[j1   ]=t1+t9;			a[j1+1   ]=t2+t10;
	    t1      =t1-t9;			t2        =t2-t10;
	    a[j1+p4]=t1 *c4 +t2 *s4;		a[j1+1+p4]=t2 *c4 -t1 *s4;

	    rt=(t11+t12)*ISRT2;			it=(t11-t12)*ISRT2;
	    t11     =t3+rt;			t12       =t4-it;
	    t3      =t3-rt;			t4        =t4+it;
	    a[j1+p1]=t11*c1 +t12*s1;		a[j1+1+p1]=t12*c1 -t11*s1;
	    a[j1+p5]=t3 *c5 +t4 *s5;		a[j1+1+p5]=t4 *c5 -t3 *s5;
#if PFETCH
add = add0+p7;
prefetch_p_doubles(add);
#endif
	    rt      =t5+t14;			it        =t6-t13;
	    t5      =t5-t14;			t6        =t6+t13;
	    a[j1+p2]=rt *c2 +it *s2;		a[j1+1+p2]=it *c2 -rt *s2;
	    a[j1+p6]=t5 *c6 +t6 *s6;		a[j1+1+p6]=t6 *c6 -t5 *s6;

	    rt=(t15-t16)*ISRT2;			it=(t15+t16)*ISRT2;
	    t15     =t7-rt;			t16       =t8-it;
	    t7      =t7+rt;			t8        =t8+it;
	    a[j1+p3]=t15*c3 +t16*s3;		a[j1+1+p3]=t16*c3 -t15*s3;
	    a[j1+p7]=t7 *c7 +t8 *s7;		a[j1+1+p7]=t8 *c7 -t7 *s7;
	  }
#ifdef OKTHREADS
	}
    }
#endif
	  k=k+1;
	  jlo=jlo+incr; jhi=jhi+incr;
	}
/*
printf("\n");
printf("exiting radix8_dit_passB with jlo = %d\n",jlo);
*/
}

