/*******************************************************************************
*                                                                              *
*   (C) 1997-2002 by Ernst W. Mayer.                                           *
*                                                                              *
*  This program is free software; you can redistribute it and/or modify it     *
*  under the terms of the GNU General Public License as published by the       *
*  Free Software Foundation; either version 2 of the License, or (at your      *
*  option) any later version.                                                  *
*                                                                              *
*  This program is distributed in the hope that it will be useful, but WITHOUT *
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
*  more details.                                                               *
*                                                                              *
*  You should have received a copy of the GNU General Public License along     *
*  with this program; see the file GPL.txt.  If not, you may view one at       *
*  http://www.fsf.org/licenses/licenses.html, or obtain one by writing to the  *
*  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     *
*  02111-1307, USA.                                                            *
*                                                                              *
*******************************************************************************/

/***************/

#include "Mlucas.h"

void radix16_wrapper_square(double a[], int n, int radix0, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int nradices_prim, int radix_prim[], int ws_i, int ws_j1, int ws_j2, int ws_j2_start, int ws_k, int ws_m, int ws_blocklen, int ws_blocklen_sum)
{

/*
!   NOTE: In the following commentary, N refers to the COMPLEX vector length (n2 in the code),
!   which is half the real vector length.
!
!...Acronyms:   DIT = Decimation In Time
!               DIF = Decimation In Frequency
!               FFT = Fast Fourier Transform, i.e. a discrete FT over the complex numbers
!
!...Complex--->Real wrapper for data returned by forward transform, pointwise squaring and inverse wrapper
!   prior to inverse transform, all performed in one fell swoop (or better, oneswell loop :)
!
!   A contains floating data resulting from a forward DIF FFT, i.e. in bit-reversed order.
!
!   Quite a few operations are eliminated via use of algebraic identities in
!   the derivation of the squaring formula: the old version takes (ordered)
!   forward FFT terms H[0,...,N-1] and uses the following formula
!   to generate the terms F[0,...,N-1] of the true DFT:
!
!   F[j] = {( H[j]+H~[N-j] ) - I*exp(+pi*I*j/N)*( H[j]+H~[N-j] )}/2,
!
!       j=0, ... , N-1.
!
!   Then one does a pointwise squaring to obtain terms G[0,...,N-1],
!   and does an inverse complex wrapper prior to entering the inverse FFT:
!
!   I[j] = {( G[j]+G~[N-j] ) + I*exp(-pi*I*j/N)*( G[j]-G~[N-j] )}/2,
!
!       j=0, ... , N-1.
!
!   By combining the 3 steps into one and doing some algebra, one
!   can get directly from the H's to the I's via
!
!   I[j] = H[j]^2 + {1 + exp(2*pi*I*j/N)}*{H[j]-H~[N-j]}^2/4,
!
!       j=0, ... , N-1,
!
!   and of course there are lots of symmetries in the calculation of
!   the data pairs I[j] and I[N-j] which can be (and are) exploited.
!
!   When the data in question are bit-reversed rather than ordered, it is
!   a nontrivial problem as to how to effect the above in a cache-friendly manner.
!   The problem reduces to figuring out what happens to (j, N-j) index pairs
!   under bit-reversal reordering. Happily this problem has an elegant solution,
!   which amounts to bit-reversing the array of roots of unity (i.e. the exp(2*pi*I*j/N)
!   in the last expression above, since this issue only affects the floating data)
!   and processing the bit-reversed data in the A-array (the floating-point H[j]'s)
!   in a sequence of smaller (but contiguous) sub-blocks - in each of the sub-blocks,
!   one array index increases monotonically, and a second (pointing to the bit-reversed
!   H[N-j]'s) decreases monotonically. Thus, things are very much like in the ordered-
!   data (j, N-j) index scheme, except that we have O(log2(N)) blocks, and the (k)th block
!   has index pairs (j, N'(k)-j). Further details on this may be found below, under the
!   heading "SOLVING THE CACHE FLOW PROBLEM...".
*/
	static int *index = 0x0, *index_ptmp = 0x0, *itmp = 0x0;	/* ptrs to Bit-reversal index array and temporary integer array; allocate later */
	int idum,j1pad,j2pad,kp,l,iroot,k1,k2;
	int i,j1,j2,j2_start,k,m,blocklen,blocklen_sum;
	static int n2,n16,nrtm1,ndivrad0m1;
	static double c = 0.9238795325112867561281831, s = 0.3826834323650897717284599;/* exp[i*(twopi/16)]     */
	double rt,it,re = 0.0, im= 0.0
	,cA1,cA2,cA3,cA4,cA5,cA6,cA7,cA8,cA9,cA10,cA11,cA12,cA13,cA14,cA15
	,sA1,sA2,sA3,sA4,sA5,sA6,sA7,sA8,sA9,sA10,sA11,sA12,sA13,sA14,sA15
	,cB1,cB2,cB3,cB4,cB5,cB6,cB7,cB8,cB9,cB10,cB11,cB12,cB13,cB14,cB15
	,sB1,sB2,sB3,sB4,sB5,sB6,sB7,sB8,sB9,sB10,sB11,sB12,sB13,sB14,sB15
	,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,t30,t31,t32
	,aj1p0r,aj1p1r,aj1p2r,aj1p3r,aj1p4r,aj1p5r,aj1p6r,aj1p7r,aj1p8r,aj1p9r,aj1p10r,aj1p11r,aj1p12r,aj1p13r,aj1p14r,aj1p15r
	,aj1p0i,aj1p1i,aj1p2i,aj1p3i,aj1p4i,aj1p5i,aj1p6i,aj1p7i,aj1p8i,aj1p9i,aj1p10i,aj1p11i,aj1p12i,aj1p13i,aj1p14i,aj1p15i
	,aj2p0r,aj2p1r,aj2p2r,aj2p3r,aj2p4r,aj2p5r,aj2p6r,aj2p7r,aj2p8r,aj2p9r,aj2p10r,aj2p11r,aj2p12r,aj2p13r,aj2p14r,aj2p15r
	,aj2p0i,aj2p1i,aj2p2i,aj2p3i,aj2p4i,aj2p5i,aj2p6i,aj2p7i,aj2p8i,aj2p9i,aj2p10i,aj2p11i,aj2p12i,aj2p13i,aj2p14i,aj2p15i;
#if PFETCH
	double *add0;
#endif

	static int first_entry=TRUE;

/*...If a new runlength, deallocate the index-array and set first_entry to true.
 */

	if(!first_entry && (n >> 4) != n16)     /* New runlength? */
	{
	  free((void *)index_ptmp);	index_ptmp=0x0;	index=0x0;
	  first_entry=TRUE;

/****************/
#if defined(USE_THREADS) && defined(_OPENMP) && defined(DEBUG_OMP)
	/* Obtain and print thread id */
	int tid, nthreads;
	tid = omp_get_thread_num();
	printf("radix16_wrapper_square: freeing memory from thread %2d\n", tid);

	/* Only master thread does this */
	if (tid == 0)
	{
		nthreads = omp_get_num_threads();
		printf("radix16_wrapper_square: Number of threads = %d\n", nthreads);
	}
	else
	{
		assert(0);	/* In shared-memory model, only master thread should init anything. */
	}
#endif
/****************/

	}

/*...initialize things upon first entry */

	if(first_entry)
	{

/****************/
#if defined(USE_THREADS) && defined(_OPENMP) && defined(DEBUG_OMP)
	/* Obtain and print thread id */
	int tid, nthreads;
	tid = omp_get_thread_num();
	printf("radix16_wrapper_square: initialization from thread %2d\n", tid);

	/* Only master thread does this */
	if (tid == 0)
	{
		nthreads = omp_get_num_threads();
		printf("radix16_wrapper_square: Number of threads = %d\n", nthreads);
	}
	else
	{
		assert(0);	/* In shared-memory model, only master thread should init anything. */
	}
#endif
/****************/

	  first_entry=FALSE;
	  n2  = n/2;
	  n16 = n/16;
	  nrtm1 = nrt - 1;
	  ndivrad0m1 = n/radix0 - 1;

/*...******Forward FFT****** final-pass trig array is here: first, calculate the needed dimension... */
/*
!...Final forward (DIF) FFT pass sincos data start out in bit-reversed order.
!   Allocate and initialize an index array containing N/16 indices...
*/
	  itmp = (int *)calloc(n2/16, sizeof(int));
	  if(!itmp){ printf("FATAL: unable to allocate array ITMP in radix16_wrapper_square.\n"); exit(EXIT_FAILURE); }

	  for(i=0; i < n2/16; i++)
	  {
	    itmp[i]=i;
	  }

/*...then bit-reverse INDEX with respect to N/16. Radices are sent to BR routine
 in opposite order from which they are processed in forward FFT. */
/*
	  radix_tmp = (int *)calloc(nradices_prim-4, sizeof(int));
	  if(!radix_tmp){ printf("FATAL: unable to allocate array RADIX_TMP in radix16_wrapper_square.\n"); exit(EXIT_FAILURE); }

	  for(i=0; i < nradices_prim-4; i++)
	  {
	    radix_tmp[i] = radix_prim[nradices_prim-5-i];
	  }

	  bit_reverse_int(itmp, n2/16, nradices_prim-4, &radix_tmp [0   ], +1);
	  free((void *)radix_tmp); radix_tmp=0x0;
 */
	  bit_reverse_int(itmp, n2/16, nradices_prim-4, &radix_prim[nradices_prim-5], -1);
/*
!...trig array for wrapper/squaring part is here. We will make use of the fact that the sincos
!   data will be in bit-reversed order, together with the symmetries of the sineand cosine
!   functions about pi/4 and pi/2, to cut our table length by a factor of 16 here. Here's how: the ordered sincos
!   data (prior to bit-reversal of the data and reduction of table length) are a set of
!   complex exponentials exp(I*j*2*pi/N), where j=0,...,N/2-1 and N is the complex FFT length.
!   That is, the argument (complex polar angle) of the exponential is in [0,pi). Within the
!   fused radix16/wrapper-square/radix16inv pass routine, we process 8 such complex exponentials
!   at a time, together with 16 complex FFT array data. Since these 8 sincos data are bit-reversal-
!   reordered, their angular arguments always have the form
!
!	x, x+pi/2, x+pi/4, x+3*pi/4, x+pi/8, x+5*pi/8, x+3*pi/8, x+7*pi/8,
!
!   which allows them to be written as four pairs, each of the form (angle, angle+pi/2):
!
!	(x0, x0+pi/2), (x1, x1+pi/2), (x2, x2+pi/2), (x3, x3+pi/2).
!
!   Given exp(I*angle) = (re,im), we use the complex symmetry  exp(I*(angle+pi/2)) = (-im,re)
!   to get the second exponential of each pair. Since x1 = x0+pi/4, x2 = x0+pi/8 and x1 = x0+3*pi/8,
!   we further let
!		  ISRT2 :=     1/sqrt(2) =    cos(pi/4) =    sin(pi/4),
!		  (c,s) := exp(I*  pi/8) = (cos(  pi/8) , sin(  pi/8)),
!	      and	  exp(I*3*pi/8) = (cos(3*pi/8) , sin(3*pi/8)) = (sin(pi/8) , cos(pi/8)) = (s,c)
!
!   Given only exp(I*x0) = (re,im), this allows us to get the remaining seven complex exponentials via
!
!		  exp(I*(x0       )) = ( re, im)
!		  exp(I*(x0+  pi/2)) = I*exp(I*x0) = (-im, re)
!
!		  exp(I*(x1       )) = (re-im, re+im)*ISRT2
!		  exp(I*(x1+  pi/2)) = I*exp(I*x1)
!
!		  exp(I*(x2       )) = ( re, im)*(c,s) = (re*c-im*s, re*s+im*c)
!		  exp(I*(x2+  pi/2)) = I*exp(I*x2)
!
!		  exp(I*(x3       )) = ( re, im)*(s,c) = (re*s-im*c, re*c+im*s)
!		  exp(I*(x3+  pi/2)) = I*exp(I*x3),
!
!   where (a,b) = a+I*b and (a,b)*(c,d) denotes a complex product, i.e. using both explicit-I* and complex-
!   as-real-pair notation, (a,b)*(c,d) = (a+I*b)*(c+I*d) = (a*c - b*d) + I*(a*d+ b*c) = (a*c - b*d, a*d + b*c).
!
!   Since we already have the constants ISRT2, c and s in registers for the radix-16 transform, we need
!   define no extra temporaries for the above. Further, by saving the four scalar products re*c, im*s, re*s, im*c,
!   we can get both exp(I*x2) and exp(I*x3) using just four multiplies and fouradds, so our total cost for
!   the eight complex sincos data is 2 loads from memory, 6 FMUL, 6 FADD and 4 negations, compared to
!   16 loads from memory (and the accompanying register pressure) for the old version.
!
!   We further use the fact that for every two blocks of 16 FFT data processed together, the wrapper sincos
!   data for the second (upper) block are just the reflection about pi/2 of those for the first block.
!   This cuts the table length in half again.
*/

/*
!...Allocate and initialize an index array containing N/16 indices to store the sequentially-rearranged
!   FFT sincos data indices. (Only need N/16 of these since we need one base sincos index for every 16 complex data).
!   We don't need a separate sincos array for the rea/complex wrapper phase, since this uses the same sincos datum
!   as is used for the the first of each of the two blocks of 16 complex FFT data.
*/
/*	index = (int *)calloc(n2/16, sizeof(int));	*/
	index_ptmp = ALLOC_INT(n2/16);
	if(!index_ptmp){ printf("FATAL: unable to allocate array INDEX in radix16_wrapper_square.\n"); exit(EXIT_FAILURE); }
	index = ALIGN_INT(index_ptmp);
/*
!...Now rearrange FFT sincos indices using the main loop structure as a template.
!   The first length-2 block is a little different, since there we process the 0-15 and 16-31 array
!   elements separately.
*/
	index[0]=itmp [0];
	index[1]=itmp [1];

	k1 =16;	/* init uncompressed wrapper sincos array index */
	k  =2;	/* init   compressed wrapper sincos array index */

	blocklen=16;
	blocklen_sum=16;
	j2_start=96;

	for(i = nradices_prim-6; i >= 0; i--)   /* Radices get processed in reverse order here as in forward FFT. */
	{
	  kp = k1 + blocklen;

	  for(m = 0; m < (blocklen-1)>>1; m += 8)	/* Since we now process TWO 16-element sets per loop execution, only execute the loop half as many times as before. */
	  {
/*...grab the next 2 FFT sincos indices: one for the lower (j1, in the actual loop) data block, one for the upper (j2) data block... */

	    index[k  ]=itmp[(k1  )>>3];
	    index[k+1]=itmp[(kp-8)>>3];

	    k1 = k1+ 8;
	    kp = kp- 8;

	    k  = k + 2;
	  }

	  k1 = k1 + (blocklen >> 1);

	  if(j2_start == n-32)break;

	  blocklen_sum = blocklen_sum + blocklen;
if(!i) { printf("ERROR 10 in radix16_wrapper_square\n"); exit(EXIT_FAILURE); }
	  blocklen = (radix_prim[i-1]-1)*blocklen_sum;

	  j2_start = j2_start+(blocklen<<2);
	}

	j1 = 0;

/*...and deallocate the temporary storage arrays. */

	free((void *)itmp);	itmp=0x0;

	}	/* end of initialization sequence. */

/*
!   SOLVING THE CACHE FLOW PROBLEM FOR BIT-REVERSED ARRAY DATA:
!
!   In a previous version of this routine, the data entering the wrapper/square
!   step were ordered, and (j, N-j) -indexed data pairs were processed
!   together with data from a simple ordered sincos array W0, which
!   stored W0(j) = [1+exp(2*I*pi*j/N)], j=0,...,N/2-1, where n is the
!   complex vector length. Life was easy in the wrapper/square step, but
!   the explicit bit-reversals needed to arrange data for the DIT forward
!   transform (and after the wrapper/square for the DIT inverse transform)
!   had a significant adverse effect on performance of the code overall.
!
!   In this version, we do a DIF forward transform, so data enter the wrapper/square
!   step in bit-reversed order. Lacking a simpler pattern, I at first processed
!   them in the same order as in the old version, i.e. simply fetched
!   data in slots (index(j),index(N-j)) and processed as before. However,
!   if one looks at the actual access patterns that result from this non-strategy,
!   things are ugly: whereas j and N-j have nice predictable behavior,
!   index(j) and index(N-j) bounce all over the place. In fact, it is hard
!   to devise a worse stategy than this, from a caching perspective.
!
!   So I thought, why not loop over index(j), rather than j?
!   This requires us to bit-reverse the data in the W0 (sincos) array
!   but that imposes no penalty since these data are precomputed anyway.
!   Thus, at least the index of the first data element of each pair
!   behaves in a predictable fashion.
!
!   Well, when one does things this way, it turns out the second index also
!   behaves in a predictable way, although not quite as simply as for ordered
!   data. We note that under a bit reversal, the index range j=0,...,N/2-1
!   maps to the evens and j=N/2,...,N-1 maps to the odds, respectively.
!   This tells us we should loop over index(j) and increment by two.
!   When we do this, a surprisingly nice pattern emerges for index(N-j).
!   Even more surprisingly, this pattern also persists for N an odd prime
!   times a power of 2, with a tiny amount of added bookeeping required.
!
!   What happens is that as the first (even) index advances monotonically
!   from 0 to N-2, the second (odd) index runs through O(log2(N)) contiguous
!   odd-index subranges, but runs through each one in REVERSE order, i.e.
!   in decrements of two. The precise number of these odd-index subranges
!   (which overlap a corresponding even-index subrange to form what I call
!   a BLOCK) is equal to log2(power-of-2 part of N) + 1, where
!
!	N = P*2^M, where P is a small prime first radix > 1 (possibly 2.)
!
!   Here are four examples, for N = 16, 24, 20 and 28, assuming a 1-D, unpadded
!   zero-offset data array. I've broken chunks of data into blocks that reveal the pattern.
!   Note the correlation between the bit pattern of N-1 (written vertically, with least
!   significant bit at the top) and the pattern of block lengths
!  (block length = half the total number of complex data processed):
!
!       N=16 (FIRST RADIX = 2):                 N=24 (FIRST RADIX = 3):                 N=20 (FIRST RADIX = 5):                 N=28 (FIRST RADIX = 7):
!
!       N - 1 = 15 = 1111_2:                    N - 1 = 23 = 10111_2:                   N - 1 = 19 = 10011_2:                   N - 1 = 27 = 11011_2:
!       j1 =    j2 =            Current bits    j1 =    j2 =            Current bits    j1 =    j2 =            Current bits    j1 =    j2 =            Current bits
!       index(j)index(N-j)      of (N - 1)_2:   index(j)index(N-j)      of (N - 1)_2:   index(j)index(N-j)      of (N - 1)_2:   index(j)index(N-j)      of (N - 1)_2:
!       ------  ------          -------------   ------  ------          -------------   ------  ------          -------------   ------  ------          -------------
!       0,1 (j=0,N/2) done separately   1       0,1 (j=0,N/2) done separately   1       0,1 (j=0,N/2) done separately   1       0,1 (j=0,N/2) done separately   1
!       ------  ------                          ------  ------                          ------  ------                          ------  ------
!       2       3       length-2 block  1       2       3       length-2 block  1       2       3       length-2 block  1       2       3       length-2 block  1
!       ------  ------                          ------  ------	                        ------  ------                          ------  ------
!       4       7       length-4 block  1       4       7       length-4 block  1       4       19                              4       27
!       6       5                               6       5                               6       17                              6       25
!       ------  ------                          ------  ------                          8       15                              8       23
!       8       15                              8       23                              10      13      length-16 block 100     10      21
!       10      13      length-8 block  1       10      21                              12      11                              12      19
!       12      11                              12      19                              14      9                               14      17      length-24 block 110
!       14      9                               14      17      length-16 block 10      16      7                               16      15
!       ------  ------                          16      15                              18      5                               18      13
!                                               18      13                              ------  ------                          20      11
!                                               20      11                                                                      22      9
!                                               22      9                                                                       24      7
!                                               ------  ------                                                                  26      5
!                                                                                                                               ------  ------
!   So the formula for calculating block sizes is as follows: At the end of each block, we
!   multiply the accumulated blocklength (which is just double the current blocklength) by
!   the next value of (current bits), which is = (radix_next - 1), to get the length
!   of the next block. To get the value of j2_start for the next block,
!   we add the length of the next block to j2_start for the current block.
!
!...Exploiting the pattern makes for a much cache-friendlier code, which follows. Note
!   that if we are processing ordered modular data alongside bit-reversed floating data,
!   we actually have two index pairs to keep track of, e.g. for N=16:
!
!   Floating index 1 (j1):  2, 4, 6, 8,10,12,14
!   Floating index 2 (j2):  3, 7, 5,15,13,11, 9 (0 and 1 done separately)
!
!   Modular  index 1 (j3):  1, 2, 3, 4, 5, 6, 7
!   Modular  index 2 (j4): 15,14,13,12,11,10, 9 (0 and N/2 done separately)
!
!   so the floating index 1 is just double its modular counterpart. The index 2's must be
!   stored separately.
!
!
!   FUSING THE WRAPPER-SQUARE STEP WITH THE RADIX-16 PASSES PRECEDING AND FOLLOWING IT:
!
!   OK, so you've made it this far, and justifiably are hoping to see some actual, honest-
!   to-goodness code. Alas, things must get just a bit more complicated before that happy
!   moment arrives. That's because in order to squeeze maximum performance out of our code,
!   we should seize any opportunities to minimize unnecessary data movement. One such
!   presents itself here. We have a radix-16 DIF pass immediately preceding and a radix-16
!   DIT pass immediately following the wrapper/square, so we do all three of the above steps
!   in a single in-place sweep, thus reducing the number of passes through the big data
!   arrays from three to one. We do the first 16 complex data (and their images under the (j,N'-j)
!   correlation) separately, after which blocklengths are always a multiple of 16, allowing us to simplify the indexing.
!
!   The general loop structure looks as follows:
!
!   loop:
!       do m = 1,blocklen
!
!         call pair_square(a(j1),a(j1+1),a(j2),a(j2+1),w(1,j),w(2,j))
!
!         j  = j +1
!         j1 = j1+4
!         j2 = j2-4
!       enddo
!       ...
!       blocklen = 2*blocklen
!       j2_start = j2_start+ishft(blocklen,2)
!       j2=j2_start
!       cycle
!   endloop
!
!   So the first 16 complex elements and their images are processed as follows:
!
!   exe0: process a(0,1) and a(2,3) separately;
!
!   exe1:
!       j=1
!       blocklen = 1
!       j1=4
!       j2_start=6
!
!         call pair_square(a( 4),a( 5),a( 6),a( 7),w(:,1))
!
!         j  = 2
!         j1 = 8
!         j2 = 2
!       enddo
!
!       blocklen = 2
!       j2_start = 14
!   end1
!
!   exe2:
!       j=2
!       blocklen = 2
!       j1=8
!       j2_start=14
!
!         call pair_square(a( 8),a( 9),a(14),a(15),w(:,2))
!         call pair_square(a(12),a(13),a(10),a(11),w(:,3))
!
!         j  = 4
!         j1 = 16
!         j2 = 6
!       enddo
!
!       blocklen = 4
!       j2_start = 30
!   end2
!
!   exe3:
!       j=4
!       blocklen = 4
!       j1=16
!       j2_start=30
!
!         call pair_square(a(16),a(17),a(30),a(31),w(:,4))
!         call pair_square(a(20),a(21),a(26),a(27),w(:,5))
!         call pair_square(a(24),a(25),a(22),a(23),w(:,6))
!         call pair_square(a(28),a(29),a(18),a(19),w(:,7))
!
!         j  = 8
!         j1 = 32
!         j2 = 14
!       enddo
!
!       blocklen = 8
!       j2_start = 62
!   end3
!
!   exe4:
!       j=8
!       blocklen = 8
!       j1=32
!       j2_start=62
!
!         call pair_square(a(32),a(33),a(62),a(63),w(:,8))
!         call pair_square(a(36),a(37),a(58),a(59),w(:,9))
!         call pair_square(a(40),a(41),a(54),a(55),w(:,10))
!         call pair_square(a(44),a(45),a(50),a(51),w(:,11))
!         call pair_square(a(48),a(49),a(46),a(47),w(:,12))
!         call pair_square(a(52),a(53),a(42),a(43),w(:,13))
!         call pair_square(a(56),a(57),a(38),a(39),w(:,14))
!         call pair_square(a(60),a(61),a(34),a(35),w(:,15))
!
!         j  = 16
!         j1 = 64
!         j2 = 30
!       enddo
!
!       blocklen = 16
!       j2_start = 126
!   end4
!
!   From here onward the blocklength will always be an integer multiple of 16, i.e. we can process each block using pairs of nonoverlapping
!   blocks of 16 complex data each, which is compatible to fusion with radix-16 pass routines. For example, the next (fifth) loop execution
!   looks like:
!
!   exe5:
!       j=16
!       blocklen = 16
!       j1=64
!       j2_start=126
!                                                                       ...and these 32 complex data can be processed as follows:
!         call pair_square(a( 64),a( 65),a(126),a(127),w(:,16))
!         call pair_square(a( 68),a( 69),a(122),a(123),w(:,17))         do radix-16 DIF pass (using DIF sincos  0-15) to get a( 64: 95) ;
!         call pair_square(a( 72),a( 73),a(118),a(119),w(:,18))         do radix-16 DIF pass (using DIF sincos 16-31) to get a( 96:127) ;
!         call pair_square(a( 76),a( 77),a(114),a(115),w(:,19))
!         call pair_square(a( 80),a( 81),a(110),a(111),w(:,20)) {need to see whether there's any nice pattern to the FFT sincos
!         call pair_square(a( 84),a( 85),a(106),a(107),w(:,21))  data here which would allow us to cut the storage of same}
!         call pair_square(a( 88),a( 89),a(102),a(103),w(:,22))
!         call pair_square(a( 92),a( 93),a( 98),a( 99),w(:,23))         combine the 32 resulting complex array data as shown at left    ;
!
!         call pair_square(a( 96),a( 97),a( 94),a( 95),w(:,24))
!         call pair_square(a(100),a(101),a( 90),a( 91),w(:,25))         do radix-16 DIT pass on a( 64: 95)      ;
!         call pair_square(a(104),a(105),a( 86),a( 87),w(:,26))         do radix-16 DIT pass on a( 96:127)      .
!         call pair_square(a(108),a(109),a( 82),a( 83),w(:,27))
!         call pair_square(a(112),a(113),a( 78),a( 79),w(:,28))
!         call pair_square(a(116),a(117),a( 74),a( 75),w(:,29))
!         call pair_square(a(120),a(121),a( 70),a( 71),w(:,30))
!         call pair_square(a(124),a(125),a( 66),a( 67),w(:,31))
!                                                                        The radix-16 passes share many register data, providing added savings.
!         j  = 32
!         j1 = 128
!         j2 = 62
!       enddo
!
!       blocklen = 32
!       j2_start = 254
!   end5
*/

/* Init the loop-control variables: */

	i            = ws_i           ;
	j1           = ws_j1          ;
	j2           = ws_j2          ;
	j2_start     = ws_j2_start    ;
	k            = ws_k           ;
	m            = ws_m           ;
	blocklen     = ws_blocklen    ;
	blocklen_sum = ws_blocklen_sum;

	/* If j1 == 0 we need to init the loop counters; otherwise, just jump
	   right in and pick up where we left off on the previous pair of blocks:
	*/
	if(j1 > 0)
		goto jump_in;

/*
!...All but the first two radix-16 blocks are done on Mr. Henry Ford's assembly line. From there onward the blocklength
!   will always be an integer multiple of 16, i.e. we can process each block using pairs of nonoverlapping blocks of 16
!   complex data each, which is compatible to fusion with radix-16 pass routines.
*/

      for(i = nradices_prim-5; i >= 0; i-- )	/* Main loop: lower bound = nradices_prim-radix_now. */
      {						/* Remember, radices get processed in reverse order here as in forward FFT. */

	for(m = 0; m < (blocklen-1)>>1; m += 8) /* Since we now process TWO 16-element sets per loop execution, only execute the loop half as many times as before. */
	{
/*
if((j1 & ndivrad0m1 == 0) && j1*radix0%n != 0) printf("WARN: %8d  %8d\n",j1 & ndivrad0m1,j1*radix0%n);
*/

if(j1 && j1*radix0%n == 0)
{
/*printf("%8d  %8d  %8d  %8d\n",j1,j1*radix0,j1*radix0%n,j1 & ndivrad0m1);*/
  return;
}

jump_in:	/* Entry point for all blocks but the first. */

	  j1pad = j1 + ( (j1 >> DAT_BITS) << PAD_BITS );	/* floating padded-array 1st element index is here */
	  j2pad = j2 + ( (j2 >> DAT_BITS) << PAD_BITS );	/* floating padded-array 2nd element index is here */

/* 1st set of inputs: */
	    idum = j1pad;
	    iroot = index[k];
	    l = iroot;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += iroot;		 /* 2*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    rt=rt1[k2].re;	it=rt1[k2].im;
	    cA1 =t1*rt-t2*it;	sA1 =t1*it+t2*rt;
/*
printf("1: %20.15lf  %20.15lf\n", cA1, sA1);
*/
	re=cA1;	im=sA1;	/* Save for the wrapper Step... */

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += (iroot << 1);	/* 4*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    rt=rt1[k2].re;	it=rt1[k2].im;
	    cA2 =t1*rt-t2*it;	sA2 =t1*it+t2*rt;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += (iroot << 2);	/* 8*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    rt=rt1[k2].re;	it=rt1[k2].im;
	    cA4 =t1*rt-t2*it;	sA4 =t1*it+t2*rt;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += (iroot << 2) + iroot;	/* 13*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    rt=rt1[k2].re;	it=rt1[k2].im;
	    cA8 =t1*rt-t2*it;	sA8 =t1*it+t2*rt;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    rt=rt1[k2].re;	it=rt1[k2].im;
	    cA13=t1*rt-t2*it;	sA13=t1*it+t2*rt;

	    /* c3,5 */
	    t1=cA1 *cA4 ;	t2=cA1 *sA4 ;	rt=sA1 *cA4 ;	it=sA1 *sA4;
	    cA3 =t1 +it;	sA3 =t2 -rt;	cA5 =t1 -it;	sA5 =t2 +rt;

	    /* c6,7,9,10 */
	    t1=cA1 *cA8 ;	t2=cA1 *sA8 ;	rt=sA1 *cA8 ;	it=sA1 *sA8;
	    cA7 =t1 +it;	sA7 =t2 -rt;	cA9 =t1 -it;	sA9 =t2 +rt;

	    t1=cA2 *cA8 ;	t2=cA2 *sA8 ;	rt=sA2 *cA8 ;	it=sA2 *sA8;
	    cA6 =t1 +it;	sA6 =t2 -rt;	cA10=t1 -it;	sA10=t2 +rt;

	    /* c11,12,14,15 */
	    t1=cA1 *cA13;	t2=cA1 *sA13;	rt=sA1 *cA13;	it=sA1 *sA13;
	    cA12=t1 +it;	sA12=t2 -rt;	cA14=t1 -it;	sA14=t2 +rt;

	    t1=cA2 *cA13;	t2=cA2 *sA13;	rt=sA2 *cA13;	it=sA2 *sA13;
	    cA11=t1 +it;	sA11=t2 -rt;	cA15=t1 -it;	sA15=t2 +rt;

/*   gather the needed data (16 64-bit complex, i.e. 32 64-bit reals) and do the first set of four length-4 transforms... */

/*...Block 1: */
	    t1 =a[idum   ];				t2 =a[idum+1 ];
	    rt =a[idum+16]*cA8 -a[idum+17]*sA8 ;	it =a[idum+17]*cA8 +a[idum+16]*sA8;
	    t3 =t1 -rt;	t1 =t1 +rt;
	    t4 =t2 -it;	t2 =t2 +it;

	    t5 =a[idum+8 ]*cA4 -a[idum+9 ]*sA4;		t6 =a[idum+9 ]*cA4 +a[idum+8 ]*sA4;
	    rt =a[idum+24]*cA12-a[idum+25]*sA12;	it =a[idum+25]*cA12+a[idum+24]*sA12;
	    t7 =t5 -rt;	t5 =t5 +rt;
	    t8 =t6 -it;	t6 =t6 +it;

	    rt =t5;	t5 =t1 -rt;	t1 =t1 +rt;
	    it =t6;	t6 =t2 -it;	t2 =t2 +it;

	    rt =t7;	t7 =t3 +t8;	t3 =t3 -t8;
			t8 =t4 -rt;	t4 =t4 +rt;

/*...Block 2: */
	    t9 =a[idum+4 ]*cA2 -a[idum+5 ]*sA2 ;	t10=a[idum+5 ]*cA2 +a[idum+4 ]*sA2 ;
	    rt =a[idum+20]*cA10-a[idum+21]*sA10;	it =a[idum+21]*cA10+a[idum+20]*sA10;
	    t11=t9 -rt;	t9 =t9 +rt;
	    t12=t10-it;	t10=t10+it;

	    t13=a[idum+12]*cA6 -a[idum+13]*sA6 ;	t14=a[idum+13]*cA6 +a[idum+12]*sA6 ;
	    rt =a[idum+28]*cA14-a[idum+29]*sA14;	it =a[idum+29]*cA14+a[idum+28]*sA14;
	    t15=t13-rt;	t13=t13+rt;
	    t16=t14-it;	t14=t14+it;

	    rt =t13;	t13=t9 -rt;	t9 =t9 +rt;
	    it =t14;	t14=t10-it;	t10=t10+it;

	    rt =t15;	t15=t11+t16;	t11=t11-t16;
			t16=t12-rt;	t12=t12+rt;

/*...Block 3: */
	    t17=a[idum+2 ]*cA1 -a[idum+3 ]*sA1 ;	t18=a[idum+3 ]*cA1 +a[idum+2 ]*sA1 ;
	    rt =a[idum+18]*cA9 -a[idum+19]*sA9 ;	it =a[idum+19]*cA9 +a[idum+18]*sA9 ;
	    t19=t17-rt;	t17=t17+rt;
	    t20=t18-it;	t18=t18+it;

	    t21=a[idum+10]*cA5 -a[idum+11]*sA5 ;	t22=a[idum+11]*cA5 +a[idum+10]*sA5 ;
	    rt =a[idum+26]*cA13-a[idum+27]*sA13;	it =a[idum+27]*cA13+a[idum+26]*sA13;
	    t23=t21-rt;	t21=t21+rt;
	    t24=t22-it;	t22=t22+it;

	    rt =t21;	t21=t17-rt;	t17=t17+rt;
	    it =t22;	t22=t18-it;	t18=t18+it;

	    rt =t23;	t23=t19+t24;	t19=t19-t24;
			t24=t20-rt;	t20=t20+rt;

/*...Block 4: */
	    t25=a[idum+6 ]*cA3 -a[idum+7 ]*sA3 ;	t26=a[idum+7 ]*cA3 +a[idum+6 ]*sA3 ;
	    rt =a[idum+22]*cA11-a[idum+23]*sA11;	it =a[idum+23]*cA11+a[idum+22]*sA11;
	    t27=t25-rt;	t25=t25+rt;
	    t28=t26-it;	t26=t26+it;

	    t29=a[idum+14]*cA7 -a[idum+15]*sA7 ;	t30=a[idum+15]*cA7 +a[idum+14]*sA7 ;
	    rt =a[idum+30]*cA15-a[idum+31]*sA15;	it =a[idum+31]*cA15+a[idum+30]*sA15;
	    t31=t29-rt;	t29=t29+rt;
	    t32=t30-it;	t30=t30+it;

	    rt =t29;	t29=t25-rt;	t25=t25+rt;
	    it =t30;	t30=t26-it;	t26=t26+it;

	    rt =t31;	t31=t27+t32;	t27=t27-t32;
			t32=t28-rt;	t28=t28+rt;
/*
!...and now do four more radix-4 transforms, including the internal twiddle factors:
!	1, exp(i* 1*twopi/16) =       ( c, s), exp(i* 2*twopi/16) = ISRT2*( 1, 1), exp(i* 3*twopi/16) =       ( s, c) (for inputs to transform block 2)
!	1, exp(i* 2*twopi/16) = ISRT2*( 1, 1), exp(i* 4*twopi/16) =       ( 0, 1), exp(i* 6*twopi/16) = ISRT2*(-1, 1) (for inputs to transform block 3)
!	1, exp(i* 3*twopi/16) =       ( s, c), exp(i* 6*twopi/16) = ISRT2*(-1, 1), exp(i* 9*twopi/16) =       (-c,-s) (for inputs to transform block 4).
!  (This is 4 real*complex and 4 complex*complex multiplies (= 24 FMUL), compared to 6 and 4, respectively (= 28 FMUL) for my old scheme.)
!   I.e. do similar as above, except inputs a[j1  +p0:15:1] are replaced by t0:30:2,
!					   a[j1+1+p0:15:1] are replaced by t1:31:2, and v.v. for outputs,
!   and only the last 3 inputs to each of the radix-4 transforms 2 through 4 are multiplied by non-unity twiddles.
*/
/*...Block 1: t1,9,17,25 */
	    rt =t9;	t9 =t1 -rt;	t1 =t1 +rt;
	    it =t10;	t10=t2 -it;	t2 =t2 +it;

	    rt =t25;	t25=t17-rt;	t17=t17+rt;
	    it =t26;	t26=t18-it;	t18=t18+it;

	    aj1p0r =t1+t17;	aj1p0i =t2+t18;
	    aj1p1r =t1-t17;	aj1p1i =t2-t18;

	    aj1p2r =t9 -t26;	aj1p2i =t10+t25;
	    aj1p3r =t9 +t26;	aj1p3i =t10-t25;

/*...Block 3: t5,13,21,29 */
	    rt =t13;	t13=t5 +t14;	t5 =t5 -t14;
			t14=t6 -rt;	t6 =t6 +rt;

	    rt =(t21-t22)*ISRT2;t22=(t21+t22)*ISRT2;	t21=rt;
	    rt =(t30+t29)*ISRT2;it =(t30-t29)*ISRT2;
	    t29=t21+rt;	t21=t21-rt;
	    t30=t22+it;	t22=t22-it;

	    aj1p4r =t5+t21;	aj1p4i =t6+t22;
	    aj1p5r =t5-t21;	aj1p5i =t6-t22;

	    aj1p6r =t13-t30;	aj1p6i =t14+t29;
	    aj1p7r =t13+t30;	aj1p7i =t14-t29;

/*...Block 2: t3,11,19,27 */
	    rt =(t11-t12)*ISRT2;it =(t11+t12)*ISRT2;
	    t11=t3 -rt;	t3 =t3 +rt;
	    t12=t4 -it;	t4 =t4 +it;

	    rt =t19*c - t20*s;	t20=t20*c + t19*s;	t19=rt;
	    rt =t27*s - t28*c;	it =t28*s + t27*c;
	    t27=t19-rt;	t19=t19+rt;
	    t28=t20-it;	t20=t20+it;

	    aj1p8r =t3+t19;	aj1p8i =t4+t20;
	    aj1p9r =t3-t19;	aj1p9i =t4-t20;

	    aj1p10r=t11-t28;	aj1p10i=t12+t27;
	    aj1p11r=t11+t28;	aj1p11i=t12-t27;

/*...Block 4: t7,15,23,31 */
	    rt =(t16+t15)*ISRT2;it =(t16-t15)*ISRT2;
	    t15=t7 +rt;	t7 =t7 -rt;
	    t16=t8 +it;	t8 =t8 -it;

	    rt =t23*s - t24*c;	t24=t24*s + t23*c;	t23=rt;
	    rt =t31*c - t32*s;	it =t32*c + t31*s;
	    t31=t23+rt;	t23=t23-rt;
	    t32=t24+it;	t24=t24-it;

	    aj1p12r=t7+t23;	aj1p12i=t8+t24;
	    aj1p13r=t7-t23;	aj1p13i=t8-t24;

	    aj1p14r=t15-t32;	aj1p14i=t16+t31;
	    aj1p15r=t15+t32;	aj1p15i=t16-t31;

	k=k+1;	/* increment sincos array index */

/* 2nd set of inputs: */
	    idum = j2pad;
	    iroot = index[k];
	    l = iroot;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += iroot;		 /* 2*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    rt=rt1[k2].re;	it=rt1[k2].im;
	    cB1 =t1*rt-t2*it;	sB1 =t1*it+t2*rt;
/*
printf("2: %20.15lf  %20.15lf\n", cB1, sB1);
*/
	if(j1 == 0){ re=cB1;	im=sB1; }   /* The j1 = 0 case is special... */

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += (iroot << 1);	/* 4*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    rt=rt1[k2].re;	it=rt1[k2].im;
	    cB2 =t1*rt-t2*it;	sB2 =t1*it+t2*rt;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += (iroot << 2);	/* 8*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    rt=rt1[k2].re;	it=rt1[k2].im;
	    cB4 =t1*rt-t2*it;	sB4 =t1*it+t2*rt;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += (iroot << 2) + iroot;	/* 13*iroot */
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    rt=rt1[k2].re;	it=rt1[k2].im;
	    cB8 =t1*rt-t2*it;	sB8 =t1*it+t2*rt;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    t1=rt0[k1].re;	t2=rt0[k1].im;
	    rt=rt1[k2].re;	it=rt1[k2].im;
	    cB13=t1*rt-t2*it;	sB13=t1*it+t2*rt;

	    /* c3,5 */
	    t1=cB1 *cB4 ;	t2=cB1 *sB4 ;	rt=sB1 *cB4 ;	it=sB1 *sB4;
	    cB3 =t1 +it;	sB3 =t2 -rt;	cB5 =t1 -it;	sB5 =t2 +rt;

	    /* c6,7,9,10 */
	    t1=cB1 *cB8 ;	t2=cB1 *sB8 ;	rt=sB1 *cB8 ;	it=sB1 *sB8;
	    cB7 =t1 +it;	sB7 =t2 -rt;	cB9 =t1 -it;	sB9 =t2 +rt;

	    t1=cB2 *cB8 ;	t2=cB2 *sB8 ;	rt=sB2 *cB8 ;	it=sB2 *sB8;
	    cB6 =t1 +it;	sB6 =t2 -rt;	cB10=t1 -it;	sB10=t2 +rt;

	    /* c11,12,14,15 */
	    t1=cB1 *cB13;	t2=cB1 *sB13;	rt=sB1 *cB13;	it=sB1 *sB13;
	    cB12=t1 +it;	sB12=t2 -rt;	cB14=t1 -it;	sB14=t2 +rt;

	    t1=cB2 *cB13;	t2=cB2 *sB13;	rt=sB2 *cB13;	it=sB2 *sB13;
	    cB11=t1 +it;	sB11=t2 -rt;	cB15=t1 -it;	sB15=t2 +rt;

/*   gather the needed data (16 64-bit complex, i.e. 32 64-bit reals) and do the first set of four length-4 transforms... */

/*...Block 1: */
	    t1 =a[idum   ];				t2 =a[idum+1 ];
	    rt =a[idum+16]*cB8 -a[idum+17]*sB8 ;	it =a[idum+17]*cB8 +a[idum+16]*sB8;
	    t3 =t1 -rt;	t1 =t1 +rt;
	    t4 =t2 -it;	t2 =t2 +it;

	    t5 =a[idum+8 ]*cB4 -a[idum+9 ]*sB4;		t6 =a[idum+9 ]*cB4 +a[idum+8 ]*sB4;
	    rt =a[idum+24]*cB12-a[idum+25]*sB12;	it =a[idum+25]*cB12+a[idum+24]*sB12;
	    t7 =t5 -rt;	t5 =t5 +rt;
	    t8 =t6 -it;	t6 =t6 +it;

	    rt =t5;	t5 =t1 -rt;	t1 =t1 +rt;
	    it =t6;	t6 =t2 -it;	t2 =t2 +it;

	    rt =t7;	t7 =t3 +t8;	t3 =t3 -t8;
			t8 =t4 -rt;	t4 =t4 +rt;

/*...Block 2: */
	    t9 =a[idum+4 ]*cB2 -a[idum+5 ]*sB2 ;	t10=a[idum+5 ]*cB2 +a[idum+4 ]*sB2 ;
	    rt =a[idum+20]*cB10-a[idum+21]*sB10;	it =a[idum+21]*cB10+a[idum+20]*sB10;
	    t11=t9 -rt;	t9 =t9 +rt;
	    t12=t10-it;	t10=t10+it;

	    t13=a[idum+12]*cB6 -a[idum+13]*sB6 ;	t14=a[idum+13]*cB6 +a[idum+12]*sB6 ;
	    rt =a[idum+28]*cB14-a[idum+29]*sB14;	it =a[idum+29]*cB14+a[idum+28]*sB14;
	    t15=t13-rt;	t13=t13+rt;
	    t16=t14-it;	t14=t14+it;

	    rt =t13;	t13=t9 -rt;	t9 =t9 +rt;
	    it =t14;	t14=t10-it;	t10=t10+it;

	    rt =t15;	t15=t11+t16;	t11=t11-t16;
			t16=t12-rt;	t12=t12+rt;

/*...Block 3: */
	    t17=a[idum+2 ]*cB1 -a[idum+3 ]*sB1 ;	t18=a[idum+3 ]*cB1 +a[idum+2 ]*sB1 ;
	    rt =a[idum+18]*cB9 -a[idum+19]*sB9 ;	it =a[idum+19]*cB9 +a[idum+18]*sB9 ;
	    t19=t17-rt;	t17=t17+rt;
	    t20=t18-it;	t18=t18+it;

	    t21=a[idum+10]*cB5 -a[idum+11]*sB5 ;	t22=a[idum+11]*cB5 +a[idum+10]*sB5 ;
	    rt =a[idum+26]*cB13-a[idum+27]*sB13;	it =a[idum+27]*cB13+a[idum+26]*sB13;
	    t23=t21-rt;	t21=t21+rt;
	    t24=t22-it;	t22=t22+it;

	    rt =t21;	t21=t17-rt;	t17=t17+rt;
	    it =t22;	t22=t18-it;	t18=t18+it;

	    rt =t23;	t23=t19+t24;	t19=t19-t24;
			t24=t20-rt;	t20=t20+rt;

/*...Block 4: */
	    t25=a[idum+6 ]*cB3 -a[idum+7 ]*sB3 ;	t26=a[idum+7 ]*cB3 +a[idum+6 ]*sB3 ;
	    rt =a[idum+22]*cB11-a[idum+23]*sB11;	it =a[idum+23]*cB11+a[idum+22]*sB11;
	    t27=t25-rt;	t25=t25+rt;
	    t28=t26-it;	t26=t26+it;

	    t29=a[idum+14]*cB7 -a[idum+15]*sB7 ;	t30=a[idum+15]*cB7 +a[idum+14]*sB7 ;
	    rt =a[idum+30]*cB15-a[idum+31]*sB15;	it =a[idum+31]*cB15+a[idum+30]*sB15;
	    t31=t29-rt;	t29=t29+rt;
	    t32=t30-it;	t30=t30+it;

	    rt =t29;	t29=t25-rt;	t25=t25+rt;
	    it =t30;	t30=t26-it;	t26=t26+it;

	    rt =t31;	t31=t27+t32;	t27=t27-t32;
			t32=t28-rt;	t28=t28+rt;
/*
!...and now do four more radix-4 transforms, including the internal twiddle factors:
*/
/*...Block 1: t1,9,17,25 */
	    rt =t9;	t9 =t1 -rt;	t1 =t1 +rt;
	    it =t10;	t10=t2 -it;	t2 =t2 +it;

	    rt =t25;	t25=t17-rt;	t17=t17+rt;
	    it =t26;	t26=t18-it;	t18=t18+it;

	    aj2p0r =t1+t17;	aj2p0i =t2+t18;
	    aj2p1r =t1-t17;	aj2p1i =t2-t18;

	    aj2p2r =t9 -t26;	aj2p2i =t10+t25;
	    aj2p3r =t9 +t26;	aj2p3i =t10-t25;

/*...Block 3: t5,13,21,29 */
	    rt =t13;	t13=t5 +t14;	t5 =t5 -t14;
			t14=t6 -rt;	t6 =t6 +rt;

	    rt =(t21-t22)*ISRT2;t22=(t21+t22)*ISRT2;	t21=rt;
	    rt =(t30+t29)*ISRT2;it =(t30-t29)*ISRT2;
	    t29=t21+rt;	t21=t21-rt;
	    t30=t22+it;	t22=t22-it;

	    aj2p4r =t5+t21;	aj2p4i =t6+t22;
	    aj2p5r =t5-t21;	aj2p5i =t6-t22;

	    aj2p6r =t13-t30;	aj2p6i =t14+t29;
	    aj2p7r =t13+t30;	aj2p7i =t14-t29;

/*...Block 2: t3,11,19,27 */
	    rt =(t11-t12)*ISRT2;it =(t11+t12)*ISRT2;
	    t11=t3 -rt;	t3 =t3 +rt;
	    t12=t4 -it;	t4 =t4 +it;

	    rt =t19*c - t20*s;	t20=t20*c + t19*s;	t19=rt;
	    rt =t27*s - t28*c;	it =t28*s + t27*c;
	    t27=t19-rt;	t19=t19+rt;
	    t28=t20-it;	t20=t20+it;

	    aj2p8r =t3+t19;	aj2p8i =t4+t20;
	    aj2p9r =t3-t19;	aj2p9i =t4-t20;

	    aj2p10r=t11-t28;	aj2p10i=t12+t27;
	    aj2p11r=t11+t28;	aj2p11i=t12-t27;

/*...Block 4: t7,15,23,31 */
	    rt =(t16+t15)*ISRT2;it =(t16-t15)*ISRT2;
	    t15=t7 +rt;	t7 =t7 -rt;
	    t16=t8 +it;	t8 =t8 -it;

	    rt =t23*s - t24*c;	t24=t24*s + t23*c;	t23=rt;
	    rt =t31*c - t32*s;	it =t32*c + t31*s;
	    t31=t23+rt;	t23=t23-rt;
	    t32=t24+it;	t24=t24-it;

	    aj2p12r=t7+t23;	aj2p12i=t8+t24;
	    aj2p13r=t7-t23;	aj2p13i=t8-t24;

	    aj2p14r=t15-t32;	aj2p14i=t16+t31;
	    aj2p15r=t15+t32;	aj2p15i=t16-t31;

	k=k+1;	/* increment sincos array index */
/*
!...send the pairs of complex elements which are to be combined and sincos temporaries needed for the squaring to a
!   small subroutine. The j1 = 0 case is again exceptional.
*/
	  if(j1==0)     /* NB: mustn't use re, im as temps in exe1-3 section, since those contain saved sincos data for exe4 block. */
	  {
/*...j=0 (for which I(0)=Re{H(0)}^2+i*Im{H(0)}^2) is done separately... */

	    rt=aj1p0r;
	    aj1p0r=(rt+aj1p0i)*(rt+aj1p0i);
	    aj1p0i=(rt-aj1p0i)*(rt-aj1p0i);
	    rt=aj1p0r;
	    aj1p0r=0.5*(rt+aj1p0i);
	    aj1p0i=0.5*(rt-aj1p0i);
/*
!...as is j=N/2 (for which I(j)=H(j)^2). Note that under bit-reversal the N/2 element gets mapped into
!   the second complex data slot, i.e. is adjacent to the starting element.
*/
	    rt =aj1p1r*aj1p1i;
	    aj1p1r=(aj1p1r+aj1p1i)*(aj1p1r-aj1p1i);
	    aj1p1i=rt+rt;

	    pair_square(&aj1p2r ,&aj1p2i ,&aj1p3r ,&aj1p3i ,0.0,1.0);
/* exe1 */


	    rt=ISRT2;	it=ISRT2;



	    pair_square(&aj1p4r ,&aj1p4i ,&aj1p7r ,&aj1p7i , rt, it);
/* exe2 */
	    pair_square(&aj1p6r ,&aj1p6i ,&aj1p5r ,&aj1p5i ,-it, rt);
/* exe2 */


	    t3=c;	t4=0;	t5=s;	t6=0;
	    rt=t3-t4;	it=t5+t6;


	    pair_square(&aj1p8r ,&aj1p8i ,&aj1p15r,&aj1p15i, rt, it);
/* exe3 */
	    pair_square(&aj1p10r,&aj1p10i,&aj1p13r,&aj1p13i,-it, rt);
/* exe3 */


	    rt=t5-t6;	it=t3+t4;


	    pair_square(&aj1p12r,&aj1p12i,&aj1p11r,&aj1p11i, rt, it);
/* exe3 */
	    pair_square(&aj1p14r,&aj1p14i,&aj1p9r ,&aj1p9i ,-it, rt);
/* exe3 */

/*   exe4: */

	    pair_square(&aj2p0r ,&aj2p0i ,&aj2p15r,&aj2p15i, re, im);
	    pair_square(&aj2p2r ,&aj2p2i ,&aj2p13r,&aj2p13i,-im, re);

	    rt=(re-im)*ISRT2;	it=(re+im)*ISRT2;

	    pair_square(&aj2p4r ,&aj2p4i ,&aj2p11r,&aj2p11i, rt, it);
	    pair_square(&aj2p6r ,&aj2p6i ,&aj2p9r ,&aj2p9i ,-it, rt);

	    t3=re*c;	t4=im*s;	t5=re*s;	t6=im*c;
	    rt=t3-t4;	it=t5+t6;

	    pair_square(&aj2p8r ,&aj2p8i ,&aj2p7r ,&aj2p7i , rt, it);
	    pair_square(&aj2p10r,&aj2p10i,&aj2p5r ,&aj2p5i ,-it, rt);

	    rt=t5-t6;	it=t3+t4;

	    pair_square(&aj2p12r,&aj2p12i,&aj2p3r ,&aj2p3i , rt, it);
	    pair_square(&aj2p14r,&aj2p14i,&aj2p1r ,&aj2p1i ,-it, rt);
	  }
	  else
	  {
	    /*
	    pair_square(&aj1p0r ,&aj1p0i ,&aj2p15r,&aj2p15i, re, im);	pair_square(&aj2p14r,&aj2p14i,&aj1p1r ,&aj1p1i ,-re, im);
	    pair_square(&aj1p2r ,&aj1p2i ,&aj2p13r,&aj2p13i,-im, re);	pair_square(&aj2p12r,&aj2p12i,&aj1p3r ,&aj1p3i , im, re);
	   */
	    pair_square2A(&aj1p0r ,&aj1p0i ,&aj2p15r,&aj2p15i,&aj1p2r ,&aj1p2i ,&aj2p13r,&aj2p13i, re, im);
	    pair_square2B(&aj2p14r,&aj2p14i,&aj1p1r ,&aj1p1i ,&aj2p12r,&aj2p12i,&aj1p3r ,&aj1p3i ,-re, im);

	    rt=(re-im)*ISRT2;	it=(re+im)*ISRT2;
	    /*
	    pair_square(&aj1p4r ,&aj1p4i ,&aj2p11r,&aj2p11i, rt, it);	pair_square(&aj2p10r,&aj2p10i,&aj1p5r ,&aj1p5i ,-rt, it);
	    pair_square(&aj1p6r ,&aj1p6i ,&aj2p9r ,&aj2p9i ,-it, rt);	pair_square(&aj2p8r ,&aj2p8i ,&aj1p7r ,&aj1p7i , it, rt);
	    */
	    pair_square2A(&aj1p4r ,&aj1p4i ,&aj2p11r,&aj2p11i,&aj1p6r ,&aj1p6i ,&aj2p9r ,&aj2p9i , rt, it);
	    pair_square2B(&aj2p10r,&aj2p10i,&aj1p5r ,&aj1p5i ,&aj2p8r ,&aj2p8i ,&aj1p7r ,&aj1p7i ,-rt, it);

	    t3=re*c;	t4=im*s;	t5=re*s;	t6=im*c;
	    rt=t3-t4;	it=t5+t6;
	    /*
	    pair_square(&aj1p8r ,&aj1p8i ,&aj2p7r ,&aj2p7i , rt, it);	pair_square(&aj2p6r ,&aj2p6i ,&aj1p9r ,&aj1p9i ,-rt, it);
	    pair_square(&aj1p10r,&aj1p10i,&aj2p5r ,&aj2p5i ,-it, rt);	pair_square(&aj2p4r ,&aj2p4i ,&aj1p11r,&aj1p11i, it, rt);
	    */
	    pair_square2A(&aj1p8r ,&aj1p8i ,&aj2p7r ,&aj2p7i ,&aj1p10r,&aj1p10i,&aj2p5r ,&aj2p5i , rt, it);
	    pair_square2B(&aj2p6r ,&aj2p6i ,&aj1p9r ,&aj1p9i ,&aj2p4r ,&aj2p4i ,&aj1p11r,&aj1p11i,-rt, it);

	    rt=t5-t6;	it=t3+t4;
	    /*
	    pair_square(&aj1p12r,&aj1p12i,&aj2p3r ,&aj2p3i , rt, it);	pair_square(&aj2p2r ,&aj2p2i ,&aj1p13r,&aj1p13i,-rt, it);
	    pair_square(&aj1p14r,&aj1p14i,&aj2p1r ,&aj2p1i ,-it, rt);	pair_square(&aj2p0r ,&aj2p0i ,&aj1p15r,&aj1p15i, it, rt);
	    */
	    pair_square2A(&aj1p12r,&aj1p12i,&aj2p3r ,&aj2p3i ,&aj1p14r,&aj1p14i,&aj2p1r ,&aj2p1i , rt, it);
	    pair_square2B(&aj2p2r ,&aj2p2i ,&aj1p13r,&aj1p13i,&aj2p0r ,&aj2p0i ,&aj1p15r,&aj1p15i,-rt, it);
	  }

/*...And do an inverse DIT radix-16 pass on the squared-data blocks. */

/* 1st set of inputs: */
	  idum = j1pad;
#if PFETCH
add0 = &a[idum+32];
#endif
/*   gather the needed data (16 64-bit complex, i.e. 32 64-bit reals) and do the first set of four length-4 IDIT transforms... */

/*...Block 1: */
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    t3 =aj1p0r -aj1p1r;	t4 =aj1p0i -aj1p1i;
	    t1 =aj1p0r +aj1p1r;	t2 =aj1p0i +aj1p1i;

	    t7 =aj1p2r -aj1p3r;	t8 =aj1p2i -aj1p3i;
	    t5 =aj1p2r +aj1p3r;	t6 =aj1p2i +aj1p3i;

	    rt =t5;	t5 =t1 -rt;	t1 =t1 +rt;
	    it =t6;	t6 =t2 -it;	t2 =t2 +it;

	    rt =t7;	t7 =t3 -t8;	t3 =t3 +t8;
			t8 =t4 +rt;	t4 =t4 -rt;

/*...Block 2: */
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    t11=aj1p4r -aj1p5r;	t12=aj1p4i-aj1p5i;
	    t9 =aj1p4r +aj1p5r;	t10=aj1p4i+aj1p5i;

	    t15=aj1p6r-aj1p7r;	t16=aj1p6i-aj1p7i;
	    t13=aj1p6r+aj1p7r;	t14=aj1p6i+aj1p7i;

	    rt =t13;	t13=t9 -rt;	t9 =t9 +rt;
	    it =t14;	t14=t10-it;	t10=t10+it;

	    rt =t15;	t15=t11-t16;	t11=t11+t16;
			t16=t12+rt;	t12=t12-rt;

/*...Block 3: */
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    t19=aj1p8r-aj1p9r;	t20=aj1p8i-aj1p9i;
	    t17=aj1p8r+aj1p9r;	t18=aj1p8i+aj1p9i;

	    t23=aj1p10r-aj1p11r;	t24=aj1p10i-aj1p11i;
	    t21=aj1p10r+aj1p11r;	t22=aj1p10i+aj1p11i;

	    rt =t21;	t21=t17-rt;	t17=t17+rt;
	    it =t22;	t22=t18-it;	t18=t18+it;

	    rt =t23;	t23=t19-t24;	t19=t19+t24;
			t24=t20+rt;	t20=t20-rt;

/*...Block 4: */
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    t27=aj1p12r-aj1p13r;	t28=aj1p12i-aj1p13i;
	    t25=aj1p12r+aj1p13r;	t26=aj1p12i+aj1p13i;

	    t31=aj1p14r-aj1p15r;	t32=aj1p14i-aj1p15i;
	    t29=aj1p14r+aj1p15r;	t30=aj1p14i+aj1p15i;

	    rt =t29;	t29=t25-rt;	t25=t25+rt;
	    it =t30;	t30=t26-it;	t26=t26+it;

	    rt =t31;	t31=t27-t32;	t27=t27+t32;
			t32=t28+rt;	t28=t28-rt;
/*
!...and now do four more radix-4 transforms, including the internal twiddle factors:
!	1, exp(-i* 1*twopi/16) =       ( c,-s), exp(-i* 2*twopi/16) = ISRT2*( 1,-1), exp(-i* 3*twopi/16) =       ( s,-c) (for inputs to transform block 2)
!	1, exp(-i* 2*twopi/16) = ISRT2*( 1,-1), exp(-i* 4*twopi/16) =       ( 0,-1), exp(-i* 6*twopi/16) = ISRT2*(-1,-1) (for inputs to transform block 3)
!	1, exp(-i* 3*twopi/16) =       ( s,-c), exp(-i* 6*twopi/16) = ISRT2*(-1,-1), exp(-i* 9*twopi/16) =       (-c, s) (for inputs to transform block 4).
!  (This is 4 real*complex and 4 complex*complex multiplies (= 24 FMUL), compared to 6 and 4, respectively (= 28 FMUL) for my old scheme.)
!   I.e. do similar as above, except inputs a[j1  +p0:15:1] are replaced by t0:30:2,
!					   a[j1+1+p0:15:1] are replaced by t1:31:2, and v.v. for outputs,
!   and only the last 3 inputs to each of the radix-4 transforms 2 through 4 are multiplied by non-unity twiddles.
*/
/*...Block 1: t1,9,17,25 */
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    rt =t9;	t9 =t1 -rt;	t1 =t1 +rt;
	    it =t10;	t10=t2 -it;	t2 =t2 +it;

	    rt =t25;	t25=t17-rt;	t17=t17+rt;
	    it =t26;	t26=t18-it;	t18=t18+it;

#if USE_NEW_DIT
	    a[idum   ]=t1+t17;			a[idum+1 ]=t2+t18;
	    t1	=t1-t17;			t2	=t2-t18;
	    a[idum+16]=t1 *cA8 +t2 *sA8 ;	a[idum+17]=t2 *cA8 -t1 *sA8;

	    rt	=t9 +t26;			it	=t10-t25;
	    t9	=t9 -t26;			t10	=t10+t25;
	    a[idum+8 ]=rt *cA4 +it *sA4 ;	a[idum+9 ]=it *cA4 -rt *sA4;
	    a[idum+24]=t9 *cA12+t10*sA12;	a[idum+25]=t10*cA12-t9 *sA12;
#else
	    a[idum   ]=t1+t17;	a[idum+1 ]=t2+t18;
	    a[idum+16]=t1-t17;	a[idum+17]=t2-t18;

	    a[idum+8 ]=t9 +t26;	a[idum+9 ]=t10-t25;
	    a[idum+24]=t9 -t26;	a[idum+25]=t10+t25;
#endif

/*...Block 3: t5,13,21,29 */
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =t13;	t13=t5 -t14;	t5 =t5 +t14;
			t14=t6 +rt;	t6 =t6 -rt;

	    rt =(t22+t21)*ISRT2;t22=(t22-t21)*ISRT2;	t21=rt;
	    rt =(t29-t30)*ISRT2;it =(t29+t30)*ISRT2;
	    t29=t21+rt;	t21=t21-rt;
	    t30=t22+it;	t22=t22-it;

#if USE_NEW_DIT
	    rt	=t5 +t21;			it	=t6 +t22;
	    t5	=t5 -t21;			t6	=t6 -t22;
	    a[idum+4 ]=rt *cA2 +it *sA2 ;	a[idum+5 ]=it *cA2 -rt *sA2 ;
	    a[idum+20]=t5 *cA10+t6 *sA10;	a[idum+21]=t6 *cA10-t5 *sA10;

	    rt	=t13+t30;			it	=t14-t29;
	    t13	=t13-t30;			t14	=t14+t29;
	    a[idum+12]=rt *cA6 +it *sA6 ;	a[idum+13]=it *cA6 -rt *sA6 ;
	    a[idum+28]=t13*cA14+t14*sA14;	a[idum+29]=t14*cA14-t13*sA14;
#else
	    a[idum+4 ]=t5+t21;	a[idum+5 ]=t6+t22;
	    a[idum+20]=t5-t21;	a[idum+21]=t6-t22;

	    a[idum+12]=t13+t30;	a[idum+13]=t14-t29;
	    a[idum+28]=t13-t30;	a[idum+29]=t14+t29;
#endif

/*...Block 2: t3,11,19,27 */
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    rt =(t12+t11)*ISRT2;it =(t12-t11)*ISRT2;
	    t11=t3 -rt;	t3 =t3 +rt;
	    t12=t4 -it;	t4 =t4 +it;

	    rt =t19*c + t20*s;	t20=t20*c - t19*s;	t19=rt;
	    rt =t27*s + t28*c;	it =t28*s - t27*c;
	    t27=t19-rt;	t19=t19+rt;
	    t28=t20-it;	t20=t20+it;

#if USE_NEW_DIT
	    rt	=t3 +t19;			it	=t4 +t20;
	    t3	=t3 -t19;			t4	=t4 -t20;
	    a[idum+2 ]=rt *cA1 +it *sA1 ;	a[idum+3 ]=it *cA1 -rt *sA1 ;
	    a[idum+18]=t3 *cA9 +t4 *sA9 ;	a[idum+19]=t4 *cA9 -t3 *sA9 ;

	    rt	=t11+t28;			it	=t12-t27;
	    t11	=t11-t28;			t12	=t12+t27;
	    a[idum+10]=rt *cA5 +it *sA5 ;	a[idum+11]=it *cA5 -rt *sA5 ;
	    a[idum+26]=t11*cA13+t12*sA13;	a[idum+27]=t12*cA13-t11*sA13;
#else
	    a[idum+2 ]=t3+t19;	a[idum+3 ]=t4+t20;
	    a[idum+18]=t3-t19;	a[idum+19]=t4-t20;

	    a[idum+10]=t11+t28;	a[idum+11]=t12-t27;
	    a[idum+26]=t11-t28;	a[idum+27]=t12+t27;
#endif

/*...Block 4: t7,15,23,31 */
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =(t15-t16)*ISRT2;it =(t15+t16)*ISRT2;
	    rt =(t15-t16)*ISRT2;it =(t15+t16)*ISRT2;
	    t15=t7 +rt;	t7 =t7 -rt;
	    t16=t8 +it;	t8 =t8 -it;

	    rt =t23*s + t24*c;	t24=t24*s - t23*c;	t23=rt;
	    rt =t31*c + t32*s;	it =t32*c - t31*s;
	    t31=t23+rt;	t23=t23-rt;
	    t32=t24+it;	t24=t24-it;

#if USE_NEW_DIT
	    rt	=t7 +t23;			it	=t8 +t24;
	    t7	=t7 -t23;			t8	=t8 -t24;
	    a[idum+6 ]=rt *cA3 +it *sA3 ;	a[idum+7 ]=it *cA3 -rt *sA3 ;
	    a[idum+22]=t7 *cA11+t8 *sA11;	a[idum+23]=t8 *cA11-t7 *sA11;

	    rt	=t15+t32;			it	=t16-t31;
	    t15	=t15-t32;			t16	=t16+t31;
	    a[idum+14]=rt *cA7 +it *sA7 ;	a[idum+15]=it *cA7 -rt *sA7 ;
	    a[idum+30]=t15*cA15+t16*sA15;	a[idum+31]=t16*cA15-t15*sA15;
#else
	    a[idum+6 ]=t7+t23;	a[idum+7 ]=t8+t24;
	    a[idum+22]=t7-t23;	a[idum+23]=t8-t24;

	    a[idum+14]=t15+t32;	a[idum+15]=t16-t31;
	    a[idum+30]=t15-t32;	a[idum+31]=t16+t31;
#endif

/* 2nd set of inputs: */
	  idum = j2pad;
#if PFETCH
add0 = &a[idum-32];
#endif
/*...Block 1: */
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    t3 =aj2p0r -aj2p1r;	t4 =aj2p0i -aj2p1i;
	    t1 =aj2p0r +aj2p1r;	t2 =aj2p0i +aj2p1i;

	    t7 =aj2p2r -aj2p3r;	t8 =aj2p2i -aj2p3i;
	    t5 =aj2p2r +aj2p3r;	t6 =aj2p2i +aj2p3i;

	    rt =t5;	t5 =t1 -rt;	t1 =t1 +rt;
	    it =t6;	t6 =t2 -it;	t2 =t2 +it;

	    rt =t7;	t7 =t3 -t8;	t3 =t3 +t8;
			t8 =t4 +rt;	t4 =t4 -rt;

/*...Block 2: */
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    t11=aj2p4r -aj2p5r;	t12=aj2p4i-aj2p5i;
	    t9 =aj2p4r +aj2p5r;	t10=aj2p4i+aj2p5i;

	    t15=aj2p6r-aj2p7r;	t16=aj2p6i-aj2p7i;
	    t13=aj2p6r+aj2p7r;	t14=aj2p6i+aj2p7i;

	    rt =t13;	t13=t9 -rt;	t9 =t9 +rt;
	    it =t14;	t14=t10-it;	t10=t10+it;

	    rt =t15;	t15=t11-t16;	t11=t11+t16;
			t16=t12+rt;	t12=t12-rt;

/*...Block 3: */
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    t19=aj2p8r-aj2p9r;	t20=aj2p8i-aj2p9i;
	    t17=aj2p8r+aj2p9r;	t18=aj2p8i+aj2p9i;

	    t23=aj2p10r-aj2p11r;	t24=aj2p10i-aj2p11i;
	    t21=aj2p10r+aj2p11r;	t22=aj2p10i+aj2p11i;

	    rt =t21;	t21=t17-rt;	t17=t17+rt;
	    it =t22;	t22=t18-it;	t18=t18+it;

	    rt =t23;	t23=t19-t24;	t19=t19+t24;
			t24=t20+rt;	t20=t20-rt;

/*...Block 4: */
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    t27=aj2p12r-aj2p13r;	t28=aj2p12i-aj2p13i;
	    t25=aj2p12r+aj2p13r;	t26=aj2p12i+aj2p13i;

	    t31=aj2p14r-aj2p15r;	t32=aj2p14i-aj2p15i;
	    t29=aj2p14r+aj2p15r;	t30=aj2p14i+aj2p15i;

	    rt =t29;	t29=t25-rt;	t25=t25+rt;
	    it =t30;	t30=t26-it;	t26=t26+it;

	    rt =t31;	t31=t27-t32;	t27=t27+t32;
			t32=t28+rt;	t28=t28-rt;

/*
!...and now do four more radix-4 transforms, including the internal twiddle factors:
*/
/*...Block 1: t1,9,17,25 */
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    rt =t9;	t9 =t1 -rt;	t1 =t1 +rt;
	    it =t10;	t10=t2 -it;	t2 =t2 +it;

	    rt =t25;	t25=t17-rt;	t17=t17+rt;
	    it =t26;	t26=t18-it;	t18=t18+it;

#if USE_NEW_DIT
	    a[idum   ]=t1+t17;			a[idum+1 ]=t2+t18;
	    t1	=t1-t17;			t2	=t2-t18;
	    a[idum+16]=t1 *cB8 +t2 *sB8 ;	a[idum+17]=t2 *cB8 -t1 *sB8;

	    rt	=t9 +t26;			it	=t10-t25;
	    t9	=t9 -t26;			t10	=t10+t25;
	    a[idum+8 ]=rt *cB4 +it *sB4;	a[idum+9 ]=it *cB4 -rt *sB4;
	    a[idum+24]=t9 *cB12+t10*sB12;	a[idum+25]=t10*cB12-t9 *sB12;
#else
	    a[idum   ]=t1+t17;	a[idum+1 ]=t2+t18;
	    a[idum+16]=t1-t17;	a[idum+17]=t2-t18;

	    a[idum+8 ]=t9 +t26;	a[idum+9 ]=t10-t25;
	    a[idum+24]=t9 -t26;	a[idum+25]=t10+t25;
#endif

/*...Block 3: t5,13,21,29 */
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =t13;	t13=t5 -t14;	t5 =t5 +t14;
			t14=t6 +rt;	t6 =t6 -rt;

	    rt =(t22+t21)*ISRT2;t22=(t22-t21)*ISRT2;	t21=rt;
	    rt =(t29-t30)*ISRT2;it =(t29+t30)*ISRT2;
	    t29=t21+rt;	t21=t21-rt;
	    t30=t22+it;	t22=t22-it;

#if USE_NEW_DIT
	    rt	=t5 +t21;			it	=t6 +t22;
	    t5	=t5 -t21;			t6	=t6 -t22;
	    a[idum+4 ]=rt *cB2 +it *sB2 ;	a[idum+5 ]=it *cB2 -rt *sB2 ;
	    a[idum+20]=t5 *cB10+t6 *sB10;	a[idum+21]=t6 *cB10-t5 *sB10;

	    rt	=t13+t30;			it	=t14-t29;
	    t13	=t13-t30;			t14	=t14+t29;
	    a[idum+12]=rt *cB6 +it *sB6 ;	a[idum+13]=it *cB6 -rt *sB6 ;
	    a[idum+28]=t13*cB14+t14*sB14;	a[idum+29]=t14*cB14-t13*sB14;
#else
	    a[idum+4 ]=t5+t21;	a[idum+5 ]=t6+t22;
	    a[idum+20]=t5-t21;	a[idum+21]=t6-t22;

	    a[idum+12]=t13+t30;	a[idum+13]=t14-t29;
	    a[idum+28]=t13-t30;	a[idum+29]=t14+t29;
#endif

/*...Block 2: t3,11,19,27 */
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    rt =(t12+t11)*ISRT2;it =(t12-t11)*ISRT2;
	    t11=t3 -rt;	t3 =t3 +rt;
	    t12=t4 -it;	t4 =t4 +it;

	    rt =t19*c + t20*s;	t20=t20*c - t19*s;	t19=rt;
	    rt =t27*s + t28*c;	it =t28*s - t27*c;
	    t27=t19-rt;	t19=t19+rt;
	    t28=t20-it;	t20=t20+it;

#if USE_NEW_DIT
	    rt	=t3 +t19;			it	=t4 +t20;
	    t3	=t3 -t19;			t4	=t4 -t20;
	    a[idum+2 ]=rt *cB1 +it *sB1 ;	a[idum+3 ]=it *cB1 -rt *sB1 ;
	    a[idum+18]=t3 *cB9 +t4 *sB9 ;	a[idum+19]=t4 *cB9 -t3 *sB9 ;

	    rt	=t11+t28;			it	=t12-t27;
	    t11	=t11-t28;			t12	=t12+t27;
	    a[idum+10]=rt *cB5 +it *sB5 ;	a[idum+11]=it *cB5 -rt *sB5 ;
	    a[idum+26]=t11*cB13+t12*sB13;	a[idum+27]=t12*cB13-t11*sB13;
#else
	    a[idum+2 ]=t3+t19;	a[idum+3 ]=t4+t20;
	    a[idum+18]=t3-t19;	a[idum+19]=t4-t20;

	    a[idum+10]=t11+t28;	a[idum+11]=t12-t27;
	    a[idum+26]=t11-t28;	a[idum+27]=t12+t27;
#endif

/*...Block 4: t7,15,23,31 */
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =(t15-t16)*ISRT2;it =(t15+t16)*ISRT2;
	    rt =(t15-t16)*ISRT2;it =(t15+t16)*ISRT2;
	    t15=t7 +rt;	t7 =t7 -rt;
	    t16=t8 +it;	t8 =t8 -it;

	    rt =t23*s + t24*c;	t24=t24*s - t23*c;	t23=rt;
	    rt =t31*c + t32*s;	it =t32*c - t31*s;
	    t31=t23+rt;	t23=t23-rt;
	    t32=t24+it;	t24=t24-it;

#if USE_NEW_DIT
	    rt	=t7 +t23;			it	=t8 +t24;
	    t7	=t7 -t23;			t8	=t8 -t24;
	    a[idum+6 ]=rt *cB3 +it *sB3 ;	a[idum+7 ]=it *cB3 -rt *sB3 ;
	    a[idum+22]=t7 *cB11+t8 *sB11;	a[idum+23]=t8 *cB11-t7 *sB11;

	    rt	=t15+t32;			it	=t16-t31;
	    t15	=t15-t32;			t16	=t16+t31;
	    a[idum+14]=rt *cB7 +it *sB7 ;	a[idum+15]=it *cB7 -rt *sB7 ;
	    a[idum+30]=t15*cB15+t16*sB15;	a[idum+31]=t16*cB15-t15*sB15;
#else
	    a[idum+6 ]=t7+t23;	a[idum+7 ]=t8+t24;
	    a[idum+22]=t7-t23;	a[idum+23]=t8-t24;

	    a[idum+14]=t15+t32;	a[idum+15]=t16-t31;
	    a[idum+30]=t15-t32;	a[idum+31]=t16+t31;
#endif

/*...And update the data (j1 and j2) array indices. */

	  j1 = j1+32;
	  j2 = j2-32;

	}
/*
!...Since the foregoing loop only gets executed half as many times as in the simple version, to properly position
!   ourselves in the data array for the start of the next block, need to bump up j1 by as much as would occur in a
!   second execution of the above loop. The exception is the first loop execution, where j1 needs to be doubled (32 x 2).
*/
	j1 = j1+(blocklen << 1);

	if(j2_start == n-32)
	{
		j1 = 0;
		return;
	}

/*
!...Since the foregoing loop only gets executed half as many times as in the simple version, to properly position ourselves in the data
!   array for the start of the next block, need to bump up j1 by as much as would occur in a second execution of the above loop.
!
!   Examples:
!
!* N=768=3*2^8                                  complex data index ranges
!*      =1100000000_2.          blocklength:    evens:  j1      odds:   j2        j2_next = j2_start + next blocklength
!* k=N-1=1011111111, lobit(k>>1) = 1     2              2               3         7
!*      k=101111111, lobit(k>>1) = 1,    4              4-6             7-5       15
!*      k= 10111111, lobit(k>>1) = 1,    8              8-14            15-9      31
!*      k=  1011111, lobit(k>>1) = 1,   16              16-30           31-17     63
!*      k=   101111, lobit(k>>1) = 1,   32              32-62           63-33     127
!*      k=    10111, lobit(k>>1) = 1,   64              64-126          127-65    255
!*      k=     1011, lobit(k>>1) = 1,  128              128-254         255-129   767
!*      k=      101, lobit(k>>1) = 0,  512 = 128*(k-1)  256-766         767-257   j2_start = N-1; STOP    <---NOTE: sum of blocklengths up to this point + 2 = N.
!*                                (The added 2 is for elements 0 and 1, processed separately.)
!
!* N=1280=5*2^8
!*      =10100000000_2: lobit(k>>1)             Sum:    j1              j2        j2 at start of next block
!* k=N-1=10011111111      1              2         4    2               3         7
!*        1001111111      1              4         8    4-6             7-5       15
!*         100111111      1              8        16    8-14            15-9      31
!*          10011111      1             16        32    16-30           31-17     63
!*           1001111      1             32        64    32-62           63-33     127
!*            100111      1             64       128    64-126          127-65    255
!*             10011      1            128       256    128-254         255-129   1279
!*              1001      0 128*(k-1)=1024      1280    256-1278        1279-257  j2_start = N-1; STOP
!
!* N=1792=7*2^8
!*      =11100000000_2: lobit(k>>1)             Sum:    j1              j2        j2 at start of next block
!* k=N-1=11011111111      1              2         4    2               3         7
!*        1101111111      1              4         8    4-6             7-5       15
!*         110111111      1              8        16    8-14            15-9      31
!*          11011111      1             16        32    16-30           31-17     63
!*           1101111      1             32        64    32-62           63-33     127
!*            110111      1             64       128    64-126          127-65    255
!*             11011      1            128       256    128-254         255-129   1791
!*              1101      0 128*(k-1)=1536      1792    256-1790        1791-257  j2_start = N-1; STOP
*/

/*...Reset half-complex-blocklength for next pass. If K >> 1 has a zero trailing bit, we multiply the blocklength by K >> 1 in preparation for the final block.
*/

	blocklen_sum = blocklen_sum + blocklen;
	blocklen = (radix_prim[i-1]-1)*blocklen_sum;

/*...Next j2_start is previous one plus the (real) length of the current block = 4*(half-complex-blocklength) */

	j2_start = j2_start+(blocklen<<2);
	j2=j2_start;			    /* Reset j2 for start of the next block. */
/*printf("newblock: blocklen = %8d blocklen_sum = %8d j2 = %8d\n",blocklen,blocklen_sum,j2);*/
      }	 /* End of Main loop */

}

void pair_square(double *x1, double *y1, double *x2, double *y2, double c, double s)
{
/*
!   Given complex scalars H[j] = (x1,y1) and H[N-j] = (x2,y2) along with complex exponential E = (c,s),
!   calculates I[j] = H[j]^2 + {1 + exp(2*pi*I*j/N)}*{H[j]-H~[N-j]}^2/4 and its complex conjugate I~,
!   returns the former in H[j] and the latter in H[N-j].
*/
	double rt0,rt1,rt2,rt3,it1,it2,it3;

	rt1=*x1;
	it1=*y1;
	rt2=*x2;
	it2=*y2;

/*   calculate cross-product terms... */

	rt3=rt1*rt2+it1*it2; rt3=rt3+rt3;
	it3=it1*rt2-rt1*it2; it3=it3+it3;

/*   now calculate square terms and store back in the same temporaries. */

	rt0=(rt1+it1)*(rt1-it1); it1=rt1*it1; it1=it1+it1; rt1=rt0;
	rt0=(rt2+it2)*(rt2-it2); it2=rt2*it2; it2=it2+it2; rt2=rt0;

/*   use that (H[j] - H~[N-j])^2 = H(j)^2 - 2*H(j)*H~(N-j) + H~(N-j)^2... */
	rt3=rt1+rt2-rt3;
	it3=it1-it2-it3;
	rt0=((c+1.0)*rt3-s*it3)*0.25;
	it3=(s*rt3+(c+1.0)*it3)*0.25;

/*...and now complete and store the results. */

	*x1 = (rt1-rt0);
	*y1 = (it1-it3);

/*...N-j terms are as above, but with the replacements: rt1<-->rt2, it1<-->it2, it3|-->-it3. */

	*x2 = (rt2-rt0);
	*y2 = (it2+it3);

}

void pair_square2A(double *x1, double *y1, double *x2, double *y2,
		   double *x3, double *y3, double *x4, double *y4,
		   double c, double s)
{
/*
!   Given complex scalars H[j] = (x1,y1) and H[N-j] = (x2,y2) along with complex exponential E = (c,s),
!   calculates I[j] = H[j]^2 + {1 + exp(2*pi*I*j/N)}*{H[j]-H~[N-j]}^2/4 and its complex conjugate I~,
!   returns the former in H[j] and the latter in H[N-j].
*/
	double rt0,rt1,rt2,rt3,it1,it2,it3;
			double st0,st1,st2,st3,jt1,jt2,jt3;

	rt1=*x1;
			st1=*x3;

	it1=*y1;
			jt1=*y3;

	rt2=*x2;
			st2=*x4;

	it2=*y2;
			jt2=*y4;

/*   calculate cross-product terms... */
		/*   calculate cross-product terms... */

	rt3=rt1*rt2+it1*it2; rt3=rt3+rt3;
			st3=st1*st2+jt1*jt2; st3=st3+st3;

	it3=it1*rt2-rt1*it2; it3=it3+it3;
			jt3=jt1*st2-st1*jt2; jt3=jt3+jt3;

/*   now calculate square terms and store back in the same temporaries. */
		/*   now calculate square terms and store back in the same temporaries. */


	rt0=(rt1+it1)*(rt1-it1); it1=rt1*it1; it1=it1+it1; rt1=rt0;
			st0=(st1+jt1)*(st1-jt1); jt1=st1*jt1; jt1=jt1+jt1; st1=st0;
	rt0=(rt2+it2)*(rt2-it2); it2=rt2*it2; it2=it2+it2; rt2=rt0;
			st0=(st2+jt2)*(st2-jt2); jt2=st2*jt2; jt2=jt2+jt2; st2=st0;


/*   use that (H[j] - H~[N-j])^2 = H(j)^2 - 2*H(j)*H~(N-j) + H~(N-j)^2... */
		/*   use that (H[j] - H~[N-j])^2 = H(j)^2 - 2*H(j)*H~(N-j) + H~(N-j)^2... */

	rt3=rt1+rt2-rt3;
			st3=st1+st2-st3;

	it3=it1-it2-it3;
			jt3=jt1-jt2-jt3;

	rt0=((c+1.0)*rt3-s*it3)*0.25;
			st0=((1.0-s)*st3-c*jt3)*0.25;

	it3=(s*rt3+(c+1.0)*it3)*0.25;
			jt3=(c*st3+(1.0-s)*jt3)*0.25;

/*...and now complete and store the results. */
		/*...and now complete and store the results. */

	*x1 = (rt1-rt0);
			*x3 = (st1-st0);

	*y1 = (it1-it3);
			*y3 = (jt1-jt3);

/*...N-j terms are as above, but with the replacements: rt1<-->rt2, it1<-->it2, it3|-->-it3. */
				/*...N-j terms are as above, but with the replacements: st1<-->st2, jt1<-->jt2, jt3|-->-jt3. */

	*x2 = (rt2-rt0);
			*x4 = (st2-st0);

	*y2 = (it2+it3);
			*y4 = (jt2+jt3);
}

void pair_square2B(double *x1, double *y1, double *x2, double *y2,
		   double *x3, double *y3, double *x4, double *y4,
		   double c, double s)
{
/*
!   Given complex scalars H[j] = (x1,y1) and H[N-j] = (x2,y2) along with complex exponential E = (c,s),
!   calculates I[j] = H[j]^2 + {1 + exp(2*pi*I*j/N)}*{H[j]-H~[N-j]}^2/4 and its complex conjugate I~,
!   returns the former in H[j] and the latter in H[N-j].
*/
	double rt0,rt1,rt2,rt3,it1,it2,it3;
			double st0,st1,st2,st3,jt1,jt2,jt3;

	rt1=*x1;
			st1=*x3;

	it1=*y1;
			jt1=*y3;

	rt2=*x2;
			st2=*x4;

	it2=*y2;
			jt2=*y4;

/*   calculate cross-product terms... */
		/*   calculate cross-product terms... */

	rt3=rt1*rt2+it1*it2; rt3=rt3+rt3;
			st3=st1*st2+jt1*jt2; st3=st3+st3;

	it3=it1*rt2-rt1*it2; it3=it3+it3;
			jt3=jt1*st2-st1*jt2; jt3=jt3+jt3;

/*   now calculate square terms and store back in the same temporaries. */
		/*   now calculate square terms and store back in the same temporaries. */

	rt0=(rt1+it1)*(rt1-it1); it1=rt1*it1; it1=it1+it1; rt1=rt0;
			st0=(st1+jt1)*(st1-jt1); jt1=st1*jt1; jt1=jt1+jt1; st1=st0;
	rt0=(rt2+it2)*(rt2-it2); it2=rt2*it2; it2=it2+it2; rt2=rt0;
			st0=(st2+jt2)*(st2-jt2); jt2=st2*jt2; jt2=jt2+jt2; st2=st0;

/*   use that (H[j] - H~[N-j])^2 = H(j)^2 - 2*H(j)*H~(N-j) + H~(N-j)^2... */
		/*   use that (H[j] - H~[N-j])^2 = H(j)^2 - 2*H(j)*H~(N-j) + H~(N-j)^2... */

	rt3=rt1+rt2-rt3;
			st3=st1+st2-st3;

	it3=it1-it2-it3;
			jt3=jt1-jt2-jt3;

	rt0=((c+1.0)*rt3-s*it3)*0.25;
			st0=((s+1.0)*st3+c*jt3)*0.25;

	it3=(s*rt3+(c+1.0)*it3)*0.25;
			jt3=((s+1.0)*jt3-c*st3)*0.25;

/*...and now complete and store the results. */
		/*...and now complete and store the results. */

	*x1 = (rt1-rt0);
			*x3 = (st1-st0);

	*y1 = (it1-it3);
			*y3 = (jt1-jt3);

/*...N-j terms are as above, but with the replacements: rt1<-->rt2, it1<-->it2, it3|-->-it3. */
				/*...N-j terms are as above, but with the replacements: st1<-->st2, jt1<-->jt2, jt3|-->-jt3. */

	*x2 = (rt2-rt0);
			*x4 = (st2-st0);

	*y2 = (it2+it3);
			*y4 = (jt2+jt3);
}

