/*******************************************************************************
*                                                                              *
*   (C) 1997-2004 by Ernst W. Mayer.                                           *
*                                                                              *
*  This program is free software; you can redistribute it and/or modify it     *
*  under the terms of the GNU General Public License as published by the       *
*  Free Software Foundation; either version 2 of the License, or (at your      *
*  option) any later version.                                                  *
*                                                                              *
*  This program is distributed in the hope that it will be useful, but WITHOUT *
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
*  more details.                                                               *
*                                                                              *
*  You should have received a copy of the GNU General Public License along     *
*  with this program; see the file GPL.txt.  If not, you may view one at       *
*  http://www.fsf.org/licenses/licenses.html, or obtain one by writing to the  *
*  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     *
*  02111-1307, USA.                                                            *
*                                                                              *
*******************************************************************************/

#include "Mlucas.h"

#define QFDBG 0		/* Set this = 1 if you suspect something is qoing awry with the DWT weights and/or FFT sincos data. */

/***************/

int mers_mod_square(double a[], int n, int ilo, int ihi, int p, uint32 *err_iter, int scrnFlag, double *tdiff)
{

/*...Subroutine to perform Mersenne-mod squaring using Crandall and Fagin's discrete weighted transform (DWT)
     on the data in the length-N real vector A.

     Acronym: DIF = Decimation In Frequency, DIT = Decimation In Time.

     One-pass combined fwd-DWT-weighting/fwd-DIF-FFT/pointwise-squaring/inv-DIT-FFT/inv-DWT-weighting routine
     for use with a packed-vector complex transform to halve the runlength of the corresponding real-vector transform.

     At the beginning of each loop execution, data are assumed to have already been forward-weighted
     and the initial-radix pass of the S-pass forward FFT done. The loop then does the following:

     (1) Performs the 2nd through (S-1)st passes of the complex DIF forward FFT;
     (2) Does the final-radix forward FFT pass, the real/complex-wrapper/pointwise squaring step,
         and the initial-radix inverse FFT pass in a single pass through the data;
     (3) Performs the 2nd through (S-1)st passes of the complex DIT inverse FFT,
         with radices processed in reverse order from the forward FFT (this is
         not necessary for power-of-2 transform lengths, but ensures that DIF radix 1
         equals DIT radix S and vice versa, which is required for steps (2) and (4));
     (4) Does the final-radix inverse FFT pass, the inverse DWT weighting, the carry
         propagation step (with fractional roundoff error check), the forward DWT weighting,
         and the initial-radix forward FFT pass in a single pass through the data.
*/
struct qfloat qmul, qwt, qt, qn;	/* qfloats used for DWT weights calculation. */
struct qfloat qtheta, qr, qi, qc, qs;	/* qfloats used for FFT sincos  calculation. */
double t1;
#if QFDBG
long double theta,mt;
double t2;
uint64 i1,i2,idiff;
#endif
	static int radix_set_save = 1000;
	static int nradices_prim,nradices_radix0,radix_prim[30];/* RADIX_PRIM stores sequence of complex FFT radices used, in terms of their prime factors.	*/
	static int *index = 0x0, *si = 0x0, *index_ptmp = 0x0, *si_ptmp = 0x0;		/* Bit-reversal index array and array storing S*I mod N values for DWT weights.	*/
	static int *block_index;				/* array storing the RADIX_VEC[0] data-block indices for pass-2 of the FFT.	*/

	/* arrays storing the index values needed for the parallel-block wrapper/square scheme: */
	static int *ws_i,*ws_j1,*ws_j2,*ws_j2_start,*ws_k,*ws_m,*ws_blocklen,*ws_blocklen_sum;
	int bimodn,simodn;					/* Mnemonic: BIMODN stands for "B times I mod N", nothing to do with bimodal.	*/
	int i,ii,ierr,incr,istart,iter,j,jhi,jstart,k,koffset,l,m,mm,k2,m2,l1,l2,l2_start,blocklen,blocklen_sum,outer;
	static int psave=0,n2,nrt,nrt_bits,nwt,nwt_bits,bw,sw,bits_small;
	const  double one_half[3] = {1.0, 0.5, 0.25};		/* Needed for small-weights-tables scheme */
	static double base[2],baseinv[2],radix_inv;
	static struct complex *rt0 = 0x0, *rt1 = 0x0, *rt0_ptmp = 0x0, *rt1_ptmp = 0x0;		/* reduced-size roots of unity arrays	*/
	static double *wt0 = 0x0, *wt1 = 0x0, *tmp = 0x0, *wt0_ptmp = 0x0, *wt1_ptmp = 0x0, *tmp_ptmp = 0x0;		/* reduced-size DWT weights arrays	*/
#if NEGACYCLIC
	static double *rt2 = 0x0, *rt2_ptmp = 0x0;
#endif
	double fracmax,wt,wtinv;
	static int first_entry=TRUE;		/* Master variable controlling whether the init sequences in subroutines get done (.true.) or not.	*/
	FILE *f1,*f2;
	clock_t clock1, clock2;
/*
!...What a bunch of characters...
	character :: char_time1*10,char_time2*10,tdiff*12
	character*12, external :: get_time_diff
*/

/*...Init clock counter:	*/

	clock1 = clock();
	*tdiff = 0.0;

/*...If a new exponent or radix set, deallocate any already-allocated allocatable arrays and set first_entry to true:	*/

	if(p != psave || RADIX_SET != radix_set_save) first_entry=TRUE;

/*...on first entry, deallocate any already-allocated allocatable arrays and init DWT weights and FFT sincos tables:	*/
	if(first_entry) {
	  psave = p;
	  if(rt0)	/* If it's a new exponent of a range test, need to deallocate these. */
	  {
	    free((void *)wt0_ptmp       ); wt0_ptmp        = 0x0; wt0   = 0x0;
	    free((void *)wt1_ptmp       ); wt1_ptmp        = 0x0; wt1   = 0x0;
	    free((void *)tmp_ptmp       ); tmp_ptmp        = 0x0; tmp   = 0x0;
	    free((void *)si_ptmp        ); si_ptmp         = 0x0; si    = 0x0;
	    free((void *)rt0_ptmp       ); rt0_ptmp        = 0x0; rt0   = 0x0;
	    free((void *)rt1_ptmp       ); rt1_ptmp        = 0x0; rt1   = 0x0;
#if NEGACYCLIC
	    free((void *)rt2_ptmp       ); rt2_ptmp        = 0x0; rt2   = 0x0;
#endif
	    free((void *)index_ptmp     ); index_ptmp      = 0x0; index = 0x0;
	    free((void *)block_index    ); block_index     = 0x0;
	    free((void *)ws_i           ); ws_i            = 0x0;
	    free((void *)ws_j1          ); ws_j1           = 0x0;
	    free((void *)ws_j2          ); ws_j2           = 0x0;
	    free((void *)ws_j2_start    ); ws_j2_start     = 0x0;
	    free((void *)ws_k           ); ws_k            = 0x0;
	    free((void *)ws_m           ); ws_m            = 0x0;
	    free((void *)ws_blocklen    ); ws_blocklen     = 0x0;
	    free((void *)ws_blocklen_sum); ws_blocklen_sum = 0x0;
	  }
	  first_entry=FALSE;

	  bw     = p%n;		/* Number of bigwords in the Crandall/Fagin mixed-radix representation = (Mersenne exponent) mod (vector length).	*/
	  sw     = n - bw;	/* Number of smallwords.	*/

	  n2 =n/2;		/* Complex vector length.	*/

	/*
	  This call sets NRADICES and the first (NRADICES) elements of RADIX_VEC:
	*/
	  get_fft_radices(n, RADIX_SET, scrnFlag);

	  /* My array padding scheme requires N/RADIX_VEC[0] to be a power of 2, and to be >= 2^DAT_BITS, where the latter
	  parameter is set in the Mdata.h file: */

	  if(n%RADIX_VEC[0] != 0){ printf("FATAL: RADIX_VEC[0] does not divide N!\n"); exit(EXIT_FAILURE); }

	  /* Quick-n-dirty hack to chek if a 32-bit int is a power of 2: */
	  i = n/RADIX_VEC[0];
	  if(leadz(i)+trailz(i) != 31){ printf("FATAL: n/RADIX_VEC[0] not a power of 2!\n"); exit(EXIT_FAILURE); }

	  if(DAT_BITS < 31)
	  {
		/* Now make sure n/RADIX_VEC[0] is sufficiently large (unless n < 2^DAT_BITS, in which case it doesn't matter): */
		if(i < (1 << DAT_BITS)){ printf("FATAL: n/RADIX_VEC[0] must be >= %u!\n", (1 << DAT_BITS)); exit(EXIT_FAILURE); }

		/* We also have a lower limit on 2^DAT_BITS set by the wrapper_square routine: */
		if((1 << DAT_BITS) < 2*RADIX_VEC[NRADICES-1]){ printf("FATAL: final FFT radix may not exceed = %u!\n", (1 << (DAT_BITS-1))); exit(EXIT_FAILURE); }
	  }

	  if(RADIX_SET != radix_set_save)
	  {
	    radix_set_save = RADIX_SET;
	    if(INTERACT)
	    {
	      printf("Using complex FFT radices"); for(i = 0; i < NRADICES; i++){ printf("%10d",RADIX_VEC[i]); }; printf("\n");
	    }
  	    else
  	    {
	      f1 = fopen(STATFILE,"a");
	      fprintf(f1,"Using complex FFT radices"); for(i = 0; i < NRADICES; i++){ fprintf(f1,"%10d",RADIX_VEC[i]); }; fprintf(f1,"\n");
  	      fclose(f1);
	      if (scrnFlag)                               /* Echo output to stddev */
	      {
	        printf("Using complex FFT radices"); for(i = 0; i < NRADICES; i++){ fprintf(f1,"%10d",RADIX_VEC[i]); }; fprintf(f1,"\n");
	      }
  	    }
	  }

	  radix_inv=(double)(1.0L/RADIX_VEC[0]);

/*...stuff for the reduced-length DWT weights arrays is here:	*/

	  nwt_bits = (int)(log(sqrt(1.0*n))/log(2.0) + 0.5);	   /* No need for a fancy NINT here.	*/
	  nwt    = 1 << nwt_bits;	/* To save on storage, we calculate the first NWT weights directly and then re-use
			                   them N/NWT times, each time multiplying the basic weights by a single scalar multiplier (and times 0.5
			                   or 1.0). Thus, the total number of weights data is NWT + (N/NWT). To minimize this, we find the positive
			                   minimum of the function f(x) = x + N/x, which occurs when f' = 1 - N/(x^2) = 0, or x = NWT = sqrt(N),
			                   which gives the total number of weights data as 2*sqrt(N). However, to make the algorithm efficient,
			                   we want NWT a power of 2, so we take NWT = 2^[nint(log2(sqrt(N)))]. In the worst case this needs two
			                   arrays, one with sqrt(2*N) elements, the other with sqrt(2*N)/2, for a total of 3*sqrt(2)*sqrt(N)/2
			                   elements, only a few percent more than the minimum possible number.
			                   The NWT basic weights are stored in the WT0 array; the N/NWT scalar multipliers are in WT1.	*/

	  if(n%nwt){ printf("FATAL: NWT does not divide N!\n"); exit(EXIT_FAILURE); }

/*...The roots arrays need only be half the dimension of the weights arrays (since we need n/2 complex roots
     vs. n real weights), but have the same total storage since each entry is complex:	*/
	  /*
	  wt0 = (double *)calloc(nwt+1         ,sizeof(double));
	  wt1 = (double *)calloc(n/nwt+RADIX_VEC[0],sizeof(double));
	  tmp = (double *)calloc(n/nwt+1       ,sizeof(double));
	  si  = (   int *)calloc(nwt+1         ,sizeof(   int));
	  */
	  wt0_ptmp = ALLOC_DOUBLE(nwt+1         );	if(!wt0_ptmp){ printf("FATAL: unable to allocate array WT0 in mers_mod_square.\n"); exit(EXIT_FAILURE); }; wt0 = ALIGN_DOUBLE(wt0_ptmp);
	  wt1_ptmp = ALLOC_DOUBLE(n/nwt+RADIX_VEC[0]);	if(!wt1_ptmp){ printf("FATAL: unable to allocate array WT1 in mers_mod_square.\n"); exit(EXIT_FAILURE); }; wt1 = ALIGN_DOUBLE(wt1_ptmp);
	  tmp_ptmp = ALLOC_DOUBLE(n/nwt+1       );	if(!tmp_ptmp){ printf("FATAL: unable to allocate array TMP in mers_mod_square.\n"); exit(EXIT_FAILURE); }; tmp = ALIGN_DOUBLE(tmp_ptmp);
	  si_ptmp  = ALLOC_INT   (nwt+1         );	if(!si_ptmp ){ printf("FATAL: unable to allocate array SI  in mers_mod_square.\n"); exit(EXIT_FAILURE); }; si  = ALIGN_INT   (si_ptmp );

/*...Find the Crandall/Fagin weighting factors and number of bits per digit. The TMP array gets rearranged
    in a cache-friendly way to obtain the WT1 array, but keep the TMP array around to simplify the weighting
    steps at the beginning and end of each iteration cycle, where speed is not crucial.	*/

/* For qfloat implementation, speed things by defining a constant multiplier qmul = 2^(s/N), repeatedly
   multiplying the previous weight factor by that, and dividing by 2 whenever simodn >= N.	*/

	qt   = i64_to_q((int64)sw);
	qn   = i64_to_q((int64) n);
	qt   = qfdiv(qt, qn);		/* s/N...	 */
	qt   = qfmul(qt, QLN2);		/* ...(s/N)*ln(2)...	*/
	qmul = qfexp(qt);		/* ...and get 2^(s/N) via exp[(s/N)*ln(2)]. */
	qwt  = QONE;	/* init weights multiplier chain. */

t1 = qfdbl(qmul);
#if QFDBG
t2 = pow((long double)2.0L, (long double)1.0L*sw/n);
i1 = *(uint64 *)&t1;
i2 = *(uint64 *)&t2;
if(i1 != i2)
{
	idiff = labs(i1-i2);
	printf("INFO: QWT1= %16llX, DWT = %16llX DIFFER BY %20llu\n", i1, i2, idiff);
}
#endif

	  simodn=0;
	  for(i=0; i<nwt; i++)
	  {
t1 = qfdbl(qwt);
#if QFDBG
t2 = pow((long double)2.0L, (long double)1.0L*simodn/n);
i1 = *(uint64 *)&t1;
i2 = *(uint64 *)&t2;
if(i1 != i2)
{
	idiff = labs(i1-i2);
	if(idiff >= (uint64)2)
	{
		printf("INFO: I = %8d: QWT = %16llX, DWT = %16llX DIFFER BY %20llu\n", i, i1, i2, idiff);
	/*	assert(0);	*/
	}
}
#endif
	    wt0  [i] = t1;	/* Ith DWT weight factor = 2^[(s*i mod N)/N], where the exponent is done using floating divide.	*/
	    si   [i] = simodn;
/*printf("I = %d; WT0 = %20.10f; SI = %d\n",i,wt0[i],si[i]);	*/
	    simodn = simodn + sw;

qwt= qfmul(qwt, qmul);

	    if(simodn >= n)
	    {
	      simodn = simodn - n;
qwt= qfmul_pow2(qwt, -1);
	    }
	  }

	  wt0[nwt] = 2*wt0[0];	/* This is so the case L = 0 comes out right.	*/
	  si [nwt] = n;		/* When L = 0, want to ensure (B*J mod N) - SI(NWT) < 0, so set SI(NWT) := N	*/

	  k=(int)(1.0*sw*nwt - 1.0*((int)(1.0*sw*nwt/n))*n);	/* The Jth scalar weights multiplier is 2^[(J*S*NWT mod N)/N].	*/
								/* Use real*8 to do the MOD, since S*NWT may be > 31 bits.	*/
	qt   = i64_to_q((int64) k);
	qn   = i64_to_q((int64) n);
	qt   = qfdiv(qt, qn);		/* k/N...	 */
	qt   = qfmul(qt, QLN2);	/* ...(k/N)*ln(2)...	*/
	qmul = qfexp(qt);		/* ...and get 2^(k/N) via exp[(k/N)*ln(2)]. */
	qwt  = QONE;	/* init weights multiplier chain. */

t1 = qfdbl(qmul);
#if QFDBG
t2 = pow((long double)2.0L, (long double)1.0L*k/n);
i1 = *(uint64 *)&t1;
i2 = *(uint64 *)&t2;
if(i1 != i2)
{
	idiff = labs(i1-i2);
	printf("INFO: QWT2= %16llX, DWT = %16llX DIFFER BY %20llu\n", i1, i2, idiff);
}
#endif

	  j=0;			/* Store I*K mod NN here. We don't directly calculate I*K, since that can overflow a 32-bit integer at large runlengths.	*/
	  for(i=0; i<n/nwt; i++)
	  {
t1 = qfdbl(qwt);
#if QFDBG
t2 = pow((long double)2.0L, (long double)1.0L*j/n);
i1 = *(uint64 *)&t1;
i2 = *(uint64 *)&t2;
if(i1 != i2)
{
	idiff = labs(i1-i2);
	if(idiff >= (uint64)2)
	{
		printf("INFO: J = %8d: QWT = %16llX, DWT = %16llX DIFFER BY %20llu\n", i, i1, i2, idiff);
	/*	assert(0);	*/
	}
}
#endif
	    tmp[i] = t1;
/*printf("I = %d; TMP = %20.10f\n",i,tmp[i]);	*/
	    j=j+k;

qwt= qfmul(qwt, qmul);

	    if(j >= n)
	    {
	      j = j - n;
qwt= qfmul_pow2(qwt, -1);
	    }
	  }
	  tmp[n/nwt] = 2*tmp[0];	/* This is so the case L = 0 comes out right.	*/

	  bits_small = p/n;			/* number of bits in a smallword.	*/
	  base   [0] = (double)(1 << bits_small);	base   [1] = (double)(2*base[0]);
	  baseinv[0] = (double)(1.0/base[0]    );	baseinv[1] = (double)(1.0/base[1]);	/* don't need extended precision for this since both bases are powers of 2.	*/

/*...In the actual radix*_ditN_cy_dif1 carry propagation routine, the elements of the second weights table
     as constructed above are accessed in strides of length n/(radix1*nwt), so it makes sense to prearrange
     them so as to replace these long strides with unit strides, and thus to be accessing contiguous data instead.	*/

	  for(i=0; i<n/(nwt*RADIX_VEC[0]); i++)
	  {
	    for(j=i*RADIX_VEC[0], k=i; j<(i+1)*RADIX_VEC[0]; j++, k += n/(nwt*RADIX_VEC[0]))
	    {
	      wt1[j] = tmp[k];	/* Gather (radix1) stride-[n/(radix1*nwt)]-separated data into a contiguous block.		*/
	    }
	  }
	  for(j=n/nwt, k=n/(nwt*RADIX_VEC[0]); j < (n/nwt+RADIX_VEC[0]); j++, k += n/(nwt*RADIX_VEC[0]))
	  {
	    wt1[j] = tmp[k];	/* This is so the case L = 0 comes out right.	*/
	  }

/*...Stuff for the reduced-length roots arrays is here:	*/

/*...The roots arrays need only be half the dimension of the weights arrays (since we need n/2 complex roots
     vs. n real weights), but have the same total storage since each entry is complex:	*/

	  nrt=nwt;	nrt_bits=nwt_bits;
	  /*
	  rt0 = (struct complex *)calloc(nwt      ,sizeof(struct complex));
	  rt1 = (struct complex *)calloc(n/(2*nwt),sizeof(struct complex));
	  */
#if NEGACYCLIC
	/* If the modulus is such that a length-n negacyclic convolution is needed, we need
	!...N/2 Nth roots of -1 (i.e. 2*Nth roots of unity), rt2(j)=exp(i*j*pi/N), j=0,...,N/2-1.
	!   We can exploit the symmetries of complex exponentials to reduce storage and improve performance.
	!
	!   Symmetry (1): Since each component of WT is simply a complex exponential, inverse weights
	!                 are just complex conjugates, and don't need to be stored separately.
	!
	!   Symmetry (2): We can use the following symmetries about pi/4:
	!                    cos(pi/2-x) = sin x,  sin(pi/2-x) = cos x
	!                 to reduce the storage by half again. Namely, just define wt(0:N/4-1),
	!                 and whenever a weight corresponding to j >= N/4 is needed, just use [ Im(wt(N/2-j) + I*Re(wt(N/2-j) ].
	*/

#endif
/*...The rt0 array stores the (0:nrt-1)th powers of the [n2]th root of unity
(i.e. will be accessed using the lower (nrt) bits of the integer sincos index):
*/
	  rt0_ptmp = ALLOC_COMPLEX(nwt      );	if(!rt0_ptmp){ printf("FATAL: unable to allocate array RT0 in mers_mod_square.\n"); exit(EXIT_FAILURE); } rt0 = ALIGN_COMPLEX(rt0_ptmp);

	qt     = i64_to_q((int64)n2);
	qtheta = qfdiv(Q2PI, qt);	/* 2*pi/(N/2) */
	qr     = qfcos(qtheta);
	qi     = qfsin(qtheta);
	qc     = QONE; qs = QZRO;	/* init sincos multiplier chain. */

	t1 = qfdbl(qr);
	#if QFDBG
	theta = qfdbl(Q2PI)/n2;
	t2 = cos(theta);
	i1 = *(uint64 *)&t1;
	i2 = *(uint64 *)&t2;
	if(i1 != i2)
	{
		idiff = labs(i1-i2);
		printf("INFO: QCOS1= %16llX, DCOS = %16llX DIFFER BY %20llu\n", i1, i2, idiff);
	}
	#endif

	t1 = qfdbl(qi);
	#if QFDBG
	t2 = sin(theta);
	i1 = *(uint64 *)&t1;
	i2 = *(uint64 *)&t2;
	if(i1 != i2)
	{
		idiff = labs(i1-i2);
		printf("INFO: QSIN1= %16llX, DSIN = %16llX DIFFER BY %20llu\n", i1, i2, idiff);
	}
	#endif

	qt = QZRO;

	  for(i=0; i<nwt; i++)
	  {
		qc = qfcos(qt);
		t1 = qfdbl(qc);
	#if QFDBG
		mt = i*theta;
		t2 = cos(mt);
		i1 = *(uint64 *)&t1;
		i2 = *(uint64 *)&t2;
		if(i1 != i2)
		{
			idiff = labs(i1-i2);
			if(idiff >= (uint64)8)
			{
				printf("INFO: I = %8d: QCOS = %16llX, DCOS = %16llX DIFFER BY %20llu\n", i, i1, i2, idiff);
			/*	assert(0);	*/
			}
		}
	#endif
	    rt0[i].re = t1;

		qs = qfsin(qt);
		t1 = qfdbl(qs);
	#if QFDBG
		t2 = sin(mt);
		i1 = *(uint64 *)&t1;
		i2 = *(uint64 *)&t2;
		if(i1 != i2)
		{
			idiff = labs(i1-i2);
			if(idiff >= (uint64)8)
			{
				printf("INFO: I = %8d: QSIN = %16llX, DSIN = %16llX DIFFER BY %20llu\n", i, i1, i2, idiff);
			/*	assert(0);	*/
			}
		}
	#endif
	    rt0[i].im = t1;

	/*printf("I = %d; RT0 = %20.10f %20.10f\n",i,rt0[i].re,rt0[i].im);	*/
	/*	    errprint_sincos(&rt0[i].re,&rt0[i].im,(double)(mt));	* Workaround for the DEC Unix 4.0 real*16 sincos bug	*/

		qt = qfadd(qt, qtheta);
		/* Get next terms of the recurrence:  qcnew = qcold*qr - qsold*qi,  qsnew = qcold*qi + qsold*qr. */
		/*qn = qfmul(qc, qr); qt = qfmul(qs, qi); qmul = qfsub(qn, qt);	* Store qcnew in qmul for now. */
		/*qn = qfmul(qc, qi); qt = qfmul(qs, qr); qs   = qfadd(qn, qt); qc = qmul;	*/
	  }

/*...The rt1 array stores the (0:(n/2)/nrt-1)th powers of the [(n/2)/nrt]th root of unity
(and will be accessed using the upper bits, <nrt:31>, of the integer sincos index):
*/
	  rt1_ptmp = ALLOC_COMPLEX(n/(2*nwt));	if(!rt1_ptmp){ printf("FATAL: unable to allocate array RT1 in mers_mod_square.\n"); exit(EXIT_FAILURE); } rt1 = ALIGN_COMPLEX(rt1_ptmp);

	qn     = i64_to_q((int64)nwt);
	qt     = i64_to_q((int64)n2);
	qt     = qfdiv(qn, qt);		/*      NWT/(N/2) */
	qtheta = qfmul(Q2PI, qt);	/* 2*pi*NWT/(N/2) */
	qr     = qfcos(qtheta);
	qi     = qfsin(qtheta);
	qc  = QONE; qs = QZRO;	/* init sincos multiplier chain. */

	t1 = qfdbl(qr);
	#if QFDBG
	theta = qfdbl(Q2PI)*nwt/n2;
	t2 = cos(theta);
	i1 = *(uint64 *)&t1;
	i2 = *(uint64 *)&t2;
	if(i1 != i2)
	{
		idiff = labs(i1-i2);
		printf("INFO: QCOS2= %16llX, DCOS = %16llX DIFFER BY %20llu\n", i1, i2, idiff);
	}
	#endif

	t1 = qfdbl(qi);
	#if QFDBG
	t2 = sin(theta);
	i1 = *(uint64 *)&t1;
	i2 = *(uint64 *)&t2;
	if(i1 != i2)
	{
		idiff = labs(i1-i2);
		printf("INFO: QSIN2= %16llX, DSIN = %16llX DIFFER BY %20llu\n", i1, i2, idiff);
	}
	#endif

	qt = QZRO;

	  for(i=0; i<(n2/nwt); i++)
	  {
		qc = qfcos(qt);
		t1 = qfdbl(qc);
	#if QFDBG
		mt = i*theta;
		t2 = cos(mt);
		i1 = *(uint64 *)&t1;
		i2 = *(uint64 *)&t2;
		if(i1 != i2)
		{
			idiff = labs(i1-i2);
			if(idiff >= (uint64)8)
			{
				printf("INFO: J = %8d: QCOS = %16llX, DCOS = %16llX DIFFER BY %20llu\n", i, i1, i2, idiff);
			/*	assert(0);	*/
			}
		}
	#endif
	    rt1[i].re = t1;

		qs = qfsin(qt);
		t1 = qfdbl(qs);
	#if QFDBG
		t2 = sin(mt);
		i1 = *(uint64 *)&t1;
		i2 = *(uint64 *)&t2;
		if(i1 != i2)
		{
			idiff = labs(i1-i2);
			if(idiff >= (uint64)8)
			{
				printf("INFO: J = %8d: QSIN = %16llX, DSIN = %16llX DIFFER BY %20llu\n", i, i1, i2, idiff);
			/*	assert(0);	*/
			}
		}
	#endif
	    rt1[i].im = t1;

	/*printf("I = %d; RT1 = %20.10f %20.10f\n",i,rt1[i].re,rt1[i].im);	*/
	/*	    errprint_sincos(&rt1[i].re,&rt1[i].im,(double)(mt));	* Workaround for the DEC Unix 4.0 real*16 sincos bug	*/

		qt = qfadd(qt, qtheta);
		/* Get next terms of the recurrence:  qcnew = qcold*qr - qsold*qi,  qsnew = qcold*qi + qsold*qr. */
		/*qn = qfmul(qc, qr); qt = qfmul(qs, qi); qmul = qfsub(qn, qt);	* Store qcnew in qmul for now. */
		/*qn = qfmul(qc, qi); qt = qfmul(qs, qr); qs   = qfadd(qn, qt); qc = qmul;	*/
	  }

#if NEGACYCLIC
/* If the modulus is such that a length-n negacyclic convolution is needed, we need
!...N (2*N)th roots of unity (i.e. Nth roots of -1), rt2(j)=exp(i*j*pi/N), j=0,...,N-1.
!   We can exploit the symmetries of complex exponentials to reduce storage and improve performance.
!
!   Symmetry (0): Since each component of WT is simply a complex exponential, inverse weights
!                 are just complex conjugates, and don't need to be stored separately.
!
!   Symmetry (1): In the context of a "right-angle" transform, i.e. one where we treat the length-N
!                 real input vector x_j as a length-N/2 complex vector with jth element = x_j + i*x_(j+N/2)
!                 and perform a length-N/2 complex transform on this vector, the x_j term has negacyclic-DWT
!                 multiplier W_j = exp(i*j*pi/N), and the x_(j+N/2) term has W_(j+N/2) = exp(i*(j+N/2)*pi/N)
!                 = exp(i*j*pi/N + i*pi/2) = i*W_j, so letting W_j = Re + i*Im, the reslting weighted input
!                 to the complex forward transform is [x_j*(Re + i*Im) + x_(j+N/2)*(-Im + i*Re)]
!                 = [x_j*Re - x_(j+N/2)*Im] + i*[x_j*Im + x_(j+N/2)*Re] .
!                 This right-angle scheme thus only needs the first N/2 complex Nth roots of -1.
!
!   Symmetry (2): We can use the following symmetries about pi/4:
!                    cos(pi/2-x) = sin x,  sin(pi/2-x) = cos x
!                 to reduce the storage by half again. Namely, just define wt(0:N/4-1),
!                 and whenever a weight corresponding to j >= N/4 is needed, just use [ Im(wt(N/2-j) + I*Re(wt(N/2-j) ].
*/
	// We store the n/2 complex weights as N doubles, to make the indexing match that of the (real) convolution outputs:
	  rt2_ptmp = ALLOC_DOUBLE(n);	if(!rt2_ptmp){ printf("FATAL: unable to allocate array rt2 in mers_mod_square.\n"); exit(EXIT_FAILURE); } rt2 = ALIGN_COMPLEX(rt2_ptmp);

	qt     = i64_to_q((int64)2*n);
	qtheta = qfdiv(Q2PI, qt);	/* 2*pi/(2*N) */
	qr     = qfcos(qtheta);
	qi     = qfsin(qtheta);
	qc     = QONE; qs = QZRO;	/* init sincos multiplier chain. */

	t1 = qfdbl(qr);
	#if QFDBG
	theta = qfdbl(Q2PI)/(2*n);
	t2 = cos(theta);
	i1 = *(uint64 *)&t1;
	i2 = *(uint64 *)&t2;
	if(i1 != i2)
	{
		idiff = labs(i1-i2);
		printf("INFO: QCOS1= %16llX, DCOS = %16llX DIFFER BY %20llu\n", i1, i2, idiff);
	}
	#endif

	t1 = qfdbl(qi);
	#if QFDBG
	t2 = sin(theta);
	i1 = *(uint64 *)&t1;
	i2 = *(uint64 *)&t2;
	if(i1 != i2)
	{
		idiff = labs(i1-i2);
		printf("INFO: QSIN1= %16llX, DSIN = %16llX DIFFER BY %20llu\n", i1, i2, idiff);
	}
	#endif

	qt = QZRO;

	  for(i = 0; i < n-1; i += 2)
	  {
		qc = qfcos(qt);
		t1 = qfdbl(qc);
	#if QFDBG
		mt = i*theta;
		t2 = cos(mt);
		i1 = *(uint64 *)&t1;
		i2 = *(uint64 *)&t2;
		if(i1 != i2)
		{
			idiff = labs(i1-i2);
			if(idiff >= (uint64)8)
			{
				printf("INFO: I = %8d: QCOS = %16llX, DCOS = %16llX DIFFER BY %20llu\n", i, i1, i2, idiff);
			/*	assert(0);	*/
			}
		}
	#endif
	    rt2[i  ] = t1;

		qs = qfsin(qt);
		t1 = qfdbl(qs);
	#if QFDBG
		t2 = sin(mt);
		i1 = *(uint64 *)&t1;
		i2 = *(uint64 *)&t2;
		if(i1 != i2)
		{
			idiff = labs(i1-i2);
			if(idiff >= (uint64)8)
			{
				printf("INFO: I = %8d: QSIN = %16llX, DSIN = %16llX DIFFER BY %20llu\n", i, i1, i2, idiff);
			/*	assert(0);	*/
			}
		}
	#endif
	    rt2[i+1] = t1;

	/*printf("I = %d; rt2 = %20.10f %20.10f\n",i,rt2[i],rt2[i+1]);	*/
	/*	    errprint_sincos(&rt2[i],&rt2[i+1],(double)(mt));	* Workaround for the DEC Unix 4.0 real*16 sincos bug	*/

		qt = qfadd(qt, qtheta);
		/* Get next terms of the recurrence:  qcnew = qcold*qr - qsold*qi,  qsnew = qcold*qi + qsold*qr. */
		/*qn = qfmul(qc, qr); qt = qfmul(qs, qi); qmul = qfsub(qn, qt);	* Store qcnew in qmul for now. */
		/*qn = qfmul(qc, qi); qt = qfmul(qs, qr); qs   = qfadd(qn, qt); qc = qmul;	*/
	  }
#endif	// endif(NEGACYCLIC)

/*...******Forward FFT****** trig arrays are here: first, calculate the needed dimension...	*/

	  k =0;
	  mm=RADIX_VEC[0];			/* First radix requires no twiddle factors.	*/

	  for(i=1; i<NRADICES-1; i++){	/* We do the final DIF FFT radix within the wrapper_square routine, so store	*/
	    k =k+mm;			/* that block of sincos data there, where they can be merged with the wrapper sincos data.	*/
	    mm=mm*RADIX_VEC[i];
	  }

	  if(mm*RADIX_VEC[NRADICES-1] != n2){ printf("FATAL: product of radices not equal to complex vector length\n"); exit(EXIT_FAILURE); }

/*	  index = (int *)calloc(k,sizeof(int));	*/
	  index_ptmp = ALLOC_INT(k);	if(!index_ptmp){ printf("FATAL: unable to allocate array INDEX in mers_mod_square.\n"); exit(EXIT_FAILURE); } index = ALIGN_INT(index_ptmp);

/*...Forward (DIF) FFT sincos data are in bit-reversed order. We define a separate last-pass twiddles
     array within the routine wrapper_square, since that allows us to merge those nicely with the wrapper sincos data.	*/

	  k =0;
	  l =0;
	  mm=1;

/*...First radix needs no twiddle factors, just need it for building the radix_prim array.	*/

	  switch(RADIX_VEC[0]){
	  case 2 :
	    nradices_radix0 = 1;
	    radix_prim[l++] = 2; break;
	  case 3 :
	    nradices_radix0 = 1;
	    radix_prim[l++] = 3; break;
	  case 4 :
	    nradices_radix0 = 2;
	    radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	  case 5 :
	    nradices_radix0 = 1;
	    radix_prim[l++] = 5; break;
	  case 6 :
	    nradices_radix0 = 2;
	    radix_prim[l++] = 3; radix_prim[l++] = 2; break;
	  case 7 :
	    nradices_radix0 = 1;
	    radix_prim[l++] = 7; break;
	  case 8 :
	    nradices_radix0 = 3;
	    radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	  case 9 :
	    nradices_radix0 = 2;
	    radix_prim[l++] = 3; radix_prim[l++] = 3; break;
	  case 10 :
	    nradices_radix0 = 2;
	    radix_prim[l++] = 5; radix_prim[l++] = 2; break;
	  case 11 :
	    nradices_radix0 = 1;
	    radix_prim[l++] = 11; break;
	  case 12 :
	    nradices_radix0 = 3;
	    radix_prim[l++] = 3; radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	  case 13 :
	    nradices_radix0 = 1;
	    radix_prim[l++] = 13; break;
	  case 14 :
	    nradices_radix0 = 2;
	    radix_prim[l++] = 7; radix_prim[l++] = 2; break;
	  case 15 :
	    nradices_radix0 = 2;
	    radix_prim[l++] = 5; radix_prim[l++] = 3; break;
	  case 16 :
	    nradices_radix0 = 4;
	    radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	  case 18:
	    nradices_radix0 = 3;
	    radix_prim[l++] = 3; radix_prim[l++] = 3; radix_prim[l++] = 2; break;
	  case 20:
	    nradices_radix0 = 3;
	    radix_prim[l++] = 5; radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	  case 22 :
	    nradices_radix0 = 2;
	    radix_prim[l++] =11; radix_prim[l++] = 2; break;
	  case 24 :
	    nradices_radix0 = 4;
	    radix_prim[l++] = 3; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	  /*
	  case 25 :
	    nradices_radix0 = 2;
	    radix_prim[l++] = 5; radix_prim[l++] = 5; break;
	  */
	  case 26 :
	    nradices_radix0 = 2;
	    radix_prim[l++] =13; radix_prim[l++] = 2; break;
	  case 28:
	    nradices_radix0 = 3;
	    radix_prim[l++] = 7; radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	  case 30:
	    nradices_radix0 = 3;
	    radix_prim[l++] = 5; radix_prim[l++] = 3; radix_prim[l++] = 2; break;
	  case 32 :
	    nradices_radix0 = 5;
	    radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	  case 36 :
	    nradices_radix0 = 4;
	    radix_prim[l++] = 3; radix_prim[l++] = 3; radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	  case 64 :
	    nradices_radix0 = 6;
	    radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	  case 128 :
	    nradices_radix0 = 7;
	    radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	  default :
	    printf("FATAL: radix %d not available. Halting...\n",RADIX_VEC[i]); exit(EXIT_FAILURE);
	  }

	  for(i=1; i < NRADICES; i++)
	  {
/*...Allocate and initialize an index array containing MM indices...	*/

	    if(i<(NRADICES-1))
	    {
	      mm=mm*RADIX_VEC[i-1];	/* MM = product of all the preceding radices	*/

	      for(m=0; m < mm; m++)
	      {
	        index[k+m]=m;
	      }

/*...then bit-reverse INDEX with respect to the accumulated radices.
     The order of radices sent to bit_reverse_int is the reverse of that in which these radices are processed
     in the forward (decimation in frequency) FFT. This is moot for a power-of-2 FFT (or any FFT whose length
     is a prime power), but necessary for general vector lengths which are a product of 2 or more distinct primes.

     If the current (Ith) radix is composite with distinct prime factors (e.g. 15 = 3*5), we must specify these
     factors here in the opposite order from that which is used in the actual FFT-pass routine. For example,
     if the radix-15 pass implementation does 5 radix-3 DFTs, followed by 3 radix-5 DFTs, then we send (3,5)
     as the corresponding reverse-ordered prime radices to the bit-reversal routine, not (5,3).	*/

	      bit_reverse_int(&index[k],mm,l,&radix_prim[l-1],-1);
/*
for(m=0; m < mm; m++)
{
  printf("%8d%8d\n",m,index[k+m]);
}
printf("\n");
printf("\n");
*/
	      k += mm;

	    }

/*...Select structure for getting prime factors of the current radix and adding them to the radix_prim array.	*/

	    switch(RADIX_VEC[i]){
	    case 2 :
	      radix_prim[l++] = 2; break;
	    case 3 :
	      radix_prim[l++] = 3; break;
	    case 4 :
	      radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	    case 5 :
	      radix_prim[l++] = 5; break;
	    case 6 :
	      radix_prim[l++] = 3; radix_prim[l++] = 2; break;
	    case 7 :
	      radix_prim[l++] = 7; break;
	    case 8 :
	      radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	    case 9 :
	      radix_prim[l++] = 3; radix_prim[l++] = 3; break;
	    case 10 :
	      radix_prim[l++] = 5; radix_prim[l++] = 2; break;
	    case 11 :
	      radix_prim[l++] = 11; break;
	    case 12 :
	      radix_prim[l++] = 3; radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	    case 13 :
	      radix_prim[l++] = 13; break;
	    case 14 :
	      radix_prim[l++] = 7; radix_prim[l++] = 2; break;
	    case 15 :
	      radix_prim[l++] = 5; radix_prim[l++] = 3; break;
	    case 16 :
	      radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	    case 18:
	      radix_prim[l++] = 3; radix_prim[l++] = 3; radix_prim[l++] = 2; break;
	    case 20:
	      radix_prim[l++] = 5; radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	    case 22 :
	      radix_prim[l++] =11; radix_prim[l++] = 2; break;
	    case 24 :
	      radix_prim[l++] = 3; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	    /*
	    case 25 :
	      radix_prim[l++] = 5; radix_prim[l++] = 5; break;
	    */
	    case 26 :
	      radix_prim[l++] =13; radix_prim[l++] = 2; break;
	    case 28:
	      radix_prim[l++] = 7; radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	    case 30:
	      radix_prim[l++] = 5; radix_prim[l++] = 3; radix_prim[l++] = 2; break;
	    case 32 :
	      radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	    case 36 :
	      radix_prim[l++] = 3; radix_prim[l++] = 3; radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	    case 64 :
	      radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; break;
	    case 128 :
	      radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2; radix_prim[l++] = 2;
	    default :
	      printf("FATAL: radix %d not available. Halting...\n",RADIX_VEC[i]); exit(EXIT_FAILURE);
	    }

	  }

	  nradices_prim = l;

/* Break the remaining portion of the FFT into RADIX_VEC[0] blocks, each of which ideally
   should operate on a dataset which fits entirely into the L2 cache of the host machine. */

/* 8/23/2004: Need to allocate anextra element here to account for the padding element that gets inserted when RADIX_VEC[0] is odd: */

	block_index     = (int *)malloc((RADIX_VEC[0]+1)*sizeof(int));	if(!block_index    ){ printf("FATAL: unable to allocate array BLOCK_INDEX     in mers_mod_square.\n"); exit(EXIT_FAILURE); }

/*
 Examples:

  radix0 = 3: want 2 sets of l1,2 pairs = {0,-},{1,2}
  radix0 = 4: want 2 sets of l1,2 pairs = {0,1},{2,3}
  radix0 = 5: want 2 sets of l1,2 pairs = {0,-},{1,4,2,3}
  radix0 = 6: want 2 sets of l1,2 pairs = {0,1},{2,5,3,4}
  radix0 = 7: want 2 sets of l1,2 pairs = {0,-},{1,6,2,5,3,4}
  radix0 = 8: want 3 sets of l1,2 pairs = {0,1},{2,3},{4,7,5,6}
  radix0 = 9: want 3 sets of l1,2 pairs = {0,-},{1,2},{3,8,4,7,5,6}
  radix0 =10: want 2 sets of l1,2 pairs = {0,1},{2,9,3,8,4,7,5,6}
  radix0 =11: want 2 sets of l1,2 pairs = {0,-},{1,10,2,9,3,8,4,7,5,6}
  radix0 =12: want 3 sets of l1,2 pairs = {0,1},{2,3},{4,11,5,10,6,9,7,8}
  radix0 =13: want 2 sets of l1,2 pairs = {0,-},{1,12,2,11,3,10,4,9,5,8,6,7}
  radix0 =14: want 2 sets of l1,2 pairs = {0,1},{2,13,3,12,4,11,5,10,6,9,7,8}		blocklen = 2,12		l2_start = 1,13		increments by 12
  radix0 =15: want 3 sets of l1,2 pairs = {0,-},{1,2},{3,14,4,13,5,12,6,11,7,10,8,9}	blocklen = 1,2,12	l2_start = 0,2,14,	increments by 2,12
  radix0 =16: want 4 sets of l1,2 pairs = {0,1},{2,3},{4,7,5,6},{8,15,9,14,10,13,11,12}	blocklen = 2,2,4,8	l2_start = 1,3,7,15,	increments by 2,4,8

 Rules:
  1) For even radix0 use [nradices_radix0] sets of blocks; for odd radices use [nradices_radix0+1] sets,
     the first of which consists of a single block.
  2) Outer loop over prime factors of radix0, in reverse order of which these subradices appear in forward FFT:
         - blocklen (number of l-values of) first set = 2 - radix0%2
         - blocklen of sets 2...nradices_radix0 = (number of blocks done in previous sets) * (current subradix - 1) .
  3) Throughout block processing, l1 increases monotonically from 0. Within each block, for each value of l1, the corresponding
     l2_start starts at 1 - radix0%2and increases by blocklen_newblock for each new block.

  In a multithreaded implementation, each thread will be responsible for 2 of the above blocks of data (unless radix0 is odd,
  in which case one of the threads only gets one block to chew on.) For example, if NTHREADS = 4 and radix0 = 14, the chunk-processing
  loop would occur in 2 parallel passes, with each of the 4 threads having the following blocks to process:

	Pass 1:
	Thread 0: ( 0, 1)
	Thread 1: ( 2,13)
	Thread 2: ( 3,12)
	Thread 3: ( 4,11)

	Pass 2:
	Thread 0: ( 5,10)
	Thread 1: ( 6, 9)
	Thread 2: ( 7, 8)
	Thread 3:   ---		(i.e. thread is idle)

  For radix0 = 15, things look as follows:

	Pass 1:
	Thread 0: ( 0, -)	(i.e. thread is only 50% utilized)
	Thread 1: ( 1, 2)
	Thread 2: ( 3,14)
	Thread 3: ( 4,13)

	Pass 2:
	Thread 0: ( 5,12)
	Thread 1: ( 6,11)
	Thread 2: ( 7,10)
	Thread 3: ( 8, 9)

  Thus, if (2*NTHREADS) does not divide radix0, there will always be one or more under-or-unutilized threads.
*/

  blocklen = 2 - (RADIX_VEC[0] & 1);	/* blocklen = 2 for even RADIX_VEC[0], 1 for odd. */
  blocklen_sum=0;
  l1=0;
  l2=blocklen - 1; l2_start=l2;		/* l2_start = 1 for even RADIX_VEC[0], 0 for odd. */

  /* Init the block_index array: */
  ii = 0;	/* Need a linear index to provide access into the block_index array: */

  for(outer = nradices_radix0 - l2; outer >= 0; outer-- )	/* This mimics the increasing-blocksize loop in the wrapper_square routines. */
  {
    /*printf("Block %d : blocklen = %d  blocklen_sum = %d\n",nradices_radix0 - outer,blocklen,blocklen_sum);*/

    for(m = 0; m <= (blocklen-1)>>1; m++)	/* Since we now process TWO blocks per loop execution, only execute the loop half as many times as before. */
    {
      /* Now execute body of loop once with l = l1, once with l = l2 (only do l1 if blocklen == 1).
      Since blocklen is even except when radix0 odd and first pass, loop from 0 to 1 - blocklen%2.
      */
      l = l1;
      /*for(j = 0; j <= 1 - (blocklen & 1); j++)*/
      /* Do two loop executions. For the case where blocklen is odd, insert a dummy padding element
      (indicated by its having a negative value) into the second (j = 1) slot of the block_index array:
      */
      for(j = 0; j < 2; j++)
      {
	if(!(l >= 0 && l < RADIX_VEC[0])) { printf("ERROR 10 in mers_mod_square.c\n"); exit(EXIT_FAILURE); }

	if((blocklen & 1) && j == 1)
	{
	  block_index[ii] = -1;
	/*printf("%3d ---\n",ii);*/
	}
	else
	{
	  block_index[ii] = l;
	/*printf("%3d %3d\n",ii,l);*/
	}

	ii++;	/* every time we execute this innermost loop (which corresponds to
		   one block of FFT data being processed), increment the linear array index */
	l += l2-l1;
      }	/* end j-loop */
      l1++;
      l2--;
    }

    if(outer == 0)break;

    blocklen_sum += blocklen;
    blocklen = (radix_prim[outer-1] - 1)*blocklen_sum;

    /*...Next j2_start is previous one plus the length of the current block */

    l1 = l2_start + 1;
    l2_start += blocklen;
    l2 = l2_start;			/* Reset j2 for start of the next block. */
  }		/* End of Main loop */

	/* arrays storing the index values needed for the parallel-block wrapper/square scheme: */
	ws_i            = (int *)malloc(RADIX_VEC[0]*sizeof(int));	if(!ws_i           ){ printf("FATAL: unable to allocate array WS_I            in mers_mod_square.\n"); exit(EXIT_FAILURE); }
	ws_j1           = (int *)malloc(RADIX_VEC[0]*sizeof(int));	if(!ws_j1          ){ printf("FATAL: unable to allocate array WS_J1           in mers_mod_square.\n"); exit(EXIT_FAILURE); }
	ws_j2           = (int *)malloc(RADIX_VEC[0]*sizeof(int));	if(!ws_j2          ){ printf("FATAL: unable to allocate array WS_J2           in mers_mod_square.\n"); exit(EXIT_FAILURE); }
	ws_j2_start     = (int *)malloc(RADIX_VEC[0]*sizeof(int));	if(!ws_j2_start    ){ printf("FATAL: unable to allocate array WS_J2_START     in mers_mod_square.\n"); exit(EXIT_FAILURE); }
	ws_k            = (int *)malloc(RADIX_VEC[0]*sizeof(int));	if(!ws_k           ){ printf("FATAL: unable to allocate array WS_K            in mers_mod_square.\n"); exit(EXIT_FAILURE); }
	ws_m            = (int *)malloc(RADIX_VEC[0]*sizeof(int));	if(!ws_m           ){ printf("FATAL: unable to allocate array WS_M            in mers_mod_square.\n"); exit(EXIT_FAILURE); }
	ws_blocklen     = (int *)malloc(RADIX_VEC[0]*sizeof(int));	if(!ws_blocklen    ){ printf("FATAL: unable to allocate array WS_BLOCKLEN     in mers_mod_square.\n"); exit(EXIT_FAILURE); }
	ws_blocklen_sum = (int *)malloc(RADIX_VEC[0]*sizeof(int));	if(!ws_blocklen_sum){ printf("FATAL: unable to allocate array WS_BLOCKLEN_SUM in mers_mod_square.\n"); exit(EXIT_FAILURE); }

	for(ii = 0; ii < RADIX_VEC[0]; ii += 2)
	{
		/*...Final DIF pass, wrapper/squaring and initial DIT pass are all done in-place.
		     This combines data from both the l1 and l2-block, except in the case ii = 0
		     for even RADIX_VEC[0], for which the l1 = 0 and l2 = 1 blocks are processed separately within
		     wrapper_square, i.e. we must call this routine a second time to process data in the l2-block.
		*/
		if(ii == 0 && !(RADIX_VEC[0] & 1))
			jhi = 2;
		else
			jhi = 1;

		for(j = 0; j < jhi; j++)
		{
			l = ii + j;	/* This will actually leave some 'holes' in the init arrays, but we
					don't care, as these are small and we only use the init'ed elements. */

			switch(RADIX_VEC[NRADICES-1])
			{
			  case 16 :
			    radix16_wrapper_ini(n, RADIX_VEC[0], l, nradices_prim, radix_prim, ws_i, ws_j1, ws_j2, ws_j2_start, ws_k, ws_m, ws_blocklen, ws_blocklen_sum);	break;
			  case 32 :
			    radix32_wrapper_ini(n, RADIX_VEC[0], l, nradices_prim, radix_prim, ws_i, ws_j1, ws_j2, ws_j2_start, ws_k, ws_m, ws_blocklen, ws_blocklen_sum);	break;
			  /*
			  case 64 :
			  */
			  default :
			    printf("FATAL: radix %d not available for wrapper_square. Halting...\n",RADIX_VEC[NRADICES-1]); exit(EXIT_FAILURE);
			}
		}
	}

/*...Notes on generation of the (radix-1) complex sincos data needed for each block of (radix) complex array data.

     Begin with two small tables: RT0 contains the first NRT = O(sqrt(N)) Nth complex roots of unity;
                                  RT1 contains all of the (N/NRT)th complex roots of unity.
     NRT is chosen to be a power of 2, to make table lookups fast. Then, to generate any desired Jth power of the primitive
     Nth root of unity (0 <= J < N), one can simply do a complex multiply of elements from each of these small tables:

  		exp(i*2*pi*J/N) = RT0(int[J/NRT]) * RT1(J mod NRT) .	(cmul)

     Since NRT is a power of 2, the integer divide-and-truncate int[J/NRT] needs just a rightward binary shift,
     and the remainder (J mod NRT) needs just a mask, i.e. a bitwise AND(J,NRT-1).  Both of these are fast,
     and the smallness of the tables means that they can completely fit into the machine's data cache, so that
     the array accesses are fast even if the index patterns are not sequential.

     But we can do even better: if we are willing to accept a (very) slight increase in the level of rounding error,
     we can drastically reduce the number of even the relatively cheap roots array accesses, and reduce the average
     number of floating-point operations needed to get each needed complex root, from 6 for the above scheme (4 FMUL, 2 FADD)
     to something closer to 4. We also improve the balance of floating multiplies and adds at the same time. Here's how:

     In each case we allow at most binary products of roots gotten via table multiply (in order to keep RO error in check),
     and we seek combinations that are fast to calculate, i.e. have a relatively small number of floating-point operations,
     and a good balance of FADD and FMUL.

     Summary of sincos generation for the various radices supported by Mlucas:

     RADIX-5: need powers 1-4:
     Use small-tables multiply to get powers 1,3:
     ------
     1x		standard cmul					4 mul, 2 add
     2=3-1	(c1,-s1)*(c3,s3) = (c1*c3+s1*s3, c1*s3-s1*c3)	4 mul, 2 add; store the 4 scalar products...
     3x		standard cmul					4 mul, 2 add
     4=3+1	(c1,+s1)*(c3,s3) = (c1*c3-s1*s3, c1*s3+s1*c3)	0 mul, 2 add ...and then need just 4 adds to get both 2,4
     							totals: 12 mul, 8 add	(5.0 ops per complex datum) + 2 table look-ups

     RADIX-6: need powers 1-5:
     Use small-tables multiply to get powers 1,4:		opcount:
     ------	----------------------------			------------
     1x		standard cmul					4 mul, 2 add + table look-up
     2=1+1	standard csqr					2 mul, 3 add
     3=4-1	get together with 5 using just...		4 mul, 4 add
     4x		standard cmul					4 mul, 2 add + table look-up
     5=4+1						totals: 14 mul, 11 add	(5.0 ops per complex datum) + 2 table look-ups

     RADIX-7: need powers 1-6:
     Use small-tables multiply to get powers 2,3:		opcount:
     ------	----------------------------			------------
     1=3-2	get together with 5 using just...		4 mul, 4 add
     2x		standard cmul					4 mul, 2 add + table look-up
     3x		standard cmul					4 mul, 2 add + table look-up
     4=2+2	standard csqr					2 mul, 3 add
     5=3+2
     6=3+3	standard csqr					2 mul, 3 add
     							totals: 16 mul, 14 add	(5.0 ops per complex datum) + 2 table look-ups

     RADIX-8: need powers 1-7:
     Use small-tables multiply to get powers 1,2,5:		opcount:
     ------	----------------------------			------------
     1x		standard cmul					4 mul, 2 add + table look-up
     2x		standard cmul					4 mul, 2 add + table look-up
     3=5-2	get together with 7 using just...		4 mul, 4 add
     4=5-1	get together with 6 using just...		4 mul, 4 add
     5x		standard cmul					4 mul, 2 add + table look-up
     6=5+1
     7=5+2						totals: 20 mul, 14 add	(4.86 ops per complex datum) + 3 table look-ups

     On the other hand, if slightly less (but still decent) accuracy is needed, we could use one less table look-up:
     1x		standard cmul					4 mul, 2 add + table look-up
     2=1+1	standard csqr					2 mul, 3 add
     3=5-2	get together with 7 using just...		4 mul, 4 add
     4=2+2	standard csqr					2 mul, 3 add
     5x		standard cmul					4 mul, 2 add + table look-up
     6=3+3	standard csqr					2 mul, 3 add
     7=5+2						totals: 18 mul, 17 add	(5.00 ops per complex datum) + 2 table look-ups
     ...but the very slight time savings this yields are usually not worth the decrease in accuracy.

     RADIX-9: need powers 1-8:
     Use small-tables multiply to get powers 1,2,6:		opcount:
     ------	----------------------------			------------
     1x		standard cmul					4 mul, 2 add + table look-up
     2x		standard cmul					4 mul, 2 add + table look-up
     3=1+2	standard cmul					4 mul, 2 add
     4=6-2	get together with 8 using just...		4 mul, 4 add
     5=6-1	get together with 7 using just...		4 mul, 4 add
     6x		standard cmul					4 mul, 2 add + table look-up
     7=6+1
     8=6+2						totals: 24 mul, 16 add	(5.0 ops per complex datum) + 3 table look-ups

     RADIX-10: need powers 1-9:
     Use small-tables multiply to get powers 1,2,7:		opcount:
     ------	----------------------------			------------
     1x		standard cmul					4 mul, 2 add + table look-up
     2x		standard cmul					4 mul, 2 add + table look-up
     3=2+1	standard cmul					4 mul, 2 add
     4=2+2	standard csqr					2 mul, 3 add
     5=7-2	get together with 9 using just...		4 mul, 4 add
     6=7-1	get together with 8 using just...		4 mul, 4 add
     7x		standard cmul					4 mul, 2 add + table look-up
     8=7+1
     9=7+2						totals: 26 mul, 19 add	(5.0 ops per complex datum) + 3 table look-ups

     RADIX-11: need powers 1-10:
     Use small-tables multiply to get powers 1,2,7:		opcount:
     ------	----------------------------			------------
     1x		standard cmul					4 mul, 2 add + table look-up
     2=3-1	get together with 4 using just...		4 mul, 4 add
     3x		standard cmul					4 mul, 2 add + table look-up
     4=3+1
     5=6-1	get together with 7 using just...		4 mul, 4 add
     6x		standard cmul					4 mul, 2 add + table look-up
     7=6+1
     8=9-1	get together with 10 using just...		4 mul, 4 add
     9x		standard cmul					4 mul, 2 add + table look-up
     10=9+1						totals: 28 mul, 20 add	(4.8 ops per complex datum) + 4 table look-ups

     RADIX-12: need powers 1-11:
     Use small-tables multiply to get powers 1,2,3,8:		opcount:
     --------	----------------------------			------------
     1x		standard cmul					4 mul, 2 add + table look-up
     2x		standard cmul					4 mul, 2 add + table look-up
     3x		standard cmul					4 mul, 2 add + table look-up
     4 =2+2	standard csqr					2 mul, 3 add
     5 =8-3	get together with 11 using just...		4 mul, 4 add
     6 =8-2	get together with 10 using just...		4 mul, 4 add
     7 =8-1	get together with 9  using just...		4 mul, 4 add
     8x		standard cmul					4 mul, 2 add + table look-up
     9 =8+1
     10=8+2
     11=8+3						totals: 30 mul, 23 add	(4.82 ops per complex datum) + 4 table look-ups

     RADIX-14: need powers 1-13:
     Use small-tables multiply to get powers 1,2,8,11:
     ---------
     1x		standard cmul					4 mul, 2 add + table look-up
     2x		standard cmul					4 mul, 2 add + table look-up
     3 =5-2	get together with 7 using just...		4 mul, 4 add
     4 =5-1	get together with 6 using just...		4 mul, 4 add
     5x		standard cmul					4 mul, 2 add + table look-up
     6 =5+1
     7 =5+2
     8=4+4	standard csqr					2 mul, 3 add
     9 =11-2	get together with 13 using just...		4 mul, 4 add
     10=11-1	get together with 12 using just...		4 mul, 4 add
     11x	standard cmul					4 mul, 2 add + table look-up
     12=11+1
     13=11+2						totals: 34 mul, 27 add	(4.69 ops per complex datum) + 4 table look-ups

     RADIX-15: need powers 1-14:
     Use small-tables multiply to get powers 1,2,3,7,12:
     ---------
     1x		standard cmul					4 mul, 2 add + table look-up
     2x		standard cmul					4 mul, 2 add + table look-up
     3x		standard cmul					4 mul, 2 add + table look-up
     4 =7-3	get together with 10 using just...		4 mul, 4 add
     5 =7-2	get together with 9 using just...		4 mul, 4 add
     6 =7-1	get together with 8 using just...		4 mul, 4 add
     7x		standard cmul					4 mul, 2 add + table look-up
     8 =7+1
     9 =7+2
     10=7+3
     11=12-1	get together with 13 using just...		4 mul, 4 add
     12x	standard cmul					4 mul, 2 add + table look-up
     13=12+1
     14=7+7	standard csqr					2 mul, 3 add
  							totals: 38 mul, 29 add	(4.79 ops per complex datum) + 5 table look-ups

     RADIX-16: need powers 1-15:
     Use small-tables multiply to get powers 1,2,8,13:
     ---------
     1x		standard cmul					4 mul, 2 add + table look-up
     2x		standard cmul					4 mul, 2 add + table look-up
     3 =2+1	standard cmul					4 mul, 2 add
     4 =2+2	standard csqr					2 mul, 3 add
     5 =13-8	standard cmul					4 mul, 2 add
     6 =8-2	get together with 10 using just...		4 mul, 4 add
     7 =8-1	get together with 9 using just...		4 mul, 4 add
     8x		standard cmul					4 mul, 2 add + table look-up
     9 =8+1
     10=8+2
     11=13-2	get together with 15 using just...		4 mul, 4 add
     12=13-1	get together with 14 using just...		4 mul, 4 add
     13x	standard cmul					4 mul, 2 add + table look-up
     14=13+1
     15=13+2						totals: 42 mul, 31 add	(4.87 ops per complex datum) + 4 table look-ups

     73 ops seems a bit steep, though - perhaps we can save quite a few by improving the symmetry of the recurrence,
     i.e. getting rid of those expensive unpaired cmuls: e.g. 8,13 partially wasted above since don't use 13+8, just 13-8.

     Use small-tables multiply to get powers 1,2,4,8,13:
     -----------
     1x		standard cmul					4 mul, 2 add + table look-up
     2x		standard cmul					4 mul, 2 add + table look-up
     3 =4-1	get together with 5 using just...		4 mul, 4 add
     4x		standard cmul					4 mul, 2 add + table look-up
     5 =4+1
     6 =8-2	get together with 10 using just...		4 mul, 4 add
     7 =8-1	get together with 9 using just...		4 mul, 4 add
     8x		standard cmul					4 mul, 2 add + table look-up
     9 =8+1
     10=8+2
     11=13-2	get together with 15 using just...		4 mul, 4 add
     12=13-1	get together with 14 using just...		4 mul, 4 add
     13x	standard cmul					4 mul, 2 add + table look-up
     14=13+1
     15=13+2						totals: 40 mul, 30 add	(4.67 ops per complex datum) + 5 table look-ups

     Thus, for one extra CMUL (4x) save ourselves 3 ops, and get slightly better accuracy.

     RADIX-25: need powers 1-24:
     Use small-tables multiply to get powers 1,2,3,7,14,21:
     -----------
     1x		standard cmul					4 mul, 2 add + table look-up
     2x		standard cmul					4 mul, 2 add + table look-up
     3x		standard cmul					4 mul, 2 add + table look-up
     4 =7-3	get together with 10 using just...		4 mul, 4 add
     5 =7-2	get together with 9  using just...		4 mul, 4 add
     6 =7-1	get together with 8  using just...		4 mul, 4 add
     7x		standard cmul					4 mul, 2 add + table look-up
     8 =7+1
     9 =7+2
     10=7+3
     11=14-3	get together with 17 using just...		4 mul, 4 add
     12=14-2	get together with 16 using just...		4 mul, 4 add
     13=14-1	get together with 15 using just...		4 mul, 4 add
     14x	standard cmul					4 mul, 2 add + table look-up
     15=14+1
     16=14+2
     17=14+3
     18=21-3	get together with 24 using just...		4 mul, 4 add
     19=21-2	get together with 23 using just...		4 mul, 4 add
     20=21-1	get together with 22 using just...		4 mul, 4 add
     21x	standard cmul					4 mul, 2 add + table look-up
     22=21+1
     23=21+2
     24=21+3						totals: 60 mul, 48 add	(4.50 ops per complex datum) + 6 table look-ups

     RADIX-32: need powers 1-31:
     Use small-tables multiply to get powers 1,2,3,7,14,21,28:
     -----------
     1x		standard cmul					4 mul, 2 add + table look-up
     2x		standard cmul					4 mul, 2 add + table look-up
     3x		standard cmul					4 mul, 2 add + table look-up
     4 =7-3	get together with 10 using just...		4 mul, 4 add
     5 =7-2	get together with 9  using just...		4 mul, 4 add
     6 =7-1	get together with 8  using just...		4 mul, 4 add
     7x		standard cmul					4 mul, 2 add + table look-up
     8 =7+1
     9 =7+2
     10=7+3
     11=14-3	get together with 17 using just...		4 mul, 4 add
     12=14-2	get together with 16 using just...		4 mul, 4 add
     13=14-1	get together with 15 using just...		4 mul, 4 add
     14x	standard cmul					4 mul, 2 add + table look-up
     15=14+1
     16=14+2
     17=14+3
     18=21-3	get together with 24 using just...		4 mul, 4 add
     19=21-2	get together with 23 using just...		4 mul, 4 add
     20=21-1	get together with 22 using just...		4 mul, 4 add
     21x	standard cmul					4 mul, 2 add + table look-up
     22=21+1
     23=21+2
     24=21+3
     25=28-3	get together with 31 using just...		4 mul, 4 add
     26=28-2	get together with 30 using just...		4 mul, 4 add
     27=28-1	get together with 29 using just...		4 mul, 4 add
     28x	standard cmul					4 mul, 2 add + table look-up
     29=28+1
     30=28+2
     31=28+3						totals: 76 mul, 62 add	(4.45 ops per complex datum) + 7 table look-ups

     RADIX-64: need powers 1-63:
     Use small-tables multiply to get powers 1,2,3,6,11,18,25,32,39,46,53,60;
     get 4 via 2^2, {5,7} via 6-+1; {8,14,9,13,10,12} via 11-+{3,2,1}, rest similar.
*/

	}	/* end of initialization sequence.	*/

/**********************************************************************/

/*...At the start of each iteration cycle, need to forward weight the array of integer residue digits...	*/

/* This is a workaround for yet another DEC/Compaq TruUnix 4.0 bug - in this case, the
4.0 C compiler causes one of the elements of the A-array to be set to some small multiple
of machine epsilon, rather tha to zero. */
/*
if(ilo == 0)
{
printf("test loop...\n");
	for(i=1; i < n; i++)	* Expect a[0] to be nonzero, so skip it. *
	{
	  j = i + ((i >> DAT_BITS) << PAD_BITS );
	  if(a[j] != 0.0)
	  {
	     printf("warning: nonzero a[%8d] = %25.16e\n",j,a[j]);
	     a[j] = 0.0;
	  }
	}
printf("done.\n");
}
*/
	simodn = 0;
	for(i=0; i < n; i++)
	{
	  j = i + ((i >> DAT_BITS) << PAD_BITS );
	  l = i & (nwt-1);
	  k = i >> nwt_bits;
	  m = (uint32)(simodn-si[l]) >> 31;	/* Cast result of subtract to unsigned int, to ensure vacated bits filled with 0 on right-shifting.	*/
	  wt= wt0[l]*tmp[k]*one_half[m];
	  a[j] *= wt;
	  simodn += sw;
	  if(simodn >= n) simodn -= n;
	}

/*...and perform the initial pass of the forward transform.	*/

/*...NOTE: If the first radix to be processed is 2, 4 or 8, it is assumed that a power-of-2 FFT is being performed,
     hence no small-prime version of the corresponding pass1 routines is needed.	*/

	switch(RADIX_VEC[0])
	{
/*	case 3 :			*/
/*	   radix3_dif_pass1(a,n)	*/
	case 5 :
	   radix5_dif_pass1(a,n); break;
	case 6 :
	   radix6_dif_pass1(a,n); break;
	case 7 :
	   radix7_dif_pass1(a,n); break;
	case 8 :
	   radix8_dif_pass1(a,n); break;
	case 9 :
	   radix9_dif_pass1(a,n); break;
	case 10 :
	  radix10_dif_pass1(a,n); break;
	case 11 :
	  radix11_dif_pass1(a,n); break;
	case 12 :
	  radix12_dif_pass1(a,n); break;
	case 13 :
	  radix13_dif_pass1(a,n); break;
	case 14 :
	  radix14_dif_pass1(a,n); break;
	case 15 :
	  radix15_dif_pass1(a,n); break;
	case 16 :
	  radix16_dif_pass1(a,n); break;
	case 18 :
	  radix18_dif_pass1(a,n); break;
	case 20 :
	  radix20_dif_pass1(a,n); break;
	case 22 :
	  radix22_dif_pass1(a,n); break;
	case 24 :
	  radix24_dif_pass1(a,n); break;
	case 26 :
	  radix26_dif_pass1(a,n); break;
	case 28 :
	  radix28_dif_pass1(a,n); break;
	case 30 :
	  radix30_dif_pass1(a,n); break;
	case 32 :
	  radix32_dif_pass1(a,n); break;
	case 36 :
	  radix36_dif_pass1(a,n); break;
	default :
	  printf("FATAL: radix %d not available for dif_pass1. Halting...\n",RADIX_VEC[0]); exit(EXIT_FAILURE);
	}

/**********************************************************************/

/*...iteration loop is here. Do forward-FFT/pointwise-square/inverse-FFT, inverse weighting,
     carry propagation, fractional error checking and forward weighting in same loop.	*/

	ierr = 0;

	for(iter=ilo+1; iter <= ihi; iter++)
	{

#if DEBUG || defined(DEBUG_OMP)
	printf("Iter = %d\n",iter);
#endif

/*...perform the FFT-based squaring:
     Do last S-1 of S forward decimation-in-frequency transform passes.	*/

    /* Process (radix0/2) pairs of same-sized data blocks.
    In a multithreaded implementation, process NTHREADS block pairs in parallel fashion:
    */
#if defined(USE_THREADS) && defined(_OPENMP)
  omp_set_num_threads(NTHREADS);
//  #pragma omp parallel for default(shared)
//  #pragma omp parallel for private(i,ii,incr,istart,j,jhi,jstart,k,koffset,l,mm) default(shared)
  #pragma omp parallel private(i,ii,incr,istart,j,jhi,jstart,k,koffset,l,mm) default(shared)
  {
    #pragma omp for schedule(static) nowait
#endif

/********** NEW STUFF!!!!!!!!!!!!!!!!!!! *************/
#if 1
    for(ii = 0; ii < RADIX_VEC[0]; ii += 2)
    {
      process_chunk(a,n,nrt,nrt_bits,rt0,rt1,index,block_index,ii,nradices_prim,radix_prim,ws_i,ws_j1,ws_j2,ws_j2_start,ws_k,ws_m,ws_blocklen,ws_blocklen_sum);
    }	/* End of ii-loop. */

#else
    for(l = 0; l < RADIX_VEC[0]; l++)
    {
  /*
  #if 0
    for(ii = 0; ii < RADIX_VEC[0]; ii += 2)
    {
      // If radix0 odd and i = 0, process just one block of data, otherwise do two:
      if(ii == 0 && (RADIX_VEC[0] & 1))
      {
      	jhi = 1;
      	assert(block_index[ii    ] >= 0);
      	assert(block_index[ii + 1] = -1);
      }
      else
      {
        jhi = 2;
      	assert(block_index[ii    ] >= 0);
      	assert(block_index[ii + 1] >= 0);
      }

	//***************
	#if defined(USE_THREADS) && defined(_OPENMP) && defined(DEBUG_OMP)
		// Obtain and print thread id
		int tid, nthreads;
		tid = omp_get_thread_num();
		printf("A: Hello World from thread = %2d with II = %2d, JHI = %2d\n", tid, ii, jhi);

		// Only master thread does this
		if (tid == 0)
		{
			nthreads = omp_get_num_threads();
			printf("Number of threads = %d\n", nthreads);
		}
	#elif DEBUG
		printf("II = %2d\n",ii);
	#endif
	//***************

      for(j = 0; j < jhi; j++)
      {
	// Get block index of the chunk of contiguous data to be processed:
	l = block_index[ii + j];
  #endif
  */
	k    = 0;		/* init permutation array index	*/
	mm   = 1;		/* mm stores the number of outer-loop executions during the upcoming transform pass.
				   For DIF, we use the same set of twiddles during each outer-loop execution, and
				   the twiddles index varies in bit-reversed order between outer-loop executions. */
	incr = n/RADIX_VEC[0];	/* Each transform pass inner-loop execution processes RADIX_VEC[i] complex data, separated
				   by real stride = incr/RADIX_VEC[i]. Each outer-loop execution thus processes a contiguous
				   block of [incr] real data, where [incr] is a power of two. */
				/* NB: mm * incr = n/RADIX_VEC[0] throughout the set of DIF passes. During the (i)th DIF pass,
				   mm = (RADIX_VEC[1]*...*RADIX_VEC[i-1]), i.e. incr = n/(RADIX_VEC[0]*RADIX_VEC[1]*...*RADIX_VEC[i-1]), so
				   that INCR = N/[PRODUCT OF RADICES DONE SO FAR]. The (i)th DIF pass (when i > 0) uses
				   { n*RADIX_VEC[i]/incr = (RADIX_VEC[0]*RADIX_VEC[1]*...*RADIX_VEC[i]) }th roots of unity as twiddles */
	/****************
	#if defined(USE_THREADS) && defined(_OPENMP) && defined(DEBUG_OMP)
		// Obtain and print thread id
		int tid, nthreads;
		tid = omp_get_thread_num();
		printf("B: Thread = %2d, ii, j, l, k, mm, incr = %2d %2d %2d %7d %2d %7d\n", tid, ii, j, l, k, mm, incr);
	#endif
	****************/

	assert(l >= 0);
	istart = l*incr;	/* Starting location of current data-block-to-be-processed within A-array. */
	jstart = istart + ((istart >> DAT_BITS) << PAD_BITS );

	for(i=1; i <= NRADICES-2; i++)
	{
	/****************
	#if defined(USE_THREADS) && defined(_OPENMP) && defined(DEBUG_OMP)
		// Obtain and print thread id
		int tid, nthreads;
		tid = omp_get_thread_num();
		printf("B: Thread = %d, i = %d\n", tid, i);
	#endif
	****************/

	/* Offset from base address of index array = L*NLOOPS = L*MM : */
	  koffset = l*mm;
	/*
	printf("radix%d_dif_pass...istart = %d  k = %d  mm = %d  incr = %d  koffset = %d\n",RADIX_VEC[i],istart,k,mm,incr,koffset);
	*/
	  switch(RADIX_VEC[i])
	  {
	    case  8 :
	        radix8_dif_pass(&a[jstart],n,nrt,nrt_bits,rt0,rt1,&index[k+koffset],mm,incr); break;
	    case 16 :
	       radix16_dif_pass(&a[jstart],n,nrt,nrt_bits,rt0,rt1,&index[k+koffset],mm,incr); break;
	    case 32 :
	       radix32_dif_pass(&a[jstart],n,nrt,nrt_bits,rt0,rt1,&index[k+koffset],mm,incr); break;
	    default :
	      printf("FATAL: radix %d not available for dif_pass. Halting...\n",RADIX_VEC[i]); exit(EXIT_FAILURE);
	  }
	/*printf("done.\n");*/

	  k    += mm*RADIX_VEC[0];
	  mm   *= RADIX_VEC[i];
	  incr /= RADIX_VEC[i];

	}	/* end i-loop. */
	//***l += l2-l1;
  #if 0
      }	/* end j-loop */
  #endif
//*** EWM: try multithreading the FFT/IFFT, but NOT the wrapper/square:
    }
//***
#if defined(USE_THREADS) && defined(_OPENMP)
  }	// end of #pragma omp parallel private
#endif

/*...Final DIF pass, wrapper/squaring and initial DIT pass are all done in-place.
     This combines data from both the l1 and l2-block, except in the case ii = 0
     for even RADIX_VEC[0], for which the l1 = 0 and l2 = 1 blocks are processed separately within
     wrapper_square, i.e. we must call this routine a second time to process data in the l2-block.
*/
//*** EWM: try multithreading the FFT/IFFT, but NOT the wrapper/square:
    for(ii = 0; ii < RADIX_VEC[0]; ii += 2)
    {
//***
      if(ii == 0 && !(RADIX_VEC[0] & 1))
      	jhi = 2;
      else
        jhi = 1;
/*
printf("radix%d_wrapper_square...\n",RADIX_VEC[NRADICES-1]);
*/
      for(j = 0; j < jhi; j++)
      {
	l = ii + j;

	switch(RADIX_VEC[NRADICES-1])
	{
	  case 16 :
	      radix16_wrapper_square(a,n,RADIX_VEC[0],nrt,nrt_bits,rt0,rt1,nradices_prim,radix_prim, ws_i[l], ws_j1[l], ws_j2[l], ws_j2_start[l], ws_k[l], ws_m[l], ws_blocklen[l], ws_blocklen_sum[l]); break;
	  case 32 :
	      radix32_wrapper_square(a,n,RADIX_VEC[0],nrt,nrt_bits,rt0,rt1,nradices_prim,radix_prim, ws_i[l], ws_j1[l], ws_j2[l], ws_j2_start[l], ws_k[l], ws_m[l], ws_blocklen[l], ws_blocklen_sum[l]); break;
	  /*
	  case 64 :
	      radix64_wrapper_square(a,n,RADIX_VEC[0],nrt,nrt_bits,rt0,rt1,nradices_prim,radix_prim, ws_i[l], ws_j1[l], ws_j2[l], ws_j2_start[l], ws_k[l], ws_m[l], ws_blocklen[l], ws_blocklen_sum[l]); break;
	  */
	  default :
	      printf("FATAL: radix %d not available for wrapper/square. Halting...\n",RADIX_VEC[NRADICES-1]); exit(EXIT_FAILURE);
	}
      }
//*** EWM: try multithreading the FFT/IFFT, but NOT the wrapper/square:
    }
//***

/*...Rest of inverse decimation-in-time (DIT) transform. Note that during IFFT we process the radices in reverse
     order. The first array sent to each pass routine is assumed to contain the bit-reversed floating data.	*/
#if defined(USE_THREADS) && defined(_OPENMP)
  omp_set_num_threads(NTHREADS);
//  #pragma omp parallel for private(i,ii,incr,istart,j,jhi,jstart,k,koffset,l,mm) default(shared)
  #pragma omp parallel private(i,ii,incr,istart,j,jhi,jstart,k,koffset,l,mm) default(shared)
  {
    #pragma omp for schedule(dynamic) nowait
#endif
//*** EWM: try multithreading the FFT/IFFT, but NOT the wrapper/square:
    for(l = 0; l < RADIX_VEC[0]; l++)
    {
#if 0
    for(ii = 0; ii < RADIX_VEC[0]; ii += 2)
    {
//***
      /* If radix0 odd and i = 0, process just one block of data, otherwise do two: */
      if(ii == 0 && (RADIX_VEC[0] & 1))
      	jhi = 1;
      else
        jhi = 2;

      for(j = 0; j < jhi; j++)
      {
	/* Get block index of the chunk of contiguous data to be processed: */
	l = block_index[ii + j];
	assert(l >= 0);
#endif
        /* Quick-n-dirty way of generating the correct starting values of k, mm and incr -
        simply use the skeleton of the forward (DIF) loop, sans the i = NRADICES-2 pass (since k, mm and incr are post-incremented):
        */
	k    = 0;
	mm   = 1;
	incr = n/RADIX_VEC[0];

	/* calculate main-array index offset here, before incr gets changed: */
	istart = l*incr;
	jstart = istart + ((istart >> DAT_BITS) << PAD_BITS );

	for(i=1; i <= NRADICES-2; i++)
	{
	  k    += mm*RADIX_VEC[0];
	  mm   *= RADIX_VEC[i];
	  incr /= RADIX_VEC[i];
	}

	/* Now do the DIT loop, running the radices (and hence the values of k, mm and incr) in reverse: */

	for(i=NRADICES-2; i >= 1; i--)
	{
	  incr *= RADIX_VEC[i];
	  mm   /= RADIX_VEC[i];
	  k    -= mm*RADIX_VEC[0];

	  koffset = l*mm;
	/*
	printf("radix%d_dit_pass...istart = %d  k = %d  mm = %d  incr = %d  koffset = %d\n",RADIX_VEC[i],istart,k,mm,incr,koffset);
	*/
	  switch(RADIX_VEC[i])
	  {
	    case  8 :
	        radix8_dit_passB(&a[jstart],n,nrt,nrt_bits,rt0,rt1,&index[k+koffset],mm,incr); break;
	    case 16 :
	       radix16_dit_passB(&a[jstart],n,nrt,nrt_bits,rt0,rt1,&index[k+koffset],mm,incr); break;
	    case 32 :
	       radix32_dit_passB(&a[jstart],n,nrt,nrt_bits,rt0,rt1,&index[k+koffset],mm,incr); break;
	    default :
	      printf("FATAL: radix %d not available for dit_pass. Halting...\n",RADIX_VEC[i]); exit(EXIT_FAILURE);
	  }
	  /*
	  if(i==1)
	  {
		assert(k    == 0);
		assert(mm   == 1);
		assert(incr == n/RADIX_VEC[0]);
	  }
	  */
	}	/* end i-loop */
	//***l += l2-l1;
#if 0
      }	/* end j-loop */
#endif
    }	/* End of ii-loop. */
#endif	// end if(0)

#if defined(USE_THREADS) && defined(_OPENMP)
  }	// end of #pragma omp parallel private
#endif

	/****************
	#if defined(USE_THREADS) && defined(_OPENMP) && defined(DEBUG_OMP)
		int tid, nthreads;
		tid = omp_get_thread_num();
		printf("C: Hello World from thread = %d\n", tid);
		// Only master thread does this
		if (tid == 0)
		{
			nthreads = omp_get_num_threads();
			printf("Number of threads = %d\n", nthreads);
			assert(nthreads == 1);
		}
	#endif
	****************/

/*...Do the final inverse FFT pass, carry propagation and initial forward FFT pass in one fell swoop, er, swell loop...	*/

	fracmax = 0.0;

      if(iter <= *err_iter)	/* Determine whether to do RO error checking in carry step, depending on iteration number.	*/
      {
	switch(RADIX_VEC[0])
	{
	  case  5 :
	    ierr =  radix5_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,&fracmax,p); break;
	  case  6 :
	    ierr =  radix6_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,&fracmax,p); break;
	  case  7 :
	    ierr =  radix7_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,&fracmax,p); break;
	  case  8 :
	    ierr =  radix8_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,&fracmax,p); break;
	  case  9 :
	    ierr =  radix9_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,&fracmax,p); break;
	  case 10 :
	    ierr = radix10_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,&fracmax,p); break;
	  case 11 :
	    ierr = radix11_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,&fracmax,p); break;
	  case 12 :
	    ierr = radix12_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,&fracmax,p); break;
	  case 13 :
	    ierr = radix13_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,&fracmax,p); break;
	  case 14 :
	    ierr = radix14_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,&fracmax,p); break;
	  case 15 :
	    ierr = radix15_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,&fracmax,p); break;
	  case 16 :
	    ierr = radix16_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,&fracmax,p); break;
#if USE_NEW_DIT
	  case 18 :
	    ierr = radix18_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,base,baseinv,iter,&fracmax,p); break;
	  case 20 :
	    ierr = radix20_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,base,baseinv,iter,&fracmax,p); break;
	  case 22 :
	    ierr = radix22_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,base,baseinv,iter,&fracmax,p); break;
	  case 24 :
	    ierr = radix24_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,base,baseinv,iter,&fracmax,p); break;
	  case 26 :
	    ierr = radix26_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,base,baseinv,iter,&fracmax,p); break;
	  case 28 :
	    ierr = radix28_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,base,baseinv,iter,&fracmax,p); break;
	  case 30 :
	    ierr = radix30_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,base,baseinv,iter,&fracmax,p); break;
	  case 36 :
	    ierr = radix36_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,base,baseinv,iter,&fracmax,p); break;
#endif
	  case 32 :
	    ierr = radix32_ditN_cy_dif1(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,&fracmax,p); break;
	  default :
	    printf("FATAL: radix %d not available for ditN_cy_dif1. Halting...\n",RADIX_VEC[0]); exit(EXIT_FAILURE);
	}
      }
      else
      {
	switch(RADIX_VEC[0])
	{
	  case  5 :
	    ierr =  radix5_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,p); break;
	  case  6 :
	    ierr =  radix6_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,p); break;
	  case  7 :
	    ierr =  radix7_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,p); break;
	  case  8 :
	    ierr =  radix8_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,p); break;
	  case  9 :
	    ierr =  radix9_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,p); break;
	  case 10 :
	    ierr = radix10_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,p); break;
	  case 11 :
	    ierr = radix11_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,p); break;
	  case 12 :
	    ierr = radix12_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,p); break;
	  case 13 :
	    ierr = radix13_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,p); break;
	  case 14 :
	    ierr = radix14_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,p); break;
  	  case 15 :
	    ierr = radix15_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,p); break;
	  case 16 :
	    ierr = radix16_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,p); break;
#if USE_NEW_DIT
	  case 18 :
	    ierr = radix18_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,base,baseinv,iter,p); break;
	  case 20 :
	    ierr = radix20_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,base,baseinv,iter,p); break;
	  case 22 :
	    ierr = radix22_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,base,baseinv,iter,p); break;
	  case 24 :
	    ierr = radix24_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,base,baseinv,iter,p); break;
	  case 26 :
	    ierr = radix26_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,base,baseinv,iter,p); break;
	  case 28 :
	    ierr = radix28_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,base,baseinv,iter,p); break;
	  case 30 :
	    ierr = radix30_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,base,baseinv,iter,p); break;
	  case 36 :
	    ierr = radix36_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,base,baseinv,iter,p); break;
#endif
	  case 32 :
	    ierr = radix32_ditN_cy_dif1_nochk(a,n,nwt,nwt_bits,wt0,wt1,si,nrt,nrt_bits,rt0,rt1,base,baseinv,iter,p); break;
	  default :
	    printf("FATAL: radix %d not available for ditN_cy_dif1_nochk. Halting...\n",RADIX_VEC[0]); exit(EXIT_FAILURE);
	}
      }

	if(ierr) return(ierr);

/* Acumulate Avg. Max. Error: */

	if(iter > AME_ITER_START)
		AME += fracmax;

/*...Now do the fractional error check. Any fractional part in [0.4,0.6] generates a warning...	*/

	if(fracmax >= 0.4)
	{
	  if(INTERACT)
	  {
	    printf("M %8d Roundoff warning on iteration %8d  maxerr = %16.12f\n",p,iter,fracmax);
	    if(fracmax >= 0.40625) *err_iter = p-1;	/*...If RO > 0.40625 warning issued at any point of the initial error-checked	*/
	    						/* segment, require error checking on each iteration, even if iter > err_iter.	*/

/*...Fractional parts close to 0.5 cause the program to quit.
     We put the interactive-mode errlimit close to 0.5, to let people really push the limits if they want to...	*/
	    if(fracmax > 0.47 )
	    {
	      printf(" FATAL ERROR...Halting execution.\n");
	      exit(EXIT_FAILURE);
	    }
	  }
	  else
	  {
	    f1 = fopen(   OFILE,"a");
	    f2 = fopen(STATFILE,"a");
	    fprintf(f1,"M %8d Roundoff warning on iteration %8d  maxerr = %16.12f\n",p,iter,fracmax);
	    fprintf(f2,"M %8d Roundoff warning on iteration %8d  maxerr = %16.12f\n",p,iter,fracmax);
	    if (scrnFlag)                               /* Echo output to stddev */
	    {
	      printf("M %8d Roundoff warning on iteration %8d  maxerr = %16.12f\n",p,iter,fracmax);
	    }

	    if(fracmax >= 0.40625) *err_iter = p-1;

/*...In range test mode, any fractional part > 0.4375 is cause for error exit.	*/
	    if(fracmax > 0.4375 ) {
	      fprintf(f1," FATAL ERROR...Halting test of exponent %8d\n",p);
	      fprintf(f2," FATAL ERROR...Halting test of exponent %8d\n",p);
	      fclose(f1);
	      fclose(f2);
	      if (scrnFlag)                               /* Echo output to stddev */
	      {
	        printf(" FATAL ERROR...Halting test of exponent %8d\n",p);
	      }
	      ierr=1;
	      return(ierr);
	    }
	    fclose(f1);
	    fclose(f2);
	  }
	}

/*...Whew - that"s a lot of stuff that just happened. Luckily, computer chips don"t understand the concept of "Miller time."	*/

	clock2 = clock();
	*tdiff += (double)(clock2 - clock1);
	clock1 = clock2;

	}	/* End of main loop	*/

/**********************************************************************/

/*...At the end of each iteration cycle, need to undo the initial DIF FFT pass...	*/

	switch(RADIX_VEC[0])
	{
/*	  case  3 :	*/
/*	     radix3_dit_pass1(a,n); break;	*/
	  case  5 :
	     radix5_dit_pass1(a,n); break;
	  case  6 :
	     radix6_dit_pass1(a,n); break;
	  case  7 :
	     radix7_dit_pass1(a,n); break;
	  case  8 :
	     radix8_dit_pass1(a,n); break;
	  case  9 :
	     radix9_dit_pass1(a,n); break;
	  case 10 :
	    radix10_dit_pass1(a,n); break;
	  case 11 :
	    radix11_dit_pass1(a,n); break;
	  case 12 :
	    radix12_dit_pass1(a,n); break;
	  case 13 :
	    radix13_dit_pass1(a,n); break;
	  case 14 :
	    radix14_dit_pass1(a,n); break;
	  case 15 :
	    radix15_dit_pass1(a,n); break;
	  case 16 :
	    radix16_dit_pass1(a,n); break;
	  case 18 :
	    radix18_dit_pass1(a,n); break;
	  case 20 :
	    radix20_dit_pass1(a,n); break;
	  case 22 :
	    radix22_dit_pass1(a,n); break;
	  case 24 :
	    radix24_dit_pass1(a,n); break;
	  case 26 :
	    radix26_dit_pass1(a,n); break;
	  case 28 :
	    radix28_dit_pass1(a,n); break;
	  case 30 :
	    radix30_dit_pass1(a,n); break;
	  case 32 :
	    radix32_dit_pass1(a,n); break;
	  case 36 :
	    radix36_dit_pass1(a,n); break;
	  default :
	    printf("FATAL: radix %d not available for dit_pass1. Halting...\n",RADIX_VEC[0]); exit(EXIT_FAILURE);
	}

/*...and unweight the data array.	*/

	bimodn = 0;
	for(i=0; i < n; i++)
	{
	  j = i + ((i >> DAT_BITS) << PAD_BITS );
	  l = i & (nwt-1);
	  k2= (n-i) >> nwt_bits;
	  m2= 1 + (((uint32)(bimodn-si[nwt-l])) >> 31);	/* Cast result of subtract to unsigned int, to ensure vacated bits filled with 0 on right-shifting.		*/
							/* Of course C has these screwed-up precedence rules which mean we've got to enclose the shift in (), too.	*/
	  wtinv=wt0[nwt-l]*tmp[k2]*one_half[m2]*radix_inv;

	  bimodn += bw;
	  if(bimodn - n >= 0)bimodn -= n;

	  a[j] = (a[j]*wtinv + RND_A) - RND_B;
	}

	*tdiff /= CLOCKS_PER_SEC;

	ierr=0;
	return(ierr);
/*...	*/
}

/***************/

void process_chunk(double a[], int n, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int index[], int block_index[], int ii, int nradices_prim, int radix_prim[], int ws_i[], int ws_j1[], int ws_j2[], int ws_j2_start[], int ws_k[], int ws_m[], int ws_blocklen[], int ws_blocklen_sum[])
{
    int i,incr,istart,j,jhi,jstart,k,koffset,l,m,mm,k2,m2;

      /* If radix0 odd and i = 0, process just one block of data, otherwise do two: */
      if(ii == 0 && (RADIX_VEC[0] & 1))
      	jhi = 1;
      else
        jhi = 2;

      for(j = 0; j < jhi; j++)
      {
	/* Get block index of the chunk of contiguous data to be processed: */
	l = block_index[ii + j];
	k    = 0;
	mm   = 1;
	incr = n/RADIX_VEC[0];

	istart = l*incr;	/* Starting location of current data-block-to-be-processed within A-array. */
	jstart = istart + ((istart >> DAT_BITS) << PAD_BITS );

	for(i=1; i <= NRADICES-2; i++)
	{
	/* Offset from base address of index array = L*NLOOPS = L*MM : */
	  koffset = l*mm;

	  switch(RADIX_VEC[i])
	  {
	    case  8 :
	        radix8_dif_pass(&a[jstart],n,nrt,nrt_bits,rt0,rt1,&index[k+koffset],mm,incr); break;
	    case 16 :
	       radix16_dif_pass(&a[jstart],n,nrt,nrt_bits,rt0,rt1,&index[k+koffset],mm,incr); break;
	    case 32 :
	       radix32_dif_pass(&a[jstart],n,nrt,nrt_bits,rt0,rt1,&index[k+koffset],mm,incr); break;
	    default :
	      printf("FATAL: radix %d not available for dif_pass. Halting...\n",RADIX_VEC[i]); exit(EXIT_FAILURE);
	  }

	  k    += mm*RADIX_VEC[0];
	  mm   *= RADIX_VEC[i];
	  incr /= RADIX_VEC[i];

	}	/* end i-loop. */
      }	/* end j-loop */

/*...Final DIF pass, wrapper/squaring and initial DIT pass are all done in-place.
     This combines data from both the l1 and l2-block, except in the case ii = 0
     for even RADIX_VEC[0], for which the l1 = 0 and l2 = 1 blocks are processed separately within
     wrapper_square, i.e. we must call this routine a second time to process data in the l2-block.
*/
      if(ii == 0 && !(RADIX_VEC[0] & 1))
      	jhi = 2;
      else
        jhi = 1;

      for(j = 0; j < jhi; j++)
      {
	l = ii + j;

	switch(RADIX_VEC[NRADICES-1])
	{
	  case 16 :
	      radix16_wrapper_square(a,n,RADIX_VEC[0],nrt,nrt_bits,rt0,rt1,nradices_prim,radix_prim, ws_i[l], ws_j1[l], ws_j2[l], ws_j2_start[l], ws_k[l], ws_m[l], ws_blocklen[l], ws_blocklen_sum[l]); break;
	  case 32 :
	      radix32_wrapper_square(a,n,RADIX_VEC[0],nrt,nrt_bits,rt0,rt1,nradices_prim,radix_prim, ws_i[l], ws_j1[l], ws_j2[l], ws_j2_start[l], ws_k[l], ws_m[l], ws_blocklen[l], ws_blocklen_sum[l]); break;
	  /*
	  case 64 :
	      radix64_wrapper_square(a,n,RADIX_VEC[0],nrt,nrt_bits,rt0,rt1,nradices_prim,radix_prim, ws_i[l], ws_j1[l], ws_j2[l], ws_j2_start[l], ws_k[l], ws_m[l], ws_blocklen[l], ws_blocklen_sum[l]); break;
	  */
	  default :
	      printf("FATAL: radix %d not available for wrapper/square. Halting...\n",RADIX_VEC[NRADICES-1]); exit(EXIT_FAILURE);
	}
      }

/*...Rest of inverse decimation-in-time (DIT) transform. Note that during IFFT we process the radices in reverse
     order. The first array sent to each pass routine is assumed to contain the bit-reversed floating data.	*/

      /* If radix0 odd and i = 0, process just one block of data, otherwise do two: */
      if(ii == 0 && (RADIX_VEC[0] & 1))
      	jhi = 1;
      else
        jhi = 2;

      for(j = 0; j < jhi; j++)
      {
	/* Get block index of the chunk of contiguous data to be processed: */
	l = block_index[ii + j];
	assert(l >= 0);

        /* Quick-n-dirty way of generating the correct starting values of k, mm and incr -
        simply use the skeleton of the forward (DIF) loop, sans the i = NRADICES-2 pass (since k, mm and incr are post-incremented):
        */
	k    = 0;
	mm   = 1;
	incr = n/RADIX_VEC[0];

	/* calculate main-array index offset here, before incr gets changed: */
	istart = l*incr;
	jstart = istart + ((istart >> DAT_BITS) << PAD_BITS );

	for(i=1; i <= NRADICES-2; i++)
	{
	  k    += mm*RADIX_VEC[0];
	  mm   *= RADIX_VEC[i];
	  incr /= RADIX_VEC[i];
	}

	/* Now do the DIT loop, running the radices (and hence the values of k, mm and incr) in reverse: */

	for(i=NRADICES-2; i >= 1; i--)
	{
	  incr *= RADIX_VEC[i];
	  mm   /= RADIX_VEC[i];
	  k    -= mm*RADIX_VEC[0];

	  koffset = l*mm;

	  switch(RADIX_VEC[i])
	  {
	    case  8 :
	        radix8_dit_passB(&a[jstart],n,nrt,nrt_bits,rt0,rt1,&index[k+koffset],mm,incr); break;
	    case 16 :
	       radix16_dit_passB(&a[jstart],n,nrt,nrt_bits,rt0,rt1,&index[k+koffset],mm,incr); break;
	    case 32 :
	       radix32_dit_passB(&a[jstart],n,nrt,nrt_bits,rt0,rt1,&index[k+koffset],mm,incr); break;
	    default :
	      printf("FATAL: radix %d not available for dit_pass. Halting...\n",RADIX_VEC[i]); exit(EXIT_FAILURE);
	  }
	}	/* end i-loop */
      }	/* end j-loop */
}

/*...Radices for processing FFT data are here. For the case N = p*2^k, we assume the odd radix p is done first.
     Especially for large lengths we can gain a small amount of speed by restricting ourselves to radices 8 and 16
     in the power-of-2 part.
*/
void	get_fft_radices(int n, int radset, int scrnFlag)
{
	int i,use_default_radices=FALSE;
	FILE*f1;
	/*
	case selection based on real vector length - product of complex radices must equal complex vector length = n2.
	Note that the first radix must be one of (5,6,7,8,9,10,11,12,13,14,15,16) and the last must be 16 or 32.
	Intermediate radices must be a combination of 8, 16 and 32.

	In general we should attempt to order the radices in ascending order, i.e. have the larger radices
	get done at the end of the DIF FFT and beginning of the DIT IFFT, since data accesses become increasingly
	local as one gets closer to the wrapper/square step. In fact, data accesses are contiguous in the DFT/IDFT
	blocks sandwiching the wrapper/square step, so we can afford to use a large-radix (say 32) DFT
	there if needed to minimize the number of radices to be processed. The one possible exception to the
	ascending-eadix rule is the final pass of the forward FFT (and hence initial pass of the inverse FFT) -
	on some architectures it is better to put the largest radix into the penultimate FFT pass and use a smaller
	one for the wrapper/square step. (Cf. the timing data for the 195MHz R10000).
	*/

	/* Set the global radix set index: */
	RADIX_SET = radset;

	switch(n)
	{
	/* Real-array length, in CS notation (1K=1024 8-byte elements):	*/
	case 512 :
	  NRADICES = 2; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; break;
	case 1024 :						/* 1K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 2; RADIX_VEC[0] = 16; RADIX_VEC[1] = 32; break;
	 case 1 :
	  NRADICES = 2; RADIX_VEC[0] = 32; RADIX_VEC[1] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 2; RADIX_VEC[0] = 16; RADIX_VEC[1] = 32; break;
	 }; break;
/***	case 1280 :*/						/* 1.25K */
/***	  NRADICES = 3; RADIX_VEC[0] = 10; RADIX_VEC[1] =  4; RADIX_VEC[2] = 16; break;*/
/***	case 1536 :*/						/* 1.5K */
/***	  NRADICES = 3; RADIX_VEC[0] =  6; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; break;*/
	case 2048 :						/* 2K */
	  NRADICES = 3; RADIX_VEC[0] =  8; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; break;
/***	case 2304 :	*/						/* 2.25K */
/***	  NRADICES = 3; RADIX_VEC[0] =  9; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; break;	*/
/***	case 2560 :	*/						/* 2.5K */
/***	  NRADICES = 3; RADIX_VEC[0] =  5; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; break;	*/
	case 3072 :						/* 3K */
	  NRADICES = 3; RADIX_VEC[0] =  6; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; break;
/***	case 3584 :	*/						/* 3.5K */
/***	  NRADICES = 3; RADIX_VEC[0] =  7; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; break;	*/
	case 4096 :						/* 4K */
	  NRADICES = 3; RADIX_VEC[0] =  8; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; break;
/***	case 4608 :	*/					/* 4.5K */
/***	  NRADICES = 3; RADIX_VEC[0] =  9; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; break;	*/
	case 5120 :						/* 5K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 3; RADIX_VEC[0] = 10; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; break;
	 case 1 :
	  NRADICES = 3; RADIX_VEC[0] =  5; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 3; RADIX_VEC[0] =  5; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; break;
	 }; break;
	case 6144 :						/* 6K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 3; RADIX_VEC[0] = 12; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; break;
	 case 1 :
	  NRADICES = 3; RADIX_VEC[0] =  6; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 3; RADIX_VEC[0] =  6; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; break;
	 }; break;
	case 7168 :						/* 7K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 3; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; break;
	 case 1 :
	  NRADICES = 3; RADIX_VEC[0] =  7; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 3; RADIX_VEC[0] =  7; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; break;
	 }; break;
	case 8192 :						/* 8K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 3; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; break;
	 case 1 :
	  NRADICES = 3; RADIX_VEC[0] =  8; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 3; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; break;
	 }; break;
	case 9216 :						/* 9K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 3; RADIX_VEC[0] =  9; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 3; RADIX_VEC[0] =  9; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; break;
	 }; break;
	case 10240 :						/* 10K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 4; RADIX_VEC[0] =  5; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; break;
	 case 1 :
	  NRADICES = 3; RADIX_VEC[0] =  5; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; break;
	 case 2 :
	  NRADICES = 3; RADIX_VEC[0] = 10; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 3; RADIX_VEC[0] =  5; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; break;
	 }; break;
	case 12288 :						/* 12K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 4; RADIX_VEC[0] =  6; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; break;
	 case 1 :
	  NRADICES = 3; RADIX_VEC[0] =  6; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; break;
	 case 2 :
	  NRADICES = 3; RADIX_VEC[0] = 12; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 3; RADIX_VEC[0] =  6; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; break;
	 }; break;
	case 14336 :						/* 14K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 4; RADIX_VEC[0] =  7; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; break;
	 case 1 :
	  NRADICES = 3; RADIX_VEC[0] =  7; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; break;
	 case 2 :
	  NRADICES = 3; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 3; RADIX_VEC[0] =  7; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; break;
	 }; break;
	case 16384 :						/* 16K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 4; RADIX_VEC[0] =  8; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; break;
	 case 1 :
	  NRADICES = 3; RADIX_VEC[0] =  8; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; break;
	 case 2 :
	  NRADICES = 3; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 3; RADIX_VEC[0] =  8; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; break;
	 }; break;
	case 18432 :						/* 18K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 4; RADIX_VEC[0] =  9; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; break;
	 case 1 :
	  NRADICES = 3; RADIX_VEC[0] =  9; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 3; RADIX_VEC[0] =  9; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; break;
	 }; break;
	case 20480 :						/* 20K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 4; RADIX_VEC[0] =  5; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 case 1 :
	  NRADICES = 3; RADIX_VEC[0] = 10; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 3; RADIX_VEC[0] = 10; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; break;
	 }; break;
	case 24576 :						/* 24K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 4; RADIX_VEC[0] =  6; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 case 1 :
	  NRADICES = 3; RADIX_VEC[0] = 12; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 3; RADIX_VEC[0] = 12; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; break;
	 }; break;
	case 28672 :						/* 28K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 4; RADIX_VEC[0] =  7; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 case 1 :
	  NRADICES = 3; RADIX_VEC[0] = 14; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 3; RADIX_VEC[0] = 14; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; break;
	 }; break;
	case 32768 :						/* 32K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 4; RADIX_VEC[0] =  8; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 case 1 :
	  NRADICES = 3; RADIX_VEC[0] = 16; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 3; RADIX_VEC[0] = 16; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; break;
	 }; break;
	case 36864 :						/* 36K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 4; RADIX_VEC[0] =  9; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] =  9; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 }; break;
	case 40960 :						/* 40K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 4; RADIX_VEC[0] =  5; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] = 10; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] =  5; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 10; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 }; break;
	case 49152 :						/* 48K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 4; RADIX_VEC[0] =  6; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] = 12; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] =  6; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 12; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 }; break;
	case 57344 :						/* 56K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 4; RADIX_VEC[0] =  7; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] = 14; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] =  7; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 7; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 }; break;
	case 65536 :						/* 64K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 4; RADIX_VEC[0] =  8; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] =  8; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 7; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 }; break;
	case 73728 :						/* 72K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 4; RADIX_VEC[0] =  9; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] =  9; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] =  9; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 }; break;
	case 81920 :						/* 80K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] =  5; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] = 10; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] =  5; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] =  5; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 10; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 }; break;
	case 98304 :						/* 96K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] =  6; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] = 12; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] =  6; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] =  6; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 12; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 }; break;
	case 114688 :						/* 112K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] =  7; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] =  7; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] =  7; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 }; break;
/*** For 128K, try to further optimize by adding radix 11 ***/
	case 131072 :						/* 128K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] =  8; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] =  8; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] =  8; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 }; break;
	case 147456 :						/* 144K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] =  9; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] =  9; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] =  9; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] =  9; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 }; break;
	case 163840 :						/* 160K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] =  5; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 10; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] = 16; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] =  5; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] = 10; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 case 4 :
	  NRADICES = 4; RADIX_VEC[0] =  5; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 case 5 :
	  NRADICES = 4; RADIX_VEC[0] = 10; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 case 6 :
	  NRADICES = 4; RADIX_VEC[0] = 10; RADIX_VEC[1] =  8; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 10; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 }; break;
	case 180224 :						/* 176K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] = 11; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] = 11; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] = 11; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] = 11; RADIX_VEC[1] =  8; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 11; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 }; break;
	case 196608 :						/* 192K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] =  6; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 12; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] = 16; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] =  6; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] = 12; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 case 4 :
	  NRADICES = 4; RADIX_VEC[0] =  6; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 case 5 :
	  NRADICES = 4; RADIX_VEC[0] = 12; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 case 6 :
	  NRADICES = 4; RADIX_VEC[0] = 12; RADIX_VEC[1] =  8; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 12; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 }; break;
	case 229376 :						/* 224K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] =  7; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 14; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] = 16; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] =  7; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 case 4 :
	  NRADICES = 4; RADIX_VEC[0] =  7; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 case 5 :
	  NRADICES = 4; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 case 6 :
	  NRADICES = 4; RADIX_VEC[0] = 14; RADIX_VEC[1] =  8; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 }; break;
/*** For= 256K, try to further optimize by adding radix 13,15 ***/
	case 262144 :						/* 256K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] =  8; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 16; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] = 16; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] =  8; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 case 4 :
	  NRADICES = 4; RADIX_VEC[0] =  8; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 case 5 :
	  NRADICES = 4; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 case 6 :
	  NRADICES = 4; RADIX_VEC[0] = 16; RADIX_VEC[1] =  8; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 7 :
	  NRADICES = 4; RADIX_VEC[0] = 32; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; break;
	 }; break;
	case 294912 :						/* 288K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] =  9; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] =  9; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] =  9; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] =  9; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 }; break;
	case 327680 :						/* 320K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] =  5; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] =  5; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] = 10; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] = 10; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 10; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 }; break;
	case 360448 :						/* 352K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] = 11; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] = 11; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] = 11; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 11; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 }; break;
	case 393216 :						/* 384K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] =  6; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] =  6; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] = 12; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] = 12; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 12; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 }; break;
  	case 425984 :						/* 416K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] = 13; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] = 13; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] = 13; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 13; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 }; break;
	case 458752 :						/* 448K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] =  7; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] =  7; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] = 14; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 }; break;
	case 491520 :						/* 480K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] = 15; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] = 15; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] = 15; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 15; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 }; break;
	case 524288 :						/* 512K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] =  8; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] =  8; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] = 16; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 }; break;
	case 589824 :						/* 576K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] =  9; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] =  9; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] = 18; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] =  9; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 }; break;
	case 655360 :						/* 640K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] =  5; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 10; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] = 10; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] = 20; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] =  5; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 }; break;
	case 720896 :						/* 704K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] = 11; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] = 11; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] = 22; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 11; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 }; break;
	case 786432 :						/* 768K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] =  6; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 12; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] = 12; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] = 24; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] =  6; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 }; break;
	case 851968 :						/* 832K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] = 13; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] = 13; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] = 26; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 13; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 }; break;
	case 917504 :						/* 896K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] =  7; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 14; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] = 14; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] = 28; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] =  7; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 }; break;
	case 983040 :						/* 960K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 5; RADIX_VEC[0] = 15; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 1 :
	  NRADICES = 4; RADIX_VEC[0] = 15; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 2 :
	  NRADICES = 4; RADIX_VEC[0] = 30; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 4; RADIX_VEC[0] = 15; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 }; break;
	case 1048576 :					/* 1M = 1024K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] =  8; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] =  8; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] =  8; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] =  8; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 case 3 :
	  NRADICES = 5; RADIX_VEC[0] = 16; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 4 :
	  NRADICES = 4; RADIX_VEC[0] = 16; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 5 :
	  NRADICES = 4; RADIX_VEC[0] = 32; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 6 :
	  NRADICES = 4; RADIX_VEC[0] = 32; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] =  8; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 }; break;
	case 1179648 :					/* 1.125M = 1152K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] =  9; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] =  8; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] =  9; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] =  9; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 case 3 :
	  NRADICES = 5; RADIX_VEC[0] = 18; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 4 :
	  NRADICES = 4; RADIX_VEC[0] = 18; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 case 5 :
	  NRADICES = 4; RADIX_VEC[0] = 36; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] =  9; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 }; break;
	case 1310720 :					/* 1.25M = 1280K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] = 10; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] =  8; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 10; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] = 10; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] = 20; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 10; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 }; break;
	case 1441792 :					/* 1.375M = 1408K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] = 11; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] =  8; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 11; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] = 11; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] = 22; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 11; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 }; break;
	case 1572864 :					/* 1.5M = 1536K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] = 12; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] =  8; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 12; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] = 12; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] = 24; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 12; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 }; break;
	case 1703936 :					/* 1.625M = 1664K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] = 13; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] =  8; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 13; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] = 13; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] = 26; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 13; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 }; break;
	case 1835008 :					/* 1.75M = 1792K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] = 14; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] =  8; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] = 14; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] = 28; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 }; break;
	case 1966080 :					/* 1.875M = 1920K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] = 15; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] =  8; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 15; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] = 15; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 case 3 :
	  NRADICES = 4; RADIX_VEC[0] = 30; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 15; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 }; break;
	case 2097152 :					/* 2M = 2048K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] =  8; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] =  8; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] =  8; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 3 :
	  NRADICES = 5; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 4:
	  NRADICES = 4; RADIX_VEC[0] = 32; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 }; break;
	case 2359296 :					/* 2.25M = 2304K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] =  9; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] =  9; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] =  9; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 3 :
	  NRADICES = 5; RADIX_VEC[0] = 18; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 case 4 :
	  NRADICES = 4; RADIX_VEC[0] = 36; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] =  9; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 }; break;
	case 2621440 :					/* 2.5M = 2560K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] = 10; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 10; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] = 10; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 3 :
	  NRADICES = 5; RADIX_VEC[0] = 20; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 10; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 }; break;
	case 2883584 :					/* 2.75M = 2816K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] = 11; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 11; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] = 11; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 3 :
	  NRADICES = 5; RADIX_VEC[0] = 22; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 11; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 }; break;
	case 3145728 :					/* 3M = 3072K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] = 12; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 12; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] = 12; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 3 :
	  NRADICES = 5; RADIX_VEC[0] = 24; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 12; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 }; break;
	case 3407872 :					/* 3.25M = 3328K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] = 13; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 13; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] = 13; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 3 :
	  NRADICES = 5; RADIX_VEC[0] = 26; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 13; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 }; break;
	case 3670016 :					/* 3.5M = 3584K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] = 14; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 3 :
	  NRADICES = 5; RADIX_VEC[0] = 28; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 }; break;
	case 3932160 :					/* 3.75M = 3840K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] = 15; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] =  8; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 15; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] = 15; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 3 :
	  NRADICES = 5; RADIX_VEC[0] = 30; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 15; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 }; break;
	case 4194304 :					 /* 4M = 4096K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] =  8; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] =  8; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] =  8; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 3 :
	  NRADICES = 5; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; break;
	 case 4 :
	  NRADICES = 5; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 5 :
	  NRADICES = 5; RADIX_VEC[0] = 32; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] =  8; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; break;
	 }; break;
	case 4718592 :					/* 4.5M = 4608K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] =  9; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] =  9; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] =  9; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 3 :
	  NRADICES = 5; RADIX_VEC[0] = 18; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 4 :
	  NRADICES = 5; RADIX_VEC[0] = 36; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] =  9; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; break;
	 }; break;
	case 5242880 :					/* 5M = 5120K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] = 10; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 10; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] = 10; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 3 :
	  NRADICES = 5; RADIX_VEC[0] = 20; RADIX_VEC[1] = 32; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 10; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; break;
	 }; break;
	case 5767168 :					/* 5.5M = 5632K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] = 11; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 11; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] = 11; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 3 :
	  NRADICES = 5; RADIX_VEC[0] = 22; RADIX_VEC[1] = 32; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 11; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; break;
	 }; break;
	case 6291456 :					/* 6M = 6144K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] = 12; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 12; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] = 12; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 3 :
	  NRADICES = 5; RADIX_VEC[0] = 24; RADIX_VEC[1] = 32; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 12; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; break;
	 }; break;
	case 6815744 :					/* 6.5M = 6656K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] = 13; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 13; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] = 13; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 3 :
	  NRADICES = 5; RADIX_VEC[0] = 26; RADIX_VEC[1] = 32; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 13; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; break;
	 }; break;
	case 7340032 :					/* 7M = 7168K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] = 14; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 3 :
	  NRADICES = 5; RADIX_VEC[0] = 28; RADIX_VEC[1] = 32; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; break;
	 }; break;
	case 7864320 :					/* 7.5M = 7680K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] = 15; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] = 15; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] = 15; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 3 :
	  NRADICES = 5; RADIX_VEC[0] = 30; RADIX_VEC[1] = 32; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 15; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; break;
	 }; break;
	case 8388608 :					 /* 8M = 8192K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 6; RADIX_VEC[0] =  8; RADIX_VEC[1] =  8; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 1 :
	  NRADICES = 5; RADIX_VEC[0] =  8; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; break;
	 case 2 :
	  NRADICES = 5; RADIX_VEC[0] =  8; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 3 :
	  NRADICES = 5; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; break;
	 case 4 :
	  NRADICES = 5; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 32; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 case 5 :
	  NRADICES = 5; RADIX_VEC[0] = 32; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 5; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; break;
	 }; break;


	case 117440512 :					 /* 112M = 114688K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 7; RADIX_VEC[0] = 14; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; RADIX_VEC[6] = 16; break;
	 case 1 :
	  NRADICES = 6; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; RADIX_VEC[5] = 32; break;
	 case 2 :
	  NRADICES = 6; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; RADIX_VEC[5] = 16; break;
	 case 3 :
	  NRADICES = 6; RADIX_VEC[0] = 14; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 4 :
	  NRADICES = 6; RADIX_VEC[0] = 28; RADIX_VEC[1] = 32; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 6; RADIX_VEC[0] = 14; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; RADIX_VEC[5] = 32; break;
	 }; break;
	case 125829120 :					 /* 120M = 122880K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 7; RADIX_VEC[0] = 15; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; RADIX_VEC[6] = 16; break;
	 case 1 :
	  NRADICES = 6; RADIX_VEC[0] = 15; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; RADIX_VEC[5] = 32; break;
	 case 2 :
	  NRADICES = 6; RADIX_VEC[0] = 15; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; RADIX_VEC[5] = 16; break;
	 case 3 :
	  NRADICES = 6; RADIX_VEC[0] = 15; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 4 :
	  NRADICES = 6; RADIX_VEC[0] = 30; RADIX_VEC[1] = 32; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 6; RADIX_VEC[0] = 15; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; RADIX_VEC[5] = 32; break;
	 }; break;
	case 134217728 :					 /* 128M = 131072K */
	 switch(RADIX_SET){
	 case 0 :
	  NRADICES = 7; RADIX_VEC[0] = 16; RADIX_VEC[1] =  8; RADIX_VEC[2] =  8; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; RADIX_VEC[6] = 16; break;
	 case 1 :
	  NRADICES = 6; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; RADIX_VEC[5] = 32; break;
	 case 2 :
	  NRADICES = 6; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 32; RADIX_VEC[4] = 32; RADIX_VEC[5] = 16; break;
	 case 3 :
	  NRADICES = 6; RADIX_VEC[0] = 16; RADIX_VEC[1] = 32; RADIX_VEC[2] = 32; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 case 4 :
	  NRADICES = 6; RADIX_VEC[0] = 32; RADIX_VEC[1] = 32; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 16; RADIX_VEC[5] = 16; break;
	 default :
	   use_default_radices=TRUE;
	   NRADICES = 6; RADIX_VEC[0] = 16; RADIX_VEC[1] = 16; RADIX_VEC[2] = 16; RADIX_VEC[3] = 16; RADIX_VEC[4] = 32; RADIX_VEC[5] = 32; break;
	 }; break;
	default :
	  printf("FATAL: length %5d K not available. Halting...\n",n>>10); exit(EXIT_FAILURE);
	}

	/* If user specified an unavailable radix set, print an informational message.	*/
	if(use_default_radices)
	{
		if(INTERACT)
		{
		  printf("WARN: radix set %10d not available - using defaults.\n",RADIX_SET);
		  printf("Using complex FFT radices"); for(i = 0; i < NRADICES; i++){ printf("%10d",RADIX_VEC[i]); }; printf("\n");
		}
		else
		{
		  f1 = fopen(STATFILE,"a");
		  fprintf(f1,"WARN: radix set %10d not available - using defaults.\n",RADIX_SET);
		  fprintf(f1,"Using complex FFT radices"); for(i = 0; i < NRADICES; i++){ fprintf(f1,"%10d",RADIX_VEC[i]); }; fprintf(f1,"\n");
		  fclose(f1);
		  if(scrnFlag)                               /* Echo output to stddev */
		  {
		    printf("WARN: radix set %10d not available - using defaults.\n",RADIX_SET);
		    printf("Using complex FFT radices"); for(i = 0; i < NRADICES; i++){ printf("%10d",RADIX_VEC[i]); }; printf("\n");
		  }
		}

	}
}

