/*******************************************************************************
*                                                                              *
*   (C) 1997-2002 by Ernst W. Mayer.                                           *
*                                                                              *
*  This program is free software; you can redistribute it and/or modify it     *
*  under the terms of the GNU General Public License as published by the       *
*  Free Software Foundation; either version 2 of the License, or (at your      *
*  option) any later version.                                                  *
*                                                                              *
*  This program is distributed in the hope that it will be useful, but WITHOUT *
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
*  more details.                                                               *
*                                                                              *
*  You should have received a copy of the GNU General Public License along     *
*  with this program; see the file GPL.txt.  If not, you may view one at       *
*  http://www.fsf.org/licenses/licenses.html, or obtain one by writing to the  *
*  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     *
*  02111-1307, USA.                                                            *
*                                                                              *
*******************************************************************************/

/***************/

#include "Mlucas.h"

void radix32_wrapper_square(double a[], int n, int radix0, int nrt, int nrt_bits, struct complex rt0[], struct complex rt1[], int nradices_prim, int radix_prim[], int ws_i, int ws_j1, int ws_j2, int ws_j2_start, int ws_k, int ws_m, int ws_blocklen, int ws_blocklen_sum)
{

/*
!   This version dated 01 July 2000.
!
!   NOTE: In the following commentary, N refers to the COMPLEX vector length (n2 in the code),
!   which is half the real vector length.
!
!...Acronyms:	DIT = Decimation In Time
!		DIF = Decimation In Frequency
!		FFT = Fast Fourier Transform, i.e. a discrete FT over the complex numbers
!
!...Complex--->Real wrapper for data returned by forward transform, pointwise squaring and inverse wrapper
!   prior to inverse transform, all performed in one fell swoop (or better, one swell loop :)
!
!   On entry, the A-array contains floating data resulting from a forward DIF FFT, i.e. in bit-reversed order.
!
!   Quite a few operations are eliminated via use of algebraic identities in the derivation of the squaring
!   formula: the standard version of the algorithm (see, e.g., Numerical Recipes)
!   takes (ordered) forward FFT outputs H[0,...,N-1] and uses the following formula
!   to generate the N nonredundant terms F[0,...,N-1] of the DFT of the (true) real input signal:
!
!   F[j] = {( H[j]+H~[N-j] ) - I*exp(+pi*I*j/N)*( H[j]+H~[N-j] )}/2,
!
!       j=0, ... , N-1.        (Complex forward wrapper step)
!
!   Then one does a pointwise squaring to obtain terms G[0,...,N-1],
!   and does an inverse complex wrapper prior to entering the inverse FFT:
!
!   I[j] = {( G[j]+G~[N-j] ) + I*exp(-pi*I*j/N)*( G[j]-G~[N-j] )}/2,
!
!       j=0, ... , N-1.
!
!   By combining the 3 steps into one and doing some algebra, one
!   can get directly from the H's to the I's via
!
!   I[j] = H[j]^2 + {1 + exp(2*pi*I*j/N)}*{H[j]-H~[N-j]}^2/4,
!
!       j=0, ... , N-1,
!
!   and of course there are lots of symmetries in the calculation of
!   the data pairs I[j] and I[N-j] which can be (and are) exploited.
!
!   When the data in question are bit-reversed rather than ordered, it is
!   a nontrivial problem as to how to effect the above in a cache-friendly manner.
!   The problem reduces to figuring out what happens to (j, N-j) index pairs
!   under bit-reversal reordering. Happily this problem has an elegant solution,
!   which amounts to bit-reversing the array of roots of unity (i.e. the exp(2*pi*I*j/N)
!   in the last expression above, since this issue only affects the floating data)
!   and processing the bit-reversed data in the A-array (the floating-point H[j]'s)
!   in a sequence of smaller (but contiguous) sub-blocks - in each of the sub-blocks,
!   one array index increases monotonically, and a second (pointing to the bit-reversed
!   H[N-j]'s) decreases monotonically. Thus, things are very much like in the ordered-
!   data (j, N-j) index scheme, except that we have O(log2(N)) blocks, and the (k)th block
!   has index pairs (j, N'(k)-j). Further details on this may be found below, under the
!   heading "SOLVING THE CACHE FLOW PROBLEM...".
*/
	static int *index = 0x0, *index_ptmp = 0x0, *itmp = 0x0;	/* ptrs to Bit-reversal index array and temporary integer array; allocated later */
	int idum,j1pad,j2pad,kp,l,iroot,k1,k2;
	int i,j1,j2,j2_start,k,m,blocklen,blocklen_sum;
	static int n2,n32,nrtm1;
	static double    c     = 0.92387953251128675613, s     = 0.38268343236508977173	/* exp[  i*(twopi/16)]	*/
			,c32_1 = 0.98078528040323044912, s32_1 = 0.19509032201612826784	/* exp(  i*twopi/32), the radix-32 fundamental sincos datum	*/
			,c32_3 = 0.83146961230254523708, s32_3 = 0.55557023301960222473;/* exp(3*i*twopi/32)	*/

/*...Now define 320 (that's right, go ahead and count 'em: three hundred and twenty) real*8 temporaries:	*/
	double rt,it,re = 0.0, im = 0.0
	    ,c01,c02,c03,c04,c05,c06,c07,c08,c09,c0A,c0B,c0C,c0D,c0E,c0F
	,c10,c11,c12,c13,c14,c15,c16,c17,c18,c19,c1A,c1B,c1C,c1D,c1E,c1F
	    ,s01,s02,s03,s04,s05,s06,s07,s08,s09,s0A,s0B,s0C,s0D,s0E,s0F
	,s10,s11,s12,s13,s14,s15,s16,s17,s18,s19,s1A,s1B,s1C,s1D,s1E,s1F
	    ,c21,c22,c23,c24,c25,c26,c27,c28,c29,c2A,c2B,c2C,c2D,c2E,c2F
	,c30,c31,c32,c33,c34,c35,c36,c37,c38,c39,c3A,c3B,c3C,c3D,c3E,c3F
	    ,s21,s22,s23,s24,s25,s26,s27,s28,s29,s2A,s2B,s2C,s2D,s2E,s2F
	,s30,s31,s32,s33,s34,s35,s36,s37,s38,s39,s3A,s3B,s3C,s3D,s3E,s3F
	,t00,t01,t02,t03,t04,t05,t06,t07,t08,t09,t0A,t0B,t0C,t0D,t0E,t0F
	,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,t1A,t1B,t1C,t1D,t1E,t1F
	,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,t2A,t2B,t2C,t2D,t2E,t2F
	,t30,t31,t32,t33,t34,t35,t36,t37,t38,t39,t3A,t3B,t3C,t3D,t3E,t3F
	,a1p00r,a1p01r,a1p02r,a1p03r,a1p04r,a1p05r,a1p06r,a1p07r,a1p08r,a1p09r,a1p0Ar,a1p0Br,a1p0Cr,a1p0Dr,a1p0Er,a1p0Fr
	,a1p10r,a1p11r,a1p12r,a1p13r,a1p14r,a1p15r,a1p16r,a1p17r,a1p18r,a1p19r,a1p1Ar,a1p1Br,a1p1Cr,a1p1Dr,a1p1Er,a1p1Fr
	,a1p00i,a1p01i,a1p02i,a1p03i,a1p04i,a1p05i,a1p06i,a1p07i,a1p08i,a1p09i,a1p0Ai,a1p0Bi,a1p0Ci,a1p0Di,a1p0Ei,a1p0Fi
	,a1p10i,a1p11i,a1p12i,a1p13i,a1p14i,a1p15i,a1p16i,a1p17i,a1p18i,a1p19i,a1p1Ai,a1p1Bi,a1p1Ci,a1p1Di,a1p1Ei,a1p1Fi
	,a2p00r,a2p01r,a2p02r,a2p03r,a2p04r,a2p05r,a2p06r,a2p07r,a2p08r,a2p09r,a2p0Ar,a2p0Br,a2p0Cr,a2p0Dr,a2p0Er,a2p0Fr
	,a2p10r,a2p11r,a2p12r,a2p13r,a2p14r,a2p15r,a2p16r,a2p17r,a2p18r,a2p19r,a2p1Ar,a2p1Br,a2p1Cr,a2p1Dr,a2p1Er,a2p1Fr
	,a2p00i,a2p01i,a2p02i,a2p03i,a2p04i,a2p05i,a2p06i,a2p07i,a2p08i,a2p09i,a2p0Ai,a2p0Bi,a2p0Ci,a2p0Di,a2p0Ei,a2p0Fi
	,a2p10i,a2p11i,a2p12i,a2p13i,a2p14i,a2p15i,a2p16i,a2p17i,a2p18i,a2p19i,a2p1Ai,a2p1Bi,a2p1Ci,a2p1Di,a2p1Ei,a2p1Fi;
#if PFETCH
	double *add0;
#endif
	static int first_entry=TRUE;

/*...If a new runlength, deallocate the index-array and set first_entry to true. */

	if(!first_entry && (n >> 5) != n32)	/* New runlength? */
	{
	  free((void *)index_ptmp); index_ptmp=0x0; index=0x0;
	  first_entry=TRUE;
	}

/*...initialize things upon first entry */

	if(first_entry)
	{
	  first_entry=FALSE;
	  n2  = n/2;
	  n32 = n/32;
	  nrtm1 = nrt - 1;

/*...******Forward FFT****** final-pass trig array is here: first, calculate the needed dimension... */
/*
!...Final forward (DIF) FFT pass sincos data start out in bit-reversed order.
!   Allocate and initialize an index array containing N/16 indices...
*/
	  itmp = (int *)calloc(n2/32, sizeof(int));
	  if(!itmp){ printf("FATAL: unable to allocate array ITMP in radix32_wrapper_square.\n"); exit(EXIT_FAILURE); }

	  for(i=0; i < n2/32; i++)
	  {
	    itmp[i]=i;
	  }

/*...then bit-reverse INDEX with respect to N/16. Radices are sent to BR routine in opposite order from which they are processed in forward FFT. */
/*
	  radix_tmp = (int *)calloc(nradices_prim-5, sizeof(int));	if(!radix_tmp){ printf("FATAL: unable to allocate array RADIX_TMP in radix32_wrapper_square.\n"); exit(EXIT_FAILURE); }

	  for(i=0; i < nradices_prim-5; i++)
	  {
	    radix_tmp[i] = radix_prim[nradices_prim-6-i];
	  }

	  bit_reverse_int(itmp, n2/32, nradices_prim-5, &radix_tmp[0], +1);
	  free((void *)radix_tmp); radix_tmp=0x0;
 */
	  bit_reverse_int(itmp, n2/32, nradices_prim-5, &radix_prim[nradices_prim-6], -1);
/*
!...trig array for wrapper/squaring part is here. We will make use of the fact that the sincos
!   data will be in bit-reversed order, together with the symmetries of the sin and cosine
!   functions about pi/4 and pi/2, to cut our table length by a factor of 16 here. Here's how:
!   the ordered sincos data (prior to bit-reversal of the data and reduction of table length) are a set of
!   complex exponentials exp(I*j*2*pi/N), where j=0,...,N/2-1 and N is the complex FFT length.
!   That is, the argument (complex polar angle) of the exponential is in [0,pi). Within the
!   fused radix32/wrapper-square/radix32inv pass routine, we process 16 such complex exponentials
!   at a time, together with 32 complex FFT array data. Since these 16 sincos data are bit-reversal-
!   reordered, we can treat them as eight pairs, each of the form (angle, angle+pi/2):
!
!   x0+(0,pi/2), x1+(0,pi/2), x2+(0,pi/2), x3+(0,pi/2), x4+(0,pi/2), x5+(0,pi/2), x6+(0,pi/2), x7+(0,pi/2).
!
!   Given exp(I*angle) = (re,im), we use the complex symmetry  exp(I*(angle+pi/2)) = (-im,re)
!   to get the second exponential of each pair. Since, for given x0, the other sincos arguments have form
!
!   x1=x0+pi/4, x2=x0+pi/8, x3=x0+3*pi/8, x4=x0+pi/16, x5=x0+5*pi/16, x=x0+3*pi/16, x=x0+7*pi/16
!
!   (i.e. x_{0,1,2,3,4,5,6,7} = x0 + {0,4,2,6,1,5,3,7)}*pi/16), we further let
!
!      exp(I*  pi/8 ) = (cos(  pi/4 ) , sin(  pi/4 )) = (1/sqrt(2), 1/sqrt(2)):= (ISRT2, ISRT2);
!
!      exp(I*  pi/8 ) = (cos(  pi/8 ) , sin(  pi/8 )):= (c    ,s    );
!      exp(I*3*pi/8 ) = (cos(3*pi/8 ) , sin(3*pi/8 )) = (s    ,c    );
!
!      exp(I*  pi/16) = (cos(  pi/16) , sin(  pi/16)):= (c32_1,s32_1);
!      exp(I*5*pi/16) = (cos(5*pi/16) , sin(5*pi/16)):= (s32_3,c32_3);
!      exp(I*3*pi/16) = (cos(3*pi/16) , sin(3*pi/16)):= (c32_3,s32_3);
!      exp(I*7*pi/16) = (cos(7*pi/16) , sin(7*pi/16)):= (s32_1,c32_1) .
!
!   Given only exp(I*x0) = (re,im), this allows us to get the remaining seven complex exponentials via
!
!      exp(I*(x0       )) = ( re, im)
!      exp(I*(x0+  pi/2)) = I*exp(I*x0) = (-im, re)
!
!      exp(I*(x1       )) = (re-im, re+im)*ISRT2
!      exp(I*(x1+  pi/2)) = I*exp(I*x1)
!
!      exp(I*(x2       )) = ( re, im)*(c    ,s    ) = (re*c    -im*s    , re*s    +im*c    )
!      exp(I*(x2+  pi/2)) = I*exp(I*x2)
!
!      exp(I*(x3       )) = ( re, im)*(s    ,c    ) = (re*s    -im*c    , re*c    +im*s    )
!      exp(I*(x3+  pi/2)) = I*exp(I*x3),
!
!      exp(I*(x4       )) = ( re, im)*(c32_1,s32_1) = (re*c32_1-im*s32_1, re*s32_1+im*c32_1)
!      exp(I*(x4+  pi/2)) = I*exp(I*x4)
!
!      exp(I*(x5       )) = ( re, im)*(s32_3,c32_3) = (re*s32_3-im*c32_3, re*c32_3+im*s32_3)
!      exp(I*(x5+  pi/2)) = I*exp(I*x5),
!
!      exp(I*(x6       )) = ( re, im)*(c32_3,s32_3) = (re*c32_3-im*s32_3, re*s32_3+im*c32_3)
!      exp(I*(x6+  pi/2)) = I*exp(I*x6)
!
!      exp(I*(x7       )) = ( re, im)*(s32_1,c32_1) = (re*s32_1-im*c32_1, re*c32_1+im*s32_1)
!      exp(I*(x7+  pi/2)) = I*exp(I*x7),
!
!   where (a,b) = a+I*b and (a,b)*(c,d) denotes a complex product, i.e. using both explicit-I* and complex-
!   as-real-pair notation, (a,b)*(c,d) = (a+I*b)*(c+I*d) = (a*c - b*d) + I*(a*d + b*c) = (a*c - b*d, a*d + b*c).
!
!   We already have the constants ISRT2, c, s, c32_1, s32_1, c32_3, s32_3 in registers for the radix-32 transform.
!   By saving the 12 scalar products re*c, im*s, re*s, im*c, re*c32_1, im*s32_1, re*s32_1, im*c32_1, re*c32_3, im*s32_3, re*s32_3, im*c32_3,
!   we can get each of exp(I*x2|x3), exp(I*x4|x5), exp(I*x6|x7) using just 4 FMUL and 4 FADD, so our total
!   cost for the 16 complex sincos data is 2 loads from memory, 14 FMUL, 14 FADD and 8 negations.
!
!   We further use the fact that for every two blocks of 32 FFT data processed together, the wrapper sincos
!   data for the second (upper) block are just the reflection about pi/2 of those for the first block.
!   This cuts the table length in half again.
*/

/*
!...Allocate and initialize an index array containing N/32 indices to store the sequentially-rearranged
!   FFT sincos data indices. (Only need N/32 of these since we need one base sincos index for every 32 complex data).
!   We don't need a separate sincos array for the real/complex wrapper phase, since this uses the same sincos datum
!   as is used for the the first of each of the two blocks of 32 complex FFT data.
*/
/*	index = (int *)calloc(n2/32, sizeof(int));	*/
	index_ptmp = ALLOC_INT(n2/32);
	if(!index_ptmp){ printf("FATAL: unable to allocate array INDEX in radix32_wrapper_square.\n"); exit(EXIT_FAILURE); }
	index      = ALIGN_INT(index_ptmp);
/*
!...Now rearrange FFT sincos indices using the main loop structure as a template.
!   The first length-2 block is a little different, since there we process the 0-31 and 32-63 array
!   elements separately.
*/
	index[0]=itmp [0];
	index[1]=itmp [1];

	k1 =32;	/* init uncompressed wrapper sincos array index */
	k  =2;	/* init   compressed wrapper sincos array index */

	blocklen=32;
	blocklen_sum=32;
	j2_start=192;

	for(i = nradices_prim-7; i >= 0; i--)	/* Radices get processed in reverse order here as in forward FFT. */
	{
	  kp = k1 + blocklen;

	  for(m = 0; m < (blocklen-1)>>1; m +=16)	/* Since now process TWO 32-element sets per loop execution, only execute the loop half as many times as before. */
	  {
/*...grab the next 2 FFT sincos indices: one for the lower (j1, in the actual loop) data block, one for the upper (j2) data block... */

	    index[k  ]=itmp[(k1   )>>4];
	    index[k+1]=itmp[(kp-16)>>4];

	    k1 = k1+16;
	    kp = kp-16;

	    k  = k + 2;
	  }

	  k1 = k1 + (blocklen >> 1);

	  if(j2_start == n-64)break;

	  blocklen_sum = blocklen_sum + blocklen;
if(!i) { printf("ERROR 10 in radix32_wrapper_square\n"); exit(EXIT_FAILURE); }
	  blocklen = (radix_prim[i-1]-1)*blocklen_sum;

	  j2_start = j2_start+(blocklen<<2);
	}

	j1 = 0;

/*...and deallocate the temporary storage arrays. */

	free((void *)itmp); itmp=0x0;

	}	/* end of initialization sequence. */

/*
!   SOLVING THE CACHE FLOW PROBLEM FOR BIT-REVERSED ARRAY DATA:
!
!   In a previous version of this routine, the data entering the wrapper/square
!   step were ordered, and (j, N-j) -indexed data pairs were processed
!   together with data from a simple ordered sincos array W0, which
!   stored W0(j) = [1+exp(2*I*pi*j/N)], j=0,...,N/2-1, where n is the
!   complex vector length. Life was easy in the wrapper/square step, but
!   the explicit bit-reversals needed to arrange data for the DIT forward
!   transform (and after the wrapper/square for the DIT inverse transform)
!   had a significant adverse effect on performance of the code overall.
!
!   In this version, we do a DIF forward transform, so data enter the wrapper/square
!   step in bit-reversed order. Lacking a simpler pattern, I at first processed
!   them in the same order as in the old version, i.e. simply fetched
!   data in slots (index(j),index(N-j)) and processed as before. However,
!   if one looks at the actual access patterns that result from this non-strategy,
!   things are ugly: whereas j and N-j have nice predictable behavior,
!   index(j) and index(N-j) bounce all over the place. In fact, it is hard
!   to devise a worse stategy than this, from a caching perspective.
!
!   So I thought, why not loop over index(j), rather than j?
!   This requires us to bit-reverse the data in the W0 (sincos) array
!   but that imposes no penalty since these data are precomputed anyway.
!   Thus, at least the index of the first data element of each pair
!   behaves in a predictable fashion.
!
!   Well, when one does things this way, it turns out the second index also
!   behaves in a predictable way, although not quite as simply as for ordered
!   data. We note that under a bit reversal, the index range j=0,...,N/2-1
!   maps to the evens and j=N/2,...,N-1 maps to the odds, respectively.
!   This tells us we should loop over index(j) and increment by two.
!   When we do this, a surprisingly nice pattern emerges for index(N-j).
!   Even more surprisingly, this pattern also persists for N an odd prime
!   times a power of 2, with a tiny amount of added bookeeping required.
!
!   What happens is that as the first (even) index advances monotonically
!   from 0 to N-2, the second (odd) index runs through O(log2(N)) contiguous
!   odd-index subranges, but runs through each one in REVERSE order, i.e.
!   in decrements of two. The precise number of these odd-index subranges
!   (which overlap a corresponding even-index subrange to form what I call
!   a BLOCK) is equal to log2(power-of-2 part of N) + 1, where
!
!	N = P*2^M, where P is a small prime first radix > 1 (possibly 2.)
!
!   Here are four examples, for N = 16, 24, 20 and 28, assuming a 1-D, unpadded
!   zero-offset data array. I've broken chunks of data into blocks that reveal the pattern.
!   Note the correlation between the bit pattern of N-1 (written vertically, with least
!   significant bit at the top) and the pattern of block lengths
!  (block length = half the total number of complex data processed):
!
!	N=16 (FIRST RADIX = 2):			N=24 (FIRST RADIX = 3):			N=20 (FIRST RADIX = 5):			N=28 (FIRST RADIX = 7):
!
!	N - 1 = 15 = 1111_2:			N - 1 = 23 = 10111_2:			N - 1 = 19 = 10011_2:			N - 1 = 27 = 11011_2:
!	j1 =	j2 =		Current bits	j1 =	j2 =		Current bits	j1 =	j2 =		Current bits	j1 =	j2 =		Current bits
!	index(j)index(N-j)	of (N - 1)_2:	index(j)index(N-j)	of (N - 1)_2:	index(j)index(N-j)	of (N - 1)_2:	index(j)index(N-j)	of (N - 1)_2:
!	------	------		-------------	------	------		-------------	------	------		-------------	------	------		-------------
!	0,1 (j=0,N/2) done separately	1	0,1 (j=0,N/2) done separately	1	0,1 (j=0,N/2) done separately	1	0,1 (j=0,N/2) done separately	1
!	------	------				------	------				------	------				------	------
!	2	3	length-2 block	1	2	3	length-2 block	1	2	3	length-2 block	1	2	3	length-2 block	1
!	------	------				------	------				------	------				------	------
!	4	7	length-4 block	1	4	7	length-4 block	1	4	19				4	27
!	6	5				6	5				6	17				6	25
!	------	------				------	------				8	15				8	23
!	8	15				8	23				10	13	length-16 block	100	10	21
!	10	13	length-8 block	1	10	21				12	11				12	19
!	12	11				12	19				14	9				14	17	length-24 block	110
!	14	9				14	17	length-16 block	10	16	7				16	15
!	------	------				16	15				18	5				18	13
!						18	13				------	------				20	11
!						20	11									22	9
!						22	9									24	7
!						------	------									26	5
!																------	------
!   So the formula for calculating block sizes is as follows: At the end of each block, we
!   multiply the accumulated blocklength (which is just double the current blocklength) by
!   the next value of (current bits), which is = (radix_next - 1), to get the length
!   of the next block. To get the value of j2_start for the next block,
!   we add the length of the next block to j2_start for the current block.
!
!...Exploiting the pattern makes for a much cache-friendlier code, which follows. Note
!   that if we are processing ordered modular data alongside bit-reversed floating data,
!   we actually have two index pairs to keep track of, e.g. for N=16:
!
!   Floating index 1 (j1):  2, 4, 6, 8,10,12,14
!   Floating index 2 (j2):  3, 7, 5,15,13,11, 9	(0 and 1 done separately)
!
!   Modular  index 1 (j3):  1, 2, 3, 4, 5, 6, 7
!   Modular  index 2 (j4): 15,14,13,12,11,10, 9	(0 and N/2 done separately)
!
!   so the floating index 1 is just double its modular counterpart. The index 2's must be
!   stored separately.
!
!
!   FUSING THE WRAPPER-SQUARE STEP WITH THE RADIX-16 PASSES PRECEDING AND FOLLOWING IT:
!
!   OK, so you've made it this far, and justifiably are hoping to see some actual, honest-
!   to-goodness code. Alas, things must get just a bit more complicated before that happy
!   moment arrives. That's because in order to squeeze maximum performance out of our code,
!   we should seize any opportunities to minimize unnecessary data movement. One such
!   presents itself here. We have a radix-16 DIF pass immediately preceding and a radix-16
!   DIT pass immediately following the wrapper/square, so we do all three of the above steps
!   in a single in-place sweep, thus reducing the number of passes through the big data
!   arrays from three to one. We do the first 16 complex data (and their images under the (j,N'-j)
!   correlation) separately, after which blocklengths are always a multiple of 16, allowing us to simplify the indexing.
!
!   The general loop structure looks as follows:
!
!   loop:
!   	do m = 1,blocklen
!
!   	  call pair_square(a(j1),a(j1+1),a(j2),a(j2+1),w(1,j),w(2,j))
!
!   	  j  = j +1
!   	  j1 = j1+4
!   	  j2 = j2-4
!   	enddo
!   	...
!   	blocklen = 2*blocklen
!   	j2_start = j2_start+ishft(blocklen,2)
!   	j2=j2_start
!   	cycle
!   endloop
!
!   So the first 16 complex elements and their images are processed as follows:
!
!   exe0: process a(0,1) and a(2,3) separately;
!
!   exe1:
!   	j=1
!   	blocklen = 1
!   	j1=4
!   	j2_start=6
!
!   	  call pair_square(a( 4),a( 5),a( 6),a( 7),w(:,1))
!
!   	  j  = 2
!   	  j1 = 8
!   	  j2 = 2
!   	enddo
!
!   	blocklen = 2
!   	j2_start = 14
!   end1
!
!   exe2:
!   	j=2
!   	blocklen = 2
!   	j1=8
!   	j2_start=14
!
!   	  call pair_square(a( 8),a( 9),a(14),a(15),w(:,2))
!   	  call pair_square(a(12),a(13),a(10),a(11),w(:,3))
!
!   	  j  = 4
!   	  j1 = 16
!   	  j2 = 6
!   	enddo
!
!   	blocklen = 4
!   	j2_start = 30
!   end2
!
!   exe3:
!   	j=4
!   	blocklen = 4
!   	j1=16
!   	j2_start=30
!
!   	  call pair_square(a(16),a(17),a(30),a(31),w(:,4))
!   	  call pair_square(a(20),a(21),a(26),a(27),w(:,5))
!   	  call pair_square(a(24),a(25),a(22),a(23),w(:,6))
!   	  call pair_square(a(28),a(29),a(18),a(19),w(:,7))
!
!   	  j  = 8
!   	  j1 = 32
!   	  j2 = 14
!   	enddo
!
!   	blocklen = 8
!   	j2_start = 62
!   end3
!
!   exe4:
!   	j=8
!   	blocklen = 8
!   	j1=32
!   	j2_start=62
!
!   	  call pair_square(a(32),a(33),a(62),a(63),w(:,8))
!   	  call pair_square(a(36),a(37),a(58),a(59),w(:,9))
!   	  call pair_square(a(40),a(41),a(54),a(55),w(:,10))
!   	  call pair_square(a(44),a(45),a(50),a(51),w(:,11))
!   	  call pair_square(a(48),a(49),a(46),a(47),w(:,12))
!   	  call pair_square(a(52),a(53),a(42),a(43),w(:,13))
!   	  call pair_square(a(56),a(57),a(38),a(39),w(:,14))
!   	  call pair_square(a(60),a(61),a(34),a(35),w(:,15))
!
!   	  j  = 16
!   	  j1 = 64
!   	  j2 = 30
!   	enddo
!
!   	blocklen = 16
!   	j2_start = 126
!   end4
!
!   exe5:
!   	j=16
!   	blocklen = 16
!   	j1=64
!   	j2_start=126
!
!   	  call pair_square(a( 64),a( 65),a(126),a(127),w(:,16))
!   	  call pair_square(a( 68),a( 69),a(122),a(123),w(:,17))
!   	  call pair_square(a( 72),a( 73),a(118),a(119),w(:,18))
!   	  call pair_square(a( 76),a( 77),a(114),a(115),w(:,19))
!   	  call pair_square(a( 80),a( 81),a(110),a(111),w(:,20))
!   	  call pair_square(a( 84),a( 85),a(106),a(107),w(:,21))
!   	  call pair_square(a( 88),a( 89),a(102),a(103),w(:,22))
!   	  call pair_square(a( 92),a( 93),a( 98),a( 99),w(:,23))
!   	  call pair_square(a( 96),a( 97),a( 94),a( 95),w(:,24))
!   	  call pair_square(a(100),a(101),a( 90),a( 91),w(:,25))
!   	  call pair_square(a(104),a(105),a( 86),a( 87),w(:,26))
!   	  call pair_square(a(108),a(109),a( 82),a( 83),w(:,27))
!   	  call pair_square(a(112),a(113),a( 78),a( 79),w(:,28))
!   	  call pair_square(a(116),a(117),a( 74),a( 75),w(:,29))
!   	  call pair_square(a(120),a(121),a( 70),a( 71),w(:,30))
!   	  call pair_square(a(124),a(125),a( 66),a( 67),w(:,31))
!
!   	  j  = 32
!   	  j1 = 128
!   	  j2 = 62
!   	enddo
!
!   	blocklen = 32
!   	j2_start = 254
!   end5
!
!   From here onward the blocklength will always be an integer multiple of 32, i.e. we can process each block using pairs of nonoverlapping
!   blocks of 32 complex data each, which is compatible to fusion with radix-32 pass routines. For example, the next (fifth) loop execution
!   looks like:
!
!   exe6:
!   	j=32
!   	blocklen = 32
!   	j1=128
!   	j2_start=254
!
!   	  call pair_square(a(128),a(129),a(254),a(255),w(:,32))
!   	  call pair_square(a(132),a(133),a(250),a(251),w(:,33))
!   	  call pair_square(a(136),a(137),a(246),a(247),w(:,34))
!   	  call pair_square(a(140),a(141),a(242),a(243),w(:,35))
!   	  call pair_square(a(144),a(145),a(238),a(239),w(:,36))
!   	  call pair_square(a(148),a(149),a(234),a(235),w(:,37))
!   	  call pair_square(a(152),a(153),a(230),a(231),w(:,38))
!   	  call pair_square(a(156),a(157),a(226),a(227),w(:,39))
!   	  call pair_square(a(160),a(161),a(222),a(223),w(:,40))
!   	  call pair_square(a(164),a(165),a(218),a(219),w(:,41))
!   	  call pair_square(a(168),a(169),a(214),a(215),w(:,42))
!   	  call pair_square(a(172),a(173),a(210),a(211),w(:,43))
!   	  call pair_square(a(176),a(177),a(206),a(207),w(:,44))
!   	  call pair_square(a(180),a(181),a(202),a(203),w(:,45))
!   	  call pair_square(a(184),a(185),a(198),a(199),w(:,46))
!   	  call pair_square(a(188),a(189),a(194),a(195),w(:,47))
!
!   	  call pair_square(a(192),a(193),a(190),a(191),w(:,48))
!   	  call pair_square(a(196),a(197),a(186),a(187),w(:,49))
!   	  call pair_square(a(200),a(201),a(182),a(183),w(:,50))
!   	  call pair_square(a(204),a(205),a(178),a(179),w(:,51))
!   	  call pair_square(a(208),a(209),a(174),a(175),w(:,52))
!   	  call pair_square(a(212),a(213),a(170),a(171),w(:,53))
!   	  call pair_square(a(216),a(217),a(166),a(167),w(:,54))
!   	  call pair_square(a(220),a(221),a(162),a(163),w(:,55))
!   	  call pair_square(a(224),a(225),a(158),a(159),w(:,56))
!   	  call pair_square(a(228),a(229),a(154),a(155),w(:,57))
!   	  call pair_square(a(232),a(233),a(150),a(151),w(:,58))
!   	  call pair_square(a(236),a(237),a(146),a(147),w(:,59))
!   	  call pair_square(a(240),a(241),a(142),a(143),w(:,60))
!   	  call pair_square(a(244),a(245),a(138),a(139),w(:,61))
!   	  call pair_square(a(248),a(249),a(134),a(135),w(:,62))
!   	  call pair_square(a(252),a(253),a(130),a(131),w(:,63))
!
!   	...and these 64 complex data can be processed as follows:
!	1) do radix-32 DIF (using sincos data  0-31) to get a(128:191);
!	2) do radix-32 DIF (using sincos data 32-63) to get a(192:255);
!	3) combine the 64 resulting complex array data as shown above;
!	4) do radix-32 DIT on a(128:191);
!	5) do radix-32 DIT on a(192:255).
!   	The radix-32 passes share many register data, providing added savings.
!   	  j  = 64
!   	  j1 = 256
!   	  j2 = 126
!   	enddo
!
!   	blocklen = 64
!   	j2_start = 510
!   end6
*/

/* Init the loop-control variables: */

	i            = ws_i           ;
	j1           = ws_j1          ;
	j2           = ws_j2          ;
	j2_start     = ws_j2_start    ;
	k            = ws_k           ;
	m            = ws_m           ;
	blocklen     = ws_blocklen    ;
	blocklen_sum = ws_blocklen_sum;

	/* If j1 == 0 we need to init the loop counters; otherwise, just jump
	   right in and pick up where we left off on the previous pair of blocks:
	*/
	if(j1 > 0)
		goto jump_in;

/*
!...All but the first two radix-32 blocks are done on Mr. Henry Ford's assembly line. From there onward the blocklength
!   will always be an integer multiple of 16, i.e. we can process each block using pairs of nonoverlapping blocks of 32
!   complex data each, which is compatible to fusion with radix-32 pass routines.
*/

      for(i = nradices_prim-6; i >= 0; i-- )	/* Main loop: lower bound = nradices_prim-radix_now. */
      {						/* Remember, radices get processed in reverse order here as in forward FFT. */

	for(m = 0; m < (blocklen-1)>>1; m +=16)	/* Since we now process TWO 32-element sets per loop execution, only execute the loop half as many times as before. */
	{

if(j1 && j1*radix0%n == 0) return;

jump_in:	/* Entry point for all blocks but the first. */

/*printf("%8d %8d\n",j1,j2);*/

	  j1pad = j1 + ( (j1 >> DAT_BITS) << PAD_BITS );	/* floating padded-array 1st element index is here */
	  j2pad = j2 + ( (j2 >> DAT_BITS) << PAD_BITS );	/* floating padded-array 2nd element index is here */

/* 1st set of inputs: */
	    idum = j1pad;
	    iroot = index[k];
	    l = iroot;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += iroot;			/* 2*iroot	*/
	    t00=rt0[k1].re;	t01=rt0[k1].im;
	    rt =rt1[k2].re;	it =rt1[k2].im;
	    c01=t00*rt -t01*it;	s01=t00*it +t01*rt;

	re=c01;	im=s01;	/* Save for the wrapper Step... */

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += iroot;				/* 3*iroot	*/
	    t00=rt0[k1].re;	t01=rt0[k1].im;
	    rt =rt1[k2].re;	it =rt1[k2].im;
	    c02=t00*rt -t01*it;	s02=t00*it +t01*rt;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += 4*iroot;			/* 7*iroot	*/
	    iroot = l;				/* 7*iroot	*/
	    t00=rt0[k1].re;	t01=rt0[k1].im;
	    rt =rt1[k2].re;	it =rt1[k2].im;
	    c03=t00*rt -t01*it;	s03=t00*it +t01*rt;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += iroot;				/* 14*iroot	*/
	    t00=rt0[k1].re;	t01=rt0[k1].im;
	    rt =rt1[k2].re;	it =rt1[k2].im;
	    c07=t00*rt -t01*it;	s07=t00*it +t01*rt;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += iroot;				/* 21*iroot	*/
	    t00=rt0[k1].re;	t01=rt0[k1].im;
	    rt =rt1[k2].re;	it =rt1[k2].im;
	    c0E=t00*rt -t01*it;	s0E=t00*it +t01*rt;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += iroot;				/* 28*iroot	*/
	    t00=rt0[k1].re;	t01=rt0[k1].im;
	    rt =rt1[k2].re;	it =rt1[k2].im;
	    c15=t00*rt -t01*it;	s15=t00*it +t01*rt;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    t00=rt0[k1].re;	t01=rt0[k1].im;
	    rt =rt1[k2].re;	it =rt1[k2].im;
	    c1C=t00*rt -t01*it;	s1C=t00*it +t01*rt;

/*(c,s)4-10 (decimal indices), 4-0A (hexadecimal indices):	*/
	    t00=c01*c07; t01=c01*s07; t02=s01*c07; t03=s01*s07;
	    c06=t00+t03; s06=t01-t02; c08=t00-t03; s08=t01+t02;

	    t00=c02*c07; t01=c02*s07; t02=s02*c07; t03=s02*s07;
	    c05=t00+t03; s05=t01-t02; c09=t00-t03; s09=t01+t02;

	    t00=c03*c07; t01=c03*s07; t02=s03*c07; t03=s03*s07;
	    c04=t00+t03; s04=t01-t02; c0A=t00-t03; s0A=t01+t02;

/*(c,s)11-17 (decimal indices), 0B-11 (hexadecimal indices):;	*/
	    t00=c01*c0E; t01=c01*s0E; t02=s01*c0E; t03=s01*s0E;
	    c0D=t00+t03; s0D=t01-t02; c0F=t00-t03; s0F=t01+t02;

	    t00=c02*c0E; t01=c02*s0E; t02=s02*c0E; t03=s02*s0E;
	    c0C=t00+t03; s0C=t01-t02; c10=t00-t03; s10=t01+t02;

	    t00=c03*c0E; t01=c03*s0E; t02=s03*c0E; t03=s03*s0E;
	    c0B=t00+t03; s0B=t01-t02; c11=t00-t03; s11=t01+t02;

/*(c,s)18-24 (decimal indices), 12-18 (hexadecimal indices):	*/
	    t00=c01*c15; t01=c01*s15; t02=s01*c15; t03=s01*s15;
	    c14=t00+t03; s14=t01-t02; c16=t00-t03; s16=t01+t02;

	    t00=c02*c15; t01=c02*s15; t02=s02*c15; t03=s02*s15;
	    c13=t00+t03; s13=t01-t02; c17=t00-t03; s17=t01+t02;

	    t00=c03*c15; t01=c03*s15; t02=s03*c15; t03=s03*s15;
	    c12=t00+t03; s12=t01-t02; c18=t00-t03; s18=t01+t02;

/*(c,s)25-31 (decimal indices), 19-1F (hexadecimal indices):	*/
	    t00=c01*c1C; t01=c01*s1C; t02=s01*c1C; t03=s01*s1C;
	    c1B=t00+t03; s1B=t01-t02; c1D=t00-t03; s1D=t01+t02;

	    t00=c02*c1C; t01=c02*s1C; t02=s02*c1C; t03=s02*s1C;
	    c1A=t00+t03; s1A=t01-t02; c1E=t00-t03; s1E=t01+t02;

	    t00=c03*c1C; t01=c03*s1C; t02=s03*c1C; t03=s03*s1C;
	    c19=t00+t03; s19=t01-t02; c1F=t00-t03; s1F=t01+t02;

/*       gather the needed data (32 64-bit complex, i.e. 64 64-bit reals) and do the first set of four length-8 transforms...
         We process the sincos data in bit-reversed order.	*/

/*...Block 1:	*/
	    t00=a[idum   ];				t01=a[idum+1   ];
	    rt =a[idum+32]*c10 - a[idum+1+32]*s10;	it =a[idum+1+32]*c10 + a[idum+32]*s10;
	    t02=t00-rt;		t03=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    t04=a[idum+16]*c08 - a[idum+1+16]*s08;	t05=a[idum+1+16]*c08 + a[idum+16]*s08;
	    rt =a[idum+48]*c18 - a[idum+1+48]*s18;	it =a[idum+1+48]*c18 + a[idum+48]*s18;
	    t06=t04-rt;		t07=t05-it;
	    t04=t04+rt;		t05=t05+it;

	    rt =t04;		it =t05;
	    t04=t00-rt;		t05=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t06;		it =t07;
	    t06=t02+it;		t07=t03-rt;
	    t02=t02-it;		t03=t03+rt;

	    t08=a[idum+8 ]*c04 - a[idum+1+8 ]*s04;	t09=a[idum+1+8 ]*c04 + a[idum+8 ]*s04;
	    rt =a[idum+40]*c14 - a[idum+1+40]*s14;	it =a[idum+1+40]*c14 + a[idum+40]*s14;
	    t0A=t08-rt;		t0B=t09-it;
	    t08=t08+rt;		t09=t09+it;

	    t0C=a[idum+24]*c0C - a[idum+1+24]*s0C;	t0D=a[idum+1+24]*c0C + a[idum+24]*s0C;
	    rt =a[idum+56]*c1C - a[idum+1+56]*s1C;	it =a[idum+1+56]*c1C + a[idum+56]*s1C;
	    t0E=t0C-rt;		t0F=t0D-it;
	    t0C=t0C+rt;		t0D=t0D+it;

	    rt =t0C;		it =t0D;
	    t0C=t08-rt;		t0D=t09-it;
	    t08=t08+rt;		t09=t09+it;

	    rt =t0E;		it =t0F;
	    t0E=t0A+it;		t0F=t0B-rt;
	    t0A=t0A-it;		t0B=t0B+rt;

	    rt =t08;		it =t09;
	    t08=t00-rt;		t09=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t0C;		it =t0D;
	    t0C=t04+it;		t0D=t05-rt;
	    t04=t04-it;		t05=t05+rt;

	    rt =(t0A-t0B)*ISRT2;it =(t0A+t0B)*ISRT2;
	    t0A=t02-rt;		t0B=t03-it;
	    t02=t02+rt;		t03=t03+it;

	    rt =(t0E+t0F)*ISRT2;it =(t0F-t0E)*ISRT2;
	    t0E=t06+rt;		t0F=t07+it;
	    t06=t06-rt;		t07=t07-it;

/*...Block 2:	*/
	    t10=a[idum+4 ]*c02 - a[idum+1+4 ]*s02;	t11=a[idum+1+4 ]*c02 + a[idum+4 ]*s02;
	    rt =a[idum+36]*c12 - a[idum+1+36]*s12;	it =a[idum+1+36]*c12 + a[idum+36]*s12;
	    t12=t10-rt;		t13=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    t14=a[idum+20]*c0A - a[idum+1+20]*s0A;	t15=a[idum+1+20]*c0A + a[idum+20]*s0A;
	    rt =a[idum+52]*c1A - a[idum+1+52]*s1A;	it =a[idum+1+52]*c1A + a[idum+52]*s1A;
	    t16=t14-rt;		t17=t15-it;
	    t14=t14+rt;		t15=t15+it;

	    rt =t14;		it =t15;
	    t14=t10-rt;		t15=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t16;		it =t17;
	    t16=t12+it;		t17=t13-rt;
	    t12=t12-it;		t13=t13+rt;

	    t18=a[idum+12]*c06 - a[idum+1+12]*s06;	t19=a[idum+1+12]*c06 + a[idum+12]*s06;
	    rt =a[idum+44]*c16 - a[idum+1+44]*s16;	it =a[idum+1+44]*c16 + a[idum+44]*s16;
	    t1A=t18-rt;		t1B=t19-it;
	    t18=t18+rt;		t19=t19+it;

	    t1C=a[idum+28]*c0E - a[idum+1+28]*s0E;	t1D=a[idum+1+28]*c0E + a[idum+28]*s0E;
	    rt =a[idum+60]*c1E - a[idum+1+60]*s1E;	it =a[idum+1+60]*c1E + a[idum+60]*s1E;
	    t1E=t1C-rt;		t1F=t1D-it;
	    t1C=t1C+rt;		t1D=t1D+it;

	    rt =t1C;		it =t1D;
	    t1C=t18-rt;		t1D=t19-it;
	    t18=t18+rt;		t19=t19+it;

	    rt =t1E;		it =t1F;
	    t1E=t1A+it;		t1F=t1B-rt;
	    t1A=t1A-it;		t1B=t1B+rt;

	    rt =t18;		it =t19;
	    t18=t10-rt;		t19=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t1C;		it =t1D;
	    t1C=t14+it;		t1D=t15-rt;
	    t14=t14-it;		t15=t15+rt;

	    rt =(t1A-t1B)*ISRT2;it =(t1A+t1B)*ISRT2;
	    t1A=t12-rt;		t1B=t13-it;
	    t12=t12+rt;		t13=t13+it;

	    rt =(t1E+t1F)*ISRT2;it =(t1F-t1E)*ISRT2;
	    t1E=t16+rt;		t1F=t17+it;
	    t16=t16-rt;		t17=t17-it;

/*...Block 3:	*/
	    t20=a[idum+2 ]*c01 - a[idum+1+2 ]*s01;	t21=a[idum+1+2 ]*c01 + a[idum+2 ]*s01;
	    rt =a[idum+34]*c11 - a[idum+1+34]*s11;	it =a[idum+1+34]*c11 + a[idum+34]*s11;
	    t22=t20-rt;		t23=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    t24=a[idum+18]*c09 - a[idum+1+18]*s09;	t25=a[idum+1+18]*c09 + a[idum+18]*s09;
	    rt =a[idum+50]*c19 - a[idum+1+50]*s19;	it =a[idum+1+50]*c19 + a[idum+50]*s19;
	    t26=t24-rt;		t27=t25-it;
	    t24=t24+rt;		t25=t25+it;

	    rt =t24;		it =t25;
	    t24=t20-rt;		t25=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t26;		it =t27;
	    t26=t22+it;		t27=t23-rt;
	    t22=t22-it;		t23=t23+rt;

	    t28=a[idum+10]*c05 - a[idum+1+10]*s05;	t29=a[idum+1+10]*c05 + a[idum+10]*s05;
	    rt =a[idum+42]*c15 - a[idum+1+42]*s15;	it =a[idum+1+42]*c15 + a[idum+42]*s15;
	    t2A=t28-rt;		t2B=t29-it;
	    t28=t28+rt;		t29=t29+it;

	    t2C=a[idum+26]*c0D - a[idum+1+26]*s0D;	t2D=a[idum+1+26]*c0D + a[idum+26]*s0D;
	    rt =a[idum+58]*c1D - a[idum+1+58]*s1D;	it =a[idum+1+58]*c1D + a[idum+58]*s1D;
	    t2E=t2C-rt;		t2F=t2D-it;
	    t2C=t2C+rt;		t2D=t2D+it;

	    rt =t2C;		it =t2D;
	    t2C=t28-rt;		t2D=t29-it;
	    t28=t28+rt;		t29=t29+it;

	    rt =t2E;		it =t2F;
	    t2E=t2A+it;		t2F=t2B-rt;
	    t2A=t2A-it;		t2B=t2B+rt;

	    rt =t28;		it =t29;
	    t28=t20-rt;		t29=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t2C;		it =t2D;
	    t2C=t24+it;		t2D=t25-rt;
	    t24=t24-it;		t25=t25+rt;

	    rt =(t2A-t2B)*ISRT2;it =(t2A+t2B)*ISRT2;
	    t2A=t22-rt;		t2B=t23-it;
	    t22=t22+rt;		t23=t23+it;

	    rt =(t2E+t2F)*ISRT2;it =(t2F-t2E)*ISRT2;
	    t2E=t26+rt;		t2F=t27+it;
	    t26=t26-rt;		t27=t27-it;

/*...Block 4:	*/
	    t30=a[idum+6 ]*c03 - a[idum+1+6 ]*s03;	t31=a[idum+1+6 ]*c03 + a[idum+6 ]*s03;
	    rt =a[idum+38]*c13 - a[idum+1+38]*s13;	it =a[idum+1+38]*c13 + a[idum+38]*s13;
	    t32=t30-rt;		t33=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    t34=a[idum+22]*c0B - a[idum+1+22]*s0B;	t35=a[idum+1+22]*c0B + a[idum+22]*s0B;
	    rt =a[idum+54]*c1B - a[idum+1+54]*s1B;	it =a[idum+1+54]*c1B + a[idum+54]*s1B;
	    t36=t34-rt;		t37=t35-it;
	    t34=t34+rt;		t35=t35+it;

	    rt =t34;		it =t35;
	    t34=t30-rt;		t35=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t36;		it =t37;
	    t36=t32+it;		t37=t33-rt;
	    t32=t32-it;		t33=t33+rt;

	    t38=a[idum+14]*c07 - a[idum+1+14]*s07;	t39=a[idum+1+14]*c07 + a[idum+14]*s07;
	    rt =a[idum+46]*c17 - a[idum+1+46]*s17;	it =a[idum+1+46]*c17 + a[idum+46]*s17;
	    t3A=t38-rt;		t3B=t39-it;
	    t38=t38+rt;		t39=t39+it;

	    t3C=a[idum+30]*c0F - a[idum+1+30]*s0F;	t3D=a[idum+1+30]*c0F + a[idum+30]*s0F;
	    rt =a[idum+62]*c1F - a[idum+1+62]*s1F;	it =a[idum+1+62]*c1F + a[idum+62]*s1F;
	    t3E=t3C-rt;		t3F=t3D-it;
	    t3C=t3C+rt;		t3D=t3D+it;

	    rt =t3C;		it =t3D;
	    t3C=t38-rt;		t3D=t39-it;
	    t38=t38+rt;		t39=t39+it;

	    rt =t3E;		it =t3F;
	    t3E=t3A+it;		t3F=t3B-rt;
	    t3A=t3A-it;		t3B=t3B+rt;

	    rt =t38;		it =t39;
	    t38=t30-rt;		t39=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t3C;		it =t3D;
	    t3C=t34+it;		t3D=t35-rt;
	    t34=t34-it;		t35=t35+rt;

	    rt =(t3A-t3B)*ISRT2;it =(t3A+t3B)*ISRT2;
	    t3A=t32-rt;		t3B=t33-it;
	    t32=t32+rt;		t33=t33+it;

	    rt =(t3E+t3F)*ISRT2;it =(t3F-t3E)*ISRT2;
	    t3E=t36+rt;		t3F=t37+it;
	    t36=t36-rt;		t37=t37-it;

/*...and now do eight radix-4 transforms, including the internal twiddle factors:
	1, exp(i* 1*twopi/32) =       ( c32_1, s32_1), exp(i* 2*twopi/32) =       ( c    , s    ), exp(i* 3*twopi/32) =       ( c32_3, s32_3) (for inputs to transform block 2),
	1, exp(i* 2*twopi/32) =       ( c    , s    ), exp(i* 4*twopi/32) = ISRT2*( 1    , 1    ), exp(i* 3*twopi/32) =       ( s    , c    ) (for inputs to transform block 3),
	1, exp(i* 3*twopi/32) =       ( c32_3, s32_3), exp(i* 6*twopi/32) =       ( s    , c    ), exp(i* 9*twopi/32) =       (-s32_1, c32_1) (for inputs to transform block 4),
	1, exp(i* 4*twopi/32) = ISRT2*( 1    , 1    ), exp(i* 8*twopi/32) =       ( 0    , 1    ), exp(i*12*twopi/32) = ISRT2*(-1    , 1    ) (for inputs to transform block 5),
	1, exp(i* 5*twopi/32) =       ( s32_3, c32_3), exp(i*10*twopi/32) =       (-s    , c    ), exp(i*15*twopi/32) =       (-c32_1, s32_1) (for inputs to transform block 6),
	1, exp(i* 6*twopi/32) =       ( s    , c    ), exp(i*12*twopi/32) = ISRT2*(-1    , 1    ), exp(i*18*twopi/32) =       (-c    ,-s    ) (for inputs to transform block 7),
	1, exp(i* 7*twopi/32) =       ( s32_1, c32_1), exp(i*14*twopi/32) =       (-c    , s    ), exp(i*21*twopi/32) =       (-s32_3,-c32_3) (for inputs to transform block 8),
   and only the last 3 inputs to each of the radix-4 transforms 2 through 8 are multiplied by non-unity twiddles.
   Within each block we process the 8 needed twiddles in bit-reversed order.	*/

/*...Block 1: t00,t10,t20,t30	*/
	    rt =t10;	t10=t00-rt;	t00=t00+rt;
	    it =t11;	t11=t01-it;	t01=t01+it;

	    rt =t30;	t30=t20-rt;	t20=t20+rt;
	    it =t31;	t31=t21-it;	t21=t21+it;

	    a1p00r = t00+t20;		a1p00i = t01+t21;
	    a1p01r = t00-t20;		a1p01i = t01-t21;

	    a1p02r = t10-t31;		a1p02i = t11+t30;
	    a1p03r = t10+t31;		a1p03i = t11-t30;

/*...Block 5: t08,t18,t28,t38	*/
	    rt =t18;	t18=t08+t19;	t08=t08-t19;
			t19=t09-rt;	t09=t09+rt;

	    rt =(t28-t29)*ISRT2;	t29=(t28+t29)*ISRT2;		t28=rt;
	    rt =(t39+t38)*ISRT2;	it =(t39-t38)*ISRT2;
	    t38=t28+rt;			t28=t28-rt;
	    t39=t29+it;			t29=t29-it;

	    a1p04r = t08+t28;		a1p04i = t09+t29;
	    a1p05r = t08-t28;		a1p05i = t09-t29;

	    a1p06r = t18-t39;		a1p06i = t19+t38;
	    a1p07r = t18+t39;		a1p07i = t19-t38;

/*...Block 3: t04,t14,t24,t34	*/
	    rt =(t14-t15)*ISRT2;	it =(t14+t15)*ISRT2;
	    t14=t04-rt;			t04=t04+rt;
	    t15=t05-it;			t05=t05+it;

	    rt =t24*c - t25*s;		t25=t25*c + t24*s;		t24=rt;
	    rt =t34*s - t35*c;		it =t35*s + t34*c;
	    t34=t24-rt;			t24=t24+rt;
	    t35=t25-it;			t25=t25+it;

	    a1p08r = t04+t24;		a1p08i = t05+t25;
	    a1p09r = t04-t24;		a1p09i = t05-t25;

	    a1p0Ar = t14-t35;		a1p0Ai = t15+t34;
	    a1p0Br = t14+t35;		a1p0Bi = t15-t34;

/*...Block 7: t0C,t1C,t2C,t3C	*/
	    rt =(t1D+t1C)*ISRT2;	it =(t1D-t1C)*ISRT2;
	    t1C=t0C+rt;			t0C=t0C-rt;
	    t1D=t0D+it;			t0D=t0D-it;

	    rt =t2C*s - t2D*c;		t2D=t2D*s + t2C*c;		t2C=rt;
	    rt =t3C*c - t3D*s;		it =t3D*c + t3C*s;
	    t3C=t2C+rt;			t2C=t2C-rt;
	    t3D=t2D+it;			t2D=t2D-it;

	    a1p0Cr = t0C+t2C;		a1p0Ci = t0D+t2D;
	    a1p0Dr = t0C-t2C;		a1p0Di = t0D-t2D;

	    a1p0Er = t1C-t3D;		a1p0Ei = t1D+t3C;
	    a1p0Fr = t1C+t3D;		a1p0Fi = t1D-t3C;

/*...Block 2: t02,t12,t22,t32	*/
	    rt =t12*c - t13*s;		it =t13*c + t12*s;
	    t12=t02-rt;			t02=t02+rt;
	    t13=t03-it;			t03=t03+it;

	    rt =t22*c32_1 - t23*s32_1;	t23=t23*c32_1 + t22*s32_1;	t22=rt;
	    rt =t32*c32_3 - t33*s32_3;	it =t33*c32_3 + t32*s32_3;
	    t32=t22-rt;			t22=t22+rt;
	    t33=t23-it;			t23=t23+it;

	    a1p10r = t02+t22;		a1p10i = t03+t23;
	    a1p11r = t02-t22;		a1p11i = t03-t23;

	    a1p12r = t12-t33;		a1p12i = t13+t32;
	    a1p13r = t12+t33;		a1p13i = t13-t32;

/*...Block 6: t0A,t1A,t2A,t3A	*/
	    rt =t1A*s + t1B*c;		it =t1B*s - t1A*c;
	    t1A=t0A+rt;			t0A =t0A-rt;
	    t1B=t0B+it;			t0B =t0B-it;

	    rt =t2A*s32_3 - t2B*c32_3;	t2B=t2B*s32_3 + t2A*c32_3;	t2A=rt;
	    rt =t3A*c32_1 + t3B*s32_1;	it =t3B*c32_1 - t3A*s32_1;
	    t3A=t2A+rt;			t2A=t2A-rt;
	    t3B=t2B+it;			t2B=t2B-it;

	    a1p14r = t0A+t2A;		a1p14i = t0B+t2B;
	    a1p15r = t0A-t2A;		a1p15i = t0B-t2B;

	    a1p16r = t1A-t3B;		a1p16i = t1B+t3A;
	    a1p17r = t1A+t3B;		a1p17i = t1B-t3A;

/*...Block 4: t06,t16,t26,t36	*/
	    rt =t16*s - t17*c;		it =t17*s + t16*c;
	    t16=t06-rt;			t06 =t06+rt;
	    t17=t07-it;			t07 =t07+it;

	    rt =t26*c32_3 - t27*s32_3;	t27=t27*c32_3 + t26*s32_3;	t26=rt;
	    rt =t36*s32_1 + t37*c32_1;	it =t37*s32_1 - t36*c32_1;
	    t36=t26+rt;			t26=t26-rt;
	    t37=t27+it;			t27=t27-it;

	    a1p18r = t06+t26;		a1p18i = t07+t27;
	    a1p19r = t06-t26;		a1p19i = t07-t27;

	    a1p1Ar = t16-t37;		a1p1Ai = t17+t36;
	    a1p1Br = t16+t37;		a1p1Bi = t17-t36;

/*...Block 8: t0E,t1E,t2E,t3E	*/
	    rt =t1E*c + t1F*s;		it =t1F*c - t1E*s;
	    t1E=t0E+rt;			t0E =t0E-rt;
	    t1F=t0F+it;			t0F =t0F-it;

	    rt =t2E*s32_1 - t2F*c32_1;	t2F=t2F*s32_1 + t2E*c32_1;	t2E=rt;
	    rt =t3E*s32_3 - t3F*c32_3;	it =t3F*s32_3 + t3E*c32_3;
	    t3E=t2E+rt;			t2E=t2E-rt;
	    t3F=t2F+it;			t2F=t2F-it;

	    a1p1Cr = t0E+t2E;		a1p1Ci = t0F+t2F;
	    a1p1Dr = t0E-t2E;		a1p1Di = t0F-t2F;

	    a1p1Er = t1E-t3F;		a1p1Ei = t1F+t3E;
	    a1p1Fr = t1E+t3F;		a1p1Fi = t1F-t3E;

	k=k+1;	/* increment sincos array index */

/* 2nd set of inputs: */
	    idum = j2pad;
	    iroot = index[k];
	    l = iroot;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += iroot;			/* 2*iroot	*/
	    t00=rt0[k1].re;	t01=rt0[k1].im;
	    rt =rt1[k2].re;	it =rt1[k2].im;
	    c21=t00*rt -t01*it;	s21=t00*it +t01*rt;

	if(j1 == 0){ re=c21;	im=s21; }   /* The j1 = 0 case is special... */

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += iroot;				/* 3*iroot	*/
	    t00=rt0[k1].re;	t01=rt0[k1].im;
	    rt =rt1[k2].re;	it =rt1[k2].im;
	    c22=t00*rt -t01*it;	s22=t00*it +t01*rt;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += 4*iroot;			/* 7*iroot	*/
	    iroot = l;				/* 7*iroot	*/
	    t00=rt0[k1].re;	t01=rt0[k1].im;
	    rt =rt1[k2].re;	it =rt1[k2].im;
	    c23=t00*rt -t01*it;	s23=t00*it +t01*rt;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += iroot;				/* 14*iroot	*/
	    t00=rt0[k1].re;	t01=rt0[k1].im;
	    rt =rt1[k2].re;	it =rt1[k2].im;
	    c27=t00*rt -t01*it;	s27=t00*it +t01*rt;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += iroot;				/* 21*iroot	*/
	    t00=rt0[k1].re;	t01=rt0[k1].im;
	    rt =rt1[k2].re;	it =rt1[k2].im;
	    c2E=t00*rt -t01*it;	s2E=t00*it +t01*rt;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    l += iroot;				/* 28*iroot	*/
	    t00=rt0[k1].re;	t01=rt0[k1].im;
	    rt =rt1[k2].re;	it =rt1[k2].im;
	    c35=t00*rt -t01*it;	s35=t00*it +t01*rt;

	    k1=(l & nrtm1);
	    k2=(l >> nrt_bits);
	    t00=rt0[k1].re;	t01=rt0[k1].im;
	    rt =rt1[k2].re;	it =rt1[k2].im;
	    c3C=t00*rt -t01*it;	s3C=t00*it +t01*rt;

/*(c,s)4-10 (decimal indices), 4-0A (hexadecimal indices):	*/
	    t00=c21*c27; t01=c21*s27; t02=s21*c27; t03=s21*s27;
	    c26=t00+t03; s26=t01-t02; c28=t00-t03; s28=t01+t02;

	    t00=c22*c27; t01=c22*s27; t02=s22*c27; t03=s22*s27;
	    c25=t00+t03; s25=t01-t02; c29=t00-t03; s29=t01+t02;

	    t00=c23*c27; t01=c23*s27; t02=s23*c27; t03=s23*s27;
	    c24=t00+t03; s24=t01-t02; c2A=t00-t03; s2A=t01+t02;

/*(c,s)11-17 (decimal indices), 0B-11 (hexadecimal indices):;	*/
	    t00=c21*c2E; t01=c21*s2E; t02=s21*c2E; t03=s21*s2E;
	    c2D=t00+t03; s2D=t01-t02; c2F=t00-t03; s2F=t01+t02;

	    t00=c22*c2E; t01=c22*s2E; t02=s22*c2E; t03=s22*s2E;
	    c2C=t00+t03; s2C=t01-t02; c30=t00-t03; s30=t01+t02;

	    t00=c23*c2E; t01=c23*s2E; t02=s23*c2E; t03=s23*s2E;
	    c2B=t00+t03; s2B=t01-t02; c31=t00-t03; s31=t01+t02;

/*(c,s)18-24 (decimal indices), 12-18 (hexadecimal indices):	*/
	    t00=c21*c35; t01=c21*s35; t02=s21*c35; t03=s21*s35;
	    c34=t00+t03; s34=t01-t02; c36=t00-t03; s36=t01+t02;

	    t00=c22*c35; t01=c22*s35; t02=s22*c35; t03=s22*s35;
	    c33=t00+t03; s33=t01-t02; c37=t00-t03; s37=t01+t02;

	    t00=c23*c35; t01=c23*s35; t02=s23*c35; t03=s23*s35;
	    c32=t00+t03; s32=t01-t02; c38=t00-t03; s38=t01+t02;

/*(c,s)25-31 (decimal indices), 19-1F (hexadecimal indices):	*/
	    t00=c21*c3C; t01=c21*s3C; t02=s21*c3C; t03=s21*s3C;
	    c3B=t00+t03; s3B=t01-t02; c3D=t00-t03; s3D=t01+t02;

	    t00=c22*c3C; t01=c22*s3C; t02=s22*c3C; t03=s22*s3C;
	    c3A=t00+t03; s3A=t01-t02; c3E=t00-t03; s3E=t01+t02;

	    t00=c23*c3C; t01=c23*s3C; t02=s23*c3C; t03=s23*s3C;
	    c39=t00+t03; s39=t01-t02; c3F=t00-t03; s3F=t01+t02;

/*       gather the needed data (32 64-bit complex, i.e. 64 64-bit reals) and do the first set of four length-8 transforms...
         We process the sincos data in bit-reversed order.	*/

/*...Block 1:	*/
	    t00=a[idum   ];				t01=a[idum+1   ];
	    rt =a[idum+32]*c30 - a[idum+1+32]*s30;	it =a[idum+1+32]*c30 + a[idum+32]*s30;
	    t02=t00-rt;		t03=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    t04=a[idum+16]*c28 - a[idum+1+16]*s28;	t05=a[idum+1+16]*c28 + a[idum+16]*s28;
	    rt =a[idum+48]*c38 - a[idum+1+48]*s38;	it =a[idum+1+48]*c38 + a[idum+48]*s38;
	    t06=t04-rt;		t07=t05-it;
	    t04=t04+rt;		t05=t05+it;

	    rt =t04;		it =t05;
	    t04=t00-rt;		t05=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t06;		it =t07;
	    t06=t02+it;		t07=t03-rt;
	    t02=t02-it;		t03=t03+rt;

	    t08=a[idum+8 ]*c24 - a[idum+1+8 ]*s24;	t09=a[idum+1+8 ]*c24 + a[idum+8 ]*s24;
	    rt =a[idum+40]*c34 - a[idum+1+40]*s34;	it =a[idum+1+40]*c34 + a[idum+40]*s34;
	    t0A=t08-rt;		t0B=t09-it;
	    t08=t08+rt;		t09=t09+it;

	    t0C=a[idum+24]*c2C - a[idum+1+24]*s2C;	t0D=a[idum+1+24]*c2C + a[idum+24]*s2C;
	    rt =a[idum+56]*c3C - a[idum+1+56]*s3C;	it =a[idum+1+56]*c3C + a[idum+56]*s3C;
	    t0E=t0C-rt;		t0F=t0D-it;
	    t0C=t0C+rt;		t0D=t0D+it;

	    rt =t0C;		it =t0D;
	    t0C=t08-rt;		t0D=t09-it;
	    t08=t08+rt;		t09=t09+it;

	    rt =t0E;		it =t0F;
	    t0E=t0A+it;		t0F=t0B-rt;
	    t0A=t0A-it;		t0B=t0B+rt;

	    rt =t08;		it =t09;
	    t08=t00-rt;		t09=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t0C;		it =t0D;
	    t0C=t04+it;		t0D=t05-rt;
	    t04=t04-it;		t05=t05+rt;

	    rt =(t0A-t0B)*ISRT2;it =(t0A+t0B)*ISRT2;
	    t0A=t02-rt;		t0B=t03-it;
	    t02=t02+rt;		t03=t03+it;

	    rt =(t0E+t0F)*ISRT2;it =(t0F-t0E)*ISRT2;
	    t0E=t06+rt;		t0F=t07+it;
	    t06=t06-rt;		t07=t07-it;

/*...Block 2:	*/
	    t10=a[idum+4 ]*c22 - a[idum+1+4 ]*s22;	t11=a[idum+1+4 ]*c22 + a[idum+4 ]*s22;
	    rt =a[idum+36]*c32 - a[idum+1+36]*s32;	it =a[idum+1+36]*c32 + a[idum+36]*s32;
	    t12=t10-rt;		t13=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    t14=a[idum+20]*c2A - a[idum+1+20]*s2A;	t15=a[idum+1+20]*c2A + a[idum+20]*s2A;
	    rt =a[idum+52]*c3A - a[idum+1+52]*s3A;	it =a[idum+1+52]*c3A + a[idum+52]*s3A;
	    t16=t14-rt;		t17=t15-it;
	    t14=t14+rt;		t15=t15+it;

	    rt =t14;		it =t15;
	    t14=t10-rt;		t15=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t16;		it =t17;
	    t16=t12+it;		t17=t13-rt;
	    t12=t12-it;		t13=t13+rt;

	    t18=a[idum+12]*c26 - a[idum+1+12]*s26;	t19=a[idum+1+12]*c26 + a[idum+12]*s26;
	    rt =a[idum+44]*c36 - a[idum+1+44]*s36;	it =a[idum+1+44]*c36 + a[idum+44]*s36;
	    t1A=t18-rt;		t1B=t19-it;
	    t18=t18+rt;		t19=t19+it;

	    t1C=a[idum+28]*c2E - a[idum+1+28]*s2E;	t1D=a[idum+1+28]*c2E + a[idum+28]*s2E;
	    rt =a[idum+60]*c3E - a[idum+1+60]*s3E;	it =a[idum+1+60]*c3E + a[idum+60]*s3E;
	    t1E=t1C-rt;		t1F=t1D-it;
	    t1C=t1C+rt;		t1D=t1D+it;

	    rt =t1C;		it =t1D;
	    t1C=t18-rt;		t1D=t19-it;
	    t18=t18+rt;		t19=t19+it;

	    rt =t1E;		it =t1F;
	    t1E=t1A+it;		t1F=t1B-rt;
	    t1A=t1A-it;		t1B=t1B+rt;

	    rt =t18;		it =t19;
	    t18=t10-rt;		t19=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t1C;		it =t1D;
	    t1C=t14+it;		t1D=t15-rt;
	    t14=t14-it;		t15=t15+rt;

	    rt =(t1A-t1B)*ISRT2;it =(t1A+t1B)*ISRT2;
	    t1A=t12-rt;		t1B=t13-it;
	    t12=t12+rt;		t13=t13+it;

	    rt =(t1E+t1F)*ISRT2;it =(t1F-t1E)*ISRT2;
	    t1E=t16+rt;		t1F=t17+it;
	    t16=t16-rt;		t17=t17-it;

/*...Block 3:	*/
	    t20=a[idum+2 ]*c21 - a[idum+1+2 ]*s21;	t21=a[idum+1+2 ]*c21 + a[idum+2 ]*s21;
	    rt =a[idum+34]*c31 - a[idum+1+34]*s31;	it =a[idum+1+34]*c31 + a[idum+34]*s31;
	    t22=t20-rt;		t23=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    t24=a[idum+18]*c29 - a[idum+1+18]*s29;	t25=a[idum+1+18]*c29 + a[idum+18]*s29;
	    rt =a[idum+50]*c39 - a[idum+1+50]*s39;	it =a[idum+1+50]*c39 + a[idum+50]*s39;
	    t26=t24-rt;		t27=t25-it;
	    t24=t24+rt;		t25=t25+it;

	    rt =t24;		it =t25;
	    t24=t20-rt;		t25=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t26;		it =t27;
	    t26=t22+it;		t27=t23-rt;
	    t22=t22-it;		t23=t23+rt;

	    t28=a[idum+10]*c25 - a[idum+1+10]*s25;	t29=a[idum+1+10]*c25 + a[idum+10]*s25;
	    rt =a[idum+42]*c35 - a[idum+1+42]*s35;	it =a[idum+1+42]*c35 + a[idum+42]*s35;
	    t2A=t28-rt;		t2B=t29-it;
	    t28=t28+rt;		t29=t29+it;

	    t2C=a[idum+26]*c2D - a[idum+1+26]*s2D;	t2D=a[idum+1+26]*c2D + a[idum+26]*s2D;
	    rt =a[idum+58]*c3D - a[idum+1+58]*s3D;	it =a[idum+1+58]*c3D + a[idum+58]*s3D;
	    t2E=t2C-rt;		t2F=t2D-it;
	    t2C=t2C+rt;		t2D=t2D+it;

	    rt =t2C;		it =t2D;
	    t2C=t28-rt;		t2D=t29-it;
	    t28=t28+rt;		t29=t29+it;

	    rt =t2E;		it =t2F;
	    t2E=t2A+it;		t2F=t2B-rt;
	    t2A=t2A-it;		t2B=t2B+rt;

	    rt =t28;		it =t29;
	    t28=t20-rt;		t29=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t2C;		it =t2D;
	    t2C=t24+it;		t2D=t25-rt;
	    t24=t24-it;		t25=t25+rt;

	    rt =(t2A-t2B)*ISRT2;it =(t2A+t2B)*ISRT2;
	    t2A=t22-rt;		t2B=t23-it;
	    t22=t22+rt;		t23=t23+it;

	    rt =(t2E+t2F)*ISRT2;it =(t2F-t2E)*ISRT2;
	    t2E=t26+rt;		t2F=t27+it;
	    t26=t26-rt;		t27=t27-it;

/*...Block 4:	*/
	    t30=a[idum+6 ]*c23 - a[idum+1+6 ]*s23;	t31=a[idum+1+6 ]*c23 + a[idum+6 ]*s23;
	    rt =a[idum+38]*c33 - a[idum+1+38]*s33;	it =a[idum+1+38]*c33 + a[idum+38]*s33;
	    t32=t30-rt;		t33=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    t34=a[idum+22]*c2B - a[idum+1+22]*s2B;	t35=a[idum+1+22]*c2B + a[idum+22]*s2B;
	    rt =a[idum+54]*c3B - a[idum+1+54]*s3B;	it =a[idum+1+54]*c3B + a[idum+54]*s3B;
	    t36=t34-rt;		t37=t35-it;
	    t34=t34+rt;		t35=t35+it;

	    rt =t34;		it =t35;
	    t34=t30-rt;		t35=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t36;		it =t37;
	    t36=t32+it;		t37=t33-rt;
	    t32=t32-it;		t33=t33+rt;

	    t38=a[idum+14]*c27 - a[idum+1+14]*s27;	t39=a[idum+1+14]*c27 + a[idum+14]*s27;
	    rt =a[idum+46]*c37 - a[idum+1+46]*s37;	it =a[idum+1+46]*c37 + a[idum+46]*s37;
	    t3A=t38-rt;		t3B=t39-it;
	    t38=t38+rt;		t39=t39+it;

	    t3C=a[idum+30]*c2F - a[idum+1+30]*s2F;	t3D=a[idum+1+30]*c2F + a[idum+30]*s2F;
	    rt =a[idum+62]*c3F - a[idum+1+62]*s3F;	it =a[idum+1+62]*c3F + a[idum+62]*s3F;
	    t3E=t3C-rt;		t3F=t3D-it;
	    t3C=t3C+rt;		t3D=t3D+it;

	    rt =t3C;		it =t3D;
	    t3C=t38-rt;		t3D=t39-it;
	    t38=t38+rt;		t39=t39+it;

	    rt =t3E;		it =t3F;
	    t3E=t3A+it;		t3F=t3B-rt;
	    t3A=t3A-it;		t3B=t3B+rt;

	    rt =t38;		it =t39;
	    t38=t30-rt;		t39=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t3C;		it =t3D;
	    t3C=t34+it;		t3D=t35-rt;
	    t34=t34-it;		t35=t35+rt;

	    rt =(t3A-t3B)*ISRT2;it =(t3A+t3B)*ISRT2;
	    t3A=t32-rt;		t3B=t33-it;
	    t32=t32+rt;		t33=t33+it;

	    rt =(t3E+t3F)*ISRT2;it =(t3F-t3E)*ISRT2;
	    t3E=t36+rt;		t3F=t37+it;
	    t36=t36-rt;		t37=t37-it;
/*
!...and now do eight radix-4 transforms, including the internal twiddle factors:
*/

/*...Block 1: t00,t10,t20,t30	*/
	    rt =t10;	t10=t00-rt;	t00=t00+rt;
	    it =t11;	t11=t01-it;	t01=t01+it;

	    rt =t30;	t30=t20-rt;	t20=t20+rt;
	    it =t31;	t31=t21-it;	t21=t21+it;

	    a2p00r = t00+t20;		a2p00i = t01+t21;
	    a2p01r = t00-t20;		a2p01i = t01-t21;

	    a2p02r = t10-t31;		a2p02i = t11+t30;
	    a2p03r = t10+t31;		a2p03i = t11-t30;

/*...Block 5: t08,t18,t28,t38	*/
	    rt =t18;	t18=t08+t19;	t08=t08-t19;
			t19=t09-rt;	t09=t09+rt;

	    rt =(t28-t29)*ISRT2;	t29=(t28+t29)*ISRT2;		t28=rt;
	    rt =(t39+t38)*ISRT2;	it =(t39-t38)*ISRT2;
	    t38=t28+rt;			t28=t28-rt;
	    t39=t29+it;			t29=t29-it;

	    a2p04r = t08+t28;		a2p04i = t09+t29;
	    a2p05r = t08-t28;		a2p05i = t09-t29;

	    a2p06r = t18-t39;		a2p06i = t19+t38;
	    a2p07r = t18+t39;		a2p07i = t19-t38;

/*...Block 3: t04,t14,t24,t34	*/
	    rt =(t14-t15)*ISRT2;	it =(t14+t15)*ISRT2;
	    t14=t04-rt;			t04=t04+rt;
	    t15=t05-it;			t05=t05+it;

	    rt =t24*c - t25*s;		t25=t25*c + t24*s;		t24=rt;
	    rt =t34*s - t35*c;		it =t35*s + t34*c;
	    t34=t24-rt;			t24=t24+rt;
	    t35=t25-it;			t25=t25+it;

	    a2p08r = t04+t24;		a2p08i = t05+t25;
	    a2p09r = t04-t24;		a2p09i = t05-t25;

	    a2p0Ar = t14-t35;		a2p0Ai = t15+t34;
	    a2p0Br = t14+t35;		a2p0Bi = t15-t34;

/*...Block 7: t0C,t1C,t2C,t3C	*/
	    rt =(t1D+t1C)*ISRT2;	it =(t1D-t1C)*ISRT2;
	    t1C=t0C+rt;			t0C=t0C-rt;
	    t1D=t0D+it;			t0D=t0D-it;

	    rt =t2C*s - t2D*c;		t2D=t2D*s + t2C*c;		t2C=rt;
	    rt =t3C*c - t3D*s;		it =t3D*c + t3C*s;
	    t3C=t2C+rt;			t2C=t2C-rt;
	    t3D=t2D+it;			t2D=t2D-it;

	    a2p0Cr = t0C+t2C;		a2p0Ci = t0D+t2D;
	    a2p0Dr = t0C-t2C;		a2p0Di = t0D-t2D;

	    a2p0Er = t1C-t3D;		a2p0Ei = t1D+t3C;
	    a2p0Fr = t1C+t3D;		a2p0Fi = t1D-t3C;

/*...Block 2: t02,t12,t22,t32	*/
	    rt =t12*c - t13*s;		it =t13*c + t12*s;
	    t12=t02-rt;			t02=t02+rt;
	    t13=t03-it;			t03=t03+it;

	    rt =t22*c32_1 - t23*s32_1;	t23=t23*c32_1 + t22*s32_1;	t22=rt;
	    rt =t32*c32_3 - t33*s32_3;	it =t33*c32_3 + t32*s32_3;
	    t32=t22-rt;			t22=t22+rt;
	    t33=t23-it;			t23=t23+it;

	    a2p10r = t02+t22;		a2p10i = t03+t23;
	    a2p11r = t02-t22;		a2p11i = t03-t23;

	    a2p12r = t12-t33;		a2p12i = t13+t32;
	    a2p13r = t12+t33;		a2p13i = t13-t32;

/*...Block 6: t0A,t1A,t2A,t3A	*/
	    rt =t1A*s + t1B*c;		it =t1B*s - t1A*c;
	    t1A=t0A+rt;			t0A =t0A-rt;
	    t1B=t0B+it;			t0B =t0B-it;

	    rt =t2A*s32_3 - t2B*c32_3;	t2B=t2B*s32_3 + t2A*c32_3;	t2A=rt;
	    rt =t3A*c32_1 + t3B*s32_1;	it =t3B*c32_1 - t3A*s32_1;
	    t3A=t2A+rt;			t2A=t2A-rt;
	    t3B=t2B+it;			t2B=t2B-it;

	    a2p14r = t0A+t2A;		a2p14i = t0B+t2B;
	    a2p15r = t0A-t2A;		a2p15i = t0B-t2B;

	    a2p16r = t1A-t3B;		a2p16i = t1B+t3A;
	    a2p17r = t1A+t3B;		a2p17i = t1B-t3A;

/*...Block 4: t06,t16,t26,t36	*/
	    rt =t16*s - t17*c;		it =t17*s + t16*c;
	    t16=t06-rt;			t06 =t06+rt;
	    t17=t07-it;			t07 =t07+it;

	    rt =t26*c32_3 - t27*s32_3;	t27=t27*c32_3 + t26*s32_3;	t26=rt;
	    rt =t36*s32_1 + t37*c32_1;	it =t37*s32_1 - t36*c32_1;
	    t36=t26+rt;			t26=t26-rt;
	    t37=t27+it;			t27=t27-it;

	    a2p18r = t06+t26;		a2p18i = t07+t27;
	    a2p19r = t06-t26;		a2p19i = t07-t27;

	    a2p1Ar = t16-t37;		a2p1Ai = t17+t36;
	    a2p1Br = t16+t37;		a2p1Bi = t17-t36;

/*...Block 8: t0E,t1E,t2E,t3E	*/
	    rt =t1E*c + t1F*s;		it =t1F*c - t1E*s;
	    t1E=t0E+rt;			t0E =t0E-rt;
	    t1F=t0F+it;			t0F =t0F-it;

	    rt =t2E*s32_1 - t2F*c32_1;	t2F=t2F*s32_1 + t2E*c32_1;	t2E=rt;
	    rt =t3E*s32_3 - t3F*c32_3;	it =t3F*s32_3 + t3E*c32_3;
	    t3E=t2E+rt;			t2E=t2E-rt;
	    t3F=t2F+it;			t2F=t2F-it;

	    a2p1Cr = t0E+t2E;		a2p1Ci = t0F+t2F;
	    a2p1Dr = t0E-t2E;		a2p1Di = t0F-t2F;

	    a2p1Er = t1E-t3F;		a2p1Ei = t1F+t3E;
	    a2p1Fr = t1E+t3F;		a2p1Fi = t1F-t3E;

	k=k+1;	/* increment sincos array index */

/*...send the pairs of complex elements which are to be combined and sincos temporaries needed for the squaring to a
     small subroutine. The j1 = 0 case is again exceptional.	*/

	  if(j1==0)    	/* NB: mustn't use re, im as temps in exe1-3 section, since those contain saved sincos data for exe4 block.	*/
	  {
/*...j=0 (for which I(0)=Re{H(0)}^2+i*Im{H(0)}^2) is done separately...	*/

	    rt=a1p00r;
	    a1p00r=(rt+a1p00i)*(rt+a1p00i);
	    a1p00i=(rt-a1p00i)*(rt-a1p00i);
	    rt=a1p00r;
	    a1p00r=0.5*(rt+a1p00i);
	    a1p00i=0.5*(rt-a1p00i);

/*...as is j=N/2 (for which I(j)=H(j)^2). Note that under bit-reversal the N/2 element gets mapped into
     the second complex data slot, i.e. is adjacent to the starting element.	*/

	    rt =a1p01r*a1p01i;
	    a1p01r=(a1p01r+a1p01i)*(a1p01r-a1p01i);
	    a1p01i=rt+rt;

	    pair_square(&a1p02r,&a1p02i,&a1p03r,&a1p03i,0.0,1.0);		/* exe1	*/

	    rt=ISRT2;	it=ISRT2;

	    pair_square(&a1p04r,&a1p04i,&a1p07r,&a1p07i, rt, it);		/* exe2	*/
	    pair_square(&a1p06r,&a1p06i,&a1p05r,&a1p05i,-it, rt);		/* exe2	*/

	    t03=c; t04=0; t05=s; t06=0;
	    rt=t03-t04;		it=t05+t06;

	    pair_square(&a1p08r,&a1p08i,&a1p0Fr,&a1p0Fi, rt, it);		/* exe3	*/
	    pair_square(&a1p0Ar,&a1p0Ai,&a1p0Dr,&a1p0Di,-it, rt);		/* exe3	*/

	    rt=t05-t06;		it=t03+t04;

	    pair_square(&a1p0Cr,&a1p0Ci,&a1p0Br,&a1p0Bi, rt, it);		/* exe3	*/
	    pair_square(&a1p0Er,&a1p0Ei,&a1p09r,&a1p09i,-it, rt);		/* exe3	*/

	    t03=c32_1; t04=0; t05=s32_1; t06=0;
	    rt=t03-t04;		it=t05+t06;

	    pair_square(&a1p10r,&a1p10i,&a1p1Fr,&a1p1Fi, rt, it);		/* exe4	*/
	    pair_square(&a1p12r,&a1p12i,&a1p1Dr,&a1p1Di,-it, rt);		/* exe4	*/

	    t07=c32_3; t08=0; t09=s32_3; t0A=0;
	    rt=t09-t0A;		it=t07+t08;

	    pair_square(&a1p14r,&a1p14i,&a1p1Br,&a1p1Bi, rt, it);		/* exe4	*/
	    pair_square(&a1p16r,&a1p16i,&a1p19r,&a1p19i,-it, rt);		/* exe4	*/

	    rt=t07-t08;		it=t09+t0A;

	    pair_square(&a1p18r,&a1p18i,&a1p17r,&a1p17i, rt, it);		/* exe4	*/
	    pair_square(&a1p1Ar,&a1p1Ai,&a1p15r,&a1p15i,-it, rt);		/* exe4	*/

	    rt=t05-t06;		it=t03+t04;

	    pair_square(&a1p1Cr,&a1p1Ci,&a1p13r,&a1p13i, rt, it);		/* exe4	*/
	    pair_square(&a1p1Er,&a1p1Ei,&a1p11r,&a1p11i,-it, rt);		/* exe4	*/

/*   exe5:	*/

	    pair_square(&a2p00r,&a2p00i,&a2p1Fr,&a2p1Fi, re, im);
	    pair_square(&a2p02r,&a2p02i,&a2p1Dr,&a2p1Di,-im, re);

	    rt=(re-im)*ISRT2;	it=(re+im)*ISRT2;

	    pair_square(&a2p04r,&a2p04i,&a2p1Br,&a2p1Bi, rt, it);
	    pair_square(&a2p06r,&a2p06i,&a2p19r,&a2p19i,-it, rt);

	    t03=re*c; t04=im*s; t05=re*s; t06=im*c;
	    rt=t03-t04;		it=t05+t06;

	    pair_square(&a2p08r,&a2p08i,&a2p17r,&a2p17i, rt, it);
	    pair_square(&a2p0Ar,&a2p0Ai,&a2p15r,&a2p15i,-it, rt);

	    rt=t05-t06;		it=t03+t04;

	    pair_square(&a2p0Cr,&a2p0Ci,&a2p13r,&a2p13i, rt, it);
	    pair_square(&a2p0Er,&a2p0Ei,&a2p11r,&a2p11i,-it, rt);

	    t03=re*c32_1; t04=im*s32_1; t05=re*s32_1; t06=im*c32_1;
	    rt=t03-t04;		it=t05+t06;

	    pair_square(&a2p10r,&a2p10i,&a2p0Fr,&a2p0Fi, rt, it);
	    pair_square(&a2p12r,&a2p12i,&a2p0Dr,&a2p0Di,-it, rt);

	    t07=re*c32_3; t08=im*s32_3; t09=re*s32_3; t0A=im*c32_3;
	    rt=t09-t0A;		it=t07+t08;

	    pair_square(&a2p14r,&a2p14i,&a2p0Br,&a2p0Bi, rt, it);
	    pair_square(&a2p16r,&a2p16i,&a2p09r,&a2p09i,-it, rt);

	    rt=t07-t08;		it=t09+t0A;

	    pair_square(&a2p18r,&a2p18i,&a2p07r,&a2p07i, rt, it);
	    pair_square(&a2p1Ar,&a2p1Ai,&a2p05r,&a2p05i,-it, rt);

	    rt=t05-t06;		it=t03+t04;

	    pair_square(&a2p1Cr,&a2p1Ci,&a2p03r,&a2p03i, rt, it);
	    pair_square(&a2p1Er,&a2p1Ei,&a2p01r,&a2p01i,-it, rt);
	  }
	  else
	  {
pair_square(&a1p00r,&a1p00i,&a2p1Fr,&a2p1Fi, re, im);	pair_square(&a2p1Er,&a2p1Ei,&a1p01r,&a1p01i,-re, im);
pair_square(&a1p02r,&a1p02i,&a2p1Dr,&a2p1Di,-im, re);	pair_square(&a2p1Cr,&a2p1Ci,&a1p03r,&a1p03i, im, re);

	    rt=(re-im)*ISRT2;	it=(re+im)*ISRT2;

pair_square(&a1p04r,&a1p04i,&a2p1Br,&a2p1Bi, rt, it);	pair_square(&a2p1Ar,&a2p1Ai,&a1p05r,&a1p05i,-rt, it);
pair_square(&a1p06r,&a1p06i,&a2p19r,&a2p19i,-it, rt);	pair_square(&a2p18r,&a2p18i,&a1p07r,&a1p07i, it, rt);

	    t03=re*c; t04=im*s; t05=re*s; t06=im*c;
	    rt=t03-t04;		it=t05+t06;

pair_square(&a1p08r,&a1p08i,&a2p17r,&a2p17i, rt, it);	pair_square(&a2p16r,&a2p16i,&a1p09r,&a1p09i,-rt, it);
pair_square(&a1p0Ar,&a1p0Ai,&a2p15r,&a2p15i,-it, rt);	pair_square(&a2p14r,&a2p14i,&a1p0Br,&a1p0Bi, it, rt);

	    rt=t05-t06;		it=t03+t04;

pair_square(&a1p0Cr,&a1p0Ci,&a2p13r,&a2p13i, rt, it);	pair_square(&a2p12r,&a2p12i,&a1p0Dr,&a1p0Di,-rt, it);
pair_square(&a1p0Er,&a1p0Ei,&a2p11r,&a2p11i,-it, rt);	pair_square(&a2p10r,&a2p10i,&a1p0Fr,&a1p0Fi, it, rt);

	    t03=re*c32_1; t04=im*s32_1; t05=re*s32_1; t06=im*c32_1;
	    rt=t03-t04;		it=t05+t06;

pair_square(&a1p10r,&a1p10i,&a2p0Fr,&a2p0Fi, rt, it);	pair_square(&a2p0Er,&a2p0Ei,&a1p11r,&a1p11i,-rt, it);
pair_square(&a1p12r,&a1p12i,&a2p0Dr,&a2p0Di,-it, rt);	pair_square(&a2p0Cr,&a2p0Ci,&a1p13r,&a1p13i, it, rt);

	    t07=re*c32_3; t08=im*s32_3; t09=re*s32_3; t0A=im*c32_3;
	    rt=t09-t0A;		it=t07+t08;

pair_square(&a1p14r,&a1p14i,&a2p0Br,&a2p0Bi, rt, it);	pair_square(&a2p0Ar,&a2p0Ai,&a1p15r,&a1p15i,-rt, it);
pair_square(&a1p16r,&a1p16i,&a2p09r,&a2p09i,-it, rt);	pair_square(&a2p08r,&a2p08i,&a1p17r,&a1p17i, it, rt);

	    rt=t07-t08;		it=t09+t0A;

pair_square(&a1p18r,&a1p18i,&a2p07r,&a2p07i, rt, it);	pair_square(&a2p06r,&a2p06i,&a1p19r,&a1p19i,-rt, it);
pair_square(&a1p1Ar,&a1p1Ai,&a2p05r,&a2p05i,-it, rt);	pair_square(&a2p04r,&a2p04i,&a1p1Br,&a1p1Bi, it, rt);

	    rt=t05-t06;		it=t03+t04;

pair_square(&a1p1Cr,&a1p1Ci,&a2p03r,&a2p03i, rt, it);	pair_square(&a2p02r,&a2p02i,&a1p1Dr,&a1p1Di,-rt, it);
pair_square(&a1p1Er,&a1p1Ei,&a2p01r,&a2p01i,-it, rt);	pair_square(&a2p00r,&a2p00i,&a1p1Fr,&a1p1Fi, it, rt);

	  }	/* endif(j1==0)	*/

/*...And do an inverse DIT radix-32 pass on the squared-data blocks.	*/

/* 1st set of inputs: */
	  idum = j1pad;
#if PFETCH
add0 = &a[idum+64];
#endif
/*   gather the needed data (32 64-bit complex, i.e. 64 64-bit reals) and do the first set of four length-8 IDIT transforms... */

/*...Block 1: */
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    t02=a1p00r-a1p01r;	t03=a1p00i-a1p01i;
	    t00=a1p00r+a1p01r;	t01=a1p00i+a1p01i;

	    t06=a1p02r-a1p03r;	t07=a1p02i-a1p03i;
	    t04=a1p02r+a1p03r;	t05=a1p02i+a1p03i;

	    rt =t04;		it =t05;
	    t04=t00-rt;		t05=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t06;		it =t07;
	    t06=t02-it;		t07=t03+rt;
	    t02=t02+it;		t03=t03-rt;

	    t0A=a1p04r-a1p05r;	t0B=a1p04i-a1p05i;
	    t08=a1p04r+a1p05r;	t09=a1p04i+a1p05i;

	    t0E=a1p06r-a1p07r;	t0F=a1p06i-a1p07i;
	    t0C=a1p06r+a1p07r;	t0D=a1p06i+a1p07i;

	    rt =t0C;		it =t0D;
	    t0C=t08-rt;		t0D=t09-it;
	    t08=t08+rt;		t09=t09+it;

	    rt =t0E;		it =t0F;
	    t0E=t0A-it;		t0F=t0B+rt;
	    t0A=t0A+it;		t0B=t0B-rt;
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =t08;		it =t09;
	    t08=t00-rt;		t09=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t0C;		it =t0D;
	    t0C=t04-it;		t0D=t05+rt;
	    t04=t04+it;		t05=t05-rt;

	    rt =(t0A+t0B)*ISRT2;it =(t0A-t0B)*ISRT2;
	    t0A=t02-rt;		t0B=t03+it;
	    t02=t02+rt;		t03=t03-it;

	    rt =(t0E-t0F)*ISRT2;it =(t0F+t0E)*ISRT2;
	    t0E=t06+rt;		t0F=t07+it;
	    t06=t06-rt;		t07=t07-it;

/*...Block 2:	*/
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    t12=a1p08r-a1p09r;	t13=a1p08i-a1p09i;
	    t10=a1p08r+a1p09r;	t11=a1p08i+a1p09i;

	    t16=a1p0Ar-a1p0Br;	t17=a1p0Ai-a1p0Bi;
	    t14=a1p0Ar+a1p0Br;	t15=a1p0Ai+a1p0Bi;

	    rt =t14;		it =t15;
	    t14=t10-rt;		t15=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t16;		it =t17;
	    t16=t12-it;		t17=t13+rt;
	    t12=t12+it;		t13=t13-rt;

	    t1A=a1p0Cr-a1p0Dr;	t1B=a1p0Ci-a1p0Di;
	    t18=a1p0Cr+a1p0Dr;	t19=a1p0Ci+a1p0Di;

	    t1E=a1p0Er-a1p0Fr;	t1F=a1p0Ei-a1p0Fi;
	    t1C=a1p0Er+a1p0Fr;	t1D=a1p0Ei+a1p0Fi;

	    rt =t1C;		it =t1D;
	    t1C=t18-rt;		t1D=t19-it;
	    t18=t18+rt;		t19=t19+it;

	    rt =t1E;		it =t1F;
	    t1E=t1A-it;		t1F=t1B+rt;
	    t1A=t1A+it;		t1B=t1B-rt;
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =t18;		it =t19;
	    t18=t10-rt;		t19=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t1C;		it =t1D;
	    t1C=t14-it;		t1D=t15+rt;
	    t14=t14+it;		t15=t15-rt;

	    rt =(t1A+t1B)*ISRT2;it =(t1A-t1B)*ISRT2;
	    t1A=t12-rt;		t1B=t13+it;
	    t12=t12+rt;		t13=t13-it;

	    rt =(t1E-t1F)*ISRT2;it =(t1F+t1E)*ISRT2;
	    t1E=t16+rt;		t1F=t17+it;
	    t16=t16-rt;		t17=t17-it;

/*...Block 3:	*/
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    t22=a1p10r-a1p11r;	t23=a1p10i-a1p11i;
	    t20=a1p10r+a1p11r;	t21=a1p10i+a1p11i;

	    t26=a1p12r-a1p13r;	t27=a1p12i-a1p13i;
	    t24=a1p12r+a1p13r;	t25=a1p12i+a1p13i;

	    rt =t24;		it =t25;
	    t24=t20-rt;		t25=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t26;		it =t27;
	    t26=t22-it;		t27=t23+rt;
	    t22=t22+it;		t23=t23-rt;

	    t2A=a1p14r-a1p15r;	t2B=a1p14i-a1p15i;
	    t28=a1p14r+a1p15r;	t29=a1p14i+a1p15i;

	    t2E=a1p16r-a1p17r;	t2F=a1p16i-a1p17i;
	    t2C=a1p16r+a1p17r;	t2D=a1p16i+a1p17i;

	    rt =t2C;		it =t2D;
	    t2C=t28-rt;		t2D=t29-it;
	    t28=t28+rt;		t29=t29+it;

	    rt =t2E;		it =t2F;
	    t2E=t2A-it;		t2F=t2B+rt;
	    t2A=t2A+it;		t2B=t2B-rt;
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =t28;		it =t29;
	    t28=t20-rt;		t29=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t2C;		it =t2D;
	    t2C=t24-it;		t2D=t25+rt;
	    t24=t24+it;		t25=t25-rt;

	    rt =(t2A+t2B)*ISRT2;it =(t2A-t2B)*ISRT2;
	    t2A=t22-rt;		t2B=t23+it;
	    t22=t22+rt;		t23=t23-it;

	    rt =(t2E-t2F)*ISRT2;it =(t2F+t2E)*ISRT2;
	    t2E=t26+rt;		t2F=t27+it;
	    t26=t26-rt;		t27=t27-it;

/*...Block 4:	*/
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    t32=a1p18r-a1p19r;	t33=a1p18i-a1p19i;
	    t30=a1p18r+a1p19r;	t31=a1p18i+a1p19i;

	    t36=a1p1Ar-a1p1Br;	t37=a1p1Ai-a1p1Bi;
	    t34=a1p1Ar+a1p1Br;	t35=a1p1Ai+a1p1Bi;

	    rt =t34;		it =t35;
	    t34=t30-rt;		t35=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t36;		it =t37;
	    t36=t32-it;		t37=t33+rt;
	    t32=t32+it;		t33=t33-rt;

	    t3A=a1p1Cr-a1p1Dr;	t3B=a1p1Ci-a1p1Di;
	    t38=a1p1Cr+a1p1Dr;	t39=a1p1Ci+a1p1Di;

	    t3E=a1p1Er-a1p1Fr;	t3F=a1p1Ei-a1p1Fi;
	    t3C=a1p1Er+a1p1Fr;	t3D=a1p1Ei+a1p1Fi;

	    rt =t3C;		it =t3D;
	    t3C=t38-rt;		t3D=t39-it;
	    t38=t38+rt;		t39=t39+it;

	    rt =t3E;		it =t3F;
	    t3E=t3A-it;		t3F=t3B+rt;
	    t3A=t3A+it;		t3B=t3B-rt;
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =t38;		it =t39;
	    t38=t30-rt;		t39=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t3C;		it =t3D;
	    t3C=t34-it;		t3D=t35+rt;
	    t34=t34+it;		t35=t35-rt;

	    rt =(t3A+t3B)*ISRT2;it =(t3A-t3B)*ISRT2;
	    t3A=t32-rt;		t3B=t33+it;
	    t32=t32+rt;		t33=t33-it;

	    rt =(t3E-t3F)*ISRT2;it =(t3F+t3E)*ISRT2;
	    t3E=t36+rt;		t3F=t37+it;
	    t36=t36-rt;		t37=t37-it;

/*...and now do eight radix-4 transforms, including the internal twiddle factors:
	1, exp(-i* 1*twopi/32) =       ( c32_1,-s32_1), exp(-i* 2*twopi/32) =       ( c    ,-s    ), exp(-i* 3*twopi/32) =       ( c32_3,-s32_3) (for inputs to transform block 2),
	1, exp(-i* 2*twopi/32) =       ( c    ,-s    ), exp(-i* 4*twopi/32) = ISRT2*( 1    ,-1    ), exp(-i* 3*twopi/32) =       ( s    ,-c    ) (for inputs to transform block 3),
	1, exp(-i* 3*twopi/32) =       ( c32_3,-s32_3), exp(-i* 6*twopi/32) =       ( s    ,-c    ), exp(-i* 9*twopi/32) =       (-s32_1,-c32_1) (for inputs to transform block 4),
	1, exp(-i* 4*twopi/32) = ISRT2*( 1    ,-1    ), exp(-i* 8*twopi/32) =       ( 0    ,-1    ), exp(-i*12*twopi/32) = ISRT2*(-1    ,-1    ) (for inputs to transform block 5),
	1, exp(-i* 5*twopi/32) =       ( s32_3,-c32_3), exp(-i*10*twopi/32) =       (-s    ,-c    ), exp(-i*15*twopi/32) =       (-c32_1,-s32_1) (for inputs to transform block 6),
	1, exp(-i* 6*twopi/32) =       ( s    ,-c    ), exp(-i*12*twopi/32) = ISRT2*(-1    ,-1    ), exp(-i*18*twopi/32) =       (-c    , s    ) (for inputs to transform block 7),
	1, exp(-i* 7*twopi/32) =       ( s32_1,-c32_1), exp(-i*14*twopi/32) =       (-c    ,-s    ), exp(-i*21*twopi/32) =       (-s32_3, c32_3) (for inputs to transform block 8),
   and only the last 3 inputs to each of the radix-4 transforms 2 through 8 are multiplied by non-unity twiddles.	*/

/*...Block 1: t00,t10,t20,t30	*/
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    rt =t10;	t10=t00-rt;	t00=t00+rt;
	    it =t11;	t11=t01-it;	t01=t01+it;

	    rt =t30;	t30=t20-rt;	t20=t20+rt;
	    it =t31;	t31=t21-it;	t21=t21+it;

#if USE_NEW_DIT
	    a[idum   ]=t00+t20;		a[idum+1 ]=t01+t21;
	    t00       =t00-t20;		t01       =t01-t21;
	    a[idum+32]=t00*c10+t01*s10;	a[idum+33]=t01*c10-t00*s10;

	    rt        =t10+t31;		it        =t11-t30;
	    t10       =t10-t31;		t11       =t11+t30;
	    a[idum+16]=rt *c08+it *s08;	a[idum+17]=it *c08-rt *s08;
	    a[idum+48]=t10*c18+t11*s18;	a[idum+49]=t11*c18-t10*s18;
#else
	    a[idum   ]=t00+t20;		a[idum+1   ]=t01+t21;
	    a[idum+32]=t00-t20;		a[idum+1+32]=t01-t21;

	    a[idum+16]=t10+t31;		a[idum+1+16]=t11-t30;
	    a[idum+48]=t10-t31;		a[idum+1+48]=t11+t30;
#endif
/*...Block 5: t08,t18,t28,t38	*/
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =t18;	t18=t08-t19;	t08=t08+t19;
			t19=t09+rt;	t09=t09-rt;

	    rt =(t29+t28)*ISRT2;	t29=(t29-t28)*ISRT2;		t28=rt;
	    rt =(t38-t39)*ISRT2;	it =(t38+t39)*ISRT2;
	    t38=t28+rt;			t28=t28-rt;
	    t39=t29+it;			t29=t29-it;

#if USE_NEW_DIT
	    rt        =t08+t28;		it        =t09+t29;
	    t08       =t08-t28;		t09       =t09-t29;
	    a[idum+8 ]=rt *c04+it *s04;	a[idum+9 ]=it *c04-rt *s04;
	    a[idum+40]=t08*c14+t09*s14;	a[idum+41]=t09*c14-t08*s14;

	    rt        =t18+t39;		it        =t19-t38;
	    t18       =t18-t39;		t19       =t19+t38;
	    a[idum+24]=rt *c0C+it *s0C;	a[idum+25]=it *c0C-rt *s0C;
	    a[idum+56]=t18*c1C+t19*s1C;	a[idum+57]=t19*c1C-t18*s1C;
#else
	    a[idum+8 ]=t08+t28;		a[idum+1+8 ]=t09+t29;
	    a[idum+40]=t08-t28;		a[idum+1+40]=t09-t29;

	    a[idum+24]=t18+t39;		a[idum+1+24]=t19-t38;
	    a[idum+56]=t18-t39;		a[idum+1+56]=t19+t38;
#endif
/*...Block 3: t04,t14,t24,t34	*/
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    rt =(t15+t14)*ISRT2;	it =(t15-t14)*ISRT2;
	    t14=t04-rt;			t04=t04+rt;
	    t15=t05-it;			t05=t05+it;

	    rt =t24*c + t25*s;		t25=t25*c - t24*s;		t24=rt;
	    rt =t34*s + t35*c;		it =t35*s - t34*c;
	    t34=t24-rt;			t24=t24+rt;
	    t35=t25-it;			t25=t25+it;

#if USE_NEW_DIT
	    rt        =t04+t24;		it        =t05+t25;
	    t04       =t04-t24;		t05       =t05-t25;
	    a[idum+4 ]=rt *c02+it *s02;	a[idum+5 ]=it *c02-rt *s02;
	    a[idum+36]=t04*c12+t05*s12;	a[idum+37]=t05*c12-t04*s12;

	    rt        =t14+t35;		it        =t15-t34;
	    t14       =t14-t35;		t15       =t15+t34;
	    a[idum+20]=rt *c0A+it *s0A;	a[idum+21]=it *c0A-rt *s0A;
	    a[idum+52]=t14*c1A+t15*s1A;	a[idum+53]=t15*c1A-t14*s1A;
#else
	    a[idum+4 ]=t04+t24;		a[idum+1+4 ]=t05+t25;
	    a[idum+36]=t04-t24;		a[idum+1+36]=t05-t25;

	    a[idum+20]=t14+t35;		a[idum+1+20]=t15-t34;
	    a[idum+52]=t14-t35;		a[idum+1+52]=t15+t34;
#endif
/*...Block 7: t0C,t1C,t2C,t3C	*/
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =(t1C-t1D)*ISRT2;	it =(t1C+t1D)*ISRT2;
	    t1C=t0C+rt;			t0C=t0C-rt;
	    t1D=t0D+it;			t0D=t0D-it;

	    rt =t2C*s + t2D*c;		t2D=t2D*s - t2C*c;		t2C=rt;
	    rt =t3C*c + t3D*s;		it =t3D*c - t3C*s;
	    t3C=t2C+rt;			t2C=t2C-rt;
	    t3D=t2D+it;			t2D=t2D-it;

#if USE_NEW_DIT
	    rt        =t0C+t2C;		it        =t0D+t2D;
	    t0C       =t0C-t2C;		t0D       =t0D-t2D;
	    a[idum+12]=rt *c06+it *s06;	a[idum+13]=it *c06-rt *s06;
	    a[idum+44]=t0C*c16+t0D*s16;	a[idum+45]=t0D*c16-t0C*s16;

	    rt        =t1C+t3D;		it        =t1D-t3C;
	    t1C       =t1C-t3D;		t1D       =t1D+t3C;
	    a[idum+28]=rt *c0E+it *s0E;	a[idum+29]=it *c0E-rt *s0E;
	    a[idum+60]=t1C*c1E+t1D*s1E;	a[idum+61]=t1D*c1E-t1C*s1E;
#else
	    a[idum+12]=t0C+t2C;		a[idum+1+12]=t0D+t2D;
	    a[idum+44]=t0C-t2C;		a[idum+1+44]=t0D-t2D;

	    a[idum+28]=t1C+t3D;		a[idum+1+28]=t1D-t3C;
	    a[idum+60]=t1C-t3D;		a[idum+1+60]=t1D+t3C;
#endif
/*...Block 2: t02,t12,t22,t32	*/
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    rt =t12*c + t13*s;		it =t13*c - t12*s;
	    t12=t02-rt;			t02=t02+rt;
	    t13=t03-it;			t03=t03+it;

	    rt =t22*c32_1 + t23*s32_1;	t23=t23*c32_1 - t22*s32_1;	t22=rt;
	    rt =t32*c32_3 + t33*s32_3;	it =t33*c32_3 - t32*s32_3;
	    t32=t22-rt;			t22=t22+rt;
	    t33=t23-it;			t23=t23+it;

#if USE_NEW_DIT
	    rt        =t02+t22;		it        =t03+t23;
	    t02       =t02-t22;		t03       =t03-t23;
	    a[idum+2 ]=rt *c01+it *s01;	a[idum+3 ]=it *c01-rt *s01;
	    a[idum+34]=t02*c11+t03*s11;	a[idum+35]=t03*c11-t02*s11;

	    rt        =t12+t33;		it        =t13-t32;
	    t12       =t12-t33;		t13       =t13+t32;
	    a[idum+18]=rt *c09+it *s09;	a[idum+19]=it *c09-rt *s09;
	    a[idum+50]=t12*c19+t13*s19;	a[idum+51]=t13*c19-t12*s19;
#else
	    a[idum+2 ]=t02+t22;		a[idum+1+2 ]=t03+t23;
	    a[idum+34]=t02-t22;		a[idum+1+34]=t03-t23;

	    a[idum+18]=t12+t33;		a[idum+1+18]=t13-t32;
	    a[idum+50]=t12-t33;		a[idum+1+50]=t13+t32;
#endif
/*...Block 6: t0A,t1A,t2A,t3A	*/
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =t1A*s - t1B*c;		it =t1B*s + t1A*c;
	    t1A=t0A+rt;			t0A =t0A-rt;
	    t1B=t0B+it;			t0B =t0B-it;

	    rt =t2A*s32_3 + t2B*c32_3;	t2B=t2B*s32_3 - t2A*c32_3;	t2A=rt;
	    rt =t3A*c32_1 - t3B*s32_1;	it =t3B*c32_1 + t3A*s32_1;
	    t3A=t2A+rt;			t2A=t2A-rt;
	    t3B=t2B+it;			t2B=t2B-it;

#if USE_NEW_DIT
	    rt        =t0A+t2A;		it        =t0B+t2B;
	    t0A       =t0A-t2A;		t0B       =t0B-t2B;
	    a[idum+10]=rt *c05+it *s05;	a[idum+11]=it *c05-rt *s05;
	    a[idum+42]=t0A*c15+t0B*s15;	a[idum+43]=t0B*c15-t0A*s15;

	    rt        =t1A+t3B;		it        =t1B-t3A;
	    t1A       =t1A-t3B;		t1B       =t1B+t3A;
	    a[idum+26]=rt *c0D+it *s0D;	a[idum+27]=it *c0D-rt *s0D;
	    a[idum+58]=t1A*c1D+t1B*s1D;	a[idum+59]=t1B*c1D-t1A*s1D;
#else
	    a[idum+10]=t0A+t2A;		a[idum+1+10]=t0B+t2B;
	    a[idum+42]=t0A-t2A;		a[idum+1+42]=t0B-t2B;

	    a[idum+26]=t1A+t3B;		a[idum+1+26]=t1B-t3A;
	    a[idum+58]=t1A-t3B;		a[idum+1+58]=t1B+t3A;
#endif
/*...Block 4: t06,t16,t26,t36	*/
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    rt =t16*s + t17*c;		it =t17*s - t16*c;
	    t16=t06-rt;			t06 =t06+rt;
	    t17=t07-it;			t07 =t07+it;

	    rt =t26*c32_3 + t27*s32_3;	t27=t27*c32_3 - t26*s32_3;	t26=rt;
	    rt =t36*s32_1 - t37*c32_1;	it =t37*s32_1 + t36*c32_1;
	    t36=t26+rt;			t26=t26-rt;
	    t37=t27+it;			t27=t27-it;

#if USE_NEW_DIT
	    rt        =t06+t26;		it        =t07+t27;
	    t06       =t06-t26;		t07       =t07-t27;
	    a[idum+6 ]=rt *c03+it *s03;	a[idum+7 ]=it *c03-rt *s03;
	    a[idum+38]=t06*c13+t07*s13;	a[idum+39]=t07*c13-t06*s13;

	    rt        =t16+t37;		it        =t17-t36;
	    t16       =t16-t37;		t17       =t17+t36;
	    a[idum+22]=rt *c0B+it *s0B;	a[idum+23]=it *c0B-rt *s0B;
	    a[idum+54]=t16*c1B+t17*s1B;	a[idum+55]=t17*c1B-t16*s1B;
#else
	    a[idum+6 ]=t06+t26;		a[idum+1+6 ]=t07+t27;
	    a[idum+38]=t06-t26;		a[idum+1+38]=t07-t27;

	    a[idum+22]=t16+t37;		a[idum+1+22]=t17-t36;
	    a[idum+54]=t16-t37;		a[idum+1+54]=t17+t36;
#endif
/*...Block 8: t0E,t1E,t2E,t3E	*/
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =t1E*c - t1F*s;		it =t1F*c + t1E*s;
	    t1E=t0E+rt;			t0E =t0E-rt;
	    t1F=t0F+it;			t0F =t0F-it;

	    rt =t2E*s32_1 + t2F*c32_1;	t2F=t2F*s32_1 - t2E*c32_1;	t2E=rt;
	    rt =t3E*s32_3 + t3F*c32_3;	it =t3F*s32_3 - t3E*c32_3;
	    t3E=t2E+rt;			t2E=t2E-rt;
	    t3F=t2F+it;			t2F=t2F-it;

#if USE_NEW_DIT
	    rt        =t0E+t2E;		it        =t0F+t2F;
	    t0E       =t0E-t2E;		t0F       =t0F-t2F;
	    a[idum+14]=rt *c07+it *s07;	a[idum+15]=it *c07-rt *s07;
	    a[idum+46]=t0E*c17+t0F*s17;	a[idum+47]=t0F*c17-t0E*s17;

	    rt        =t1E+t3F;		it        =t1F-t3E;
	    t1E       =t1E-t3F;		t1F       =t1F+t3E;
	    a[idum+30]=rt *c0F+it *s0F;	a[idum+31]=it *c0F-rt *s0F;
	    a[idum+62]=t1E*c1F+t1F*s1F;	a[idum+63]=t1F*c1F-t1E*s1F;
#else
	    a[idum+14]=t0E+t2E;		a[idum+1+14]=t0F+t2F;
	    a[idum+46]=t0E-t2E;		a[idum+1+46]=t0F-t2F;

	    a[idum+30]=t1E+t3F;		a[idum+1+30]=t1F-t3E;
	    a[idum+62]=t1E-t3F;		a[idum+1+62]=t1F+t3E;
#endif


/* 2nd set of inputs: */
	  idum = j2pad;
#if PFETCH
add0 = &a[idum-64];
#endif
/*...Block 1: */
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
/*...Block 1:	*/
	    t02=a2p00r-a2p01r;	t03=a2p00i-a2p01i;
	    t00=a2p00r+a2p01r;	t01=a2p00i+a2p01i;

	    t06=a2p02r-a2p03r;	t07=a2p02i-a2p03i;
	    t04=a2p02r+a2p03r;	t05=a2p02i+a2p03i;

	    rt =t04;		it =t05;
	    t04=t00-rt;		t05=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t06;		it =t07;
	    t06=t02-it;		t07=t03+rt;
	    t02=t02+it;		t03=t03-rt;

	    t0A=a2p04r-a2p05r;	t0B=a2p04i-a2p05i;
	    t08=a2p04r+a2p05r;	t09=a2p04i+a2p05i;

	    t0E=a2p06r-a2p07r;	t0F=a2p06i-a2p07i;
	    t0C=a2p06r+a2p07r;	t0D=a2p06i+a2p07i;

	    rt =t0C;		it =t0D;
	    t0C=t08-rt;		t0D=t09-it;
	    t08=t08+rt;		t09=t09+it;

	    rt =t0E;		it =t0F;
	    t0E=t0A-it;		t0F=t0B+rt;
	    t0A=t0A+it;		t0B=t0B-rt;
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =t08;		it =t09;
	    t08=t00-rt;		t09=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t0C;		it =t0D;
	    t0C=t04-it;		t0D=t05+rt;
	    t04=t04+it;		t05=t05-rt;

	    rt =(t0A+t0B)*ISRT2;it =(t0A-t0B)*ISRT2;
	    t0A=t02-rt;		t0B=t03+it;
	    t02=t02+rt;		t03=t03-it;

	    rt =(t0E-t0F)*ISRT2;it =(t0F+t0E)*ISRT2;
	    t0E=t06+rt;		t0F=t07+it;
	    t06=t06-rt;		t07=t07-it;

/*...Block 2:	*/
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    t12=a2p08r-a2p09r;	t13=a2p08i-a2p09i;
	    t10=a2p08r+a2p09r;	t11=a2p08i+a2p09i;

	    t16=a2p0Ar-a2p0Br;	t17=a2p0Ai-a2p0Bi;
	    t14=a2p0Ar+a2p0Br;	t15=a2p0Ai+a2p0Bi;

	    rt =t14;		it =t15;
	    t14=t10-rt;		t15=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t16;		it =t17;
	    t16=t12-it;		t17=t13+rt;
	    t12=t12+it;		t13=t13-rt;

	    t1A=a2p0Cr-a2p0Dr;	t1B=a2p0Ci-a2p0Di;
	    t18=a2p0Cr+a2p0Dr;	t19=a2p0Ci+a2p0Di;

	    t1E=a2p0Er-a2p0Fr;	t1F=a2p0Ei-a2p0Fi;
	    t1C=a2p0Er+a2p0Fr;	t1D=a2p0Ei+a2p0Fi;

	    rt =t1C;		it =t1D;
	    t1C=t18-rt;		t1D=t19-it;
	    t18=t18+rt;		t19=t19+it;

	    rt =t1E;		it =t1F;
	    t1E=t1A-it;		t1F=t1B+rt;
	    t1A=t1A+it;		t1B=t1B-rt;
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =t18;		it =t19;
	    t18=t10-rt;		t19=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t1C;		it =t1D;
	    t1C=t14-it;		t1D=t15+rt;
	    t14=t14+it;		t15=t15-rt;

	    rt =(t1A+t1B)*ISRT2;it =(t1A-t1B)*ISRT2;
	    t1A=t12-rt;		t1B=t13+it;
	    t12=t12+rt;		t13=t13-it;

	    rt =(t1E-t1F)*ISRT2;it =(t1F+t1E)*ISRT2;
	    t1E=t16+rt;		t1F=t17+it;
	    t16=t16-rt;		t17=t17-it;

/*...Block 3:	*/
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    t22=a2p10r-a2p11r;	t23=a2p10i-a2p11i;
	    t20=a2p10r+a2p11r;	t21=a2p10i+a2p11i;

	    t26=a2p12r-a2p13r;	t27=a2p12i-a2p13i;
	    t24=a2p12r+a2p13r;	t25=a2p12i+a2p13i;

	    rt =t24;		it =t25;
	    t24=t20-rt;		t25=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t26;		it =t27;
	    t26=t22-it;		t27=t23+rt;
	    t22=t22+it;		t23=t23-rt;

	    t2A=a2p14r-a2p15r;	t2B=a2p14i-a2p15i;
	    t28=a2p14r+a2p15r;	t29=a2p14i+a2p15i;

	    t2E=a2p16r-a2p17r;	t2F=a2p16i-a2p17i;
	    t2C=a2p16r+a2p17r;	t2D=a2p16i+a2p17i;

	    rt =t2C;		it =t2D;
	    t2C=t28-rt;		t2D=t29-it;
	    t28=t28+rt;		t29=t29+it;

	    rt =t2E;		it =t2F;
	    t2E=t2A-it;		t2F=t2B+rt;
	    t2A=t2A+it;		t2B=t2B-rt;
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =t28;		it =t29;
	    t28=t20-rt;		t29=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t2C;		it =t2D;
	    t2C=t24-it;		t2D=t25+rt;
	    t24=t24+it;		t25=t25-rt;

	    rt =(t2A+t2B)*ISRT2;it =(t2A-t2B)*ISRT2;
	    t2A=t22-rt;		t2B=t23+it;
	    t22=t22+rt;		t23=t23-it;

	    rt =(t2E-t2F)*ISRT2;it =(t2F+t2E)*ISRT2;
	    t2E=t26+rt;		t2F=t27+it;
	    t26=t26-rt;		t27=t27-it;

/*...Block 4:	*/
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    t32=a2p18r-a2p19r;	t33=a2p18i-a2p19i;
	    t30=a2p18r+a2p19r;	t31=a2p18i+a2p19i;

	    t36=a2p1Ar-a2p1Br;	t37=a2p1Ai-a2p1Bi;
	    t34=a2p1Ar+a2p1Br;	t35=a2p1Ai+a2p1Bi;

	    rt =t34;		it =t35;
	    t34=t30-rt;		t35=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t36;		it =t37;
	    t36=t32-it;		t37=t33+rt;
	    t32=t32+it;		t33=t33-rt;

	    t3A=a2p1Cr-a2p1Dr;	t3B=a2p1Ci-a2p1Di;
	    t38=a2p1Cr+a2p1Dr;	t39=a2p1Ci+a2p1Di;

	    t3E=a2p1Er-a2p1Fr;	t3F=a2p1Ei-a2p1Fi;
	    t3C=a2p1Er+a2p1Fr;	t3D=a2p1Ei+a2p1Fi;

	    rt =t3C;		it =t3D;
	    t3C=t38-rt;		t3D=t39-it;
	    t38=t38+rt;		t39=t39+it;

	    rt =t3E;		it =t3F;
	    t3E=t3A-it;		t3F=t3B+rt;
	    t3A=t3A+it;		t3B=t3B-rt;
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =t38;		it =t39;
	    t38=t30-rt;		t39=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t3C;		it =t3D;
	    t3C=t34-it;		t3D=t35+rt;
	    t34=t34+it;		t35=t35-rt;

	    rt =(t3A+t3B)*ISRT2;it =(t3A-t3B)*ISRT2;
	    t3A=t32-rt;		t3B=t33+it;
	    t32=t32+rt;		t33=t33-it;

	    rt =(t3E-t3F)*ISRT2;it =(t3F+t3E)*ISRT2;
	    t3E=t36+rt;		t3F=t37+it;
	    t36=t36-rt;		t37=t37-it;

/*
!...and now do eight radix-4 transforms, including the internal twiddle factors:
*/
/*...Block 1: t00,t10,t20,t30	*/
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    rt =t10;	t10=t00-rt;	t00=t00+rt;
	    it =t11;	t11=t01-it;	t01=t01+it;

	    rt =t30;	t30=t20-rt;	t20=t20+rt;
	    it =t31;	t31=t21-it;	t21=t21+it;

#if USE_NEW_DIT
	    a[idum   ]=t00+t20;		a[idum+1 ]=t01+t21;
	    t00       =t00-t20;		t01       =t01-t21;
	    a[idum+32]=t00*c30+t01*s30;	a[idum+33]=t01*c30-t00*s30;

	    rt        =t10+t31;		it        =t11-t30;
	    t10       =t10-t31;		t11       =t11+t30;
	    a[idum+16]=rt *c28+it *s28;	a[idum+17]=it *c28-rt *s28;
	    a[idum+48]=t10*c38+t11*s38;	a[idum+49]=t11*c38-t10*s38;
#else
	    a[idum   ]=t00+t20;		a[idum+1   ]=t01+t21;
	    a[idum+32]=t00-t20;		a[idum+1+32]=t01-t21;

	    a[idum+16]=t10+t31;		a[idum+1+16]=t11-t30;
	    a[idum+48]=t10-t31;		a[idum+1+48]=t11+t30;
#endif
/*...Block 5: t08,t18,t28,t38	*/
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =t18;	t18=t08-t19;	t08=t08+t19;
			t19=t09+rt;	t09=t09-rt;

	    rt =(t29+t28)*ISRT2;	t29=(t29-t28)*ISRT2;		t28=rt;
	    rt =(t38-t39)*ISRT2;	it =(t38+t39)*ISRT2;
	    t38=t28+rt;			t28=t28-rt;
	    t39=t29+it;			t29=t29-it;

#if USE_NEW_DIT
	    rt        =t08+t28;		it        =t09+t29;
	    t08       =t08-t28;		t09       =t09-t29;
	    a[idum+8 ]=rt *c24+it *s24;	a[idum+9 ]=it *c24-rt *s24;
	    a[idum+40]=t08*c34+t09*s34;	a[idum+41]=t09*c34-t08*s34;

	    rt        =t18+t39;		it        =t19-t38;
	    t18       =t18-t39;		t19       =t19+t38;
	    a[idum+24]=rt *c2C+it *s2C;	a[idum+25]=it *c2C-rt *s2C;
	    a[idum+56]=t18*c3C+t19*s3C;	a[idum+57]=t19*c3C-t18*s3C;
#else
	    a[idum+8 ]=t08+t28;		a[idum+1+8 ]=t09+t29;
	    a[idum+40]=t08-t28;		a[idum+1+40]=t09-t29;

	    a[idum+24]=t18+t39;		a[idum+1+24]=t19-t38;
	    a[idum+56]=t18-t39;		a[idum+1+56]=t19+t38;
#endif
/*...Block 3: t04,t14,t24,t34	*/
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    rt =(t15+t14)*ISRT2;	it =(t15-t14)*ISRT2;
	    t14=t04-rt;			t04=t04+rt;
	    t15=t05-it;			t05=t05+it;

	    rt =t24*c + t25*s;		t25=t25*c - t24*s;		t24=rt;
	    rt =t34*s + t35*c;		it =t35*s - t34*c;
	    t34=t24-rt;			t24=t24+rt;
	    t35=t25-it;			t25=t25+it;

#if USE_NEW_DIT
	    rt        =t04+t24;		it        =t05+t25;
	    t04       =t04-t24;		t05       =t05-t25;
	    a[idum+4 ]=rt *c22+it *s22;	a[idum+5 ]=it *c22-rt *s22;
	    a[idum+36]=t04*c32+t05*s32;	a[idum+37]=t05*c32-t04*s32;

	    rt        =t14+t35;		it        =t15-t34;
	    t14       =t14-t35;		t15       =t15+t34;
	    a[idum+20]=rt *c2A+it *s2A;	a[idum+21]=it *c2A-rt *s2A;
	    a[idum+52]=t14*c3A+t15*s3A;	a[idum+53]=t15*c3A-t14*s3A;
#else
	    a[idum+4 ]=t04+t24;		a[idum+1+4 ]=t05+t25;
	    a[idum+36]=t04-t24;		a[idum+1+36]=t05-t25;

	    a[idum+20]=t14+t35;		a[idum+1+20]=t15-t34;
	    a[idum+52]=t14-t35;		a[idum+1+52]=t15+t34;
#endif
/*...Block 7: t0C,t1C,t2C,t3C	*/
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =(t1C-t1D)*ISRT2;	it =(t1C+t1D)*ISRT2;
	    t1C=t0C+rt;			t0C=t0C-rt;
	    t1D=t0D+it;			t0D=t0D-it;

	    rt =t2C*s + t2D*c;		t2D=t2D*s - t2C*c;		t2C=rt;
	    rt =t3C*c + t3D*s;		it =t3D*c - t3C*s;
	    t3C=t2C+rt;			t2C=t2C-rt;
	    t3D=t2D+it;			t2D=t2D-it;

#if USE_NEW_DIT
	    rt        =t0C+t2C;		it        =t0D+t2D;
	    t0C       =t0C-t2C;		t0D       =t0D-t2D;
	    a[idum+12]=rt *c26+it *s26;	a[idum+13]=it *c26-rt *s26;
	    a[idum+44]=t0C*c36+t0D*s36;	a[idum+45]=t0D*c36-t0C*s36;

	    rt        =t1C+t3D;		it        =t1D-t3C;
	    t1C       =t1C-t3D;		t1D       =t1D+t3C;
	    a[idum+28]=rt *c2E+it *s2E;	a[idum+29]=it *c2E-rt *s2E;
	    a[idum+60]=t1C*c3E+t1D*s3E;	a[idum+61]=t1D*c3E-t1C*s3E;
#else
	    a[idum+12]=t0C+t2C;		a[idum+1+12]=t0D+t2D;
	    a[idum+44]=t0C-t2C;		a[idum+1+44]=t0D-t2D;

	    a[idum+28]=t1C+t3D;		a[idum+1+28]=t1D-t3C;
	    a[idum+60]=t1C-t3D;		a[idum+1+60]=t1D+t3C;
#endif
/*...Block 2: t02,t12,t22,t32	*/
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    rt =t12*c + t13*s;		it =t13*c - t12*s;
	    t12=t02-rt;			t02=t02+rt;
	    t13=t03-it;			t03=t03+it;

	    rt =t22*c32_1 + t23*s32_1;	t23=t23*c32_1 - t22*s32_1;	t22=rt;
	    rt =t32*c32_3 + t33*s32_3;	it =t33*c32_3 - t32*s32_3;
	    t32=t22-rt;			t22=t22+rt;
	    t33=t23-it;			t23=t23+it;

#if USE_NEW_DIT
	    rt        =t02+t22;		it        =t03+t23;
	    t02       =t02-t22;		t03       =t03-t23;
	    a[idum+2 ]=rt *c21+it *s21;	a[idum+3 ]=it *c21-rt *s21;
	    a[idum+34]=t02*c31+t03*s31;	a[idum+35]=t03*c31-t02*s31;

	    rt        =t12+t33;		it        =t13-t32;
	    t12       =t12-t33;		t13       =t13+t32;
	    a[idum+18]=rt *c29+it *s29;	a[idum+19]=it *c29-rt *s29;
	    a[idum+50]=t12*c39+t13*s39;	a[idum+51]=t13*c39-t12*s39;
#else
	    a[idum+2 ]=t02+t22;		a[idum+1+2 ]=t03+t23;
	    a[idum+34]=t02-t22;		a[idum+1+34]=t03-t23;

	    a[idum+18]=t12+t33;		a[idum+1+18]=t13-t32;
	    a[idum+50]=t12-t33;		a[idum+1+50]=t13+t32;
#endif
/*...Block 6: t0A,t1A,t2A,t3A	*/
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =t1A*s - t1B*c;		it =t1B*s + t1A*c;
	    t1A=t0A+rt;			t0A =t0A-rt;
	    t1B=t0B+it;			t0B =t0B-it;

	    rt =t2A*s32_3 + t2B*c32_3;	t2B=t2B*s32_3 - t2A*c32_3;	t2A=rt;
	    rt =t3A*c32_1 - t3B*s32_1;	it =t3B*c32_1 + t3A*s32_1;
	    t3A=t2A+rt;			t2A=t2A-rt;
	    t3B=t2B+it;			t2B=t2B-it;

#if USE_NEW_DIT
	    rt        =t0A+t2A;		it        =t0B+t2B;
	    t0A       =t0A-t2A;		t0B       =t0B-t2B;
	    a[idum+10]=rt *c25+it *s25;	a[idum+11]=it *c25-rt *s25;
	    a[idum+42]=t0A*c35+t0B*s35;	a[idum+43]=t0B*c35-t0A*s35;

	    rt        =t1A+t3B;		it        =t1B-t3A;
	    t1A       =t1A-t3B;		t1B       =t1B+t3A;
	    a[idum+26]=rt *c2D+it *s2D;	a[idum+27]=it *c2D-rt *s2D;
	    a[idum+58]=t1A*c3D+t1B*s3D;	a[idum+59]=t1B*c3D-t1A*s3D;
#else
	    a[idum+10]=t0A+t2A;		a[idum+1+10]=t0B+t2B;
	    a[idum+42]=t0A-t2A;		a[idum+1+42]=t0B-t2B;

	    a[idum+26]=t1A+t3B;		a[idum+1+26]=t1B-t3A;
	    a[idum+58]=t1A-t3B;		a[idum+1+58]=t1B+t3A;
#endif
/*...Block 4: t06,t16,t26,t36	*/
#if PFETCH
prefetch_p_doubles(add0);
add0 += 4;
#endif
	    rt =t16*s + t17*c;		it =t17*s - t16*c;
	    t16=t06-rt;			t06 =t06+rt;
	    t17=t07-it;			t07 =t07+it;

	    rt =t26*c32_3 + t27*s32_3;	t27=t27*c32_3 - t26*s32_3;	t26=rt;
	    rt =t36*s32_1 - t37*c32_1;	it =t37*s32_1 + t36*c32_1;
	    t36=t26+rt;			t26=t26-rt;
	    t37=t27+it;			t27=t27-it;

#if USE_NEW_DIT
	    rt        =t06+t26;		it        =t07+t27;
	    t06       =t06-t26;		t07       =t07-t27;
	    a[idum+6 ]=rt *c23+it *s23;	a[idum+7 ]=it *c23-rt *s23;
	    a[idum+38]=t06*c33+t07*s33;	a[idum+39]=t07*c33-t06*s33;

	    rt        =t16+t37;		it        =t17-t36;
	    t16       =t16-t37;		t17       =t17+t36;
	    a[idum+22]=rt *c2B+it *s2B;	a[idum+23]=it *c2B-rt *s2B;
	    a[idum+54]=t16*c3B+t17*s3B;	a[idum+55]=t17*c3B-t16*s3B;
#else
	    a[idum+6 ]=t06+t26;		a[idum+1+6 ]=t07+t27;
	    a[idum+38]=t06-t26;		a[idum+1+38]=t07-t27;

	    a[idum+22]=t16+t37;		a[idum+1+22]=t17-t36;
	    a[idum+54]=t16-t37;		a[idum+1+54]=t17+t36;
#endif
/*...Block 8: t0E,t1E,t2E,t3E	*/
#if PFETCH
	#if(CACHE_LINE_DOUBLES == 4)
	prefetch_p_doubles(add0);
	#endif
add0 += 4;
#endif
	    rt =t1E*c - t1F*s;		it =t1F*c + t1E*s;
	    t1E=t0E+rt;			t0E =t0E-rt;
	    t1F=t0F+it;			t0F =t0F-it;

	    rt =t2E*s32_1 + t2F*c32_1;	t2F=t2F*s32_1 - t2E*c32_1;	t2E=rt;
	    rt =t3E*s32_3 + t3F*c32_3;	it =t3F*s32_3 - t3E*c32_3;
	    t3E=t2E+rt;			t2E=t2E-rt;
	    t3F=t2F+it;			t2F=t2F-it;

#if USE_NEW_DIT
	    rt        =t0E+t2E;		it        =t0F+t2F;
	    t0E       =t0E-t2E;		t0F       =t0F-t2F;
	    a[idum+14]=rt *c27+it *s27;	a[idum+15]=it *c27-rt *s27;
	    a[idum+46]=t0E*c37+t0F*s37;	a[idum+47]=t0F*c37-t0E*s37;

	    rt        =t1E+t3F;		it        =t1F-t3E;
	    t1E       =t1E-t3F;		t1F       =t1F+t3E;
	    a[idum+30]=rt *c2F+it *s2F;	a[idum+31]=it *c2F-rt *s2F;
	    a[idum+62]=t1E*c3F+t1F*s3F;	a[idum+63]=t1F*c3F-t1E*s3F;
#else
	    a[idum+14]=t0E+t2E;		a[idum+1+14]=t0F+t2F;
	    a[idum+46]=t0E-t2E;		a[idum+1+46]=t0F-t2F;

	    a[idum+30]=t1E+t3F;		a[idum+1+30]=t1F-t3E;
	    a[idum+62]=t1E-t3F;		a[idum+1+62]=t1F+t3E;
#endif


/*...And update the data (j1 and j2) array indices.	*/

	  j1 = j1+64;
	  j2 = j2-64;

	}

/*
!...Since the foregoing loop only gets executed half as many times as in the simple version, to properly position
!   ourselves in the data array for the start of the next block, need to bump up j1 by as much as would occur in a
!   second execution of the above loop. The exception is the first loop execution, where j1 needs to be doubled (64 x 2).
*/

	j1 = j1+(blocklen << 1);

	if(j2_start == n-64)
	{
		j1 = 0;
		return;
	}

/*...Reset half-complex-blocklength for next pass. If K >> 1 has a zero trailing bit, we multiply the blocklength by K >> 1 in preparation for the final block.	*/

	blocklen_sum = blocklen_sum + blocklen;
	blocklen = (radix_prim[i-1]-1)*blocklen_sum;

/*...Next j2_start is previous one plus the (real) length of the current block = 4*(half-complex-blocklength) */

	j2_start = j2_start+(blocklen<<2);
	j2=j2_start;				/* Reset j2 for start of the next block. */

      }		/* End of Main loop */

}
