/*******************************************************************************
*                                                                              *
*   (C) 1997-2004 by Ernst W. Mayer.                                           *
*                                                                              *
*  This program is free software; you can redistribute it and/or modify it     *
*  under the terms of the GNU General Public License as published by the       *
*  Free Software Foundation; either version 2 of the License, or (at your      *
*  option) any later version.                                                  *
*                                                                              *
*  This program is distributed in the hope that it will be useful, but WITHOUT *
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
*  more details.                                                               *
*                                                                              *
*  You should have received a copy of the GNU General Public License along     *
*  with this program; see the file GPL.txt.  If not, you may view one at       *
*  http://www.fsf.org/licenses/licenses.html, or obtain one by writing to the  *
*  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     *
*  02111-1307, USA.                                                            *
*                                                                              *
*******************************************************************************/

#include "Mlucas.h"

/***************/

void radix32_dif_pass1(double a[], int n)
{
/*
!...Acronym: DIF = Decimation In Frequency
!
!...Subroutine to perform an initial radix-32 complex DIF FFT pass on the data in the length-N real vector A.
!
!   The data are stored in a 1-D zero-offset array, with 2^PAD_BITS 8-byte padding elements inserted
!   between every block of 2^DAT_BITS contiguous data. The array padding is to prevent data being accessed
!   in strides that are large powers of two and thus to minimize cache thrashing
!   (in cache-based microprocessor architectures) or bank conflicts (in supercomputers.)
!
!   See the documentation in radix16_dif_pass for further details on storage and indexing.
*/
	int j,j1,jt;
	static int n32,p01,p02,p03,p04,p08,p0C,p10,p14,p18,p1C, first_entry=TRUE;
	static double    c     = 0.92387953251128675613, s     = 0.38268343236508977173	/* exp(  i*twopi/16)	*/
			,c32_1 = 0.98078528040323044912, s32_1 = 0.19509032201612826784	/* exp(  i*twopi/32)	*/
			,c32_3 = 0.83146961230254523708, s32_3 = 0.55557023301960222473;/* exp(3*i*twopi/32)	*/
	double rt,it
		,t00,t01,t02,t03,t04,t05,t06,t07,t08,t09,t0A,t0B,t0C,t0D,t0E,t0F
		,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,t1A,t1B,t1C,t1D,t1E,t1F
		,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,t2A,t2B,t2C,t2D,t2E,t2F
		,t30,t31,t32,t33,t34,t35,t36,t37,t38,t39,t3A,t3B,t3C,t3D,t3E,t3F;

	if(!first_entry && (n >> 5) != n32)	/* New runlength?	*/
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  first_entry=FALSE;
	  n32=n/32;

	  p01 = n32;
	  p02 = p01 +p01;
	  p03 = p02 +p01;
	  p04 = p03 +p01;
	  p08 = p04 +p04;
	  p0C = p08 +p04;
	  p10 = p0C +p04;
	  p14 = p10 +p04;
	  p18 = p14 +p04;
	  p1C = p18 +p04;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p0C = p0C + ( (p0C >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p14 = p14 + ( (p14 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	  p1C = p1C + ( (p1C >> DAT_BITS) << PAD_BITS );
	}

/*...The radix-32 pass is here.	*/

      for(j=0; j < n32; j += 2)
      {
	  j1 = j  + ( (j  >> DAT_BITS) << PAD_BITS );

/*       gather the needed data (32 64-bit complex, i.e. 64 64-bit reals) and do the first set of four length-8 transforms.	*/

/*...Block 1:	*/

	jt = j1;
	    t00=a[jt    ];	t01=a[jt+1    ];
	    rt =a[jt+p10];	it =a[jt+1+p10];
	    t02=t00-rt;		t03=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    t04=a[jt+p08];	t05=a[jt+1+p08];
	    rt =a[jt+p18];	it =a[jt+1+p18];
	    t06=t04-rt;		t07=t05-it;
	    t04=t04+rt;		t05=t05+it;

	    rt =t04;		it =t05;
	    t04=t00-rt;		t05=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t06;		it =t07;
	    t06=t02+it;		t07=t03-rt;
	    t02=t02-it;		t03=t03+rt;

	    t08=a[jt+p04];	t09=a[jt+1+p04];
	    rt =a[jt+p14];	it =a[jt+1+p14];
	    t0A=t08-rt;		t0B=t09-it;
	    t08=t08+rt;		t09=t09+it;

	    t0C=a[jt+p0C];	t0D=a[jt+1+p0C];
	    rt =a[jt+p1C];	it =a[jt+1+p1C];
	    t0E=t0C-rt;		t0F=t0D-it;
	    t0C=t0C+rt;		t0D=t0D+it;

	    rt =t0C;		it =t0D;
	    t0C=t08-rt;		t0D=t09-it;
	    t08=t08+rt;		t09=t09+it;

	    rt =t0E;		it =t0F;
	    t0E=t0A+it;		t0F=t0B-rt;
	    t0A=t0A-it;		t0B=t0B+rt;

	    rt =t08;		it =t09;
	    t08=t00-rt;		t09=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t0C;		it =t0D;
	    t0C=t04+it;		t0D=t05-rt;
	    t04=t04-it;		t05=t05+rt;

	    rt =(t0A-t0B)*ISRT2;it =(t0A+t0B)*ISRT2;
	    t0A=t02-rt;		t0B=t03-it;
	    t02=t02+rt;		t03=t03+it;

	    rt =(t0E+t0F)*ISRT2;it =(t0F-t0E)*ISRT2;
	    t0E=t06+rt;		t0F=t07+it;
	    t06=t06-rt;		t07=t07-it;

/*...Block 2:	*/

	jt = j1 + p02;
	    t10=a[jt    ];	t11=a[jt+1    ];
	    rt =a[jt+p10];	it =a[jt+1+p10];
	    t12=t10-rt;		t13=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    t14=a[jt+p08];	t15=a[jt+1+p08];
	    rt =a[jt+p18];	it =a[jt+1+p18];
	    t16=t14-rt;		t17=t15-it;
	    t14=t14+rt;		t15=t15+it;

	    rt =t14;		it =t15;
	    t14=t10-rt;		t15=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t16;		it =t17;
	    t16=t12+it;		t17=t13-rt;
	    t12=t12-it;		t13=t13+rt;

	    t18=a[jt+p04];	t19=a[jt+1+p04];
	    rt =a[jt+p14];	it =a[jt+1+p14];
	    t1A=t18-rt;		t1B=t19-it;
	    t18=t18+rt;		t19=t19+it;

	    t1C=a[jt+p0C];	t1D=a[jt+1+p0C];
	    rt =a[jt+p1C];	it =a[jt+1+p1C];
	    t1E=t1C-rt;		t1F=t1D-it;
	    t1C=t1C+rt;		t1D=t1D+it;

	    rt =t1C;		it =t1D;
	    t1C=t18-rt;		t1D=t19-it;
	    t18=t18+rt;		t19=t19+it;

	    rt =t1E;		it =t1F;
	    t1E=t1A+it;		t1F=t1B-rt;
	    t1A=t1A-it;		t1B=t1B+rt;

	    rt =t18;		it =t19;
	    t18=t10-rt;		t19=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t1C;		it =t1D;
	    t1C=t14+it;		t1D=t15-rt;
	    t14=t14-it;		t15=t15+rt;

	    rt =(t1A-t1B)*ISRT2;it =(t1A+t1B)*ISRT2;
	    t1A=t12-rt;		t1B=t13-it;
	    t12=t12+rt;		t13=t13+it;

	    rt =(t1E+t1F)*ISRT2;it =(t1F-t1E)*ISRT2;
	    t1E=t16+rt;		t1F=t17+it;
	    t16=t16-rt;		t17=t17-it;

/*...Block 3:	*/

	jt = j1 + p01;
	    t20=a[jt    ];	t21=a[jt+1    ];
	    rt =a[jt+p10];	it =a[jt+1+p10];
	    t22=t20-rt;		t23=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    t24=a[jt+p08];	t25=a[jt+1+p08];
	    rt =a[jt+p18];	it =a[jt+1+p18];
	    t26=t24-rt;		t27=t25-it;
	    t24=t24+rt;		t25=t25+it;

	    rt =t24;		it =t25;
	    t24=t20-rt;		t25=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t26;		it =t27;
	    t26=t22+it;		t27=t23-rt;
	    t22=t22-it;		t23=t23+rt;

	    t28=a[jt+p04];	t29=a[jt+1+p04];
	    rt =a[jt+p14];	it =a[jt+1+p14];
	    t2A=t28-rt;		t2B=t29-it;
	    t28=t28+rt;		t29=t29+it;

	    t2C=a[jt+p0C];	t2D=a[jt+1+p0C];
	    rt =a[jt+p1C];	it =a[jt+1+p1C];
	    t2E=t2C-rt;		t2F=t2D-it;
	    t2C=t2C+rt;		t2D=t2D+it;

	    rt =t2C;		it =t2D;
	    t2C=t28-rt;		t2D=t29-it;
	    t28=t28+rt;		t29=t29+it;

	    rt =t2E;		it =t2F;
	    t2E=t2A+it;		t2F=t2B-rt;
	    t2A=t2A-it;		t2B=t2B+rt;

	    rt =t28;		it =t29;
	    t28=t20-rt;		t29=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t2C;		it =t2D;
	    t2C=t24+it;		t2D=t25-rt;
	    t24=t24-it;		t25=t25+rt;

	    rt =(t2A-t2B)*ISRT2;it =(t2A+t2B)*ISRT2;
	    t2A=t22-rt;		t2B=t23-it;
	    t22=t22+rt;		t23=t23+it;

	    rt =(t2E+t2F)*ISRT2;it =(t2F-t2E)*ISRT2;
	    t2E=t26+rt;		t2F=t27+it;
	    t26=t26-rt;		t27=t27-it;

/*...Block 4:	*/

	jt = j1 + p03;
	    t30=a[jt    ];	t31=a[jt+1    ];
	    rt =a[jt+p10];	it =a[jt+1+p10];
	    t32=t30-rt;		t33=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    t34=a[jt+p08];	t35=a[jt+1+p08];
	    rt =a[jt+p18];	it =a[jt+1+p18];
	    t36=t34-rt;		t37=t35-it;
	    t34=t34+rt;		t35=t35+it;

	    rt =t34;		it =t35;
	    t34=t30-rt;		t35=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t36;		it =t37;
	    t36=t32+it;		t37=t33-rt;
	    t32=t32-it;		t33=t33+rt;

	    t38=a[jt+p04];	t39=a[jt+1+p04];
	    rt =a[jt+p14];	it =a[jt+1+p14];
	    t3A=t38-rt;		t3B=t39-it;
	    t38=t38+rt;		t39=t39+it;

	    t3C=a[jt+p0C];	t3D=a[jt+1+p0C];
	    rt =a[jt+p1C];	it =a[jt+1+p1C];
	    t3E=t3C-rt;		t3F=t3D-it;
	    t3C=t3C+rt;		t3D=t3D+it;

	    rt =t3C;		it =t3D;
	    t3C=t38-rt;		t3D=t39-it;
	    t38=t38+rt;		t39=t39+it;

	    rt =t3E;		it =t3F;
	    t3E=t3A+it;		t3F=t3B-rt;
	    t3A=t3A-it;		t3B=t3B+rt;

	    rt =t38;		it =t39;
	    t38=t30-rt;		t39=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t3C;		it =t3D;
	    t3C=t34+it;		t3D=t35-rt;
	    t34=t34-it;		t35=t35+rt;

	    rt =(t3A-t3B)*ISRT2;it =(t3A+t3B)*ISRT2;
	    t3A=t32-rt;		t3B=t33-it;
	    t32=t32+rt;		t33=t33+it;

	    rt =(t3E+t3F)*ISRT2;it =(t3F-t3E)*ISRT2;
	    t3E=t36+rt;		t3F=t37+it;
	    t36=t36-rt;		t37=t37-it;

/*...and now do eight radix-4 transforms, including the internal twiddle factors:
  	1, exp(i* 1*twopi/32) =       ( c32_1, s32_1), exp(i* 2*twopi/32) =       ( c    , s    ), exp(i* 3*twopi/32) =       ( c32_3, s32_3) (for inputs to transform block 2),
  	1, exp(i* 2*twopi/32) =       ( c    , s    ), exp(i* 4*twopi/32) = ISRT2*( 1    , 1    ), exp(i* 3*twopi/32) =       ( s    , c    ) (for inputs to transform block 3),
  	1, exp(i* 3*twopi/32) =       ( c32_3, s32_3), exp(i* 6*twopi/32) =       ( s    , c    ), exp(i* 9*twopi/32) =       (-s32_1, c32_1) (for inputs to transform block 4),
  	1, exp(i* 4*twopi/32) = ISRT2*( 1    , 1    ), exp(i* 8*twopi/32) =       ( 0    , 1    ), exp(i*12*twopi/32) = ISRT2*(-1    , 1    ) (for inputs to transform block 5),
  	1, exp(i* 5*twopi/32) =       ( s32_3, c32_3), exp(i*10*twopi/32) =       (-s    , c    ), exp(i*15*twopi/32) =       (-c32_1, s32_1) (for inputs to transform block 6),
  	1, exp(i* 6*twopi/32) =       ( s    , c    ), exp(i*12*twopi/32) = ISRT2*(-1    , 1    ), exp(i*18*twopi/32) =       (-c    ,-s    ) (for inputs to transform block 7),
  	1, exp(i* 7*twopi/32) =       ( s32_1, c32_1), exp(i*14*twopi/32) =       (-c    , s    ), exp(i*21*twopi/32) =       (-s32_3,-c32_3) (for inputs to transform block 8),
     and only the last 3 inputs to each of the radix-4 transforms 2 through 8 are multiplied by non-unity twiddles.	*/

/*...Block 1: t00,t10,t20,t30	*/

	jt = j1;

	    rt =t10;	t10=t00-rt;	t00=t00+rt;
	    it =t11;	t11=t01-it;	t01=t01+it;

	    rt =t30;	t30=t20-rt;	t20=t20+rt;
	    it =t31;	t31=t21-it;	t21=t21+it;

	    a[jt    ]=t00+t20;		a[jt+1    ]=t01+t21;
	    a[jt+p01]=t00-t20;		a[jt+1+p01]=t01-t21;

	    a[jt+p02]=t10-t31;		a[jt+1+p02]=t11+t30;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t10+t31;		a[jt+1+p03]=t11-t30;

/*...Block 5: t08,t18,t28,t38	*/

	jt = j1 + p04;

	    rt =t18;	t18=t08+t19;	t08=t08-t19;		/* twiddle mpy by E^4 = I	*/
			t19=t09-rt;	t09=t09+rt;

	    rt =(t28-t29)*ISRT2;	t29=(t28+t29)*ISRT2;		t28=rt;	/* twiddle mpy by E^8	*/
	    rt =(t39+t38)*ISRT2;	it =(t39-t38)*ISRT2;			/* twiddle mpy by -E^12 is here...	*/
	    t38=t28+rt;			t28=t28-rt;				/* ...and get E^12 by flipping signs here.	*/
	    t39=t29+it;			t29=t29-it;

	    a[jt    ]=t08+t28;		a[jt+1    ]=t09+t29;
	    a[jt+p01]=t08-t28;		a[jt+1+p01]=t09-t29;

	    a[jt+p02]=t18-t39;		a[jt+1+p02]=t19+t38;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t18+t39;		a[jt+1+p03]=t19-t38;

/*...Block 3: t04,t14,t24,t34	*/

	jt = j1 + p08;

	    rt =(t14-t15)*ISRT2;	it =(t14+t15)*ISRT2;			/* twiddle mpy by E^4	*/
	    t14=t04-rt;			t04=t04+rt;
	    t15=t05-it;			t05=t05+it;

	    rt =t24*c - t25*s;		t25=t25*c + t24*s;		t24=rt;	/* twiddle mpy by E^2	*/
	    rt =t34*s - t35*c;		it =t35*s + t34*c;			/* twiddle mpy by E^6	*/
	    t34=t24-rt;			t24=t24+rt;
	    t35=t25-it;			t25=t25+it;

	    a[jt    ]=t04+t24;		a[jt+1    ]=t05+t25;
	    a[jt+p01]=t04-t24;		a[jt+1+p01]=t05-t25;

	    a[jt+p02]=t14-t35;		a[jt+1+p02]=t15+t34;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t14+t35;		a[jt+1+p03]=t15-t34;

/*...Block 7: t0C,t1C,t2C,t3C	*/

	jt = j1 + p0C;

	    rt =(t1D+t1C)*ISRT2;	it =(t1D-t1C)*ISRT2;			/* twiddle mpy by -E^12 is here...	*/
	    t1C=t0C+rt;			t0C=t0C-rt;				/* ...and get E^12 by flipping signs here.	*/
	    t1D=t0D+it;			t0D=t0D-it;

	    rt =t2C*s - t2D*c;		t2D=t2D*s + t2C*c;		t2C=rt;	/* twiddle mpy by E^6	*/
	    rt =t3C*c - t3D*s;		it =t3D*c + t3C*s;			/* twiddle mpy by E^18 is here...	*/
	    t3C=t2C+rt;			t2C=t2C-rt;				/* ...and get E^18 by flipping signs here.	*/
	    t3D=t2D+it;			t2D=t2D-it;

	    a[jt    ]=t0C+t2C;		a[jt+1    ]=t0D+t2D;
	    a[jt+p01]=t0C-t2C;		a[jt+1+p01]=t0D-t2D;

	    a[jt+p02]=t1C-t3D;		a[jt+1+p02]=t1D+t3C;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t1C+t3D;		a[jt+1+p03]=t1D-t3C;

/*...Block 2: t02,t12,t22,t32	*/

	jt = j1 + p10;

	    rt =t12*c - t13*s;		it =t13*c + t12*s;			/* twiddle mpy by E^2	*/
	    t12=t02-rt;			t02=t02+rt;
	    t13=t03-it;			t03=t03+it;

	    rt =t22*c32_1 - t23*s32_1;	t23=t23*c32_1 + t22*s32_1;	t22=rt;	/* twiddle mpy by E^1	*/
	    rt =t32*c32_3 - t33*s32_3;	it =t33*c32_3 + t32*s32_3;		/* twiddle mpy by E^3	*/
	    t32=t22-rt;			t22=t22+rt;
	    t33=t23-it;			t23=t23+it;

	    a[jt    ]=t02+t22;		a[jt+1    ]=t03+t23;
	    a[jt+p01]=t02-t22;		a[jt+1+p01]=t03-t23;

	    a[jt+p02]=t12-t33;		a[jt+1+p02]=t13+t32;	/* mpy by E^4=i is inlined here...;	*/
	    a[jt+p03]=t12+t33;		a[jt+1+p03]=t13-t32;

/*...Block 6: t0A,t1A,t2A,t3A	*/

	jt = j1 + p14;

	    rt =t1A*s + t1B*c;		it =t1B*s - t1A*c;			/* twiddle mpy by -E^10 is here...	*/
	    t1A=t0A+rt;			t0A =t0A-rt;				/* ...and get E^10 by flipping signs here.	*/
	    t1B=t0B+it;			t0B =t0B-it;

	    rt =t2A*s32_3 - t2B*c32_3;	t2B=t2B*s32_3 + t2A*c32_3;	t2A=rt;	/* twiddle mpy by E^5	*/
	    rt =t3A*c32_1 + t3B*s32_1;	it =t3B*c32_1 - t3A*s32_1;	 	/* twiddle mpy by -E^15 is here...	*/
	    t3A=t2A+rt;			t2A=t2A-rt;				/* ...and get E^15 by flipping signs here.	*/
	    t3B=t2B+it;			t2B=t2B-it;

	    a[jt    ]=t0A+t2A;		a[jt+1    ]=t0B+t2B;
	    a[jt+p01]=t0A-t2A;		a[jt+1+p01]=t0B-t2B;

	    a[jt+p02]=t1A-t3B;		a[jt+1+p02]=t1B+t3A;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t1A+t3B;		a[jt+1+p03]=t1B-t3A;

/*...Block 4: t06,t16,t26,t36	*/

	jt = j1 + p18;

	    rt =t16*s - t17*c;		it =t17*s + t16*c;			/* twiddle mpy by E^6	*/
	    t16=t06-rt;			t06 =t06+rt;
	    t17=t07-it;			t07 =t07+it;

	    rt =t26*c32_3 - t27*s32_3;	t27=t27*c32_3 + t26*s32_3;	t26=rt;	/* twiddle mpy by E^3	*/
	    rt =t36*s32_1 + t37*c32_1;	it =t37*s32_1 - t36*c32_1;		/* twiddle mpy by -E^9 is here...	*/
	    t36=t26+rt;			t26=t26-rt;				/* ...and get E^9 by flipping signs here.	*/
	    t37=t27+it;			t27=t27-it;

	    a[jt    ]=t06+t26;		a[jt+1    ]=t07+t27;
	    a[jt+p01]=t06-t26;		a[jt+1+p01]=t07-t27;

	    a[jt+p02]=t16-t37;		a[jt+1+p02]=t17+t36;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t16+t37;		a[jt+1+p03]=t17-t36;

/*...Block 8: t0E,t1E,t2E,t3E	*/

	jt = j1 + p1C;

	    rt =t1E*c + t1F*s;		it =t1F*c - t1E*s;			/* twiddle mpy by -E^14 is here...	*/
	    t1E=t0E+rt;			t0E =t0E-rt;				/* ...and get E^14 by flipping signs here.	*/
	    t1F=t0F+it;			t0F =t0F-it;

	    rt =t2E*s32_1 - t2F*c32_1;	t2F=t2F*s32_1 + t2E*c32_1;	t2E=rt;	/* twiddle mpy by E^7	*/
	    rt =t3E*s32_3 - t3F*c32_3;	it =t3F*s32_3 + t3E*c32_3;		/* twiddle mpy by -E^21 is here...	*/
	    t3E=t2E+rt;			t2E=t2E-rt;				/* ...and get E^21 by flipping signs here.	*/
	    t3F=t2F+it;			t2F=t2F-it;

	    a[jt    ]=t0E+t2E;		a[jt+1    ]=t0F+t2F;
	    a[jt+p01]=t0E-t2E;		a[jt+1+p01]=t0F-t2F;

	    a[jt+p02]=t1E-t3F;		a[jt+1+p02]=t1F+t3E;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t1E+t3F;		a[jt+1+p03]=t1F-t3E;

	}
}

/***************/

int radix32_ditN_cy_dif1(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
                         struct complex rt1[], double base[], double baseinv[], int iter, double *fracmax, int p)
{
/*
!...Acronym: DWT = Discrete Weighted Transform, DIT = Decimation In Time, DIF = Decimation In Frequency
!
!...Performs a final radix-32 complex DIT pass, an inverse DWT weighting, a carry propagation,
!   a forward DWT weighting, and an initial radix-32 complex DIF pass on the data in the length-N real vector A.
!
!   Data enter and are returned in the A-array.
!
!   See the documentation in mers_mod_square and radix16_dif_pass for further details on the array
!   storage scheme, and radix16_ditN_cy_dif1 for details on the reduced-length weights array scheme.
*/
	int n32,i,j,j1,jt,jstart,jhi,iroot,iroot7,root_incr,k1,k2,k,khi,l,outer
	  ,bjmodn00,bjmodn01,bjmodn02,bjmodn03,bjmodn04,bjmodn05,bjmodn06,bjmodn07,bjmodn08,bjmodn09,bjmodn0A,bjmodn0B,bjmodn0C,bjmodn0D,bjmodn0E,bjmodn0F
	  ,bjmodn10,bjmodn11,bjmodn12,bjmodn13,bjmodn14,bjmodn15,bjmodn16,bjmodn17,bjmodn18,bjmodn19,bjmodn1A,bjmodn1B,bjmodn1C,bjmodn1D,bjmodn1E,bjmodn1F;
	static int psave=0,bw,sw,bjmodnini,nm1,nrtm1,p01,p02,p03,p04,p05,p06,p07,p08,p0C,p10,p14,p18,p1C;
	const  double one_half[3] = {1.0, 0.5, 0.25};	/* Needed for small-weights-tables scheme */
	static double    c     = 0.92387953251128675613, s     = 0.38268343236508977173	/* exp(  i*twopi/16)	*/
			,c32_1 = 0.98078528040323044912, s32_1 = 0.19509032201612826784	/* exp(  i*twopi/32)	*/
			,c32_3 = 0.83146961230254523708, s32_3 = 0.55557023301960222473 /* exp(3*i*twopi/32)	*/
			,radix_inv, n2inv;
	double rt,it
		,t00,t01,t02,t03,t04,t05,t06,t07,t08,t09,t0A,t0B,t0C,t0D,t0E,t0F
		,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,t1A,t1B,t1C,t1D,t1E,t1F
		,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,t2A,t2B,t2C,t2D,t2E,t2F
		,t30,t31,t32,t33,t34,t35,t36,t37,t38,t39,t3A,t3B,t3C,t3D,t3E,t3F
		,a1p00r,a1p01r,a1p02r,a1p03r,a1p04r,a1p05r,a1p06r,a1p07r,a1p08r,a1p09r,a1p0Ar,a1p0Br,a1p0Cr,a1p0Dr,a1p0Er,a1p0Fr
		,a1p10r,a1p11r,a1p12r,a1p13r,a1p14r,a1p15r,a1p16r,a1p17r,a1p18r,a1p19r,a1p1Ar,a1p1Br,a1p1Cr,a1p1Dr,a1p1Er,a1p1Fr
		,a1p00i,a1p01i,a1p02i,a1p03i,a1p04i,a1p05i,a1p06i,a1p07i,a1p08i,a1p09i,a1p0Ai,a1p0Bi,a1p0Ci,a1p0Di,a1p0Ei,a1p0Fi
		,a1p10i,a1p11i,a1p12i,a1p13i,a1p14i,a1p15i,a1p16i,a1p17i,a1p18i,a1p19i,a1p1Ai,a1p1Bi,a1p1Ci,a1p1Di,a1p1Ei,a1p1Fi
		,cy00,cy01,cy02,cy03,cy04,cy05,cy06,cy07,cy08,cy09,cy0A,cy0B,cy0C,cy0D,cy0E,cy0F
		,cy10,cy11,cy12,cy13,cy14,cy15,cy16,cy17,cy18,cy19,cy1A,cy1B,cy1C,cy1D,cy1E,cy1F
		,temp,frac,scale;
	double maxerr = 0.0;
#if PFETCH
	double *add0;
	int prefetch_offset;
#endif
	int err;
	FILE *f1,*f2;
	static int first_entry=TRUE;

/*...stuff for the reduced-length DWT weights array is here:	*/
	int n_div_nwt;
	int col,co2,co3,m,m2,n_minus_sil,n_minus_silp1,sinwt,sinwtm1;
	double wt,wtinv,wtl,wtlp1,wtn,wtnm1,wtA,wtB,wtC;

/*...change n32 and n_div_wt to non-static to work around a gcc compiler bug. */
	n32   = n/32;
	n_div_nwt = n32 >> nwt_bits;

	if((n_div_nwt << nwt_bits) != n32)
	{
	  if(INTERACT)printf("FATAL: iter = %10d; NWT_BITS does not divide N/32 in radix32_ditN_cy_dif1.\n",iter);
	  f1 = fopen(   OFILE,"a");
	  f2 = fopen(STATFILE,"a");
	  fprintf(f1,"FATAL: iter = %10d; NWT_BITS does not divide N/32 in radix32_ditN_cy_dif1.\n",iter);
	  fprintf(f2,"FATAL: iter = %10d; NWT_BITS does not divide N/32 in radix32_ditN_cy_dif1.\n",iter);
	  fclose(f1);
	  fclose(f2);
	  err=1;
	  return(err);
	}

	if(p != psave)
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  psave = p;
	  first_entry=FALSE;
	  radix_inv = (double)((long double)1.0L/(long double)32.0L);
	  n2inv     = (double)((long double)1.0L/(long double)(n/2));

	  bw    = p%n;		/* Number of bigwords in the Crandall/Fagin mixed-radix representation = (Mersenne exponent) mod (vector length).	*/
	  sw    = n - bw;	/* Number of smallwords.	*/

	  nm1   = n-1;
	  nrtm1 = nrt - 1;

/*   constant index offsets for load/stores are here.	*/

	  p01 = n32;
	  p02 = p01 +p01;
	  p03 = p02 +p01;
	  p04 = p03 +p01;
	  p05 = p04 +p01;
	  p06 = p05 +p01;
	  p07 = p06 +p01;
	  p08 = p04 +p04;
	  p0C = p08 +p04;
	  p10 = p0C +p04;
	  p14 = p10 +p04;
	  p18 = p14 +p04;
	  p1C = p18 +p04;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p05 = p05 + ( (p05 >> DAT_BITS) << PAD_BITS );
	  p06 = p06 + ( (p06 >> DAT_BITS) << PAD_BITS );
	  p07 = p07 + ( (p07 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p0C = p0C + ( (p0C >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p14 = p14 + ( (p14 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	  p1C = p1C + ( (p1C >> DAT_BITS) << PAD_BITS );

	  bjmodnini=0;
	  for(j=0; j < n32; j++)
	  {
	    bjmodnini -= sw; bjmodnini = bjmodnini + ( (-(int)((uint32)bjmodnini >> 31)) & n);
	  }
	}

/*...The radix-32 final DIT pass is here.	*/

	cy00 =-2;		/* init carries	*/
	cy01 = 0;
	cy02 = 0;
	cy03 = 0;
	cy04 = 0;
	cy05 = 0;
	cy06 = 0;
	cy07 = 0;
	cy08 = 0;
	cy09 = 0;
	cy0A = 0;
	cy0B = 0;
	cy0C = 0;
	cy0D = 0;
	cy0E = 0;
	cy0F = 0;
	cy10 = 0;
	cy11 = 0;
	cy12 = 0;
	cy13 = 0;
	cy14 = 0;
	cy15 = 0;
	cy16 = 0;
	cy17 = 0;
	cy18 = 0;
	cy19 = 0;
	cy1A = 0;
	cy1B = 0;
	cy1C = 0;
	cy1D = 0;
	cy1E = 0;
	cy1F = 0;

	*fracmax=0;	/* init max. fractional error	*/

	iroot = 0;	/* init sincos array index	*/
	root_incr = 1;	/* init sincos array index increment (set = 1 for normal carry pass, = 0 for wrapper pass)	*/
	scale = n2inv;	/* init inverse-weight scale factor  (set = 2/n for normal carry pass, = 1 for wrapper pass)	*/

	jstart = 0;
	jhi = jstart+nwt-1;
	khi = n_div_nwt;

for(outer=0; outer <= 1; outer++)
{
	i = 1;		/* Pointer to the BASE and BASEINV arrays. lowest-order digit is always a bigword (i = 1).	*/

	bjmodn00 = 0;
	bjmodn01 = bjmodnini;
	bjmodn02 = bjmodn01 + bjmodnini - n;	bjmodn02 = bjmodn02 + ( (-(int)((uint32)bjmodn02 >> 31)) & n);
	bjmodn03 = bjmodn02 + bjmodnini - n;	bjmodn03 = bjmodn03 + ( (-(int)((uint32)bjmodn03 >> 31)) & n);
	bjmodn04 = bjmodn03 + bjmodnini - n;	bjmodn04 = bjmodn04 + ( (-(int)((uint32)bjmodn04 >> 31)) & n);
	bjmodn05 = bjmodn04 + bjmodnini - n;	bjmodn05 = bjmodn05 + ( (-(int)((uint32)bjmodn05 >> 31)) & n);
	bjmodn06 = bjmodn05 + bjmodnini - n;	bjmodn06 = bjmodn06 + ( (-(int)((uint32)bjmodn06 >> 31)) & n);
	bjmodn07 = bjmodn06 + bjmodnini - n;	bjmodn07 = bjmodn07 + ( (-(int)((uint32)bjmodn07 >> 31)) & n);
	bjmodn08 = bjmodn07 + bjmodnini - n;	bjmodn08 = bjmodn08 + ( (-(int)((uint32)bjmodn08 >> 31)) & n);
	bjmodn09 = bjmodn08 + bjmodnini - n;	bjmodn09 = bjmodn09 + ( (-(int)((uint32)bjmodn09 >> 31)) & n);
	bjmodn0A = bjmodn09 + bjmodnini - n;	bjmodn0A = bjmodn0A + ( (-(int)((uint32)bjmodn0A >> 31)) & n);
	bjmodn0B = bjmodn0A + bjmodnini - n;	bjmodn0B = bjmodn0B + ( (-(int)((uint32)bjmodn0B >> 31)) & n);
	bjmodn0C = bjmodn0B + bjmodnini - n;	bjmodn0C = bjmodn0C + ( (-(int)((uint32)bjmodn0C >> 31)) & n);
	bjmodn0D = bjmodn0C + bjmodnini - n;	bjmodn0D = bjmodn0D + ( (-(int)((uint32)bjmodn0D >> 31)) & n);
	bjmodn0E = bjmodn0D + bjmodnini - n;	bjmodn0E = bjmodn0E + ( (-(int)((uint32)bjmodn0E >> 31)) & n);
	bjmodn0F = bjmodn0E + bjmodnini - n;	bjmodn0F = bjmodn0F + ( (-(int)((uint32)bjmodn0F >> 31)) & n);
	bjmodn10 = bjmodn0F + bjmodnini - n;	bjmodn10 = bjmodn10 + ( (-(int)((uint32)bjmodn10 >> 31)) & n);
	bjmodn11 = bjmodn10 + bjmodnini - n;	bjmodn11 = bjmodn11 + ( (-(int)((uint32)bjmodn11 >> 31)) & n);
	bjmodn12 = bjmodn11 + bjmodnini - n;	bjmodn12 = bjmodn12 + ( (-(int)((uint32)bjmodn12 >> 31)) & n);
	bjmodn13 = bjmodn12 + bjmodnini - n;	bjmodn13 = bjmodn13 + ( (-(int)((uint32)bjmodn13 >> 31)) & n);
	bjmodn14 = bjmodn13 + bjmodnini - n;	bjmodn14 = bjmodn14 + ( (-(int)((uint32)bjmodn14 >> 31)) & n);
	bjmodn15 = bjmodn14 + bjmodnini - n;	bjmodn15 = bjmodn15 + ( (-(int)((uint32)bjmodn15 >> 31)) & n);
	bjmodn16 = bjmodn15 + bjmodnini - n;	bjmodn16 = bjmodn16 + ( (-(int)((uint32)bjmodn16 >> 31)) & n);
	bjmodn17 = bjmodn16 + bjmodnini - n;	bjmodn17 = bjmodn17 + ( (-(int)((uint32)bjmodn17 >> 31)) & n);
	bjmodn18 = bjmodn17 + bjmodnini - n;	bjmodn18 = bjmodn18 + ( (-(int)((uint32)bjmodn18 >> 31)) & n);
	bjmodn19 = bjmodn18 + bjmodnini - n;	bjmodn19 = bjmodn19 + ( (-(int)((uint32)bjmodn19 >> 31)) & n);
	bjmodn1A = bjmodn19 + bjmodnini - n;	bjmodn1A = bjmodn1A + ( (-(int)((uint32)bjmodn1A >> 31)) & n);
	bjmodn1B = bjmodn1A + bjmodnini - n;	bjmodn1B = bjmodn1B + ( (-(int)((uint32)bjmodn1B >> 31)) & n);
	bjmodn1C = bjmodn1B + bjmodnini - n;	bjmodn1C = bjmodn1C + ( (-(int)((uint32)bjmodn1C >> 31)) & n);
	bjmodn1D = bjmodn1C + bjmodnini - n;	bjmodn1D = bjmodn1D + ( (-(int)((uint32)bjmodn1D >> 31)) & n);
	bjmodn1E = bjmodn1D + bjmodnini - n;	bjmodn1E = bjmodn1E + ( (-(int)((uint32)bjmodn1E >> 31)) & n);
	bjmodn1F = bjmodn1E + bjmodnini - n;	bjmodn1F = bjmodn1F + ( (-(int)((uint32)bjmodn1F >> 31)) & n);

	col=0;
	co2=(n >> nwt_bits)-1+32;
	co3=co2-32;		/* At the start of each new j-loop, co3=co2-radix(1)	*/

	for(k=1; k <= khi; k++)	/* Do n/(radix(1)*nwt) outer loop executions...	*/
	{
	  for(j=jstart; j<jhi; j += 2)	/* Each inner loop execution processes (radix(1)*nwt) array data.	*/
	  {
	    j1 = j + ( (j >> DAT_BITS) << PAD_BITS );	/* padded-array fetch index is here	*/

#if !USE_NEW_DIT
/*	here are the needed sincos data - these will be processed below in bit-reversed order.	*/

	    l = iroot;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      l += iroot;			/* 2*iroot	*/
	      t00=rt0[k1].re;	t01=rt0[k1].im;
	      rt =rt1[k2].re;	it =rt1[k2].im;
	      t20=t00*rt -t01*it ;	t21=t00*it +t01*rt;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      l += iroot;			/* 3*iroot	*/
	      t00=rt0[k1].re;	t01=rt0[k1].im;
	      rt =rt1[k2].re;	it =rt1[k2].im;
	      t10=t00*rt -t01*it ;	t11=t00*it +t01*rt;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      l += 4*iroot;			/* 7*iroot	*/
	      iroot7 = l;			/* 7*iroot	*/
	      t00=rt0[k1].re;	t01=rt0[k1].im;
	      rt =rt1[k2].re;	it =rt1[k2].im;
	      t30=t00*rt -t01*it ;	t31=t00*it +t01*rt;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      l += iroot7;			/* 14*iroot	*/
	      t00=rt0[k1].re;	t01=rt0[k1].im;
	      rt =rt1[k2].re;	it =rt1[k2].im;
	      t38=t00*rt -t01*it ;	t39=t00*it +t01*rt;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      l += iroot7;			/* 21*iroot	*/
	      t00=rt0[k1].re;	t01=rt0[k1].im;
	      rt =rt1[k2].re;	it =rt1[k2].im;
	      t1C=t00*rt -t01*it ;	t1D=t00*it +t01*rt;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      l += iroot7;			/* 28*iroot	*/
	      t00=rt0[k1].re;	t01=rt0[k1].im;
	      rt =rt1[k2].re;	it =rt1[k2].im;
	      t2A=t00*rt -t01*it ;	t2B=t00*it +t01*rt;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      t00=rt0[k1].re;	t01=rt0[k1].im;
	      rt =rt1[k2].re;	it =rt1[k2].im;
	      t0E=t00*rt -t01*it ;	t0F=t00*it +t01*rt;

/*(c,s)4-10 (decimal indices), 4-0A (hexadecimal indices):	*/
	      t00=t20*t38; t01=t20*t39; rt =t21*t38; it =t21*t39;
	      t18=t00+it ; t19=t01-rt ; t04=t00-it ; t05=t01+rt ;

	      t00=t10*t38; t01=t10*t39; rt =t11*t38; it =t11*t39;
	      t28=t00+it ; t29=t01-rt ; t24=t00-it ; t25=t01+rt ;

	      t00=t30*t38; t01=t30*t39; rt =t31*t38; it =t31*t39;
	      t08=t00+it ; t09=t01-rt ; t14=t00-it ; t15=t01+rt ;

/*(c,s)11-17 (decimal indices), 0B-11 (hexadecimal indices):;	*/
	      t00=t20*t1C; t01=t20*t1D; rt =t21*t1C; it =t21*t1D;
	      t2C=t00+it ; t2D=t01-rt ; t3C=t00-it ; t3D=t01+rt ;

	      t00=t10*t1C; t01=t10*t1D; rt =t11*t1C; it =t11*t1D;
	      t0C=t00+it ; t0D=t01-rt ; t02=t00-it ; t03=t01+rt ;

	      t00=t30*t1C; t01=t30*t1D; rt =t31*t1C; it =t31*t1D;
	      t34=t00+it ; t35=t01-rt ; t22=t00-it ; t23=t01+rt ;

/*(c,s)18-24 (decimal indices), 12-18 (hexadecimal indices):	*/
	      t00=t20*t2A; t01=t20*t2B; rt =t21*t2A; it =t21*t2B;
	      t0A=t00+it ; t0B=t01-rt ; t1A=t00-it ; t1B=t01+rt ;

	      t00=t10*t2A; t01=t10*t2B; rt =t11*t2A; it =t11*t2B;
	      t32=t00+it ; t33=t01-rt ; t3A=t00-it ; t3B=t01+rt ;

	      t00=t30*t2A; t01=t30*t2B; rt =t31*t2A; it =t31*t2B;
	      t12=t00+it ; t13=t01-rt ; t06=t00-it ; t07=t01+rt ;

/*(c,s)25-31 (decimal indices), 19-1F (hexadecimal indices):	*/
	      t00=t20*t0E; t01=t20*t0F; rt =t21*t0E; it =t21*t0F;
	      t36=t00+it ; t37=t01-rt ; t2E=t00-it ; t2F=t01+rt ;

	      t00=t10*t0E; t01=t10*t0F; rt =t11*t0E; it =t11*t0F;
	      t16=t00+it ; t17=t01-rt ; t1E=t00-it ; t1F=t01+rt ;

	      t00=t30*t0E; t01=t30*t0F; rt =t31*t0E; it =t31*t0F;
	      t26=t00+it ; t27=t01-rt ; t3E=t00-it ; t3F=t01+rt ;
#endif
/*       gather the needed data (32 64-bit complex, i.e. 64 64-bit reals) and do the first set of four length-8 transforms...	*/

/*...Block 1:	*/

	jt = j1;

#if USE_NEW_DIT
	    t00=a[jt    ];	t01=a[jt+1    ];
	    rt =a[jt+p01];	it =a[jt+1+p01];
#else
	    t00=a[jt    ];				t01=a[jt+1    ];
	    rt =a[jt+p01]*t02 + a[jt+1+p01]*t03;	it =a[jt+1+p01]*t02 - a[jt+p01]*t03;
#endif
	    t02=t00-rt;		t03=t01-it;
	    t00=t00+rt;		t01=t01+it;

#if USE_NEW_DIT
	    t04=a[jt+p02];	t05=a[jt+1+p02];
	    rt =a[jt+p03];	it =a[jt+1+p03];
#else
	    rt =a[jt+p02]*t04 + a[jt+1+p02]*t05;	t05=a[jt+1+p02]*t04 - a[jt+p02]*t05;	t04=rt;
	    rt =a[jt+p03]*t06 + a[jt+1+p03]*t07;	it =a[jt+1+p03]*t06 - a[jt+p03]*t07;
#endif
	    t06=t04-rt;		t07=t05-it;
	    t04=t04+rt;		t05=t05+it;

	    rt =t04;		it =t05;
	    t04=t00-rt;		t05=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t06;		it =t07;
	    t06=t02-it;		t07=t03+rt;
	    t02=t02+it;		t03=t03-rt;

#if USE_NEW_DIT
	    t08=a[jt+p04];	t09=a[jt+1+p04];
	    rt =a[jt+p05];	it =a[jt+1+p05];
#else
	    rt =a[jt+p04]*t08 + a[jt+1+p04]*t09;	t09=a[jt+1+p04]*t08 - a[jt+p04]*t09;	t08=rt;
	    rt =a[jt+p05]*t0A + a[jt+1+p05]*t0B;	it =a[jt+1+p05]*t0A - a[jt+p05]*t0B;
#endif
	    t0A=t08-rt;		t0B=t09-it;
	    t08=t08+rt;		t09=t09+it;

#if USE_NEW_DIT
	    t0C=a[jt+p06];	t0D=a[jt+1+p06];
	    rt =a[jt+p07];	it =a[jt+1+p07];
#else
	    rt =a[jt+p06]*t0C + a[jt+1+p06]*t0D;	t0D=a[jt+1+p06]*t0C - a[jt+p06]*t0D;	t0C=rt;
	    rt =a[jt+p07]*t0E + a[jt+1+p07]*t0F;	it =a[jt+1+p07]*t0E - a[jt+p07]*t0F;
#endif
	    t0E=t0C-rt;		t0F=t0D-it;
	    t0C=t0C+rt;		t0D=t0D+it;

	    rt =t0C;		it =t0D;
	    t0C=t08-rt;		t0D=t09-it;
	    t08=t08+rt;		t09=t09+it;

	    rt =t0E;		it =t0F;
	    t0E=t0A-it;		t0F=t0B+rt;
	    t0A=t0A+it;		t0B=t0B-rt;

	    rt =t08;		it =t09;
	    t08=t00-rt;		t09=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t0C;		it =t0D;
	    t0C=t04-it;		t0D=t05+rt;
	    t04=t04+it;		t05=t05-rt;

	    rt =(t0A+t0B)*ISRT2;it =(t0A-t0B)*ISRT2;
	    t0A=t02-rt;		t0B=t03+it;
	    t02=t02+rt;		t03=t03-it;

	    rt =(t0E-t0F)*ISRT2;it =(t0F+t0E)*ISRT2;
	    t0E=t06+rt;		t0F=t07+it;
	    t06=t06-rt;		t07=t07-it;

/*...Block 2:;	*/

	jt = j1 + p08;

#if USE_NEW_DIT
	    t10=a[jt    ];	t11=a[jt+1    ];
	    rt =a[jt+p01];	it =a[jt+1+p01];
#else
	    rt =a[jt    ]*t10 + a[jt+1    ]*t11;	t11=a[jt+1    ]*t10 - a[jt    ]*t11;	t10=rt;
	    rt =a[jt+p01]*t12 + a[jt+1+p01]*t13;	it =a[jt+1+p01]*t12 - a[jt+p01]*t13;
#endif
	    t12=t10-rt;		t13=t11-it;
	    t10=t10+rt;		t11=t11+it;

#if USE_NEW_DIT
	    t14=a[jt+p02];	t15=a[jt+1+p02];
	    rt =a[jt+p03];	it =a[jt+1+p03];
#else
	    rt =a[jt+p02]*t14 + a[jt+1+p02]*t15;	t15=a[jt+1+p02]*t14 - a[jt+p02]*t15;	t14=rt;
	    rt =a[jt+p03]*t16 + a[jt+1+p03]*t17;	it =a[jt+1+p03]*t16 - a[jt+p03]*t17;
#endif
	    t16=t14-rt;		t17=t15-it;
	    t14=t14+rt;		t15=t15+it;

	    rt =t14;		it =t15;
	    t14=t10-rt;		t15=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t16;		it =t17;
	    t16=t12-it;		t17=t13+rt;
	    t12=t12+it;		t13=t13-rt;

#if USE_NEW_DIT
	    t18=a[jt+p04];	t19=a[jt+1+p04];
	    rt =a[jt+p05];	it =a[jt+1+p05];
#else
	    rt =a[jt+p04]*t18 + a[jt+1+p04]*t19;	t19=a[jt+1+p04]*t18 - a[jt+p04]*t19;	t18=rt;
	    rt =a[jt+p05]*t1A + a[jt+1+p05]*t1B;	it =a[jt+1+p05]*t1A - a[jt+p05]*t1B;
#endif
	    t1A=t18-rt;		t1B=t19-it;
	    t18=t18+rt;		t19=t19+it;

#if USE_NEW_DIT
	    t1C=a[jt+p06];	t1D=a[jt+1+p06];
	    rt =a[jt+p07];	it =a[jt+1+p07];
#else
	    rt =a[jt+p06]*t1C + a[jt+1+p06]*t1D;	t1D=a[jt+1+p06]*t1C - a[jt+p06]*t1D;	t1C=rt;
	    rt =a[jt+p07]*t1E + a[jt+1+p07]*t1F;	it =a[jt+1+p07]*t1E - a[jt+p07]*t1F;
#endif
	    t1E=t1C-rt;		t1F=t1D-it;
	    t1C=t1C+rt;		t1D=t1D+it;

	    rt =t1C;		it =t1D;
	    t1C=t18-rt;		t1D=t19-it;
	    t18=t18+rt;		t19=t19+it;

	    rt =t1E;		it =t1F;
	    t1E=t1A-it;		t1F=t1B+rt;
	    t1A=t1A+it;		t1B=t1B-rt;

	    rt =t18;		it =t19;
	    t18=t10-rt;		t19=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t1C;		it =t1D;
	    t1C=t14-it;		t1D=t15+rt;
	    t14=t14+it;		t15=t15-rt;

	    rt =(t1A+t1B)*ISRT2;it =(t1A-t1B)*ISRT2;
	    t1A=t12-rt;		t1B=t13+it;
	    t12=t12+rt;		t13=t13-it;

	    rt =(t1E-t1F)*ISRT2;it =(t1F+t1E)*ISRT2;
	    t1E=t16+rt;		t1F=t17+it;
	    t16=t16-rt;		t17=t17-it;

/*...Block 3:	*/

	jt = j1 + p10;

#if USE_NEW_DIT
	    t20=a[jt    ];	t21=a[jt+1    ];
	    rt =a[jt+p01];	it =a[jt+1+p01];
#else
	    rt =a[jt    ]*t20 + a[jt+1    ]*t21;	t21=a[jt+1    ]*t20 - a[jt    ]*t21;	t20=rt;
	    rt =a[jt+p01]*t22 + a[jt+1+p01]*t23;	it =a[jt+1+p01]*t22 - a[jt+p01]*t23;
#endif
	    t22=t20-rt;		t23=t21-it;
	    t20=t20+rt;		t21=t21+it;

#if USE_NEW_DIT
	    t24=a[jt+p02];	t25=a[jt+1+p02];
	    rt =a[jt+p03];	it =a[jt+1+p03];
#else
	    rt =a[jt+p02]*t24 + a[jt+1+p02]*t25;	t25=a[jt+1+p02]*t24 - a[jt+p02]*t25;	t24=rt;
	    rt =a[jt+p03]*t26 + a[jt+1+p03]*t27;	it =a[jt+1+p03]*t26 - a[jt+p03]*t27;
#endif
	    t26=t24-rt;		t27=t25-it;
	    t24=t24+rt;		t25=t25+it;

	    rt =t24;		it =t25;
	    t24=t20-rt;		t25=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t26;		it =t27;
	    t26=t22-it;		t27=t23+rt;
	    t22=t22+it;		t23=t23-rt;

#if USE_NEW_DIT
	    t28=a[jt+p04];	t29=a[jt+1+p04];
	    rt =a[jt+p05];	it =a[jt+1+p05];
#else
	    rt =a[jt+p04]*t28 + a[jt+1+p04]*t29;	t29=a[jt+1+p04]*t28 - a[jt+p04]*t29;	t28=rt;
	    rt =a[jt+p05]*t2A + a[jt+1+p05]*t2B;	it =a[jt+1+p05]*t2A - a[jt+p05]*t2B;
#endif
	    t2A=t28-rt;		t2B=t29-it;
	    t28=t28+rt;		t29=t29+it;

#if USE_NEW_DIT
	    t2C=a[jt+p06];	t2D=a[jt+1+p06];
	    rt =a[jt+p07];	it =a[jt+1+p07];
#else
	    rt =a[jt+p06]*t2C + a[jt+1+p06]*t2D;	t2D=a[jt+1+p06]*t2C - a[jt+p06]*t2D;	t2C=rt;
	    rt =a[jt+p07]*t2E + a[jt+1+p07]*t2F;	it =a[jt+1+p07]*t2E - a[jt+p07]*t2F;
#endif
	    t2E=t2C-rt;		t2F=t2D-it;
	    t2C=t2C+rt;		t2D=t2D+it;

	    rt =t2C;		it =t2D;
	    t2C=t28-rt;		t2D=t29-it;
	    t28=t28+rt;		t29=t29+it;

	    rt =t2E;		it =t2F;
	    t2E=t2A-it;		t2F=t2B+rt;
	    t2A=t2A+it;		t2B=t2B-rt;

	    rt =t28;		it =t29;
	    t28=t20-rt;		t29=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t2C;		it =t2D;
	    t2C=t24-it;		t2D=t25+rt;
	    t24=t24+it;		t25=t25-rt;

	    rt =(t2A+t2B)*ISRT2;it =(t2A-t2B)*ISRT2;
	    t2A=t22-rt;		t2B=t23+it;
	    t22=t22+rt;		t23=t23-it;

	    rt =(t2E-t2F)*ISRT2;it =(t2F+t2E)*ISRT2;
	    t2E=t26+rt;		t2F=t27+it;
	    t26=t26-rt;		t27=t27-it;

/*...Block 4:	*/

	jt = j1 + p18;

#if USE_NEW_DIT
	    t30=a[jt    ];	t31=a[jt+1    ];
	    rt =a[jt+p01];	it =a[jt+1+p01];
#else
	    rt =a[jt    ]*t30 + a[jt+1    ]*t31;	t31=a[jt+1    ]*t30 - a[jt    ]*t31;	t30=rt;
	    rt =a[jt+p01]*t32 + a[jt+1+p01]*t33;	it =a[jt+1+p01]*t32 - a[jt+p01]*t33;
#endif
	    t32=t30-rt;		t33=t31-it;
	    t30=t30+rt;		t31=t31+it;

#if USE_NEW_DIT
	    t34=a[jt+p02];	t35=a[jt+1+p02];
	    rt =a[jt+p03];	it =a[jt+1+p03];
#else
	    rt =a[jt+p02]*t34 + a[jt+1+p02]*t35;	t35=a[jt+1+p02]*t34 - a[jt+p02]*t35;	t34=rt;
	    rt =a[jt+p03]*t36 + a[jt+1+p03]*t37;	it =a[jt+1+p03]*t36 - a[jt+p03]*t37;
#endif
	    t36=t34-rt;		t37=t35-it;
	    t34=t34+rt;		t35=t35+it;

	    rt =t34;		it =t35;
	    t34=t30-rt;		t35=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t36;		it =t37;
	    t36=t32-it;		t37=t33+rt;
	    t32=t32+it;		t33=t33-rt;

#if USE_NEW_DIT
	    t38=a[jt+p04];	t39=a[jt+1+p04];
	    rt =a[jt+p05];	it =a[jt+1+p05];
#else
	    rt =a[jt+p04]*t38 + a[jt+1+p04]*t39;	t39=a[jt+1+p04]*t38 - a[jt+p04]*t39;	t38=rt;
	    rt =a[jt+p05]*t3A + a[jt+1+p05]*t3B;	it =a[jt+1+p05]*t3A - a[jt+p05]*t3B;
#endif
	    t3A=t38-rt;		t3B=t39-it;
	    t38=t38+rt;		t39=t39+it;

#if USE_NEW_DIT
	    t3C=a[jt+p06];	t3D=a[jt+1+p06];
	    rt =a[jt+p07];	it =a[jt+1+p07];
#else
	    rt =a[jt+p06]*t3C + a[jt+1+p06]*t3D;	t3D=a[jt+1+p06]*t3C - a[jt+p06]*t3D;	t3C=rt;
	    rt =a[jt+p07]*t3E + a[jt+1+p07]*t3F;	it =a[jt+1+p07]*t3E - a[jt+p07]*t3F;
#endif
	    t3E=t3C-rt;		t3F=t3D-it;
	    t3C=t3C+rt;		t3D=t3D+it;

	    rt =t3C;		it =t3D;
	    t3C=t38-rt;		t3D=t39-it;
	    t38=t38+rt;		t39=t39+it;

	    rt =t3E;		it =t3F;
	    t3E=t3A-it;		t3F=t3B+rt;
	    t3A=t3A+it;		t3B=t3B-rt;

	    rt =t38;		it =t39;
	    t38=t30-rt;		t39=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t3C;		it =t3D;
	    t3C=t34-it;		t3D=t35+rt;
	    t34=t34+it;		t35=t35-rt;

	    rt =(t3A+t3B)*ISRT2;it =(t3A-t3B)*ISRT2;
	    t3A=t32-rt;		t3B=t33+it;
	    t32=t32+rt;		t33=t33-it;

	    rt =(t3E-t3F)*ISRT2;it =(t3F+t3E)*ISRT2;
	    t3E=t36+rt;		t3F=t37+it;
	    t36=t36-rt;		t37=t37-it;

/*...and now do eight radix-4 transforms, including the internal twiddle factors:
  	1, exp(-i* 1*twopi/32) =       ( c32_1,-s32_1), exp(-i* 2*twopi/32) =       ( c    ,-s    ), exp(-i* 3*twopi/32) =       ( c32_3,-s32_3) (for inputs to transform block 2),
  	1, exp(-i* 2*twopi/32) =       ( c    ,-s    ), exp(-i* 4*twopi/32) = ISRT2*( 1    ,-1    ), exp(-i* 3*twopi/32) =       ( s    ,-c    ) (for inputs to transform block 3),
  	1, exp(-i* 3*twopi/32) =       ( c32_3,-s32_3), exp(-i* 6*twopi/32) =       ( s    ,-c    ), exp(-i* 9*twopi/32) =       (-s32_1,-c32_1) (for inputs to transform block 4),
  	1, exp(-i* 4*twopi/32) = ISRT2*( 1    ,-1    ), exp(-i* 8*twopi/32) =       ( 0    ,-1    ), exp(-i*12*twopi/32) = ISRT2*(-1    ,-1    ) (for inputs to transform block 5),
  	1, exp(-i* 5*twopi/32) =       ( s32_3,-c32_3), exp(-i*10*twopi/32) =       (-s    ,-c    ), exp(-i*15*twopi/32) =       (-c32_1,-s32_1) (for inputs to transform block 6),
  	1, exp(-i* 6*twopi/32) =       ( s    ,-c    ), exp(-i*12*twopi/32) = ISRT2*(-1    ,-1    ), exp(-i*18*twopi/32) =       (-c    , s    ) (for inputs to transform block 7),
  	1, exp(-i* 7*twopi/32) =       ( s32_1,-c32_1), exp(-i*14*twopi/32) =       (-c    ,-s    ), exp(-i*21*twopi/32) =       (-s32_3, c32_3) (for inputs to transform block 8),
     and only the last 3 inputs to each of the radix-4 transforms 2 through 8 are multiplied by non-unity twiddles.	*/

/*...Block 1: t00,t10,t20,t30	*/

	jt = j1;

	    rt =t10;	t10=t00-rt;	t00=t00+rt;
	    it =t11;	t11=t01-it;	t01=t01+it;

	    rt =t30;	t30=t20-rt;	t20=t20+rt;
	    it =t31;	t31=t21-it;	t21=t21+it;

	    a1p00r=t00+t20;		a1p00i=t01+t21;
	    a1p10r=t00-t20;		a1p10i=t01-t21;
	    a1p08r=t10+t31;		a1p08i=t11-t30;	/* mpy by E^-4 = -I is inlined here...	*/
	    a1p18r=t10-t31;		a1p18i=t11+t30;

/*...Block 5: t08,t18,t28,t38	*/

	jt = j1 + p04;

	    rt =t18;	t18=t08-t19;	t08=t08+t19;		/* twiddle mpy by E^8 =-I	*/
			t19=t09+rt;	t09=t09-rt;

	    rt =(t29+t28)*ISRT2;	t29=(t29-t28)*ISRT2;		t28=rt;	/* twiddle mpy by E^-4	*/
	    rt =(t38-t39)*ISRT2;	it =(t38+t39)*ISRT2;			/* twiddle mpy by E^4 = -E^-12 is here...	*/
	    t38=t28+rt;			t28=t28-rt;				/* ...and get E^-12 by flipping signs here.	*/
	    t39=t29+it;			t29=t29-it;

	    a1p04r=t08+t28;		a1p04i=t09+t29;
	    a1p14r=t08-t28;		a1p14i=t09-t29;
	    a1p0Cr=t18+t39;		a1p0Ci=t19-t38;	/* mpy by E^-4 = -I is inlined here...	*/
	    a1p1Cr=t18-t39;		a1p1Ci=t19+t38;

/*...Block 3: t04,t14,t24,t34	*/

	jt = j1 + p02;

	    rt =(t15+t14)*ISRT2;	it =(t15-t14)*ISRT2;			/* twiddle mpy by E^-4	*/
	    t14=t04-rt;			t04=t04+rt;
	    t15=t05-it;			t05=t05+it;

	    rt =t24*c + t25*s;		t25=t25*c - t24*s;		t24=rt;	/* twiddle mpy by E^-2	*/
	    rt =t34*s + t35*c;		it =t35*s - t34*c;			/* twiddle mpy by E^-6	*/
	    t34=t24-rt;			t24=t24+rt;
	    t35=t25-it;			t25=t25+it;

	    a1p02r=t04+t24;		a1p02i=t05+t25;
	    a1p12r=t04-t24;		a1p12i=t05-t25;
	    a1p0Ar=t14+t35;		a1p0Ai=t15-t34;	/* mpy by E^-4 = -I is inlined here...	*/
	    a1p1Ar=t14-t35;		a1p1Ai=t15+t34;

/*...Block 7: t0C,t1C,t2C,t3C	*/

	jt = j1 + p06;

	    rt =(t1C-t1D)*ISRT2;	it =(t1C+t1D)*ISRT2;			/* twiddle mpy by E^4 = -E^-12 is here...	*/
	    t1C=t0C+rt;			t0C=t0C-rt;				/* ...and get E^-12 by flipping signs here.	*/
	    t1D=t0D+it;			t0D=t0D-it;

	    rt =t2C*s + t2D*c;		t2D=t2D*s - t2C*c;		t2C=rt;	/* twiddle mpy by E^-6	*/
	    rt =t3C*c + t3D*s;		it =t3D*c - t3C*s;			/* twiddle mpy by E^-18 is here...	*/
	    t3C=t2C+rt;			t2C=t2C-rt;				/* ...and get E^-18 by flipping signs here.	*/
	    t3D=t2D+it;			t2D=t2D-it;

	    a1p06r=t0C+t2C;		a1p06i=t0D+t2D;
	    a1p16r=t0C-t2C;		a1p16i=t0D-t2D;
	    a1p0Er=t1C+t3D;		a1p0Ei=t1D-t3C;	/* mpy by E^-4 = -I is inlined here...	*/
	    a1p1Er=t1C-t3D;		a1p1Ei=t1D+t3C;

/*...Block 2: t02,t12,t22,t32	*/

	jt = j1 + p01;

	    rt =t12*c + t13*s;		it =t13*c - t12*s;			/* twiddle mpy by E^-2	*/
	    t12=t02-rt;			t02=t02+rt;
	    t13=t03-it;			t03=t03+it;

	    rt =t22*c32_1 + t23*s32_1;	t23=t23*c32_1 - t22*s32_1;	t22=rt;	/* twiddle mpy by E^-1	*/
	    rt =t32*c32_3 + t33*s32_3;	it =t33*c32_3 - t32*s32_3;		/* twiddle mpy by E^-3	*/
	    t32=t22-rt;			t22=t22+rt;
	    t33=t23-it;			t23=t23+it;

	    a1p01r=t02+t22;		a1p01i=t03+t23;
	    a1p11r=t02-t22;		a1p11i=t03-t23;
	    a1p09r=t12+t33;		a1p09i=t13-t32;	/* mpy by E^-4 = -I is inlined here...	*/
	    a1p19r=t12-t33;		a1p19i=t13+t32;

/*...Block 6: t0A,t1A,t2A,t3A	*/

	jt = j1 + p05;

	    rt =t1A*s - t1B*c;		it =t1B*s + t1A*c;			/* twiddle mpy by -E^-10 is here...	*/
	    t1A=t0A+rt;			t0A =t0A-rt;				/* ...and get E^-10 by flipping signs here.	*/
	    t1B=t0B+it;			t0B =t0B-it;

	    rt =t2A*s32_3 + t2B*c32_3;	t2B=t2B*s32_3 - t2A*c32_3;	t2A=rt;	/* twiddle mpy by E^-5	*/
	    rt =t3A*c32_1 - t3B*s32_1;	it =t3B*c32_1 + t3A*s32_1;		/* twiddle mpy by -E^-15 is here...	*/
	    t3A=t2A+rt;			t2A=t2A-rt;				/* ...and get E^-15 by flipping signs here.	*/
	    t3B=t2B+it;			t2B=t2B-it;

	    a1p05r=t0A+t2A;		a1p05i=t0B+t2B;
	    a1p15r=t0A-t2A;		a1p15i=t0B-t2B;
	    a1p0Dr=t1A+t3B;		a1p0Di=t1B-t3A;	/* mpy by E^-4 = -I is inlined here...	*/
	    a1p1Dr=t1A-t3B;		a1p1Di=t1B+t3A;

/*...Block 4: t06,t16,t26,t36	*/

	jt = j1 + p03;

	    rt =t16*s + t17*c;		it =t17*s - t16*c;			/* twiddle mpy by E^-6	*/
	    t16=t06-rt;			t06 =t06+rt;
	    t17=t07-it;			t07 =t07+it;

	    rt =t26*c32_3 + t27*s32_3;	t27=t27*c32_3 - t26*s32_3;	t26=rt;	/* twiddle mpy by E^-3	*/
	    rt =t36*s32_1 - t37*c32_1;	it =t37*s32_1 + t36*c32_1;		/* twiddle mpy by -E^-9 is here...	*/
	    t36=t26+rt;			t26=t26-rt;				/* ...and get E^-9 by flipping signs here.	*/
	    t37=t27+it;			t27=t27-it;

	    a1p03r=t06+t26;		a1p03i=t07+t27;
	    a1p13r=t06-t26;		a1p13i=t07-t27;
	    a1p0Br=t16+t37;		a1p0Bi=t17-t36;	/* mpy by E^-4 = -I is inlined here...	*/
	    a1p1Br=t16-t37;		a1p1Bi=t17+t36;

/*...Block 8: t0E,t1E,t2E,t3E	*/

	jt = j1 + p07;

	    rt =t1E*c - t1F*s;		it =t1F*c + t1E*s;			/* twiddle mpy by -E^-14 is here...	*/
	    t1E=t0E+rt;			t0E =t0E-rt;				/* ...and get E^-14 by flipping signs here.	*/
	    t1F=t0F+it;			t0F =t0F-it;

	    rt =t2E*s32_1 + t2F*c32_1;	t2F=t2F*s32_1 - t2E*c32_1;	t2E=rt;	/* twiddle mpy by E^-7	*/
	    rt =t3E*s32_3 + t3F*c32_3;	it =t3F*s32_3 - t3E*c32_3;		/* twiddle mpy by -E^-21 is here...	*/
	    t3E=t2E+rt;			t2E=t2E-rt;				/* ...and get E^-21 by flipping signs here.	*/
	    t3F=t2F+it;			t2F=t2F-it;

	    a1p07r=t0E+t2E;		a1p07i=t0F+t2F;
	    a1p17r=t0E-t2E;		a1p17i=t0F-t2F;
	    a1p0Fr=t1E+t3F;		a1p0Fi=t1F-t3E;	/* mpy by E^-4 = -I is inlined here...	*/
	    a1p1Fr=t1E-t3F;		a1p1Fi=t1F+t3E;

/*...and combine those to complete the radix-32 transform and do the carries. Since the outputs would
     normally be getting dispatched to 32 separate blocks of the A-array, we need 32 separate carries.	*/

	    l= j & (nwt-1);			/* We want (S*J mod N) - SI(L) for all 32 carries, so precompute	*/
	    n_minus_sil   = n-si[l  ];		/* N - SI(L) and for each J, find N - (B*J mod N) - SI(L)		*/
	    n_minus_silp1 = n-si[l+1];		/* For the inverse weight, want (S*(N - J) mod N) - SI(NWT - L) =	*/
	    sinwt   = si[nwt-l  ];		/*	= N - (S*J mod N) - SI(NWT - L) = (B*J mod N) - SI(NWT - L).	*/
	    sinwtm1 = si[nwt-l-1];

	    wtl     =wt0[    l  ];
	    wtn     =wt0[nwt-l  ]*scale;	/* Include 1/(n/2) scale factor of inverse transform here...	*/
	    wtlp1   =wt0[    l+1];
	    wtnm1   =wt0[nwt-l-1]*scale;	/* ...and here.	*/

/*...set0 is slightly different from others:	*/
	   cmplx_carry_norm_pow2_errcheck0(a1p00r,a1p00i,cy00,bjmodn00);
	    cmplx_carry_norm_pow2_errcheck(a1p01r,a1p01i,cy01,bjmodn01,0x01);
	    cmplx_carry_norm_pow2_errcheck(a1p02r,a1p02i,cy02,bjmodn02,0x02);
	    cmplx_carry_norm_pow2_errcheck(a1p03r,a1p03i,cy03,bjmodn03,0x03);
	    cmplx_carry_norm_pow2_errcheck(a1p04r,a1p04i,cy04,bjmodn04,0x04);
	    cmplx_carry_norm_pow2_errcheck(a1p05r,a1p05i,cy05,bjmodn05,0x05);
	    cmplx_carry_norm_pow2_errcheck(a1p06r,a1p06i,cy06,bjmodn06,0x06);
	    cmplx_carry_norm_pow2_errcheck(a1p07r,a1p07i,cy07,bjmodn07,0x07);
	    cmplx_carry_norm_pow2_errcheck(a1p08r,a1p08i,cy08,bjmodn08,0x08);
	    cmplx_carry_norm_pow2_errcheck(a1p09r,a1p09i,cy09,bjmodn09,0x09);
	    cmplx_carry_norm_pow2_errcheck(a1p0Ar,a1p0Ai,cy0A,bjmodn0A,0x0A);
	    cmplx_carry_norm_pow2_errcheck(a1p0Br,a1p0Bi,cy0B,bjmodn0B,0x0B);
	    cmplx_carry_norm_pow2_errcheck(a1p0Cr,a1p0Ci,cy0C,bjmodn0C,0x0C);
	    cmplx_carry_norm_pow2_errcheck(a1p0Dr,a1p0Di,cy0D,bjmodn0D,0x0D);
	    cmplx_carry_norm_pow2_errcheck(a1p0Er,a1p0Ei,cy0E,bjmodn0E,0x0E);
	    cmplx_carry_norm_pow2_errcheck(a1p0Fr,a1p0Fi,cy0F,bjmodn0F,0x0F);
	    cmplx_carry_norm_pow2_errcheck(a1p10r,a1p10i,cy10,bjmodn10,0x10);
	    cmplx_carry_norm_pow2_errcheck(a1p11r,a1p11i,cy11,bjmodn11,0x11);
	    cmplx_carry_norm_pow2_errcheck(a1p12r,a1p12i,cy12,bjmodn12,0x12);
	    cmplx_carry_norm_pow2_errcheck(a1p13r,a1p13i,cy13,bjmodn13,0x13);
	    cmplx_carry_norm_pow2_errcheck(a1p14r,a1p14i,cy14,bjmodn14,0x14);
	    cmplx_carry_norm_pow2_errcheck(a1p15r,a1p15i,cy15,bjmodn15,0x15);
	    cmplx_carry_norm_pow2_errcheck(a1p16r,a1p16i,cy16,bjmodn16,0x16);
	    cmplx_carry_norm_pow2_errcheck(a1p17r,a1p17i,cy17,bjmodn17,0x17);
	    cmplx_carry_norm_pow2_errcheck(a1p18r,a1p18i,cy18,bjmodn18,0x18);
	    cmplx_carry_norm_pow2_errcheck(a1p19r,a1p19i,cy19,bjmodn19,0x19);
	    cmplx_carry_norm_pow2_errcheck(a1p1Ar,a1p1Ai,cy1A,bjmodn1A,0x1A);
	    cmplx_carry_norm_pow2_errcheck(a1p1Br,a1p1Bi,cy1B,bjmodn1B,0x1B);
	    cmplx_carry_norm_pow2_errcheck(a1p1Cr,a1p1Ci,cy1C,bjmodn1C,0x1C);
	    cmplx_carry_norm_pow2_errcheck(a1p1Dr,a1p1Di,cy1D,bjmodn1D,0x1D);
	    cmplx_carry_norm_pow2_errcheck(a1p1Er,a1p1Ei,cy1E,bjmodn1E,0x1E);
	    cmplx_carry_norm_pow2_errcheck(a1p1Fr,a1p1Fi,cy1F,bjmodn1F,0x1F);

	    i =((uint32)(sw - bjmodn00) >> 31);	/* get ready for the next set...	*/
	    co2=co3;	/* For all data but the first set in each j-block, co2=co3. Thus, after the first block of data is done
			   and only then: for all subsequent blocks it's superfluous), this assignment decrements co2 by radix(1).	*/

/*...The radix-32 DIF pass is here:	*/

/*       gather the needed data (32 64-bit complex, i.e. 64 64-bit reals) and do the first set of four length-8 transforms.	*/

/*...Block 1:	*/

	jt = j1;

	    t02=a1p00r-a1p10r;	t03=a1p00i-a1p10i;
	    t00=a1p00r+a1p10r;	t01=a1p00i+a1p10i;

	    t06=a1p08r-a1p18r;	t07=a1p08i-a1p18i;
	    t04=a1p08r+a1p18r;	t05=a1p08i+a1p18i;

	    rt =t04;		it =t05;
	    t04=t00-rt;		t05=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t06;		it =t07;
	    t06=t02+it;		t07=t03-rt;
	    t02=t02-it;		t03=t03+rt;

	    t0A=a1p04r-a1p14r;	t0B=a1p04i-a1p14i;
	    t08=a1p04r+a1p14r;	t09=a1p04i+a1p14i;

	    t0E=a1p0Cr-a1p1Cr;	t0F=a1p0Ci-a1p1Ci;
	    t0C=a1p0Cr+a1p1Cr;	t0D=a1p0Ci+a1p1Ci;

	    rt =t0C;		it =t0D;
	    t0C=t08-rt;		t0D=t09-it;
	    t08=t08+rt;		t09=t09+it;

	    rt =t0E;		it =t0F;
	    t0E=t0A+it;		t0F=t0B-rt;
	    t0A=t0A-it;		t0B=t0B+rt;

	    rt =t08;		it =t09;
	    t08=t00-rt;		t09=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t0C;		it =t0D;
	    t0C=t04+it;		t0D=t05-rt;
	    t04=t04-it;		t05=t05+rt;

	    rt =(t0A-t0B)*ISRT2;it =(t0A+t0B)*ISRT2;
	    t0A=t02-rt;		t0B=t03-it;
	    t02=t02+rt;		t03=t03+it;

	    rt =(t0E+t0F)*ISRT2;it =(t0F-t0E)*ISRT2;
	    t0E=t06+rt;		t0F=t07+it;
	    t06=t06-rt;		t07=t07-it;

/*...Block 2:	*/

	jt = j1 + p02;

	    t12=a1p02r-a1p12r;		t13=a1p02i-a1p12i;
	    t10=a1p02r+a1p12r;		t11=a1p02i+a1p12i;

	    t16=a1p0Ar-a1p1Ar;		t17=a1p0Ai-a1p1Ai;
	    t14=a1p0Ar+a1p1Ar;		t15=a1p0Ai+a1p1Ai;

	    rt =t14;		it =t15;
	    t14=t10-rt;		t15=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t16;		it =t17;
	    t16=t12+it;		t17=t13-rt;
	    t12=t12-it;		t13=t13+rt;

	    t1A=a1p06r-a1p16r;		t1B=a1p06i-a1p16i;
	    t18=a1p06r+a1p16r;		t19=a1p06i+a1p16i;

	    t1E=a1p0Er-a1p1Er;		t1F=a1p0Ei-a1p1Ei;
	    t1C=a1p0Er+a1p1Er;		t1D=a1p0Ei+a1p1Ei;

	    rt =t1C;		it =t1D;
	    t1C=t18-rt;		t1D=t19-it;
	    t18=t18+rt;		t19=t19+it;

	    rt =t1E;		it =t1F;
	    t1E=t1A+it;		t1F=t1B-rt;
	    t1A=t1A-it;		t1B=t1B+rt;

	    rt =t18;		it =t19;
	    t18=t10-rt;		t19=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t1C;		it =t1D;
	    t1C=t14+it;		t1D=t15-rt;
	    t14=t14-it;		t15=t15+rt;

	    rt =(t1A-t1B)*ISRT2;it =(t1A+t1B)*ISRT2;
	    t1A=t12-rt;		t1B=t13-it;
	    t12=t12+rt;		t13=t13+it;

	    rt =(t1E+t1F)*ISRT2;it =(t1F-t1E)*ISRT2;
	    t1E=t16+rt;		t1F=t17+it;
	    t16=t16-rt;		t17=t17-it;

/*...Block 3:	*/

	jt = j1 + p01;

	    t22=a1p01r-a1p11r;		t23=a1p01i-a1p11i;
	    t20=a1p01r+a1p11r;		t21=a1p01i+a1p11i;

	    t26=a1p09r-a1p19r;		t27=a1p09i-a1p19i;
	    t24=a1p09r+a1p19r;		t25=a1p09i+a1p19i;

	    rt =t24;		it =t25;
	    t24=t20-rt;		t25=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t26;		it =t27;
	    t26=t22+it;		t27=t23-rt;
	    t22=t22-it;		t23=t23+rt;

	    t2A=a1p05r-a1p15r;		t2B=a1p05i-a1p15i;
	    t28=a1p05r+a1p15r;		t29=a1p05i+a1p15i;

	    t2E=a1p0Dr-a1p1Dr;		t2F=a1p0Di-a1p1Di;
	    t2C=a1p0Dr+a1p1Dr;		t2D=a1p0Di+a1p1Di;

	    rt =t2C;		it =t2D;
	    t2C=t28-rt;		t2D=t29-it;
	    t28=t28+rt;		t29=t29+it;

	    rt =t2E;		it =t2F;
	    t2E=t2A+it;		t2F=t2B-rt;
	    t2A=t2A-it;		t2B=t2B+rt;

	    rt =t28;		it =t29;
	    t28=t20-rt;		t29=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t2C;		it =t2D;
	    t2C=t24+it;		t2D=t25-rt;
	    t24=t24-it;		t25=t25+rt;

	    rt =(t2A-t2B)*ISRT2;it =(t2A+t2B)*ISRT2;
	    t2A=t22-rt;		t2B=t23-it;
	    t22=t22+rt;		t23=t23+it;

	    rt =(t2E+t2F)*ISRT2;it =(t2F-t2E)*ISRT2;
	    t2E=t26+rt;		t2F=t27+it;
	    t26=t26-rt;		t27=t27-it;

/*...Block 4:	*/

	jt = j1 + p03;

	    t32=a1p03r-a1p13r;		t33=a1p03i-a1p13i;
	    t30=a1p03r+a1p13r;		t31=a1p03i+a1p13i;

	    t36=a1p0Br-a1p1Br;		t37=a1p0Bi-a1p1Bi;
	    t34=a1p0Br+a1p1Br;		t35=a1p0Bi+a1p1Bi;

	    rt =t34;		it =t35;
	    t34=t30-rt;		t35=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t36;		it =t37;
	    t36=t32+it;		t37=t33-rt;
	    t32=t32-it;		t33=t33+rt;

	    t3A=a1p07r-a1p17r;		t3B=a1p07i-a1p17i;
	    t38=a1p07r+a1p17r;		t39=a1p07i+a1p17i;

	    t3E=a1p0Fr-a1p1Fr;		t3F=a1p0Fi-a1p1Fi;
	    t3C=a1p0Fr+a1p1Fr;		t3D=a1p0Fi+a1p1Fi;

	    rt =t3C;		it =t3D;
	    t3C=t38-rt;		t3D=t39-it;
	    t38=t38+rt;		t39=t39+it;

	    rt =t3E;		it =t3F;
	    t3E=t3A+it;		t3F=t3B-rt;
	    t3A=t3A-it;		t3B=t3B+rt;

	    rt =t38;		it =t39;
	    t38=t30-rt;		t39=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t3C;		it =t3D;
	    t3C=t34+it;		t3D=t35-rt;
	    t34=t34-it;		t35=t35+rt;

	    rt =(t3A-t3B)*ISRT2;it =(t3A+t3B)*ISRT2;
	    t3A=t32-rt;		t3B=t33-it;
	    t32=t32+rt;		t33=t33+it;

	    rt =(t3E+t3F)*ISRT2;it =(t3F-t3E)*ISRT2;
	    t3E=t36+rt;		t3F=t37+it;
	    t36=t36-rt;		t37=t37-it;

/*...and now do eight radix-4 transforms, including the internal twiddle factors:
  	1, exp(i* 1*twopi/32) =       ( c32_1, s32_1), exp(i* 2*twopi/32) =       ( c    , s    ), exp(i* 3*twopi/32) =       ( c32_3, s32_3) (for inputs to transform block 2),
  	1, exp(i* 2*twopi/32) =       ( c    , s    ), exp(i* 4*twopi/32) = ISRT2*( 1    , 1    ), exp(i* 3*twopi/32) =       ( s    , c    ) (for inputs to transform block 3),
  	1, exp(i* 3*twopi/32) =       ( c32_3, s32_3), exp(i* 6*twopi/32) =       ( s    , c    ), exp(i* 9*twopi/32) =       (-s32_1, c32_1) (for inputs to transform block 4),
  	1, exp(i* 4*twopi/32) = ISRT2*( 1    , 1    ), exp(i* 8*twopi/32) =       ( 0    , 1    ), exp(i*12*twopi/32) = ISRT2*(-1    , 1    ) (for inputs to transform block 5),
  	1, exp(i* 5*twopi/32) =       ( s32_3, c32_3), exp(i*10*twopi/32) =       (-s    , c    ), exp(i*15*twopi/32) =       (-c32_1, s32_1) (for inputs to transform block 6),
  	1, exp(i* 6*twopi/32) =       ( s    , c    ), exp(i*12*twopi/32) = ISRT2*(-1    , 1    ), exp(i*18*twopi/32) =       (-c    ,-s    ) (for inputs to transform block 7),
  	1, exp(i* 7*twopi/32) =       ( s32_1, c32_1), exp(i*14*twopi/32) =       (-c    , s    ), exp(i*21*twopi/32) =       (-s32_3,-c32_3) (for inputs to transform block 8),
     and only the last 3 inputs to each of the radix-4 transforms 2 through 8 are multiplied by non-unity twiddles.	*/

#if PFETCH
prefetch_offset = ((j >> 1) & 3)*p01;
#endif
/*...Block 1: t00,t10,t20,t30	*/

	jt = j1;

	    rt =t10;	t10=t00-rt;	t00=t00+rt;
	    it =t11;	t11=t01-it;	t01=t01+it;
#if PFETCH
add0 = &a[jt+prefetch_offset];
prefetch_p_doubles(add0);
#endif
	    rt =t30;	t30=t20-rt;	t20=t20+rt;
	    it =t31;	t31=t21-it;	t21=t21+it;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt    ]=t00+t20;		a[jt+1    ]=t01+t21;
	    a[jt+p01]=t00-t20;		a[jt+1+p01]=t01-t21;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt+p02]=t10-t31;		a[jt+1+p02]=t11+t30;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t10+t31;		a[jt+1+p03]=t11-t30;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
/*...Block 5: t08,t18,t28,t38	*/

	jt = j1 + p04;

	    rt =t18;	t18=t08+t19;	t08=t08-t19;		/* twiddle mpy by E^4 = I	*/
			t19=t09-rt;	t09=t09+rt;
#if PFETCH
add0 = &a[jt+prefetch_offset];
prefetch_p_doubles(add0);
#endif
	    rt =(t28-t29)*ISRT2;	t29=(t28+t29)*ISRT2;		t28=rt;	/* twiddle mpy by E^8	*/
	    rt =(t39+t38)*ISRT2;	it =(t39-t38)*ISRT2;			/* twiddle mpy by -E^12 is here...	*/
	    t38=t28+rt;			t28=t28-rt;				/* ...and get E^12 by flipping signs here.	*/
	    t39=t29+it;			t29=t29-it;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt    ]=t08+t28;		a[jt+1    ]=t09+t29;
	    a[jt+p01]=t08-t28;		a[jt+1+p01]=t09-t29;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt+p02]=t18-t39;		a[jt+1+p02]=t19+t38;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t18+t39;		a[jt+1+p03]=t19-t38;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
/*...Block 3: t04,t14,t24,t34	*/

	jt = j1 + p08;

	    rt =(t14-t15)*ISRT2;	it =(t14+t15)*ISRT2;			/* twiddle mpy by E^4	*/
	    t14=t04-rt;			t04=t04+rt;
	    t15=t05-it;			t05=t05+it;
#if PFETCH
add0 = &a[jt+prefetch_offset];
prefetch_p_doubles(add0);
#endif
	    rt =t24*c - t25*s;		t25=t25*c + t24*s;		t24=rt;	/* twiddle mpy by E^2	*/
	    rt =t34*s - t35*c;		it =t35*s + t34*c;			/* twiddle mpy by E^6	*/
	    t34=t24-rt;			t24=t24+rt;
	    t35=t25-it;			t25=t25+it;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt    ]=t04+t24;		a[jt+1    ]=t05+t25;
	    a[jt+p01]=t04-t24;		a[jt+1+p01]=t05-t25;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt+p02]=t14-t35;		a[jt+1+p02]=t15+t34;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t14+t35;		a[jt+1+p03]=t15-t34;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
/*...Block 7: t0C,t1C,t2C,t3C	*/

	jt = j1 + p0C;

	    rt =(t1D+t1C)*ISRT2;	it =(t1D-t1C)*ISRT2;			/* twiddle mpy by -E^12 is here...	*/
	    t1C=t0C+rt;			t0C=t0C-rt;				/* ...and get E^12 by flipping signs here.	*/
	    t1D=t0D+it;			t0D=t0D-it;
#if PFETCH
add0 = &a[jt+prefetch_offset];
prefetch_p_doubles(add0);
#endif
	    rt =t2C*s - t2D*c;		t2D=t2D*s + t2C*c;		t2C=rt;	/* twiddle mpy by E^6	*/
	    rt =t3C*c - t3D*s;		it =t3D*c + t3C*s;			/* twiddle mpy by E^18 is here...	*/
	    t3C=t2C+rt;			t2C=t2C-rt;				/* ...and get E^18 by flipping signs here.	*/
	    t3D=t2D+it;			t2D=t2D-it;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt    ]=t0C+t2C;		a[jt+1    ]=t0D+t2D;
	    a[jt+p01]=t0C-t2C;		a[jt+1+p01]=t0D-t2D;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt+p02]=t1C-t3D;		a[jt+1+p02]=t1D+t3C;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t1C+t3D;		a[jt+1+p03]=t1D-t3C;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
/*...Block 2: t02,t12,t22,t32	*/

	jt = j1 + p10;

	    rt =t12*c - t13*s;		it =t13*c + t12*s;			/* twiddle mpy by E^2	*/
	    t12=t02-rt;			t02=t02+rt;
	    t13=t03-it;			t03=t03+it;
#if PFETCH
add0 = &a[jt+prefetch_offset];
prefetch_p_doubles(add0);
#endif
	    rt =t22*c32_1 - t23*s32_1;	t23=t23*c32_1 + t22*s32_1;	t22=rt;	/* twiddle mpy by E^1	*/
	    rt =t32*c32_3 - t33*s32_3;	it =t33*c32_3 + t32*s32_3;		/* twiddle mpy by E^3	*/
	    t32=t22-rt;			t22=t22+rt;
	    t33=t23-it;			t23=t23+it;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt    ]=t02+t22;		a[jt+1    ]=t03+t23;
	    a[jt+p01]=t02-t22;		a[jt+1+p01]=t03-t23;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt+p02]=t12-t33;		a[jt+1+p02]=t13+t32;	/* mpy by E^4=i is inlined here...;	*/
	    a[jt+p03]=t12+t33;		a[jt+1+p03]=t13-t32;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
/*...Block 6: t0A,t1A,t2A,t3A	*/

	jt = j1 + p14;

	    rt =t1A*s + t1B*c;		it =t1B*s - t1A*c;			/* twiddle mpy by -E^10 is here...	*/
	    t1A=t0A+rt;			t0A =t0A-rt;				/* ...and get E^10 by flipping signs here.	*/
	    t1B=t0B+it;			t0B =t0B-it;
#if PFETCH
add0 = &a[jt+prefetch_offset];
prefetch_p_doubles(add0);
#endif
	    rt =t2A*s32_3 - t2B*c32_3;	t2B=t2B*s32_3 + t2A*c32_3;	t2A=rt;	/* twiddle mpy by E^5	*/
	    rt =t3A*c32_1 + t3B*s32_1;	it =t3B*c32_1 - t3A*s32_1;	 	/* twiddle mpy by -E^15 is here...	*/
	    t3A=t2A+rt;			t2A=t2A-rt;				/* ...and get E^15 by flipping signs here.	*/
	    t3B=t2B+it;			t2B=t2B-it;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt    ]=t0A+t2A;		a[jt+1    ]=t0B+t2B;
	    a[jt+p01]=t0A-t2A;		a[jt+1+p01]=t0B-t2B;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt+p02]=t1A-t3B;		a[jt+1+p02]=t1B+t3A;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t1A+t3B;		a[jt+1+p03]=t1B-t3A;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
/*...Block 4: t06,t16,t26,t36	*/

	jt = j1 + p18;

	    rt =t16*s - t17*c;		it =t17*s + t16*c;			/* twiddle mpy by E^6	*/
	    t16=t06-rt;			t06 =t06+rt;
	    t17=t07-it;			t07 =t07+it;
#if PFETCH
add0 = &a[jt+prefetch_offset];
prefetch_p_doubles(add0);
#endif
	    rt =t26*c32_3 - t27*s32_3;	t27=t27*c32_3 + t26*s32_3;	t26=rt;	/* twiddle mpy by E^3	*/
	    rt =t36*s32_1 + t37*c32_1;	it =t37*s32_1 - t36*c32_1;		/* twiddle mpy by -E^9 is here...	*/
	    t36=t26+rt;			t26=t26-rt;				/* ...and get E^9 by flipping signs here.	*/
	    t37=t27+it;			t27=t27-it;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt    ]=t06+t26;		a[jt+1    ]=t07+t27;
	    a[jt+p01]=t06-t26;		a[jt+1+p01]=t07-t27;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt+p02]=t16-t37;		a[jt+1+p02]=t17+t36;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t16+t37;		a[jt+1+p03]=t17-t36;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
/*...Block 8: t0E,t1E,t2E,t3E	*/

	jt = j1 + p1C;

	    rt =t1E*c + t1F*s;		it =t1F*c - t1E*s;			/* twiddle mpy by -E^14 is here...	*/
	    t1E=t0E+rt;			t0E =t0E-rt;				/* ...and get E^14 by flipping signs here.	*/
	    t1F=t0F+it;			t0F =t0F-it;
#if PFETCH
add0 = &a[jt+prefetch_offset];
prefetch_p_doubles(add0);
#endif
	    rt =t2E*s32_1 - t2F*c32_1;	t2F=t2F*s32_1 + t2E*c32_1;	t2E=rt;	/* twiddle mpy by E^7	*/
	    rt =t3E*s32_3 - t3F*c32_3;	it =t3F*s32_3 + t3E*c32_3;		/* twiddle mpy by -E^21 is here...	*/
	    t3E=t2E+rt;			t2E=t2E-rt;				/* ...and get E^21 by flipping signs here.	*/
	    t3F=t2F+it;			t2F=t2F-it;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt    ]=t0E+t2E;		a[jt+1    ]=t0F+t2F;
	    a[jt+p01]=t0E-t2E;		a[jt+1+p01]=t0F-t2F;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt+p02]=t1E-t3F;		a[jt+1+p02]=t1F+t3E;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t1E+t3F;		a[jt+1+p03]=t1F-t3E;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	  iroot += root_incr;		/* increment sincos index.	*/

	  }

	  jstart += nwt;
	  jhi    += nwt;
	  col += 32;
	  co3 -= 32;

	}

	if(root_incr==0)break;

/*   Wraparound carry cleanup loop is here:

     The cleanup carries from the end of each length-N/32 block into the begining of the next
     can all be neatly processed as follows:

     (1) Invert the radix-32 forward DIF FFT of the first block of 32 complex elements in A and unweight;
     (2) Propagate cleanup carries among the real and imaginary parts of the 32 outputs of (1);
     (3) Reweight and perform a radix-32 forward DIF FFT on the result of (2);
     (4) If any of the exit carries from (2) are nonzero, advance to the next 32 elements and repeat (1-4).
*/
	t00 = cy1F;
	cy1F= cy1E;
	cy1E= cy1D;
	cy1D= cy1C;
	cy1C= cy1B;
	cy1B= cy1A;
	cy1A= cy19;
	cy19= cy18;
	cy18= cy17;
	cy17= cy16;
	cy16= cy15;
	cy15= cy14;
	cy14= cy13;
	cy13= cy12;
	cy12= cy11;
	cy11= cy10;
	cy10= cy0F;
	cy0F= cy0E;
	cy0E= cy0D;
	cy0D= cy0C;
	cy0C= cy0B;
	cy0B= cy0A;
	cy0A= cy09;
	cy09= cy08;
	cy08= cy07;
	cy07= cy06;
	cy06= cy05;
	cy05= cy04;
	cy04= cy03;
	cy03= cy02;
	cy02= cy01;
	cy01= cy00;
	cy00= t00;

	iroot = 0;
	root_incr = 0;
	scale = 1;

	jstart = 0;
	jhi = 3;
	khi = 1;

	for(j=0; j<=jhi; j++)
	{
	  jt = j;
	  a[jt    ] *= radix_inv;
	  a[jt+p01] *= radix_inv;
	  a[jt+p02] *= radix_inv;
	  a[jt+p03] *= radix_inv;
	  a[jt+p04] *= radix_inv;
	  a[jt+p05] *= radix_inv;
	  a[jt+p06] *= radix_inv;
	  a[jt+p07] *= radix_inv;
	  jt += p08;
	  a[jt    ] *= radix_inv;
	  a[jt+p01] *= radix_inv;
	  a[jt+p02] *= radix_inv;
	  a[jt+p03] *= radix_inv;
	  a[jt+p04] *= radix_inv;
	  a[jt+p05] *= radix_inv;
	  a[jt+p06] *= radix_inv;
	  a[jt+p07] *= radix_inv;
	  jt += p08;
	  a[jt    ] *= radix_inv;
	  a[jt+p01] *= radix_inv;
	  a[jt+p02] *= radix_inv;
	  a[jt+p03] *= radix_inv;
	  a[jt+p04] *= radix_inv;
	  a[jt+p05] *= radix_inv;
	  a[jt+p06] *= radix_inv;
	  a[jt+p07] *= radix_inv;
	  jt += p08;
	  a[jt    ] *= radix_inv;
	  a[jt+p01] *= radix_inv;
	  a[jt+p02] *= radix_inv;
	  a[jt+p03] *= radix_inv;
	  a[jt+p04] *= radix_inv;
	  a[jt+p05] *= radix_inv;
	  a[jt+p06] *= radix_inv;
	  a[jt+p07] *= radix_inv;
	}
}

	if(fabs(cy00)+fabs(cy01)+fabs(cy02)+fabs(cy03)+fabs(cy04)+fabs(cy05)+fabs(cy06)+fabs(cy07)+fabs(cy08)+fabs(cy09)+fabs(cy0A)+fabs(cy0B)+fabs(cy0C)+fabs(cy0D)+fabs(cy0E)+fabs(cy0F)
	  +fabs(cy10)+fabs(cy11)+fabs(cy12)+fabs(cy13)+fabs(cy14)+fabs(cy15)+fabs(cy16)+fabs(cy17)+fabs(cy18)+fabs(cy19)+fabs(cy1A)+fabs(cy1B)+fabs(cy1C)+fabs(cy1D)+fabs(cy1E)+fabs(cy1F)
	   != 0.0)
	{
	    if(INTERACT)printf("FATAL: iter = %10d; nonzero exit carry in radix32_ditN_cy_dif1.\n",iter);
	    f1 = fopen(   OFILE,"a");
	    f2 = fopen(STATFILE,"a");
	    fprintf(f1,"FATAL: iter = %10d; nonzero exit carry in radix32_ditN_cy_dif1.\n",iter);
	    fprintf(f2,"FATAL: iter = %10d; nonzero exit carry in radix32_ditN_cy_dif1.\n",iter);
	    fclose(f1);
	    fclose(f2);
	    err=1;
	    return(err);
	}
	*fracmax = maxerr;
	return(0);
}

/***************/

int radix32_ditN_cy_dif1_nochk(double a[], int n, int nwt, int nwt_bits, double wt0[], double wt1[], int si[], int nrt, int nrt_bits, struct complex rt0[],
                         struct complex rt1[], double base[], double baseinv[], int iter                 , int p)
{
/*
!...Acronym: DWT = Discrete Weighted Transform, DIT = Decimation In Time, DIF = Decimation In Frequency
!
!...Performs a final radix-32 complex DIT pass, an inverse DWT weighting, a carry propagation,
!   a forward DWT weighting, and an initial radix-32 complex DIF pass on the data in the length-N real vector A.
!
!   Data enter and are returned in the A-array.
!
!   See the documentation in mers_mod_square and radix16_dif_pass for further details on the array
!   storage scheme, and radix16_ditN_cy_dif1 for details on the reduced-length weights array scheme.
*/
	int n32,i,j,j1,jt,jstart,jhi,iroot,iroot7,root_incr,k1,k2,k,khi,l,outer
	  ,bjmodn00,bjmodn01,bjmodn02,bjmodn03,bjmodn04,bjmodn05,bjmodn06,bjmodn07,bjmodn08,bjmodn09,bjmodn0A,bjmodn0B,bjmodn0C,bjmodn0D,bjmodn0E,bjmodn0F
	  ,bjmodn10,bjmodn11,bjmodn12,bjmodn13,bjmodn14,bjmodn15,bjmodn16,bjmodn17,bjmodn18,bjmodn19,bjmodn1A,bjmodn1B,bjmodn1C,bjmodn1D,bjmodn1E,bjmodn1F;
	static int psave=0,bw,sw,bjmodnini,nm1,nrtm1,p01,p02,p03,p04,p05,p06,p07,p08,p0C,p10,p14,p18,p1C;
	const  double one_half[3] = {1.0, 0.5, 0.25};	/* Needed for small-weights-tables scheme */
	static double    c     = 0.92387953251128675613, s     = 0.38268343236508977173	/* exp(  i*twopi/16)	*/
			,c32_1 = 0.98078528040323044912, s32_1 = 0.19509032201612826784	/* exp(  i*twopi/32)	*/
			,c32_3 = 0.83146961230254523708, s32_3 = 0.55557023301960222473 /* exp(3*i*twopi/32)	*/
			,radix_inv, n2inv;
	double rt,it
		,t00,t01,t02,t03,t04,t05,t06,t07,t08,t09,t0A,t0B,t0C,t0D,t0E,t0F
		,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,t1A,t1B,t1C,t1D,t1E,t1F
		,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,t2A,t2B,t2C,t2D,t2E,t2F
		,t30,t31,t32,t33,t34,t35,t36,t37,t38,t39,t3A,t3B,t3C,t3D,t3E,t3F
		,a1p00r,a1p01r,a1p02r,a1p03r,a1p04r,a1p05r,a1p06r,a1p07r,a1p08r,a1p09r,a1p0Ar,a1p0Br,a1p0Cr,a1p0Dr,a1p0Er,a1p0Fr
		,a1p10r,a1p11r,a1p12r,a1p13r,a1p14r,a1p15r,a1p16r,a1p17r,a1p18r,a1p19r,a1p1Ar,a1p1Br,a1p1Cr,a1p1Dr,a1p1Er,a1p1Fr
		,a1p00i,a1p01i,a1p02i,a1p03i,a1p04i,a1p05i,a1p06i,a1p07i,a1p08i,a1p09i,a1p0Ai,a1p0Bi,a1p0Ci,a1p0Di,a1p0Ei,a1p0Fi
		,a1p10i,a1p11i,a1p12i,a1p13i,a1p14i,a1p15i,a1p16i,a1p17i,a1p18i,a1p19i,a1p1Ai,a1p1Bi,a1p1Ci,a1p1Di,a1p1Ei,a1p1Fi
		,cy00,cy01,cy02,cy03,cy04,cy05,cy06,cy07,cy08,cy09,cy0A,cy0B,cy0C,cy0D,cy0E,cy0F
		,cy10,cy11,cy12,cy13,cy14,cy15,cy16,cy17,cy18,cy19,cy1A,cy1B,cy1C,cy1D,cy1E,cy1F
		,temp,scale;
#if PFETCH
	double *add0;
	int prefetch_offset;
#endif
	int err;
	FILE *f1,*f2;
	static int first_entry=TRUE;

/*...stuff for the reduced-length DWT weights array is here:	*/
	int n_div_nwt;
	int col,co2,co3,m,m2,n_minus_sil,n_minus_silp1,sinwt,sinwtm1;
	double wt,wtinv,wtl,wtlp1,wtn,wtnm1,wtA,wtB,wtC;

/*...change n32 and n_div_wt to non-static to work around a gcc compiler bug. */
	n32   = n/32;
	n_div_nwt = n32 >> nwt_bits;

	if((n_div_nwt << nwt_bits) != n32)
	{
	  if(INTERACT)printf("FATAL: iter = %10d; NWT_BITS does not divide N/32 in radix32_ditN_cy_dif1.\n",iter);
	  f1 = fopen(   OFILE,"a");
	  f2 = fopen(STATFILE,"a");
	  fprintf(f1,"FATAL: iter = %10d; NWT_BITS does not divide N/32 in radix32_ditN_cy_dif1.\n",iter);
	  fprintf(f2,"FATAL: iter = %10d; NWT_BITS does not divide N/32 in radix32_ditN_cy_dif1.\n",iter);
	  fclose(f1);
	  fclose(f2);
	  err=1;
	  return(err);
	}

	if(p != psave)
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  psave = p;
	  first_entry=FALSE;
	  radix_inv = (double)((long double)1.0L/(long double)32.0L);
	  n2inv     = (double)((long double)1.0L/(long double)(n/2));

	  bw    = p%n;		/* Number of bigwords in the Crandall/Fagin mixed-radix representation = (Mersenne exponent) mod (vector length).	*/
	  sw    = n - bw;	/* Number of smallwords.	*/

	  nm1   = n-1;
	  nrtm1 = nrt - 1;

/*   constant index offsets for load/stores are here.	*/

	  p01 = n32;
	  p02 = p01 +p01;
	  p03 = p02 +p01;
	  p04 = p03 +p01;
	  p05 = p04 +p01;
	  p06 = p05 +p01;
	  p07 = p06 +p01;
	  p08 = p04 +p04;
	  p0C = p08 +p04;
	  p10 = p0C +p04;
	  p14 = p10 +p04;
	  p18 = p14 +p04;
	  p1C = p18 +p04;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p05 = p05 + ( (p05 >> DAT_BITS) << PAD_BITS );
	  p06 = p06 + ( (p06 >> DAT_BITS) << PAD_BITS );
	  p07 = p07 + ( (p07 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p0C = p0C + ( (p0C >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p14 = p14 + ( (p14 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	  p1C = p1C + ( (p1C >> DAT_BITS) << PAD_BITS );

	  bjmodnini=0;
	  for(j=0; j < n32; j++)
	  {
	    bjmodnini -= sw; bjmodnini = bjmodnini + ( (-(int)((uint32)bjmodnini >> 31)) & n);
	  }
	}

/*...The radix-32 final DIT pass is here.	*/

	cy00 =-2;		/* init carries	*/
	cy01 = 0;
	cy02 = 0;
	cy03 = 0;
	cy04 = 0;
	cy05 = 0;
	cy06 = 0;
	cy07 = 0;
	cy08 = 0;
	cy09 = 0;
	cy0A = 0;
	cy0B = 0;
	cy0C = 0;
	cy0D = 0;
	cy0E = 0;
	cy0F = 0;
	cy10 = 0;
	cy11 = 0;
	cy12 = 0;
	cy13 = 0;
	cy14 = 0;
	cy15 = 0;
	cy16 = 0;
	cy17 = 0;
	cy18 = 0;
	cy19 = 0;
	cy1A = 0;
	cy1B = 0;
	cy1C = 0;
	cy1D = 0;
	cy1E = 0;
	cy1F = 0;

	iroot = 0;	/* init sincos array index	*/
	root_incr = 1;	/* init sincos array index increment (set = 1 for normal carry pass, = 0 for wrapper pass)	*/
	scale = n2inv;	/* init inverse-weight scale factor  (set = 2/n for normal carry pass, = 1 for wrapper pass)	*/

	jstart = 0;
	jhi = jstart+nwt-1;
	khi = n_div_nwt;

for(outer=0; outer <= 1; outer++)
{
	i = 1;		/* Pointer to the BASE and BASEINV arrays. lowest-order digit is always a bigword (i = 1).	*/

	bjmodn00 = 0;
	bjmodn01 = bjmodnini;
	bjmodn02 = bjmodn01 + bjmodnini - n;	bjmodn02 = bjmodn02 + ( (-(int)((uint32)bjmodn02 >> 31)) & n);
	bjmodn03 = bjmodn02 + bjmodnini - n;	bjmodn03 = bjmodn03 + ( (-(int)((uint32)bjmodn03 >> 31)) & n);
	bjmodn04 = bjmodn03 + bjmodnini - n;	bjmodn04 = bjmodn04 + ( (-(int)((uint32)bjmodn04 >> 31)) & n);
	bjmodn05 = bjmodn04 + bjmodnini - n;	bjmodn05 = bjmodn05 + ( (-(int)((uint32)bjmodn05 >> 31)) & n);
	bjmodn06 = bjmodn05 + bjmodnini - n;	bjmodn06 = bjmodn06 + ( (-(int)((uint32)bjmodn06 >> 31)) & n);
	bjmodn07 = bjmodn06 + bjmodnini - n;	bjmodn07 = bjmodn07 + ( (-(int)((uint32)bjmodn07 >> 31)) & n);
	bjmodn08 = bjmodn07 + bjmodnini - n;	bjmodn08 = bjmodn08 + ( (-(int)((uint32)bjmodn08 >> 31)) & n);
	bjmodn09 = bjmodn08 + bjmodnini - n;	bjmodn09 = bjmodn09 + ( (-(int)((uint32)bjmodn09 >> 31)) & n);
	bjmodn0A = bjmodn09 + bjmodnini - n;	bjmodn0A = bjmodn0A + ( (-(int)((uint32)bjmodn0A >> 31)) & n);
	bjmodn0B = bjmodn0A + bjmodnini - n;	bjmodn0B = bjmodn0B + ( (-(int)((uint32)bjmodn0B >> 31)) & n);
	bjmodn0C = bjmodn0B + bjmodnini - n;	bjmodn0C = bjmodn0C + ( (-(int)((uint32)bjmodn0C >> 31)) & n);
	bjmodn0D = bjmodn0C + bjmodnini - n;	bjmodn0D = bjmodn0D + ( (-(int)((uint32)bjmodn0D >> 31)) & n);
	bjmodn0E = bjmodn0D + bjmodnini - n;	bjmodn0E = bjmodn0E + ( (-(int)((uint32)bjmodn0E >> 31)) & n);
	bjmodn0F = bjmodn0E + bjmodnini - n;	bjmodn0F = bjmodn0F + ( (-(int)((uint32)bjmodn0F >> 31)) & n);
	bjmodn10 = bjmodn0F + bjmodnini - n;	bjmodn10 = bjmodn10 + ( (-(int)((uint32)bjmodn10 >> 31)) & n);
	bjmodn11 = bjmodn10 + bjmodnini - n;	bjmodn11 = bjmodn11 + ( (-(int)((uint32)bjmodn11 >> 31)) & n);
	bjmodn12 = bjmodn11 + bjmodnini - n;	bjmodn12 = bjmodn12 + ( (-(int)((uint32)bjmodn12 >> 31)) & n);
	bjmodn13 = bjmodn12 + bjmodnini - n;	bjmodn13 = bjmodn13 + ( (-(int)((uint32)bjmodn13 >> 31)) & n);
	bjmodn14 = bjmodn13 + bjmodnini - n;	bjmodn14 = bjmodn14 + ( (-(int)((uint32)bjmodn14 >> 31)) & n);
	bjmodn15 = bjmodn14 + bjmodnini - n;	bjmodn15 = bjmodn15 + ( (-(int)((uint32)bjmodn15 >> 31)) & n);
	bjmodn16 = bjmodn15 + bjmodnini - n;	bjmodn16 = bjmodn16 + ( (-(int)((uint32)bjmodn16 >> 31)) & n);
	bjmodn17 = bjmodn16 + bjmodnini - n;	bjmodn17 = bjmodn17 + ( (-(int)((uint32)bjmodn17 >> 31)) & n);
	bjmodn18 = bjmodn17 + bjmodnini - n;	bjmodn18 = bjmodn18 + ( (-(int)((uint32)bjmodn18 >> 31)) & n);
	bjmodn19 = bjmodn18 + bjmodnini - n;	bjmodn19 = bjmodn19 + ( (-(int)((uint32)bjmodn19 >> 31)) & n);
	bjmodn1A = bjmodn19 + bjmodnini - n;	bjmodn1A = bjmodn1A + ( (-(int)((uint32)bjmodn1A >> 31)) & n);
	bjmodn1B = bjmodn1A + bjmodnini - n;	bjmodn1B = bjmodn1B + ( (-(int)((uint32)bjmodn1B >> 31)) & n);
	bjmodn1C = bjmodn1B + bjmodnini - n;	bjmodn1C = bjmodn1C + ( (-(int)((uint32)bjmodn1C >> 31)) & n);
	bjmodn1D = bjmodn1C + bjmodnini - n;	bjmodn1D = bjmodn1D + ( (-(int)((uint32)bjmodn1D >> 31)) & n);
	bjmodn1E = bjmodn1D + bjmodnini - n;	bjmodn1E = bjmodn1E + ( (-(int)((uint32)bjmodn1E >> 31)) & n);
	bjmodn1F = bjmodn1E + bjmodnini - n;	bjmodn1F = bjmodn1F + ( (-(int)((uint32)bjmodn1F >> 31)) & n);

	col=0;
	co2=(n >> nwt_bits)-1+32;
	co3=co2-32;		/* At the start of each new j-loop, co3=co2-radix(1)	*/

	for(k=1; k <= khi; k++)	/* Do n/(radix(1)*nwt) outer loop executions...	*/
	{
	  for(j=jstart; j<jhi; j += 2)	/* Each inner loop execution processes (radix(1)*nwt) array data.	*/
	  {
	    j1 = j + ( (j >> DAT_BITS) << PAD_BITS );	/* padded-array fetch index is here	*/

#if !USE_NEW_DIT
/*	here are the needed sincos data - these will be processed below in bit-reversed order.	*/

	    l = iroot;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      l += iroot;			/* 2*iroot	*/
	      t00=rt0[k1].re;	t01=rt0[k1].im;
	      rt =rt1[k2].re;	it =rt1[k2].im;
	      t20=t00*rt -t01*it ;	t21=t00*it +t01*rt;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      l += iroot;			/* 3*iroot	*/
	      t00=rt0[k1].re;	t01=rt0[k1].im;
	      rt =rt1[k2].re;	it =rt1[k2].im;
	      t10=t00*rt -t01*it ;	t11=t00*it +t01*rt;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      l += 4*iroot;			/* 7*iroot	*/
	      iroot7 = l;			/* 7*iroot	*/
	      t00=rt0[k1].re;	t01=rt0[k1].im;
	      rt =rt1[k2].re;	it =rt1[k2].im;
	      t30=t00*rt -t01*it ;	t31=t00*it +t01*rt;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      l += iroot7;			/* 14*iroot	*/
	      t00=rt0[k1].re;	t01=rt0[k1].im;
	      rt =rt1[k2].re;	it =rt1[k2].im;
	      t38=t00*rt -t01*it ;	t39=t00*it +t01*rt;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      l += iroot7;			/* 21*iroot	*/
	      t00=rt0[k1].re;	t01=rt0[k1].im;
	      rt =rt1[k2].re;	it =rt1[k2].im;
	      t1C=t00*rt -t01*it ;	t1D=t00*it +t01*rt;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      l += iroot7;			/* 28*iroot	*/
	      t00=rt0[k1].re;	t01=rt0[k1].im;
	      rt =rt1[k2].re;	it =rt1[k2].im;
	      t2A=t00*rt -t01*it ;	t2B=t00*it +t01*rt;

	      k1=(l & nrtm1);
	      k2=(l >> nrt_bits);
	      t00=rt0[k1].re;	t01=rt0[k1].im;
	      rt =rt1[k2].re;	it =rt1[k2].im;
	      t0E=t00*rt -t01*it ;	t0F=t00*it +t01*rt;

/*(c,s)4-10 (decimal indices), 4-0A (hexadecimal indices):	*/
	      t00=t20*t38; t01=t20*t39; rt =t21*t38; it =t21*t39;
	      t18=t00+it ; t19=t01-rt ; t04=t00-it ; t05=t01+rt ;

	      t00=t10*t38; t01=t10*t39; rt =t11*t38; it =t11*t39;
	      t28=t00+it ; t29=t01-rt ; t24=t00-it ; t25=t01+rt ;

	      t00=t30*t38; t01=t30*t39; rt =t31*t38; it =t31*t39;
	      t08=t00+it ; t09=t01-rt ; t14=t00-it ; t15=t01+rt ;

/*(c,s)11-17 (decimal indices), 0B-11 (hexadecimal indices):;	*/
	      t00=t20*t1C; t01=t20*t1D; rt =t21*t1C; it =t21*t1D;
	      t2C=t00+it ; t2D=t01-rt ; t3C=t00-it ; t3D=t01+rt ;

	      t00=t10*t1C; t01=t10*t1D; rt =t11*t1C; it =t11*t1D;
	      t0C=t00+it ; t0D=t01-rt ; t02=t00-it ; t03=t01+rt ;

	      t00=t30*t1C; t01=t30*t1D; rt =t31*t1C; it =t31*t1D;
	      t34=t00+it ; t35=t01-rt ; t22=t00-it ; t23=t01+rt ;

/*(c,s)18-24 (decimal indices), 12-18 (hexadecimal indices):	*/
	      t00=t20*t2A; t01=t20*t2B; rt =t21*t2A; it =t21*t2B;
	      t0A=t00+it ; t0B=t01-rt ; t1A=t00-it ; t1B=t01+rt ;

	      t00=t10*t2A; t01=t10*t2B; rt =t11*t2A; it =t11*t2B;
	      t32=t00+it ; t33=t01-rt ; t3A=t00-it ; t3B=t01+rt ;

	      t00=t30*t2A; t01=t30*t2B; rt =t31*t2A; it =t31*t2B;
	      t12=t00+it ; t13=t01-rt ; t06=t00-it ; t07=t01+rt ;

/*(c,s)25-31 (decimal indices), 19-1F (hexadecimal indices):	*/
	      t00=t20*t0E; t01=t20*t0F; rt =t21*t0E; it =t21*t0F;
	      t36=t00+it ; t37=t01-rt ; t2E=t00-it ; t2F=t01+rt ;

	      t00=t10*t0E; t01=t10*t0F; rt =t11*t0E; it =t11*t0F;
	      t16=t00+it ; t17=t01-rt ; t1E=t00-it ; t1F=t01+rt ;

	      t00=t30*t0E; t01=t30*t0F; rt =t31*t0E; it =t31*t0F;
	      t26=t00+it ; t27=t01-rt ; t3E=t00-it ; t3F=t01+rt ;
#endif
/*       gather the needed data (32 64-bit complex, i.e. 64 64-bit reals) and do the first set of four length-8 transforms...	*/

/*...Block 1:	*/

	jt = j1;

#if USE_NEW_DIT
	    t00=a[jt    ];	t01=a[jt+1    ];
	    rt =a[jt+p01];	it =a[jt+1+p01];
#else
	    t00=a[jt    ];				t01=a[jt+1    ];
	    rt =a[jt+p01]*t02 + a[jt+1+p01]*t03;	it =a[jt+1+p01]*t02 - a[jt+p01]*t03;
#endif
	    t02=t00-rt;		t03=t01-it;
	    t00=t00+rt;		t01=t01+it;

#if USE_NEW_DIT
	    t04=a[jt+p02];	t05=a[jt+1+p02];
	    rt =a[jt+p03];	it =a[jt+1+p03];
#else
	    rt =a[jt+p02]*t04 + a[jt+1+p02]*t05;	t05=a[jt+1+p02]*t04 - a[jt+p02]*t05;	t04=rt;
	    rt =a[jt+p03]*t06 + a[jt+1+p03]*t07;	it =a[jt+1+p03]*t06 - a[jt+p03]*t07;
#endif
	    t06=t04-rt;		t07=t05-it;
	    t04=t04+rt;		t05=t05+it;

	    rt =t04;		it =t05;
	    t04=t00-rt;		t05=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t06;		it =t07;
	    t06=t02-it;		t07=t03+rt;
	    t02=t02+it;		t03=t03-rt;

#if USE_NEW_DIT
	    t08=a[jt+p04];	t09=a[jt+1+p04];
	    rt =a[jt+p05];	it =a[jt+1+p05];
#else
	    rt =a[jt+p04]*t08 + a[jt+1+p04]*t09;	t09=a[jt+1+p04]*t08 - a[jt+p04]*t09;	t08=rt;
	    rt =a[jt+p05]*t0A + a[jt+1+p05]*t0B;	it =a[jt+1+p05]*t0A - a[jt+p05]*t0B;
#endif
	    t0A=t08-rt;		t0B=t09-it;
	    t08=t08+rt;		t09=t09+it;

#if USE_NEW_DIT
	    t0C=a[jt+p06];	t0D=a[jt+1+p06];
	    rt =a[jt+p07];	it =a[jt+1+p07];
#else
	    rt =a[jt+p06]*t0C + a[jt+1+p06]*t0D;	t0D=a[jt+1+p06]*t0C - a[jt+p06]*t0D;	t0C=rt;
	    rt =a[jt+p07]*t0E + a[jt+1+p07]*t0F;	it =a[jt+1+p07]*t0E - a[jt+p07]*t0F;
#endif
	    t0E=t0C-rt;		t0F=t0D-it;
	    t0C=t0C+rt;		t0D=t0D+it;

	    rt =t0C;		it =t0D;
	    t0C=t08-rt;		t0D=t09-it;
	    t08=t08+rt;		t09=t09+it;

	    rt =t0E;		it =t0F;
	    t0E=t0A-it;		t0F=t0B+rt;
	    t0A=t0A+it;		t0B=t0B-rt;

	    rt =t08;		it =t09;
	    t08=t00-rt;		t09=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t0C;		it =t0D;
	    t0C=t04-it;		t0D=t05+rt;
	    t04=t04+it;		t05=t05-rt;

	    rt =(t0A+t0B)*ISRT2;it =(t0A-t0B)*ISRT2;
	    t0A=t02-rt;		t0B=t03+it;
	    t02=t02+rt;		t03=t03-it;

	    rt =(t0E-t0F)*ISRT2;it =(t0F+t0E)*ISRT2;
	    t0E=t06+rt;		t0F=t07+it;
	    t06=t06-rt;		t07=t07-it;

/*...Block 2:;	*/

	jt = j1 + p08;

#if USE_NEW_DIT
	    t10=a[jt    ];	t11=a[jt+1    ];
	    rt =a[jt+p01];	it =a[jt+1+p01];
#else
	    rt =a[jt    ]*t10 + a[jt+1    ]*t11;	t11=a[jt+1    ]*t10 - a[jt    ]*t11;	t10=rt;
	    rt =a[jt+p01]*t12 + a[jt+1+p01]*t13;	it =a[jt+1+p01]*t12 - a[jt+p01]*t13;
#endif
	    t12=t10-rt;		t13=t11-it;
	    t10=t10+rt;		t11=t11+it;

#if USE_NEW_DIT
	    t14=a[jt+p02];	t15=a[jt+1+p02];
	    rt =a[jt+p03];	it =a[jt+1+p03];
#else
	    rt =a[jt+p02]*t14 + a[jt+1+p02]*t15;	t15=a[jt+1+p02]*t14 - a[jt+p02]*t15;	t14=rt;
	    rt =a[jt+p03]*t16 + a[jt+1+p03]*t17;	it =a[jt+1+p03]*t16 - a[jt+p03]*t17;
#endif
	    t16=t14-rt;		t17=t15-it;
	    t14=t14+rt;		t15=t15+it;

	    rt =t14;		it =t15;
	    t14=t10-rt;		t15=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t16;		it =t17;
	    t16=t12-it;		t17=t13+rt;
	    t12=t12+it;		t13=t13-rt;

#if USE_NEW_DIT
	    t18=a[jt+p04];	t19=a[jt+1+p04];
	    rt =a[jt+p05];	it =a[jt+1+p05];
#else
	    rt =a[jt+p04]*t18 + a[jt+1+p04]*t19;	t19=a[jt+1+p04]*t18 - a[jt+p04]*t19;	t18=rt;
	    rt =a[jt+p05]*t1A + a[jt+1+p05]*t1B;	it =a[jt+1+p05]*t1A - a[jt+p05]*t1B;
#endif
	    t1A=t18-rt;		t1B=t19-it;
	    t18=t18+rt;		t19=t19+it;

#if USE_NEW_DIT
	    t1C=a[jt+p06];	t1D=a[jt+1+p06];
	    rt =a[jt+p07];	it =a[jt+1+p07];
#else
	    rt =a[jt+p06]*t1C + a[jt+1+p06]*t1D;	t1D=a[jt+1+p06]*t1C - a[jt+p06]*t1D;	t1C=rt;
	    rt =a[jt+p07]*t1E + a[jt+1+p07]*t1F;	it =a[jt+1+p07]*t1E - a[jt+p07]*t1F;
#endif
	    t1E=t1C-rt;		t1F=t1D-it;
	    t1C=t1C+rt;		t1D=t1D+it;

	    rt =t1C;		it =t1D;
	    t1C=t18-rt;		t1D=t19-it;
	    t18=t18+rt;		t19=t19+it;

	    rt =t1E;		it =t1F;
	    t1E=t1A-it;		t1F=t1B+rt;
	    t1A=t1A+it;		t1B=t1B-rt;

	    rt =t18;		it =t19;
	    t18=t10-rt;		t19=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t1C;		it =t1D;
	    t1C=t14-it;		t1D=t15+rt;
	    t14=t14+it;		t15=t15-rt;

	    rt =(t1A+t1B)*ISRT2;it =(t1A-t1B)*ISRT2;
	    t1A=t12-rt;		t1B=t13+it;
	    t12=t12+rt;		t13=t13-it;

	    rt =(t1E-t1F)*ISRT2;it =(t1F+t1E)*ISRT2;
	    t1E=t16+rt;		t1F=t17+it;
	    t16=t16-rt;		t17=t17-it;

/*...Block 3:	*/

	jt = j1 + p10;

#if USE_NEW_DIT
	    t20=a[jt    ];	t21=a[jt+1    ];
	    rt =a[jt+p01];	it =a[jt+1+p01];
#else
	    rt =a[jt    ]*t20 + a[jt+1    ]*t21;	t21=a[jt+1    ]*t20 - a[jt    ]*t21;	t20=rt;
	    rt =a[jt+p01]*t22 + a[jt+1+p01]*t23;	it =a[jt+1+p01]*t22 - a[jt+p01]*t23;
#endif
	    t22=t20-rt;		t23=t21-it;
	    t20=t20+rt;		t21=t21+it;

#if USE_NEW_DIT
	    t24=a[jt+p02];	t25=a[jt+1+p02];
	    rt =a[jt+p03];	it =a[jt+1+p03];
#else
	    rt =a[jt+p02]*t24 + a[jt+1+p02]*t25;	t25=a[jt+1+p02]*t24 - a[jt+p02]*t25;	t24=rt;
	    rt =a[jt+p03]*t26 + a[jt+1+p03]*t27;	it =a[jt+1+p03]*t26 - a[jt+p03]*t27;
#endif
	    t26=t24-rt;		t27=t25-it;
	    t24=t24+rt;		t25=t25+it;

	    rt =t24;		it =t25;
	    t24=t20-rt;		t25=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t26;		it =t27;
	    t26=t22-it;		t27=t23+rt;
	    t22=t22+it;		t23=t23-rt;

#if USE_NEW_DIT
	    t28=a[jt+p04];	t29=a[jt+1+p04];
	    rt =a[jt+p05];	it =a[jt+1+p05];
#else
	    rt =a[jt+p04]*t28 + a[jt+1+p04]*t29;	t29=a[jt+1+p04]*t28 - a[jt+p04]*t29;	t28=rt;
	    rt =a[jt+p05]*t2A + a[jt+1+p05]*t2B;	it =a[jt+1+p05]*t2A - a[jt+p05]*t2B;
#endif
	    t2A=t28-rt;		t2B=t29-it;
	    t28=t28+rt;		t29=t29+it;

#if USE_NEW_DIT
	    t2C=a[jt+p06];	t2D=a[jt+1+p06];
	    rt =a[jt+p07];	it =a[jt+1+p07];
#else
	    rt =a[jt+p06]*t2C + a[jt+1+p06]*t2D;	t2D=a[jt+1+p06]*t2C - a[jt+p06]*t2D;	t2C=rt;
	    rt =a[jt+p07]*t2E + a[jt+1+p07]*t2F;	it =a[jt+1+p07]*t2E - a[jt+p07]*t2F;
#endif
	    t2E=t2C-rt;		t2F=t2D-it;
	    t2C=t2C+rt;		t2D=t2D+it;

	    rt =t2C;		it =t2D;
	    t2C=t28-rt;		t2D=t29-it;
	    t28=t28+rt;		t29=t29+it;

	    rt =t2E;		it =t2F;
	    t2E=t2A-it;		t2F=t2B+rt;
	    t2A=t2A+it;		t2B=t2B-rt;

	    rt =t28;		it =t29;
	    t28=t20-rt;		t29=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t2C;		it =t2D;
	    t2C=t24-it;		t2D=t25+rt;
	    t24=t24+it;		t25=t25-rt;

	    rt =(t2A+t2B)*ISRT2;it =(t2A-t2B)*ISRT2;
	    t2A=t22-rt;		t2B=t23+it;
	    t22=t22+rt;		t23=t23-it;

	    rt =(t2E-t2F)*ISRT2;it =(t2F+t2E)*ISRT2;
	    t2E=t26+rt;		t2F=t27+it;
	    t26=t26-rt;		t27=t27-it;

/*...Block 4:	*/

	jt = j1 + p18;

#if USE_NEW_DIT
	    t30=a[jt    ];	t31=a[jt+1    ];
	    rt =a[jt+p01];	it =a[jt+1+p01];
#else
	    rt =a[jt    ]*t30 + a[jt+1    ]*t31;	t31=a[jt+1    ]*t30 - a[jt    ]*t31;	t30=rt;
	    rt =a[jt+p01]*t32 + a[jt+1+p01]*t33;	it =a[jt+1+p01]*t32 - a[jt+p01]*t33;
#endif
	    t32=t30-rt;		t33=t31-it;
	    t30=t30+rt;		t31=t31+it;

#if USE_NEW_DIT
	    t34=a[jt+p02];	t35=a[jt+1+p02];
	    rt =a[jt+p03];	it =a[jt+1+p03];
#else
	    rt =a[jt+p02]*t34 + a[jt+1+p02]*t35;	t35=a[jt+1+p02]*t34 - a[jt+p02]*t35;	t34=rt;
	    rt =a[jt+p03]*t36 + a[jt+1+p03]*t37;	it =a[jt+1+p03]*t36 - a[jt+p03]*t37;
#endif
	    t36=t34-rt;		t37=t35-it;
	    t34=t34+rt;		t35=t35+it;

	    rt =t34;		it =t35;
	    t34=t30-rt;		t35=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t36;		it =t37;
	    t36=t32-it;		t37=t33+rt;
	    t32=t32+it;		t33=t33-rt;

#if USE_NEW_DIT
	    t38=a[jt+p04];	t39=a[jt+1+p04];
	    rt =a[jt+p05];	it =a[jt+1+p05];
#else
	    rt =a[jt+p04]*t38 + a[jt+1+p04]*t39;	t39=a[jt+1+p04]*t38 - a[jt+p04]*t39;	t38=rt;
	    rt =a[jt+p05]*t3A + a[jt+1+p05]*t3B;	it =a[jt+1+p05]*t3A - a[jt+p05]*t3B;
#endif
	    t3A=t38-rt;		t3B=t39-it;
	    t38=t38+rt;		t39=t39+it;

#if USE_NEW_DIT
	    t3C=a[jt+p06];	t3D=a[jt+1+p06];
	    rt =a[jt+p07];	it =a[jt+1+p07];
#else
	    rt =a[jt+p06]*t3C + a[jt+1+p06]*t3D;	t3D=a[jt+1+p06]*t3C - a[jt+p06]*t3D;	t3C=rt;
	    rt =a[jt+p07]*t3E + a[jt+1+p07]*t3F;	it =a[jt+1+p07]*t3E - a[jt+p07]*t3F;
#endif
	    t3E=t3C-rt;		t3F=t3D-it;
	    t3C=t3C+rt;		t3D=t3D+it;

	    rt =t3C;		it =t3D;
	    t3C=t38-rt;		t3D=t39-it;
	    t38=t38+rt;		t39=t39+it;

	    rt =t3E;		it =t3F;
	    t3E=t3A-it;		t3F=t3B+rt;
	    t3A=t3A+it;		t3B=t3B-rt;

	    rt =t38;		it =t39;
	    t38=t30-rt;		t39=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t3C;		it =t3D;
	    t3C=t34-it;		t3D=t35+rt;
	    t34=t34+it;		t35=t35-rt;

	    rt =(t3A+t3B)*ISRT2;it =(t3A-t3B)*ISRT2;
	    t3A=t32-rt;		t3B=t33+it;
	    t32=t32+rt;		t33=t33-it;

	    rt =(t3E-t3F)*ISRT2;it =(t3F+t3E)*ISRT2;
	    t3E=t36+rt;		t3F=t37+it;
	    t36=t36-rt;		t37=t37-it;

/*...and now do eight radix-4 transforms, including the internal twiddle factors:
  	1, exp(-i* 1*twopi/32) =       ( c32_1,-s32_1), exp(-i* 2*twopi/32) =       ( c    ,-s    ), exp(-i* 3*twopi/32) =       ( c32_3,-s32_3) (for inputs to transform block 2),
  	1, exp(-i* 2*twopi/32) =       ( c    ,-s    ), exp(-i* 4*twopi/32) = ISRT2*( 1    ,-1    ), exp(-i* 3*twopi/32) =       ( s    ,-c    ) (for inputs to transform block 3),
  	1, exp(-i* 3*twopi/32) =       ( c32_3,-s32_3), exp(-i* 6*twopi/32) =       ( s    ,-c    ), exp(-i* 9*twopi/32) =       (-s32_1,-c32_1) (for inputs to transform block 4),
  	1, exp(-i* 4*twopi/32) = ISRT2*( 1    ,-1    ), exp(-i* 8*twopi/32) =       ( 0    ,-1    ), exp(-i*12*twopi/32) = ISRT2*(-1    ,-1    ) (for inputs to transform block 5),
  	1, exp(-i* 5*twopi/32) =       ( s32_3,-c32_3), exp(-i*10*twopi/32) =       (-s    ,-c    ), exp(-i*15*twopi/32) =       (-c32_1,-s32_1) (for inputs to transform block 6),
  	1, exp(-i* 6*twopi/32) =       ( s    ,-c    ), exp(-i*12*twopi/32) = ISRT2*(-1    ,-1    ), exp(-i*18*twopi/32) =       (-c    , s    ) (for inputs to transform block 7),
  	1, exp(-i* 7*twopi/32) =       ( s32_1,-c32_1), exp(-i*14*twopi/32) =       (-c    ,-s    ), exp(-i*21*twopi/32) =       (-s32_3, c32_3) (for inputs to transform block 8),
     and only the last 3 inputs to each of the radix-4 transforms 2 through 8 are multiplied by non-unity twiddles.	*/

/*...Block 1: t00,t10,t20,t30	*/

	jt = j1;

	    rt =t10;	t10=t00-rt;	t00=t00+rt;
	    it =t11;	t11=t01-it;	t01=t01+it;

	    rt =t30;	t30=t20-rt;	t20=t20+rt;
	    it =t31;	t31=t21-it;	t21=t21+it;

	    a1p00r=t00+t20;		a1p00i=t01+t21;
	    a1p10r=t00-t20;		a1p10i=t01-t21;
	    a1p08r=t10+t31;		a1p08i=t11-t30;	/* mpy by E^-4 = -I is inlined here...	*/
	    a1p18r=t10-t31;		a1p18i=t11+t30;

/*...Block 5: t08,t18,t28,t38	*/

	jt = j1 + p04;

	    rt =t18;	t18=t08-t19;	t08=t08+t19;		/* twiddle mpy by E^8 =-I	*/
			t19=t09+rt;	t09=t09-rt;

	    rt =(t29+t28)*ISRT2;	t29=(t29-t28)*ISRT2;		t28=rt;	/* twiddle mpy by E^-4	*/
	    rt =(t38-t39)*ISRT2;	it =(t38+t39)*ISRT2;			/* twiddle mpy by E^4 = -E^-12 is here...	*/
	    t38=t28+rt;			t28=t28-rt;				/* ...and get E^-12 by flipping signs here.	*/
	    t39=t29+it;			t29=t29-it;

	    a1p04r=t08+t28;		a1p04i=t09+t29;
	    a1p14r=t08-t28;		a1p14i=t09-t29;
	    a1p0Cr=t18+t39;		a1p0Ci=t19-t38;	/* mpy by E^-4 = -I is inlined here...	*/
	    a1p1Cr=t18-t39;		a1p1Ci=t19+t38;

/*...Block 3: t04,t14,t24,t34	*/

	jt = j1 + p02;

	    rt =(t15+t14)*ISRT2;	it =(t15-t14)*ISRT2;			/* twiddle mpy by E^-4	*/
	    t14=t04-rt;			t04=t04+rt;
	    t15=t05-it;			t05=t05+it;

	    rt =t24*c + t25*s;		t25=t25*c - t24*s;		t24=rt;	/* twiddle mpy by E^-2	*/
	    rt =t34*s + t35*c;		it =t35*s - t34*c;			/* twiddle mpy by E^-6	*/
	    t34=t24-rt;			t24=t24+rt;
	    t35=t25-it;			t25=t25+it;

	    a1p02r=t04+t24;		a1p02i=t05+t25;
	    a1p12r=t04-t24;		a1p12i=t05-t25;
	    a1p0Ar=t14+t35;		a1p0Ai=t15-t34;	/* mpy by E^-4 = -I is inlined here...	*/
	    a1p1Ar=t14-t35;		a1p1Ai=t15+t34;

/*...Block 7: t0C,t1C,t2C,t3C	*/

	jt = j1 + p06;

	    rt =(t1C-t1D)*ISRT2;	it =(t1C+t1D)*ISRT2;			/* twiddle mpy by E^4 = -E^-12 is here...	*/
	    t1C=t0C+rt;			t0C=t0C-rt;				/* ...and get E^-12 by flipping signs here.	*/
	    t1D=t0D+it;			t0D=t0D-it;

	    rt =t2C*s + t2D*c;		t2D=t2D*s - t2C*c;		t2C=rt;	/* twiddle mpy by E^-6	*/
	    rt =t3C*c + t3D*s;		it =t3D*c - t3C*s;			/* twiddle mpy by E^-18 is here...	*/
	    t3C=t2C+rt;			t2C=t2C-rt;				/* ...and get E^-18 by flipping signs here.	*/
	    t3D=t2D+it;			t2D=t2D-it;

	    a1p06r=t0C+t2C;		a1p06i=t0D+t2D;
	    a1p16r=t0C-t2C;		a1p16i=t0D-t2D;
	    a1p0Er=t1C+t3D;		a1p0Ei=t1D-t3C;	/* mpy by E^-4 = -I is inlined here...	*/
	    a1p1Er=t1C-t3D;		a1p1Ei=t1D+t3C;

/*...Block 2: t02,t12,t22,t32	*/

	jt = j1 + p01;

	    rt =t12*c + t13*s;		it =t13*c - t12*s;			/* twiddle mpy by E^-2	*/
	    t12=t02-rt;			t02=t02+rt;
	    t13=t03-it;			t03=t03+it;

	    rt =t22*c32_1 + t23*s32_1;	t23=t23*c32_1 - t22*s32_1;	t22=rt;	/* twiddle mpy by E^-1	*/
	    rt =t32*c32_3 + t33*s32_3;	it =t33*c32_3 - t32*s32_3;		/* twiddle mpy by E^-3	*/
	    t32=t22-rt;			t22=t22+rt;
	    t33=t23-it;			t23=t23+it;

	    a1p01r=t02+t22;		a1p01i=t03+t23;
	    a1p11r=t02-t22;		a1p11i=t03-t23;
	    a1p09r=t12+t33;		a1p09i=t13-t32;	/* mpy by E^-4 = -I is inlined here...	*/
	    a1p19r=t12-t33;		a1p19i=t13+t32;

/*...Block 6: t0A,t1A,t2A,t3A	*/

	jt = j1 + p05;

	    rt =t1A*s - t1B*c;		it =t1B*s + t1A*c;			/* twiddle mpy by -E^-10 is here...	*/
	    t1A=t0A+rt;			t0A =t0A-rt;				/* ...and get E^-10 by flipping signs here.	*/
	    t1B=t0B+it;			t0B =t0B-it;

	    rt =t2A*s32_3 + t2B*c32_3;	t2B=t2B*s32_3 - t2A*c32_3;	t2A=rt;	/* twiddle mpy by E^-5	*/
	    rt =t3A*c32_1 - t3B*s32_1;	it =t3B*c32_1 + t3A*s32_1;		/* twiddle mpy by -E^-15 is here...	*/
	    t3A=t2A+rt;			t2A=t2A-rt;				/* ...and get E^-15 by flipping signs here.	*/
	    t3B=t2B+it;			t2B=t2B-it;

	    a1p05r=t0A+t2A;		a1p05i=t0B+t2B;
	    a1p15r=t0A-t2A;		a1p15i=t0B-t2B;
	    a1p0Dr=t1A+t3B;		a1p0Di=t1B-t3A;	/* mpy by E^-4 = -I is inlined here...	*/
	    a1p1Dr=t1A-t3B;		a1p1Di=t1B+t3A;

/*...Block 4: t06,t16,t26,t36	*/

	jt = j1 + p03;

	    rt =t16*s + t17*c;		it =t17*s - t16*c;			/* twiddle mpy by E^-6	*/
	    t16=t06-rt;			t06 =t06+rt;
	    t17=t07-it;			t07 =t07+it;

	    rt =t26*c32_3 + t27*s32_3;	t27=t27*c32_3 - t26*s32_3;	t26=rt;	/* twiddle mpy by E^-3	*/
	    rt =t36*s32_1 - t37*c32_1;	it =t37*s32_1 + t36*c32_1;		/* twiddle mpy by -E^-9 is here...	*/
	    t36=t26+rt;			t26=t26-rt;				/* ...and get E^-9 by flipping signs here.	*/
	    t37=t27+it;			t27=t27-it;

	    a1p03r=t06+t26;		a1p03i=t07+t27;
	    a1p13r=t06-t26;		a1p13i=t07-t27;
	    a1p0Br=t16+t37;		a1p0Bi=t17-t36;	/* mpy by E^-4 = -I is inlined here...	*/
	    a1p1Br=t16-t37;		a1p1Bi=t17+t36;

/*...Block 8: t0E,t1E,t2E,t3E	*/

	jt = j1 + p07;

	    rt =t1E*c - t1F*s;		it =t1F*c + t1E*s;			/* twiddle mpy by -E^-14 is here...	*/
	    t1E=t0E+rt;			t0E =t0E-rt;				/* ...and get E^-14 by flipping signs here.	*/
	    t1F=t0F+it;			t0F =t0F-it;

	    rt =t2E*s32_1 + t2F*c32_1;	t2F=t2F*s32_1 - t2E*c32_1;	t2E=rt;	/* twiddle mpy by E^-7	*/
	    rt =t3E*s32_3 + t3F*c32_3;	it =t3F*s32_3 - t3E*c32_3;		/* twiddle mpy by -E^-21 is here...	*/
	    t3E=t2E+rt;			t2E=t2E-rt;				/* ...and get E^-21 by flipping signs here.	*/
	    t3F=t2F+it;			t2F=t2F-it;

	    a1p07r=t0E+t2E;		a1p07i=t0F+t2F;
	    a1p17r=t0E-t2E;		a1p17i=t0F-t2F;
	    a1p0Fr=t1E+t3F;		a1p0Fi=t1F-t3E;	/* mpy by E^-4 = -I is inlined here...	*/
	    a1p1Fr=t1E-t3F;		a1p1Fi=t1F+t3E;

/*...and combine those to complete the radix-32 transform and do the carries. Since the outputs would
     normally be getting dispatched to 32 separate blocks of the A-array, we need 32 separate carries.	*/

	    l= j & (nwt-1);			/* We want (S*J mod N) - SI(L) for all 32 carries, so precompute	*/
	    n_minus_sil   = n-si[l  ];		/* N - SI(L) and for each J, find N - (B*J mod N) - SI(L)		*/
	    n_minus_silp1 = n-si[l+1];		/* For the inverse weight, want (S*(N - J) mod N) - SI(NWT - L) =	*/
	    sinwt   = si[nwt-l  ];		/*	= N - (S*J mod N) - SI(NWT - L) = (B*J mod N) - SI(NWT - L).	*/
	    sinwtm1 = si[nwt-l-1];

	    wtl     =wt0[    l  ];
	    wtn     =wt0[nwt-l  ]*scale;	/* Include 1/(n/2) scale factor of inverse transform here...	*/
	    wtlp1   =wt0[    l+1];
	    wtnm1   =wt0[nwt-l-1]*scale;	/* ...and here.	*/

/*...set0 is slightly different from others:	*/
	   cmplx_carry_norm_pow2_nocheck0(a1p00r,a1p00i,cy00,bjmodn00);
	    cmplx_carry_norm_pow2_nocheck(a1p01r,a1p01i,cy01,bjmodn01,0x01);
	    cmplx_carry_norm_pow2_nocheck(a1p02r,a1p02i,cy02,bjmodn02,0x02);
	    cmplx_carry_norm_pow2_nocheck(a1p03r,a1p03i,cy03,bjmodn03,0x03);
	    cmplx_carry_norm_pow2_nocheck(a1p04r,a1p04i,cy04,bjmodn04,0x04);
	    cmplx_carry_norm_pow2_nocheck(a1p05r,a1p05i,cy05,bjmodn05,0x05);
	    cmplx_carry_norm_pow2_nocheck(a1p06r,a1p06i,cy06,bjmodn06,0x06);
	    cmplx_carry_norm_pow2_nocheck(a1p07r,a1p07i,cy07,bjmodn07,0x07);
	    cmplx_carry_norm_pow2_nocheck(a1p08r,a1p08i,cy08,bjmodn08,0x08);
	    cmplx_carry_norm_pow2_nocheck(a1p09r,a1p09i,cy09,bjmodn09,0x09);
	    cmplx_carry_norm_pow2_nocheck(a1p0Ar,a1p0Ai,cy0A,bjmodn0A,0x0A);
	    cmplx_carry_norm_pow2_nocheck(a1p0Br,a1p0Bi,cy0B,bjmodn0B,0x0B);
	    cmplx_carry_norm_pow2_nocheck(a1p0Cr,a1p0Ci,cy0C,bjmodn0C,0x0C);
	    cmplx_carry_norm_pow2_nocheck(a1p0Dr,a1p0Di,cy0D,bjmodn0D,0x0D);
	    cmplx_carry_norm_pow2_nocheck(a1p0Er,a1p0Ei,cy0E,bjmodn0E,0x0E);
	    cmplx_carry_norm_pow2_nocheck(a1p0Fr,a1p0Fi,cy0F,bjmodn0F,0x0F);
	    cmplx_carry_norm_pow2_nocheck(a1p10r,a1p10i,cy10,bjmodn10,0x10);
	    cmplx_carry_norm_pow2_nocheck(a1p11r,a1p11i,cy11,bjmodn11,0x11);
	    cmplx_carry_norm_pow2_nocheck(a1p12r,a1p12i,cy12,bjmodn12,0x12);
	    cmplx_carry_norm_pow2_nocheck(a1p13r,a1p13i,cy13,bjmodn13,0x13);
	    cmplx_carry_norm_pow2_nocheck(a1p14r,a1p14i,cy14,bjmodn14,0x14);
	    cmplx_carry_norm_pow2_nocheck(a1p15r,a1p15i,cy15,bjmodn15,0x15);
	    cmplx_carry_norm_pow2_nocheck(a1p16r,a1p16i,cy16,bjmodn16,0x16);
	    cmplx_carry_norm_pow2_nocheck(a1p17r,a1p17i,cy17,bjmodn17,0x17);
	    cmplx_carry_norm_pow2_nocheck(a1p18r,a1p18i,cy18,bjmodn18,0x18);
	    cmplx_carry_norm_pow2_nocheck(a1p19r,a1p19i,cy19,bjmodn19,0x19);
	    cmplx_carry_norm_pow2_nocheck(a1p1Ar,a1p1Ai,cy1A,bjmodn1A,0x1A);
	    cmplx_carry_norm_pow2_nocheck(a1p1Br,a1p1Bi,cy1B,bjmodn1B,0x1B);
	    cmplx_carry_norm_pow2_nocheck(a1p1Cr,a1p1Ci,cy1C,bjmodn1C,0x1C);
	    cmplx_carry_norm_pow2_nocheck(a1p1Dr,a1p1Di,cy1D,bjmodn1D,0x1D);
	    cmplx_carry_norm_pow2_nocheck(a1p1Er,a1p1Ei,cy1E,bjmodn1E,0x1E);
	    cmplx_carry_norm_pow2_nocheck(a1p1Fr,a1p1Fi,cy1F,bjmodn1F,0x1F);

	    i =((uint32)(sw - bjmodn00) >> 31);	/* get ready for the next set...	*/
	    co2=co3;	/* For all data but the first set in each j-block, co2=co3. Thus, after the first block of data is done
			   and only then: for all subsequent blocks it's superfluous), this assignment decrements co2 by radix(1).	*/

/*...The radix-32 DIF pass is here:	*/

/*       gather the needed data (32 64-bit complex, i.e. 64 64-bit reals) and do the first set of four length-8 transforms.	*/

/*...Block 1:	*/

	jt = j1;

	    t02=a1p00r-a1p10r;	t03=a1p00i-a1p10i;
	    t00=a1p00r+a1p10r;	t01=a1p00i+a1p10i;

	    t06=a1p08r-a1p18r;	t07=a1p08i-a1p18i;
	    t04=a1p08r+a1p18r;	t05=a1p08i+a1p18i;

	    rt =t04;		it =t05;
	    t04=t00-rt;		t05=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t06;		it =t07;
	    t06=t02+it;		t07=t03-rt;
	    t02=t02-it;		t03=t03+rt;

	    t0A=a1p04r-a1p14r;	t0B=a1p04i-a1p14i;
	    t08=a1p04r+a1p14r;	t09=a1p04i+a1p14i;

	    t0E=a1p0Cr-a1p1Cr;	t0F=a1p0Ci-a1p1Ci;
	    t0C=a1p0Cr+a1p1Cr;	t0D=a1p0Ci+a1p1Ci;

	    rt =t0C;		it =t0D;
	    t0C=t08-rt;		t0D=t09-it;
	    t08=t08+rt;		t09=t09+it;

	    rt =t0E;		it =t0F;
	    t0E=t0A+it;		t0F=t0B-rt;
	    t0A=t0A-it;		t0B=t0B+rt;

	    rt =t08;		it =t09;
	    t08=t00-rt;		t09=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t0C;		it =t0D;
	    t0C=t04+it;		t0D=t05-rt;
	    t04=t04-it;		t05=t05+rt;

	    rt =(t0A-t0B)*ISRT2;it =(t0A+t0B)*ISRT2;
	    t0A=t02-rt;		t0B=t03-it;
	    t02=t02+rt;		t03=t03+it;

	    rt =(t0E+t0F)*ISRT2;it =(t0F-t0E)*ISRT2;
	    t0E=t06+rt;		t0F=t07+it;
	    t06=t06-rt;		t07=t07-it;

/*...Block 2:	*/

	jt = j1 + p02;

	    t12=a1p02r-a1p12r;		t13=a1p02i-a1p12i;
	    t10=a1p02r+a1p12r;		t11=a1p02i+a1p12i;

	    t16=a1p0Ar-a1p1Ar;		t17=a1p0Ai-a1p1Ai;
	    t14=a1p0Ar+a1p1Ar;		t15=a1p0Ai+a1p1Ai;

	    rt =t14;		it =t15;
	    t14=t10-rt;		t15=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t16;		it =t17;
	    t16=t12+it;		t17=t13-rt;
	    t12=t12-it;		t13=t13+rt;

	    t1A=a1p06r-a1p16r;		t1B=a1p06i-a1p16i;
	    t18=a1p06r+a1p16r;		t19=a1p06i+a1p16i;

	    t1E=a1p0Er-a1p1Er;		t1F=a1p0Ei-a1p1Ei;
	    t1C=a1p0Er+a1p1Er;		t1D=a1p0Ei+a1p1Ei;

	    rt =t1C;		it =t1D;
	    t1C=t18-rt;		t1D=t19-it;
	    t18=t18+rt;		t19=t19+it;

	    rt =t1E;		it =t1F;
	    t1E=t1A+it;		t1F=t1B-rt;
	    t1A=t1A-it;		t1B=t1B+rt;

	    rt =t18;		it =t19;
	    t18=t10-rt;		t19=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t1C;		it =t1D;
	    t1C=t14+it;		t1D=t15-rt;
	    t14=t14-it;		t15=t15+rt;

	    rt =(t1A-t1B)*ISRT2;it =(t1A+t1B)*ISRT2;
	    t1A=t12-rt;		t1B=t13-it;
	    t12=t12+rt;		t13=t13+it;

	    rt =(t1E+t1F)*ISRT2;it =(t1F-t1E)*ISRT2;
	    t1E=t16+rt;		t1F=t17+it;
	    t16=t16-rt;		t17=t17-it;

/*...Block 3:	*/

	jt = j1 + p01;

	    t22=a1p01r-a1p11r;		t23=a1p01i-a1p11i;
	    t20=a1p01r+a1p11r;		t21=a1p01i+a1p11i;

	    t26=a1p09r-a1p19r;		t27=a1p09i-a1p19i;
	    t24=a1p09r+a1p19r;		t25=a1p09i+a1p19i;

	    rt =t24;		it =t25;
	    t24=t20-rt;		t25=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t26;		it =t27;
	    t26=t22+it;		t27=t23-rt;
	    t22=t22-it;		t23=t23+rt;

	    t2A=a1p05r-a1p15r;		t2B=a1p05i-a1p15i;
	    t28=a1p05r+a1p15r;		t29=a1p05i+a1p15i;

	    t2E=a1p0Dr-a1p1Dr;		t2F=a1p0Di-a1p1Di;
	    t2C=a1p0Dr+a1p1Dr;		t2D=a1p0Di+a1p1Di;

	    rt =t2C;		it =t2D;
	    t2C=t28-rt;		t2D=t29-it;
	    t28=t28+rt;		t29=t29+it;

	    rt =t2E;		it =t2F;
	    t2E=t2A+it;		t2F=t2B-rt;
	    t2A=t2A-it;		t2B=t2B+rt;

	    rt =t28;		it =t29;
	    t28=t20-rt;		t29=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t2C;		it =t2D;
	    t2C=t24+it;		t2D=t25-rt;
	    t24=t24-it;		t25=t25+rt;

	    rt =(t2A-t2B)*ISRT2;it =(t2A+t2B)*ISRT2;
	    t2A=t22-rt;		t2B=t23-it;
	    t22=t22+rt;		t23=t23+it;

	    rt =(t2E+t2F)*ISRT2;it =(t2F-t2E)*ISRT2;
	    t2E=t26+rt;		t2F=t27+it;
	    t26=t26-rt;		t27=t27-it;

/*...Block 4:	*/

	jt = j1 + p03;

	    t32=a1p03r-a1p13r;		t33=a1p03i-a1p13i;
	    t30=a1p03r+a1p13r;		t31=a1p03i+a1p13i;

	    t36=a1p0Br-a1p1Br;		t37=a1p0Bi-a1p1Bi;
	    t34=a1p0Br+a1p1Br;		t35=a1p0Bi+a1p1Bi;

	    rt =t34;		it =t35;
	    t34=t30-rt;		t35=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t36;		it =t37;
	    t36=t32+it;		t37=t33-rt;
	    t32=t32-it;		t33=t33+rt;

	    t3A=a1p07r-a1p17r;		t3B=a1p07i-a1p17i;
	    t38=a1p07r+a1p17r;		t39=a1p07i+a1p17i;

	    t3E=a1p0Fr-a1p1Fr;		t3F=a1p0Fi-a1p1Fi;
	    t3C=a1p0Fr+a1p1Fr;		t3D=a1p0Fi+a1p1Fi;

	    rt =t3C;		it =t3D;
	    t3C=t38-rt;		t3D=t39-it;
	    t38=t38+rt;		t39=t39+it;

	    rt =t3E;		it =t3F;
	    t3E=t3A+it;		t3F=t3B-rt;
	    t3A=t3A-it;		t3B=t3B+rt;

	    rt =t38;		it =t39;
	    t38=t30-rt;		t39=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t3C;		it =t3D;
	    t3C=t34+it;		t3D=t35-rt;
	    t34=t34-it;		t35=t35+rt;

	    rt =(t3A-t3B)*ISRT2;it =(t3A+t3B)*ISRT2;
	    t3A=t32-rt;		t3B=t33-it;
	    t32=t32+rt;		t33=t33+it;

	    rt =(t3E+t3F)*ISRT2;it =(t3F-t3E)*ISRT2;
	    t3E=t36+rt;		t3F=t37+it;
	    t36=t36-rt;		t37=t37-it;

/*...and now do eight radix-4 transforms, including the internal twiddle factors:
  	1, exp(i* 1*twopi/32) =       ( c32_1, s32_1), exp(i* 2*twopi/32) =       ( c    , s    ), exp(i* 3*twopi/32) =       ( c32_3, s32_3) (for inputs to transform block 2),
  	1, exp(i* 2*twopi/32) =       ( c    , s    ), exp(i* 4*twopi/32) = ISRT2*( 1    , 1    ), exp(i* 3*twopi/32) =       ( s    , c    ) (for inputs to transform block 3),
  	1, exp(i* 3*twopi/32) =       ( c32_3, s32_3), exp(i* 6*twopi/32) =       ( s    , c    ), exp(i* 9*twopi/32) =       (-s32_1, c32_1) (for inputs to transform block 4),
  	1, exp(i* 4*twopi/32) = ISRT2*( 1    , 1    ), exp(i* 8*twopi/32) =       ( 0    , 1    ), exp(i*12*twopi/32) = ISRT2*(-1    , 1    ) (for inputs to transform block 5),
  	1, exp(i* 5*twopi/32) =       ( s32_3, c32_3), exp(i*10*twopi/32) =       (-s    , c    ), exp(i*15*twopi/32) =       (-c32_1, s32_1) (for inputs to transform block 6),
  	1, exp(i* 6*twopi/32) =       ( s    , c    ), exp(i*12*twopi/32) = ISRT2*(-1    , 1    ), exp(i*18*twopi/32) =       (-c    ,-s    ) (for inputs to transform block 7),
  	1, exp(i* 7*twopi/32) =       ( s32_1, c32_1), exp(i*14*twopi/32) =       (-c    , s    ), exp(i*21*twopi/32) =       (-s32_3,-c32_3) (for inputs to transform block 8),
     and only the last 3 inputs to each of the radix-4 transforms 2 through 8 are multiplied by non-unity twiddles.	*/

#if PFETCH
prefetch_offset = ((j >> 1) & 3)*p01;
#endif
/*...Block 1: t00,t10,t20,t30	*/

	jt = j1;

	    rt =t10;	t10=t00-rt;	t00=t00+rt;
	    it =t11;	t11=t01-it;	t01=t01+it;
#if PFETCH
add0 = &a[jt+prefetch_offset];
prefetch_p_doubles(add0);
#endif
	    rt =t30;	t30=t20-rt;	t20=t20+rt;
	    it =t31;	t31=t21-it;	t21=t21+it;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt    ]=t00+t20;		a[jt+1    ]=t01+t21;
	    a[jt+p01]=t00-t20;		a[jt+1+p01]=t01-t21;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt+p02]=t10-t31;		a[jt+1+p02]=t11+t30;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t10+t31;		a[jt+1+p03]=t11-t30;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
/*...Block 5: t08,t18,t28,t38	*/

	jt = j1 + p04;

	    rt =t18;	t18=t08+t19;	t08=t08-t19;		/* twiddle mpy by E^4 = I	*/
			t19=t09-rt;	t09=t09+rt;
#if PFETCH
add0 = &a[jt+prefetch_offset];
prefetch_p_doubles(add0);
#endif
	    rt =(t28-t29)*ISRT2;	t29=(t28+t29)*ISRT2;		t28=rt;	/* twiddle mpy by E^8	*/
	    rt =(t39+t38)*ISRT2;	it =(t39-t38)*ISRT2;			/* twiddle mpy by -E^12 is here...	*/
	    t38=t28+rt;			t28=t28-rt;				/* ...and get E^12 by flipping signs here.	*/
	    t39=t29+it;			t29=t29-it;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt    ]=t08+t28;		a[jt+1    ]=t09+t29;
	    a[jt+p01]=t08-t28;		a[jt+1+p01]=t09-t29;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt+p02]=t18-t39;		a[jt+1+p02]=t19+t38;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t18+t39;		a[jt+1+p03]=t19-t38;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
/*...Block 3: t04,t14,t24,t34	*/

	jt = j1 + p08;

	    rt =(t14-t15)*ISRT2;	it =(t14+t15)*ISRT2;			/* twiddle mpy by E^4	*/
	    t14=t04-rt;			t04=t04+rt;
	    t15=t05-it;			t05=t05+it;
#if PFETCH
add0 = &a[jt+prefetch_offset];
prefetch_p_doubles(add0);
#endif
	    rt =t24*c - t25*s;		t25=t25*c + t24*s;		t24=rt;	/* twiddle mpy by E^2	*/
	    rt =t34*s - t35*c;		it =t35*s + t34*c;			/* twiddle mpy by E^6	*/
	    t34=t24-rt;			t24=t24+rt;
	    t35=t25-it;			t25=t25+it;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt    ]=t04+t24;		a[jt+1    ]=t05+t25;
	    a[jt+p01]=t04-t24;		a[jt+1+p01]=t05-t25;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt+p02]=t14-t35;		a[jt+1+p02]=t15+t34;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t14+t35;		a[jt+1+p03]=t15-t34;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
/*...Block 7: t0C,t1C,t2C,t3C	*/

	jt = j1 + p0C;

	    rt =(t1D+t1C)*ISRT2;	it =(t1D-t1C)*ISRT2;			/* twiddle mpy by -E^12 is here...	*/
	    t1C=t0C+rt;			t0C=t0C-rt;				/* ...and get E^12 by flipping signs here.	*/
	    t1D=t0D+it;			t0D=t0D-it;
#if PFETCH
add0 = &a[jt+prefetch_offset];
prefetch_p_doubles(add0);
#endif
	    rt =t2C*s - t2D*c;		t2D=t2D*s + t2C*c;		t2C=rt;	/* twiddle mpy by E^6	*/
	    rt =t3C*c - t3D*s;		it =t3D*c + t3C*s;			/* twiddle mpy by E^18 is here...	*/
	    t3C=t2C+rt;			t2C=t2C-rt;				/* ...and get E^18 by flipping signs here.	*/
	    t3D=t2D+it;			t2D=t2D-it;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt    ]=t0C+t2C;		a[jt+1    ]=t0D+t2D;
	    a[jt+p01]=t0C-t2C;		a[jt+1+p01]=t0D-t2D;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt+p02]=t1C-t3D;		a[jt+1+p02]=t1D+t3C;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t1C+t3D;		a[jt+1+p03]=t1D-t3C;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
/*...Block 2: t02,t12,t22,t32	*/

	jt = j1 + p10;

	    rt =t12*c - t13*s;		it =t13*c + t12*s;			/* twiddle mpy by E^2	*/
	    t12=t02-rt;			t02=t02+rt;
	    t13=t03-it;			t03=t03+it;
#if PFETCH
add0 = &a[jt+prefetch_offset];
prefetch_p_doubles(add0);
#endif
	    rt =t22*c32_1 - t23*s32_1;	t23=t23*c32_1 + t22*s32_1;	t22=rt;	/* twiddle mpy by E^1	*/
	    rt =t32*c32_3 - t33*s32_3;	it =t33*c32_3 + t32*s32_3;		/* twiddle mpy by E^3	*/
	    t32=t22-rt;			t22=t22+rt;
	    t33=t23-it;			t23=t23+it;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt    ]=t02+t22;		a[jt+1    ]=t03+t23;
	    a[jt+p01]=t02-t22;		a[jt+1+p01]=t03-t23;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt+p02]=t12-t33;		a[jt+1+p02]=t13+t32;	/* mpy by E^4=i is inlined here...;	*/
	    a[jt+p03]=t12+t33;		a[jt+1+p03]=t13-t32;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
/*...Block 6: t0A,t1A,t2A,t3A	*/

	jt = j1 + p14;

	    rt =t1A*s + t1B*c;		it =t1B*s - t1A*c;			/* twiddle mpy by -E^10 is here...	*/
	    t1A=t0A+rt;			t0A =t0A-rt;				/* ...and get E^10 by flipping signs here.	*/
	    t1B=t0B+it;			t0B =t0B-it;
#if PFETCH
add0 = &a[jt+prefetch_offset];
prefetch_p_doubles(add0);
#endif
	    rt =t2A*s32_3 - t2B*c32_3;	t2B=t2B*s32_3 + t2A*c32_3;	t2A=rt;	/* twiddle mpy by E^5	*/
	    rt =t3A*c32_1 + t3B*s32_1;	it =t3B*c32_1 - t3A*s32_1;	 	/* twiddle mpy by -E^15 is here...	*/
	    t3A=t2A+rt;			t2A=t2A-rt;				/* ...and get E^15 by flipping signs here.	*/
	    t3B=t2B+it;			t2B=t2B-it;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt    ]=t0A+t2A;		a[jt+1    ]=t0B+t2B;
	    a[jt+p01]=t0A-t2A;		a[jt+1+p01]=t0B-t2B;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt+p02]=t1A-t3B;		a[jt+1+p02]=t1B+t3A;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t1A+t3B;		a[jt+1+p03]=t1B-t3A;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
/*...Block 4: t06,t16,t26,t36	*/

	jt = j1 + p18;

	    rt =t16*s - t17*c;		it =t17*s + t16*c;			/* twiddle mpy by E^6	*/
	    t16=t06-rt;			t06 =t06+rt;
	    t17=t07-it;			t07 =t07+it;
#if PFETCH
add0 = &a[jt+prefetch_offset];
prefetch_p_doubles(add0);
#endif
	    rt =t26*c32_3 - t27*s32_3;	t27=t27*c32_3 + t26*s32_3;	t26=rt;	/* twiddle mpy by E^3	*/
	    rt =t36*s32_1 + t37*c32_1;	it =t37*s32_1 - t36*c32_1;		/* twiddle mpy by -E^9 is here...	*/
	    t36=t26+rt;			t26=t26-rt;				/* ...and get E^9 by flipping signs here.	*/
	    t37=t27+it;			t27=t27-it;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt    ]=t06+t26;		a[jt+1    ]=t07+t27;
	    a[jt+p01]=t06-t26;		a[jt+1+p01]=t07-t27;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt+p02]=t16-t37;		a[jt+1+p02]=t17+t36;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t16+t37;		a[jt+1+p03]=t17-t36;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
/*...Block 8: t0E,t1E,t2E,t3E	*/

	jt = j1 + p1C;

	    rt =t1E*c + t1F*s;		it =t1F*c - t1E*s;			/* twiddle mpy by -E^14 is here...	*/
	    t1E=t0E+rt;			t0E =t0E-rt;				/* ...and get E^14 by flipping signs here.	*/
	    t1F=t0F+it;			t0F =t0F-it;
#if PFETCH
add0 = &a[jt+prefetch_offset];
prefetch_p_doubles(add0);
#endif
	    rt =t2E*s32_1 - t2F*c32_1;	t2F=t2F*s32_1 + t2E*c32_1;	t2E=rt;	/* twiddle mpy by E^7	*/
	    rt =t3E*s32_3 - t3F*c32_3;	it =t3F*s32_3 + t3E*c32_3;		/* twiddle mpy by -E^21 is here...	*/
	    t3E=t2E+rt;			t2E=t2E-rt;				/* ...and get E^21 by flipping signs here.	*/
	    t3F=t2F+it;			t2F=t2F-it;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt    ]=t0E+t2E;		a[jt+1    ]=t0F+t2F;
	    a[jt+p01]=t0E-t2E;		a[jt+1+p01]=t0F-t2F;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	    a[jt+p02]=t1E-t3F;		a[jt+1+p02]=t1F+t3E;	/* mpy by E^4=i is inlined here...	*/
	    a[jt+p03]=t1E+t3F;		a[jt+1+p03]=t1F-t3E;
#if PFETCH_X86
add0 += p01;
prefetch_p_doubles(add0);
#endif
	  iroot += root_incr;		/* increment sincos index.	*/

	  }

	  jstart += nwt;
	  jhi    += nwt;
	  col += 32;
	  co3 -= 32;

	}

	if(root_incr==0)break;

/*   Wraparound carry cleanup loop is here:

     The cleanup carries from the end of each length-N/32 block into the begining of the next
     can all be neatly processed as follows:

     (1) Invert the radix-32 forward DIF FFT of the first block of 32 complex elements in A and unweight;
     (2) Propagate cleanup carries among the real and imaginary parts of the 32 outputs of (1);
     (3) Reweight and perform a radix-32 forward DIF FFT on the result of (2);
     (4) If any of the exit carries from (2) are nonzero, advance to the next 32 elements and repeat (1-4).
*/
	t00 = cy1F;
	cy1F= cy1E;
	cy1E= cy1D;
	cy1D= cy1C;
	cy1C= cy1B;
	cy1B= cy1A;
	cy1A= cy19;
	cy19= cy18;
	cy18= cy17;
	cy17= cy16;
	cy16= cy15;
	cy15= cy14;
	cy14= cy13;
	cy13= cy12;
	cy12= cy11;
	cy11= cy10;
	cy10= cy0F;
	cy0F= cy0E;
	cy0E= cy0D;
	cy0D= cy0C;
	cy0C= cy0B;
	cy0B= cy0A;
	cy0A= cy09;
	cy09= cy08;
	cy08= cy07;
	cy07= cy06;
	cy06= cy05;
	cy05= cy04;
	cy04= cy03;
	cy03= cy02;
	cy02= cy01;
	cy01= cy00;
	cy00= t00;

	iroot = 0;
	root_incr = 0;
	scale = 1;

	jstart = 0;
	jhi = 3;
	khi = 1;

	for(j=0; j<=jhi; j++)
	{
	  jt = j;
	  a[jt    ] *= radix_inv;
	  a[jt+p01] *= radix_inv;
	  a[jt+p02] *= radix_inv;
	  a[jt+p03] *= radix_inv;
	  a[jt+p04] *= radix_inv;
	  a[jt+p05] *= radix_inv;
	  a[jt+p06] *= radix_inv;
	  a[jt+p07] *= radix_inv;
	  jt += p08;
	  a[jt    ] *= radix_inv;
	  a[jt+p01] *= radix_inv;
	  a[jt+p02] *= radix_inv;
	  a[jt+p03] *= radix_inv;
	  a[jt+p04] *= radix_inv;
	  a[jt+p05] *= radix_inv;
	  a[jt+p06] *= radix_inv;
	  a[jt+p07] *= radix_inv;
	  jt += p08;
	  a[jt    ] *= radix_inv;
	  a[jt+p01] *= radix_inv;
	  a[jt+p02] *= radix_inv;
	  a[jt+p03] *= radix_inv;
	  a[jt+p04] *= radix_inv;
	  a[jt+p05] *= radix_inv;
	  a[jt+p06] *= radix_inv;
	  a[jt+p07] *= radix_inv;
	  jt += p08;
	  a[jt    ] *= radix_inv;
	  a[jt+p01] *= radix_inv;
	  a[jt+p02] *= radix_inv;
	  a[jt+p03] *= radix_inv;
	  a[jt+p04] *= radix_inv;
	  a[jt+p05] *= radix_inv;
	  a[jt+p06] *= radix_inv;
	  a[jt+p07] *= radix_inv;
	}
}

	if(fabs(cy00)+fabs(cy01)+fabs(cy02)+fabs(cy03)+fabs(cy04)+fabs(cy05)+fabs(cy06)+fabs(cy07)+fabs(cy08)+fabs(cy09)+fabs(cy0A)+fabs(cy0B)+fabs(cy0C)+fabs(cy0D)+fabs(cy0E)+fabs(cy0F)
	  +fabs(cy10)+fabs(cy11)+fabs(cy12)+fabs(cy13)+fabs(cy14)+fabs(cy15)+fabs(cy16)+fabs(cy17)+fabs(cy18)+fabs(cy19)+fabs(cy1A)+fabs(cy1B)+fabs(cy1C)+fabs(cy1D)+fabs(cy1E)+fabs(cy1F)
	   != 0.0)
	{
	    if(INTERACT)printf("FATAL: iter = %10d; nonzero exit carry in radix32_ditN_cy_dif1.\n",iter);
	    f1 = fopen(   OFILE,"a");
	    f2 = fopen(STATFILE,"a");
	    fprintf(f1,"FATAL: iter = %10d; nonzero exit carry in radix32_ditN_cy_dif1.\n",iter);
	    fprintf(f2,"FATAL: iter = %10d; nonzero exit carry in radix32_ditN_cy_dif1.\n",iter);
	    fclose(f1);
	    fclose(f2);
	    err=1;
	    return(err);
	}

	return(0);
}

/**************/

void radix32_dit_pass1(double a[], int n)
{
/*
!...Acronym: DIT = Decimation In Time
!
!...Subroutine to perform an initial radix-32 complex inverse DIT FFT pass on the data in the length-N real vector A.
!
!   See the documentation in radix16_dif_pass for further details.
*/
	int j,j1,jt;
	static int n32,p01,p02,p03,p04,p05,p06,p07,p08,p10,p18, first_entry=TRUE;
	static double    c     = 0.92387953251128675613, s     = 0.38268343236508977173	/* exp(  i*twopi/16)	*/
			,c32_1 = 0.98078528040323044912, s32_1 = 0.19509032201612826784	/* exp(  i*twopi/32)	*/
			,c32_3 = 0.83146961230254523708, s32_3 = 0.55557023301960222473;/* exp(3*i*twopi/32)	*/
	double rt,it
		,t00,t01,t02,t03,t04,t05,t06,t07,t08,t09,t0A,t0B,t0C,t0D,t0E,t0F
		,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,t1A,t1B,t1C,t1D,t1E,t1F
		,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,t2A,t2B,t2C,t2D,t2E,t2F
		,t30,t31,t32,t33,t34,t35,t36,t37,t38,t39,t3A,t3B,t3C,t3D,t3E,t3F;

	if(!first_entry && (n >> 5) != n32)	/* New runlength?	*/
	{
	  first_entry=TRUE;
	}

/*...initialize things upon first entry	*/

	if(first_entry)
	{
	  first_entry=FALSE;
	  n32=n/32;

	  p01 = n32;
	  p02 = p01 +p01;
	  p03 = p02 +p01;
	  p04 = p03 +p01;
	  p05 = p04 +p01;
	  p06 = p05 +p01;
	  p07 = p06 +p01;
	  p08 = p07 +p01;
	  p10 = p08 +p08;
	  p18 = p10 +p08;

	  p01 = p01 + ( (p01 >> DAT_BITS) << PAD_BITS );
	  p02 = p02 + ( (p02 >> DAT_BITS) << PAD_BITS );
	  p03 = p03 + ( (p03 >> DAT_BITS) << PAD_BITS );
	  p04 = p04 + ( (p04 >> DAT_BITS) << PAD_BITS );
	  p05 = p05 + ( (p05 >> DAT_BITS) << PAD_BITS );
	  p06 = p06 + ( (p06 >> DAT_BITS) << PAD_BITS );
	  p07 = p07 + ( (p07 >> DAT_BITS) << PAD_BITS );
	  p08 = p08 + ( (p08 >> DAT_BITS) << PAD_BITS );
	  p10 = p10 + ( (p10 >> DAT_BITS) << PAD_BITS );
	  p18 = p18 + ( (p18 >> DAT_BITS) << PAD_BITS );
	}

/*...The radix-32 pass is here.	*/

      for(j=0; j < n32; j += 2)
      {
	  j1 = j  + ( (j  >> DAT_BITS) << PAD_BITS );

/*       gather the needed data (32 64-bit complex, i.e. 64 64-bit reals) and do the first set of four length-8 transforms.	*/

/*...Block 1:	*/

	jt = j1;

	    t00=a[jt    ];	t01=a[jt+1    ];
	    rt =a[jt+p01];	it =a[jt+1+p01];
	    t02=t00-rt;		t03=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    t04=a[jt+p02];	t05=a[jt+1+p02];
	    rt =a[jt+p03];	it =a[jt+1+p03];
	    t06=t04-rt;		t07=t05-it;
	    t04=t04+rt;		t05=t05+it;

	    rt =t04;		it =t05;
	    t04=t00-rt;		t05=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t06;		it =t07;
	    t06=t02-it;		t07=t03+rt;
	    t02=t02+it;		t03=t03-rt;

	    t08=a[jt+p04];	t09=a[jt+1+p04];
	    rt =a[jt+p05];	it =a[jt+1+p05];
	    t0A=t08-rt;		t0B=t09-it;
	    t08=t08+rt;		t09=t09+it;

	    t0C=a[jt+p06];	t0D=a[jt+1+p06];
	    rt =a[jt+p07];	it =a[jt+1+p07];
	    t0E=t0C-rt;		t0F=t0D-it;
	    t0C=t0C+rt;		t0D=t0D+it;

	    rt =t0C;		it =t0D;
	    t0C=t08-rt;		t0D=t09-it;
	    t08=t08+rt;		t09=t09+it;

	    rt =t0E;		it =t0F;
	    t0E=t0A-it;		t0F=t0B+rt;
	    t0A=t0A+it;		t0B=t0B-rt;

	    rt =t08;		it =t09;
	    t08=t00-rt;		t09=t01-it;
	    t00=t00+rt;		t01=t01+it;

	    rt =t0C;		it =t0D;
	    t0C=t04-it;		t0D=t05+rt;
	    t04=t04+it;		t05=t05-rt;

	    rt =(t0A+t0B)*ISRT2;it =(t0A-t0B)*ISRT2;
	    t0A=t02-rt;		t0B=t03+it;
	    t02=t02+rt;		t03=t03-it;

	    rt =(t0E-t0F)*ISRT2;it =(t0F+t0E)*ISRT2;
	    t0E=t06+rt;		t0F=t07+it;
	    t06=t06-rt;		t07=t07-it;

/*...Block 2:;	*/

	jt = j1 + p08;

	    t10=a[jt    ];	t11=a[jt+1    ];
	    rt =a[jt+p01];	it =a[jt+1+p01];
	    t12=t10-rt;		t13=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    t14=a[jt+p02];	t15=a[jt+1+p02];
	    rt =a[jt+p03];	it =a[jt+1+p03];
	    t16=t14-rt;		t17=t15-it;
	    t14=t14+rt;		t15=t15+it;

	    rt =t14;		it =t15;
	    t14=t10-rt;		t15=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t16;		it =t17;
	    t16=t12-it;		t17=t13+rt;
	    t12=t12+it;		t13=t13-rt;

	    t18=a[jt+p04];	t19=a[jt+1+p04];
	    rt =a[jt+p05];	it =a[jt+1+p05];
	    t1A=t18-rt;		t1B=t19-it;
	    t18=t18+rt;		t19=t19+it;

	    t1C=a[jt+p06];	t1D=a[jt+1+p06];
	    rt =a[jt+p07];	it =a[jt+1+p07];
	    t1E=t1C-rt;		t1F=t1D-it;
	    t1C=t1C+rt;		t1D=t1D+it;

	    rt =t1C;		it =t1D;
	    t1C=t18-rt;		t1D=t19-it;
	    t18=t18+rt;		t19=t19+it;

	    rt =t1E;		it =t1F;
	    t1E=t1A-it;		t1F=t1B+rt;
	    t1A=t1A+it;		t1B=t1B-rt;

	    rt =t18;		it =t19;
	    t18=t10-rt;		t19=t11-it;
	    t10=t10+rt;		t11=t11+it;

	    rt =t1C;		it =t1D;
	    t1C=t14-it;		t1D=t15+rt;
	    t14=t14+it;		t15=t15-rt;

	    rt =(t1A+t1B)*ISRT2;it =(t1A-t1B)*ISRT2;
	    t1A=t12-rt;		t1B=t13+it;
	    t12=t12+rt;		t13=t13-it;

	    rt =(t1E-t1F)*ISRT2;it =(t1F+t1E)*ISRT2;
	    t1E=t16+rt;		t1F=t17+it;
	    t16=t16-rt;		t17=t17-it;

/*...Block 3:	*/

	jt = j1 + p10;

	    t20=a[jt    ];	t21=a[jt+1    ];
	    rt =a[jt+p01];	it =a[jt+1+p01];
	    t22=t20-rt;		t23=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    t24=a[jt+p02];	t25=a[jt+1+p02];
	    rt =a[jt+p03];	it =a[jt+1+p03];
	    t26=t24-rt;		t27=t25-it;
	    t24=t24+rt;		t25=t25+it;

	    rt =t24;		it =t25;
	    t24=t20-rt;		t25=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t26;		it =t27;
	    t26=t22-it;		t27=t23+rt;
	    t22=t22+it;		t23=t23-rt;

	    t28=a[jt+p04];	t29=a[jt+1+p04];
	    rt =a[jt+p05];	it =a[jt+1+p05];
	    t2A=t28-rt;		t2B=t29-it;
	    t28=t28+rt;		t29=t29+it;

	    t2C=a[jt+p06];	t2D=a[jt+1+p06];
	    rt =a[jt+p07];	it =a[jt+1+p07];
	    t2E=t2C-rt;		t2F=t2D-it;
	    t2C=t2C+rt;		t2D=t2D+it;

	    rt =t2C;		it =t2D;
	    t2C=t28-rt;		t2D=t29-it;
	    t28=t28+rt;		t29=t29+it;

	    rt =t2E;		it =t2F;
	    t2E=t2A-it;		t2F=t2B+rt;
	    t2A=t2A+it;		t2B=t2B-rt;

	    rt =t28;		it =t29;
	    t28=t20-rt;		t29=t21-it;
	    t20=t20+rt;		t21=t21+it;

	    rt =t2C;		it =t2D;
	    t2C=t24-it;		t2D=t25+rt;
	    t24=t24+it;		t25=t25-rt;

	    rt =(t2A+t2B)*ISRT2;it =(t2A-t2B)*ISRT2;
	    t2A=t22-rt;		t2B=t23+it;
	    t22=t22+rt;		t23=t23-it;

	    rt =(t2E-t2F)*ISRT2;it =(t2F+t2E)*ISRT2;
	    t2E=t26+rt;		t2F=t27+it;
	    t26=t26-rt;		t27=t27-it;

/*...Block 4:	*/

	jt = j1 + p18;

	    t30=a[jt    ];	t31=a[jt+1    ];
	    rt =a[jt+p01];	it =a[jt+1+p01];
	    t32=t30-rt;		t33=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    t34=a[jt+p02];	t35=a[jt+1+p02];
	    rt =a[jt+p03];	it =a[jt+1+p03];
	    t36=t34-rt;		t37=t35-it;
	    t34=t34+rt;		t35=t35+it;

	    rt =t34;		it =t35;
	    t34=t30-rt;		t35=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t36;		it =t37;
	    t36=t32-it;		t37=t33+rt;
	    t32=t32+it;		t33=t33-rt;

	    t38=a[jt+p04];	t39=a[jt+1+p04];
	    rt =a[jt+p05];	it =a[jt+1+p05];
	    t3A=t38-rt;		t3B=t39-it;
	    t38=t38+rt;		t39=t39+it;

	    t3C=a[jt+p06];	t3D=a[jt+1+p06];
	    rt =a[jt+p07];	it =a[jt+1+p07];
	    t3E=t3C-rt;		t3F=t3D-it;
	    t3C=t3C+rt;		t3D=t3D+it;

	    rt =t3C;		it =t3D;
	    t3C=t38-rt;		t3D=t39-it;
	    t38=t38+rt;		t39=t39+it;

	    rt =t3E;		it =t3F;
	    t3E=t3A-it;		t3F=t3B+rt;
	    t3A=t3A+it;		t3B=t3B-rt;

	    rt =t38;		it =t39;
	    t38=t30-rt;		t39=t31-it;
	    t30=t30+rt;		t31=t31+it;

	    rt =t3C;		it =t3D;
	    t3C=t34-it;		t3D=t35+rt;
	    t34=t34+it;		t35=t35-rt;

	    rt =(t3A+t3B)*ISRT2;it =(t3A-t3B)*ISRT2;
	    t3A=t32-rt;		t3B=t33+it;
	    t32=t32+rt;		t33=t33-it;

	    rt =(t3E-t3F)*ISRT2;it =(t3F+t3E)*ISRT2;
	    t3E=t36+rt;		t3F=t37+it;
	    t36=t36-rt;		t37=t37-it;

/*...and now do eight radix-4 transforms, including the internal twiddle factors:
  	1, exp(-i* 1*twopi/32) =       ( c32_1,-s32_1), exp(-i* 2*twopi/32) =       ( c    ,-s    ), exp(-i* 3*twopi/32) =       ( c32_3,-s32_3) (for inputs to transform block 2),
  	1, exp(-i* 2*twopi/32) =       ( c    ,-s    ), exp(-i* 4*twopi/32) = ISRT2*( 1    ,-1    ), exp(-i* 3*twopi/32) =       ( s    ,-c    ) (for inputs to transform block 3),
  	1, exp(-i* 3*twopi/32) =       ( c32_3,-s32_3), exp(-i* 6*twopi/32) =       ( s    ,-c    ), exp(-i* 9*twopi/32) =       (-s32_1,-c32_1) (for inputs to transform block 4),
  	1, exp(-i* 4*twopi/32) = ISRT2*( 1    ,-1    ), exp(-i* 8*twopi/32) =       ( 0    ,-1    ), exp(-i*12*twopi/32) = ISRT2*(-1    ,-1    ) (for inputs to transform block 5),
  	1, exp(-i* 5*twopi/32) =       ( s32_3,-c32_3), exp(-i*10*twopi/32) =       (-s    ,-c    ), exp(-i*15*twopi/32) =       (-c32_1,-s32_1) (for inputs to transform block 6),
  	1, exp(-i* 6*twopi/32) =       ( s    ,-c    ), exp(-i*12*twopi/32) = ISRT2*(-1    ,-1    ), exp(-i*18*twopi/32) =       (-c    , s    ) (for inputs to transform block 7),
  	1, exp(-i* 7*twopi/32) =       ( s32_1,-c32_1), exp(-i*14*twopi/32) =       (-c    ,-s    ), exp(-i*21*twopi/32) =       (-s32_3, c32_3) (for inputs to transform block 8),
     and only the last 3 inputs to each of the radix-4 transforms 2 through 8 are multiplied by non-unity twiddles.	*/

/*...Block 1: t00,t10,t20,t30	*/

	jt = j1;

	    rt =t10;	t10=t00-rt;	t00=t00+rt;
	    it =t11;	t11=t01-it;	t01=t01+it;

	    rt =t30;	t30=t20-rt;	t20=t20+rt;
	    it =t31;	t31=t21-it;	t21=t21+it;

	    a[jt    ]=t00+t20;		a[jt+1    ]=t01+t21;
	    a[jt+p10]=t00-t20;		a[jt+1+p10]=t01-t21;

	    a[jt+p08]=t10+t31;		a[jt+1+p08]=t11-t30;	/* mpy by E^-4 = -I is inlined here...	*/
	    a[jt+p18]=t10-t31;		a[jt+1+p18]=t11+t30;

/*...Block 5: t08,t18,t28,t38	*/

	jt = j1 + p04;

	    rt =t18;	t18=t08-t19;	t08=t08+t19;		/* twiddle mpy by E^8 =-I	*/
			t19=t09+rt;	t09=t09-rt;

	    rt =(t29+t28)*ISRT2;	t29=(t29-t28)*ISRT2;		t28=rt;	/* twiddle mpy by E^-4	*/
	    rt =(t38-t39)*ISRT2;	it =(t38+t39)*ISRT2;			/* twiddle mpy by E^4 = -E^-12 is here...	*/
	    t38=t28+rt;			t28=t28-rt;				/* ...and get E^-12 by flipping signs here.	*/
	    t39=t29+it;			t29=t29-it;

	    a[jt    ]=t08+t28;		a[jt+1    ]=t09+t29;
	    a[jt+p10]=t08-t28;		a[jt+1+p10]=t09-t29;

	    a[jt+p08]=t18+t39;		a[jt+1+p08]=t19-t38;	/* mpy by E^-4 = -I is inlined here...	*/
	    a[jt+p18]=t18-t39;		a[jt+1+p18]=t19+t38;

/*...Block 3: t04,t14,t24,t34	*/

	jt = j1 + p02;

	    rt =(t15+t14)*ISRT2;	it =(t15-t14)*ISRT2;			/* twiddle mpy by E^-4	*/
	    t14=t04-rt;			t04=t04+rt;
	    t15=t05-it;			t05=t05+it;

	    rt =t24*c + t25*s;		t25=t25*c - t24*s;		t24=rt;	/* twiddle mpy by E^-2	*/
	    rt =t34*s + t35*c;		it =t35*s - t34*c;			/* twiddle mpy by E^-6	*/
	    t34=t24-rt;			t24=t24+rt;
	    t35=t25-it;			t25=t25+it;

	    a[jt    ]=t04+t24;		a[jt+1    ]=t05+t25;
	    a[jt+p10]=t04-t24;		a[jt+1+p10]=t05-t25;

	    a[jt+p08]=t14+t35;		a[jt+1+p08]=t15-t34;	/* mpy by E^-4 = -I is inlined here...	*/
	    a[jt+p18]=t14-t35;		a[jt+1+p18]=t15+t34;

/*...Block 7: t0C,t1C,t2C,t3C	*/

	jt = j1 + p06;

	    rt =(t1C-t1D)*ISRT2;	it =(t1C+t1D)*ISRT2;			/* twiddle mpy by E^4 = -E^-12 is here...	*/
	    t1C=t0C+rt;			t0C=t0C-rt;				/* ...and get E^-12 by flipping signs here.	*/
	    t1D=t0D+it;			t0D=t0D-it;

	    rt =t2C*s + t2D*c;		t2D=t2D*s - t2C*c;		t2C=rt;	/* twiddle mpy by E^-6	*/
	    rt =t3C*c + t3D*s;		it =t3D*c - t3C*s;			/* twiddle mpy by E^-18 is here...	*/
	    t3C=t2C+rt;			t2C=t2C-rt;				/* ...and get E^-18 by flipping signs here.	*/
	    t3D=t2D+it;			t2D=t2D-it;

	    a[jt    ]=t0C+t2C;		a[jt+1    ]=t0D+t2D;
	    a[jt+p10]=t0C-t2C;		a[jt+1+p10]=t0D-t2D;

	    a[jt+p08]=t1C+t3D;		a[jt+1+p08]=t1D-t3C;	/* mpy by E^-4 = -I is inlined here...	*/
	    a[jt+p18]=t1C-t3D;		a[jt+1+p18]=t1D+t3C;

/*...Block 2: t02,t12,t22,t32	*/

	jt = j1 + p01;

	    rt =t12*c + t13*s;		it =t13*c - t12*s;			/* twiddle mpy by E^-2	*/
	    t12=t02-rt;			t02=t02+rt;
	    t13=t03-it;			t03=t03+it;

	    rt =t22*c32_1 + t23*s32_1;	t23=t23*c32_1 - t22*s32_1;	t22=rt;	/* twiddle mpy by E^-1	*/
	    rt =t32*c32_3 + t33*s32_3;	it =t33*c32_3 - t32*s32_3;		/* twiddle mpy by E^-3	*/
	    t32=t22-rt;			t22=t22+rt;
	    t33=t23-it;			t23=t23+it;

	    a[jt    ]=t02+t22;		a[jt+1    ]=t03+t23;
	    a[jt+p10]=t02-t22;		a[jt+1+p10]=t03-t23;

	    a[jt+p08]=t12+t33;		a[jt+1+p08]=t13-t32;	/* mpy by E^-4 = -I is inlined here...	*/
	    a[jt+p18]=t12-t33;		a[jt+1+p18]=t13+t32;

/*...Block 6: t0A,t1A,t2A,t3A	*/

	jt = j1 + p05;

	    rt =t1A*s - t1B*c;		it =t1B*s + t1A*c;			/* twiddle mpy by -E^-10 is here...	*/
	    t1A=t0A+rt;			t0A =t0A-rt;				/* ...and get E^-10 by flipping signs here.	*/
	    t1B=t0B+it;			t0B =t0B-it;

	    rt =t2A*s32_3 + t2B*c32_3;	t2B=t2B*s32_3 - t2A*c32_3;	t2A=rt;	/* twiddle mpy by E^-5	*/
	    rt =t3A*c32_1 - t3B*s32_1;	it =t3B*c32_1 + t3A*s32_1;		/* twiddle mpy by -E^-15 is here...	*/
	    t3A=t2A+rt;			t2A=t2A-rt;				/* ...and get E^-15 by flipping signs here.	*/
	    t3B=t2B+it;			t2B=t2B-it;

	    a[jt    ]=t0A+t2A;		a[jt+1    ]=t0B+t2B;
	    a[jt+p10]=t0A-t2A;		a[jt+1+p10]=t0B-t2B;

	    a[jt+p08]=t1A+t3B;		a[jt+1+p08]=t1B-t3A;	/* mpy by E^-4 = -I is inlined here...	*/
	    a[jt+p18]=t1A-t3B;		a[jt+1+p18]=t1B+t3A;

/*...Block 4: t06,t16,t26,t36	*/

	jt = j1 + p03;

	    rt =t16*s + t17*c;		it =t17*s - t16*c;			/* twiddle mpy by E^-6	*/
	    t16=t06-rt;			t06 =t06+rt;
	    t17=t07-it;			t07 =t07+it;

	    rt =t26*c32_3 + t27*s32_3;	t27=t27*c32_3 - t26*s32_3;	t26=rt;	/* twiddle mpy by E^-3	*/
	    rt =t36*s32_1 - t37*c32_1;	it =t37*s32_1 + t36*c32_1;		/* twiddle mpy by -E^-9 is here...	*/
	    t36=t26+rt;			t26=t26-rt;				/* ...and get E^-9 by flipping signs here.	*/
	    t37=t27+it;			t27=t27-it;

	    a[jt    ]=t06+t26;		a[jt+1    ]=t07+t27;
	    a[jt+p10]=t06-t26;		a[jt+1+p10]=t07-t27;

	    a[jt+p08]=t16+t37;		a[jt+1+p08]=t17-t36;	/* mpy by E^-4 = -I is inlined here...	*/
	    a[jt+p18]=t16-t37;		a[jt+1+p18]=t17+t36;

/*...Block 8: t0E,t1E,t2E,t3E	*/

	jt = j1 + p07;

	    rt =t1E*c - t1F*s;		it =t1F*c + t1E*s;			/* twiddle mpy by -E^-14 is here...	*/
	    t1E=t0E+rt;			t0E =t0E-rt;				/* ...and get E^-14 by flipping signs here.	*/
	    t1F=t0F+it;			t0F =t0F-it;

	    rt =t2E*s32_1 + t2F*c32_1;	t2F=t2F*s32_1 - t2E*c32_1;	t2E=rt;	/* twiddle mpy by E^-7	*/
	    rt =t3E*s32_3 + t3F*c32_3;	it =t3F*s32_3 - t3E*c32_3;		/* twiddle mpy by -E^-21 is here...	*/
	    t3E=t2E+rt;			t2E=t2E-rt;				/* ...and get E^-21 by flipping signs here.	*/
	    t3F=t2F+it;			t2F=t2F-it;

	    a[jt    ]=t0E+t2E;		a[jt+1    ]=t0F+t2F;
	    a[jt+p10]=t0E-t2E;		a[jt+1+p10]=t0F-t2F;

	    a[jt+p08]=t1E+t3F;		a[jt+1+p08]=t1F-t3E;	/* mpy by E^-4 = -I is inlined here...	*/
	    a[jt+p18]=t1E-t3F;		a[jt+1+p18]=t1F+t3E;

	}
}

